#!BPY
""" Registration info for Blender menus:
Name: 'X-Plane Annotation...'
Blender: 249
Group: 'Import'
Tooltip: 'Import an OBJ annotation.'
"""
__author__ = "Ben Supnik"
__email__ = "Ben Supnik, bsupnik at xsquawkbox dot net"
__url__ = "X-Plane, wiki.x-plane.com"
__version__ = "1.00"
__bpydoc__ = """\
This script is a derivative work of Jonathan Harris' X-Plane blender scripts, and is covered under the terms of the GPL.
"""

import sys
import os
import Blender
from Blender import Armature, Object, Mesh, NMesh, Lamp, Image, Material, Texture, Draw, Window, Text, Registry, Modifier
from Blender.Mathutils import Matrix, RotationMatrix, TranslationMatrix, Vector
from XPlaneUtils import *
from Blender.Window import GetCursorPos
from XPlaneExport import ExportError
from XPlaneImport_util import OBJimport, ParseError
from XPlaneLib import *

def set_pref_key(container, key, value):
	dict = Blender.Registry.GetKey(container)
	if not dict:
		dict={}
	dict[key]=value
	Blender.Registry.SetKey(container,dict)
	
def get_pref_key(container, key):
	dict = Blender.Registry.GetKey(container)
	if not dict:
		return None
	if not key in dict:
		return None
	return dict[key]
	
def get_obj_xmin(object):
	if object.type != 'Mesh':
		return 0.0
	xmin = 9999.0
	for v in object.getBoundBox(0):
		xmin = min(xmin, v[0])
	return xmin
	
def get_obj_xmax(object):
	if object.type != 'Mesh':
		return 0.0
	xmax = -9999.0
	for v in object.getBoundBox(0):
		xmax = max(xmax, v[0])
	return xmax
	
def makeBearable(ob):

	""" Alex sez : imported obj annotations often look horrible for various reasons. This hack attempts to make them look vaguely ok 
	"""
	scn = Blender.Scene.GetCurrent()
	# Lamps: since the importer moves all lamps to the obj origin on import, there's little point in keeping them - all they do is clutter the scene and foul up texture baking activites.
	
	if ob.type == 'Mesh':
	
		# Transparency: by default, mesh is imported opaque for some reason. Changing all mesh to alpha mode inproves things a bit:
		mesh = ob.getData()
		changed = False
		for face in mesh.faces:
			if face.mode & Mesh.FaceModes.TEX:
				if face.transp == Mesh.FaceTranspModes.SOLID:
					face.transp = Mesh.FaceTranspModes.ALPHA
					changed = True
		if changed:
			mesh.update()
		
		# Edge splits: splits in smooth-shaded mesh are not imported correctly - generally there are too FEW splits (i.e. some are missing) in the import.
		# Applying a  40 degree 'edge split' modifier to the mesh fixes most uglies:
		
		splitmod = ob.modifiers.append(Modifier.Types.EDGESPLIT)
		splitmod[Modifier.Settings.EDGESPLIT_ANGLE] = 40.0
		splitmod[Modifier.Settings.EDGESPLIT_FROM_ANGLE] = True
		ob.makeDisplayList()
	
	
def make_relative_path(path_to_child, path_to_parent_dir):
	child = path_to_child.split(os.sep)
	par = path_to_parent_dir.split(os.sep)
	while len(par) > 0 and len(child) > 0 and par[0] == child[0]:
		child.pop(0)
		par.pop(0)
	ret = []
	for x in par:
		ret.append('..')
	ret.extend(child)
	return os.sep.join(ret)

def append_to_inner(list, item):
	if len(list) == 0:
		list.append(item)
	elif type(list[-1][1]) != type([]):
		list.append(item)
	else:
		append_to_inner(list[-1][1], item)		

def extend_pop_for(items, segments):
	if len(segments) != 0:
		if len(items) == 0 or type(items[-1][1]) != type([]) or segments[0] != items[-1][0]:
			items.append((segments[0], []))
		extend_pop_for(items[-1][1],segments[1:])
	
try:

	# First go pull all local and library items from disk.  They come as a list of tuples:
	# annotation path, absolute physical path, type ('lcl' or 'lib')
	# "items" is the library items, and "objs" is the in-pack objects.

	baseFileName=Blender.Get('filename')
	l = baseFileName.lower().rfind('.blend')
	if l==-1: raise ExportError('Save this .blend file first')
	root = locate_root(baseFileName)
	packs = get_scenery_packs(root[0])
	objs = get_local_assets(root[1],'.obj')
	items = get_library(root[0],'.obj')

	# Now go through and make menu items for every folder (recursively including sub-folders).
	# We'll do that by, for each item, "filing" it under each partial path contained in its full path.
	# Our menu item format is
	# (
	#	[(annotation path, absolute physical path, type) ...]
	#  menu_item_name
	# )
	# where menu_item_name is a path to the menu item with parent menu items divided by / and
	# the annotations are stored in a list.	
	
	all_by_dir={}
	for i in items:
		parts=i[0].split('/')[:-1]
		while len(parts) > 0:
			path='/'.join(parts)
			if not path in all_by_dir:
				all_by_dir[path]=[]
			all_by_dir[path].append([i[0],i[1],i[2]])
			parts=parts[:-1]
	
	all_items=[]
	
	for partial, all_assets in all_by_dir.iteritems():
		title="%s/%d objects" % (partial, len(all_assets))
		all_items.append([all_assets, title])
		
	# Append local objects to lib after culling the lib for folders, then transform
	# from library item format to menu item format.
	
	items.extend(objs)	

	items = [([(i[0],i[1],i[2])],i[0]) for i in items]

	# Finally add the folder items and sort it to put the menu into sorted order.
	
	items.extend(all_items)
	
	items.sort(key=lambda item: item[1])

#	Recent item is disabled for now

#	last_item =	get_pref_key('xplane.tools','annotate.last_item')
#	if last_item:
#		#pop_list.append((last_item[0],len(items)))
#		#items.append(last_item)
#		vpath='/'.join(last_item[0].split('/')[:-1])
#		new_items = []
#		for i in items:
#			if i[0].startswith(vpath):
#				new_items.append([i[0],i[1],i[2],"recent items/"+i[0].split('/')[-1]])
#		items.extend(new_items)


	pop_list=[]
	
	for i in items:
		path=i[1].split('/')[:-1]
		extend_pop_for(pop_list,path)
		if len(path):
			append_to_inner(pop_list,(i[1].split('/')[-1],items.index(i)))
		else:
			pop_list.append((i[1].split('/')[-1],items.index(i)))
	

	r = Draw.PupTreeMenu(pop_list)
	if r >= 0:
		# R is the "command ID" from the menu, which is really a numeric index into our items. 
		# from that we can get the LIST of 1 or more assets to import.
		cursor_posn = Window.GetCursorPos()
		
		cur = Window.GetCursorPos()
		for asset in items[r][0]:
			print asset
			#set_pref_key('xplane.tools','annotate.last_item',items[r])
			mm = TranslationMatrix(Vector([0,0,0])).resize4x4()
			#mm = TranslationMatrix(Vector(Window.GetCursorPos())).resize4x4()
			importer=OBJimport(asset[1],mm)
			importer.verbose=0
			try:
				sel = Blender.Scene.GetCurrent().objects.selected
				old_objs = set(Blender.Scene.GetCurrent().objects)
				obj_list = importer.doimport()
				new_objs = set(Blender.Scene.GetCurrent().objects)
				wrapper=Blender.Object.New('Empty','OBJ%s' % strip_suffix(asset[0].split('/')[-1]))		# strip suffix to nuke the redundant ".obj" suffix which causes blender name-length issues
				Blender.Scene.GetCurrent().objects.link(wrapper)
				added = new_objs-old_objs
				# AG sez - don't parent children of armatures, they're already attached!
				added_non_anim = [ob for ob in added if not ob.parent]
				wrapper.makeParent(list(added_non_anim),0,0)
				if len(sel) == 1:
					sel[0].makeParent([wrapper],0,0)
					base = sel[0].getLocation('worldspace')
					cur[0] -= base[0]
					cur[1] -= base[1]
					cur[2] -= base[2]
					
				#### Alex G hacks: #####
				
				# Massage to reduce the number of artefacts on import:
				
				for ob in added:
					makeBearable(ob)
			
				# Update our location offset, based on the size of the mesh:
				
				ob_xmin = 0.0
				for ob in [ob for ob in added if ob.type == 'Mesh']:
					ob_xmin = min(ob_xmin, get_obj_xmin(ob))
				cur[0] += round(5.0 - ob_xmin)
				wrapper.setLocation(cur[0], cur[1], cur[2])

				ob_xmax = 0.0
				for ob in added:
					ob_xmax = max(ob_xmax, get_obj_xmax(ob))
				cur[0] += round(ob_xmax)
				
				# Start a new row if the line of objs is too long for blender's viewport (max = +- 10,000m each axis) or the author's sanity:
				print cursor_posn, cur
				if (cur[0] - cursor_posn[0] >=  1000.0) or (cur[0] > 9980.0):
					cur[0] = cursor_posn[0]
					cur[1] +=100.0
				
				rel_path = asset[0]
				if asset[2]=='lcl':
					rel_path = make_relative_path(asset[1],os.path.dirname(baseFileName))
				wrapper.addProperty('external',rel_path,'STRING')

			#for ob in obj_list:
			#	ob.setLocation(cur[0], cur[1], cur[2])

			except ParseError, e:
				Window.WaitCursor(0)
				Window.DrawProgressBar(0, 'ERROR')
				if e.type == ParseError.HEADER:
					msg='This is not a valid X-Plane v6, v7 or v8 OBJ file'
				elif e.type == ParseError.PANEL:
					msg='Cannot read cockpit panel texture'
				elif e.type == ParseError.NAME:
					msg='Missing dataref or light name at line %s\n' % obj.lineno
				elif e.type == ParseError.MISC:
					msg='%s at line %s' % (e.value, importer.lineno)
				else:
					thing=ParseError.TEXT[e.type]
					if e.value:
						msg='Expecting a %s, found "%s" at line %s' % (thing, e.value, obj.lineno)
					else:
						msg='Missing %s at line %s' % (thing, obj.lineno)
				print "ERROR:\t%s\n" % msg
				Draw.PupMenu("ERROR%%t|%s" % msg)
				Window.RedrawAll()
				Window.DrawProgressBar(1, 'ERROR')

	
	#print result
	Window.RedrawAll()
#Window.DrawProgressBar(1, 'ERROR')
except ExportError, e:
	print "%s" % e.msg
	Draw.PupMenu("ERROR: %s" % e.msg)
