#!BPY
"""
Name: 'XPlaneAGTools'
Blender: 249
Group: 'Object'
Tooltip: 'Tools to automate repetitve tasks in Autogen and OBJ authoring workflows.'
"""
__author__ = "Alex Gifford"
__email__ = ""
__url__ = ""
__version__ = "1.60"
__bpydoc__ = """\

XPlaneAGTools offers a set of tools aimed at authors of X-Plane related art assets.
The script can be launched from the 'Object' submenu and left open, to be used when required.
It consists of a GUI containing multiple pages, each selected by clicking on the relevant tab
at the top of the GUI. User settings are preserved when switching between pages.

The available pages are currently as follows:

*
* AGS
*
*

This page contains some tools useful during the process of creating AGString autogen art assets.
It works like this:
-	Use the blue buttons to enable one or more functions. No actions are performed at this stage.
-   Use the various input fields to customize settings as desired.
-	Select the objects you want to modify, and click on the 'APPLY' button.

The following functionality is provided:

- 	Footprint tile renaming. All selected meshes will be renamed in numerical order,
	starting from the specified number. The order is determined by the x-axis location of the tiles,
	from left to right.

-	Relocation of tile centers. It is often convenient to have the center of all tile quads located
 	at their lower left corner. Use 'move tile centers' to automatically do this.

-	Tile wrapping. The 'wrap tiles' button automatically adds 'empty'-type object 'wrappers' at each
	selected tile's origin, auto-names it and make it parent to the respective tile. 'WRAPPER LAYER'
	specifies which scene layer to add the wrappers to.

-	AGS wrapping. 'wrap AGS' does the following:
	- A new AGS top-level wrapper object of type 'Empty' is added at the 3d cursor location.
	- All selected tiles wrappers, and/or wrappers of selected tile meshes are parented to this object.
	- A standard set of AGString properties is added to the new wrapper object, for later completion.
	It is legal for the initial selection to contain other scene elements (obj, trees, facades etc),
	they will be ignored.

-	Meta-pin addition. The center section of the AGS page contains a number of controls relating to
	AGString 'meta-pins'. Meta-pins are used to define how tiles should align with road edges, other
	tiles, and random obstacles encountered in X-Plane. Adding them manually can be a tedious process,
	so a little automation is welcome! Typically, the final position of the various pins will still
	require some manual tweaking, but at least the donkey work can now be done automatically.

	- 	The 'ground point' button adds a ground reference point pin to each tile, at a nominal location.
		These pins are only necessary if the tile contains graded (non-draped) obj annotations.

	-	The 'share y' button adds a property to the respective tile wrapper to denote that the ground
		point is to be shared between adjacent tiles in AGS spellings.

	- 	'pivot pins' will add leading and trailing pivot pins to each tile, for articulated corners.

	- 	The 'anchor pin' button adds an anchor pin at the specifed distance from the front edge of each
		tile. Required by all tiles except the middle tile of three-tile articulated corners.

	- 	'Stop pins' are required by all AGS tiles, to prevent other tiles encroaching on them. When
		enabled, the 'stop pins' button will add a line of equi-spaced stop-pins around the perimeter
		of every tile, inset from each edge by the specified distances. The 'spacing' setting is used
		as a basis for the spacing between stop-pins - the final spacing will depend on each
		specific case.

-	Texture projection. The right hand section of the AGS page contains controls to enable the author
	to project the specified texture on to multiple tile objects in a single operation. A typical work-
	flow is to:

	1 	Create the draped, footprint texture in Photoshop or similar.

	2 	Add a horizontal quad to the blender scene and adjust its size to match that of the texture.

	3 	Create multiple tile quads on top of it, using the texture as a guide to their shapes and sizes
	 	(it can help to set their drawtypes to 'wire' in the Blender buttons/object/draw panel).

	4 	Set the blender ortho 3d view to vertically downwards, and select all of the new tile meshes
		(but not the guide quad). You can set the tiles back to 'shaded' drawtype now.

	5 	With the footprint texture visible in the UV/Image editor window, click the 'GET' button on
	  	the XPlaneAGTools AGS panel to load it. Match the 'size' and 'loc x/y' control settings to the
	  	size and lower-left corner location of the guide quad. Click on 'RESET' to disable all functions,
	  	and then click on 'project texture' to enable it. Click 'APPLY' to action the settings.

	This function can also be used at a later stage to 'touch-up' a tile's UV map if it has proven
	necessary to tweak its shape or size.

Typically, the majority, if not all, of the AGS functions will be performed in one operation, by filling
in the desired values and enabling all the blue function buttons, selecting the tile meshes, positioning
the 3-d cursor a few meters off to one side and clicking 'APPLY'. It's like magic!

*
*
* AGB
*

Under development - nothing here yet!

*
*
* OBJ
*

# Docs coming soon #

*
*
* Mesh
*

The 'Mesh' page focuses on managing the face modes and face transparency modes of Blender mesh objects.
Traditionally, the face modes (e.g. 'tiled', 'twoside', 'collision' etc) of a mesh can only be viewed
on a face-by-face basis, by selecting each face in turn. This is often a very laborious and error-prone
operation, particularly when dealing with multiple, potentially complex meshes.

XPlaneAGTools provides the following functionality:

-	Single face-mode search.
	E.g. Find and select all 'hard' faces in all visible objects.

-	Multiple face-mode search.
	E.g. find and select all faces which are both 'tiled' AND 'hard' AND NOT 'two-sided'.
	E.g. find and select all faces which are either 'tiled' OR 'hard'.

-	Mixed alpha search.
	Find and select all objects in the scene containing both 'Alpha' (ATTR_blend in X-Plane) and 'Clip'
	(ATTR_no_blend) faces. This is generally something to be avoided when creating meshes for X-Plane,
	but also unfortunately one which is both easy to induce and previously hard to detect.

-	Single and multiple face-mode application.
	One or more face modes can be applied to selected faces in the selected objects, without affecting
	the state of the other face-modes.
	E.g. Set all selected faces to 'Collision off' (for AGB anti-slop tile) without affecting their
	individual 'two-side' (AGB slop tile) setting. The existing copy function in the Blender face-mode
	buttons panel copies the states of ALL facemodes at once, overwriting their previous values.

Use should be fairly self-explanatory. The top row of blue buttons is used to specify mode values.
The 'target' buttons are used to specify which face-modes to include when searching for or applying
settings. Some examples:

Key:	+ 	= ON
		-	= OFF
		. 	= don't care

1. Search for all two-sided faces in visible layers, regardless of the settings of their other modes:

          Tiles      Twoside    Collision   Shadow     Invisible    Opaque       Alpha        Clip
values:     .           +           .           .           .           .           .           .
targets:    -           +           -           -           -                       -

Click 'Search Visible Objs'

2. Search selected objects for alpha-blended, tiles faces whose other face-modes are all turned off:

          Tiles      Twoside    Collision   Shadow     Invisible    Opaque       Alpha        Clip
values:     +           -           -           -           -           -           +           -
targets:    +           +           +           +           +                       +
enable the 'match all' button
Click 'Search Selected Objs'

3.	Search visible layers for two-sided faces and alpha transparent faces (they can be either or both) regardless of their
	other settings,

          Tiles      Twoside    Collision   Shadow     Invisible    Opaque       Alpha       Clip
values:     .           +           .           .           .           .           +           .
targets:    -           +           -           -           -                       +
enable the 'match ANY' button
Click 'Search Visible Objs'

4.	To turn off 'collision' on all selected faces in selected objects, without messing with their other modes:

          Tiles      Twoside    Collision   Shadow     Invisible    Opaque       Alpha       Clip
values:     .           .           -           .           .           .           .           .
targets:    -           -           +           -           -                       -
Click 'APPLY TO SELECTED'

4.	To reset all selected faces in selected objects to default face and transparency modes:

          Tiles      Twoside    Collision   Shadow     Invisible    Opaque       Alpha       Clip
values:     -           -           +           -           -           -           .           .
targets:    +           +           +           +           +                       +
Click 'APPLY TO SELECTED'


*
*
* Annotation and Spelling
*

# Docs coming soon #

*
*
* Lights
*

# Docs coming soon #

"""


import bpy
import heapq
import Blender, math, copy, random
from Blender import *
from os.path import exists, join
from Blender.Draw  import *
from Blender.BGL import *
from Blender.Image import *
from Blender.Window import Theme
from Blender.Scene import Render
from math import *
from Blender.Geometry import*
from Blender.Mathutils import Matrix, RotationMatrix, TranslationMatrix, MatMultVec, Vector, Euler
from types import*

import datetime


CL_BLACK		= [0.00, 0.00, 0.00]
CL_WHITE		= [1.00, 1.00, 1.00]
CL_TEXT_DK		= [0.20, 0.20, 0.20]
CL_TEXT_MD		= [0.25, 0.25, 0.25]
CL_TEXT_LT		= [0.32, 0.32, 0.33]
CL_ACT_PINK		= [0.92, 0.85, 0.92]
CL_SEL_PINK		= [0.92, 0.65, 0.92]
CL_GREY95		= [0.95, 0.95, 0.95]
CL_GREY75		= [0.75, 0.75, 0.75]
CL_GREY70		= [0.70, 0.70, 0.70]
CL_GREY65		= [0.65, 0.65, 0.65]
CL_GREY60		= [0.60, 0.60, 0.60]
CL_GREY50		= [0.50, 0.50, 0.50]
CL_GREY56		= [0.56, 0.56, 0.56]
CL_GREY40		= [0.40, 0.40, 0.40]
CL_GREY30		= [0.30, 0.30, 0.30]
CL_MACRO		= [0.83, 0.83, 0.83]
CL_RED			= [1.00, 0.35, 0.22]
CL_ORANGE		= [1.00, 0.76, 0.24]
CL_CYAN			= [0.45, 0.89, 0.75]
CL_GUI_ORANGE	= [1.00, 0.95, 0.85]
CL_GUI_ORANGE2	= [0.85, 0.80, 0.75]
CL_VECT_ORANGE  = [1.00, 0.75, 0.32]
CL_GUI_BLUE		= [0.15, 0.40, 0.40]
CL_GUI_MD_GREEN	= [0.62, 0.64, 0.64]
CL_GUI_GREEN	= [0.36, 0.44, 0.45]
CL_GUI_LT_BLUE	= [0.66, 0.75, 0.76]
CL_ACT_BUT		= [0.81, 0.76, 0.71]
CL_ACT_DIM		= [0.65, 0.62, 0.59]
CL_ANN_GREEN	= [0.00, 1.00, 0.21]
CL_ANN_GREEN2	= [0.00, 0.92, 0.10]
CL_ANN_RED		= [0.84, 0.10, 0.10]
CL_ANN_AMBER	= [1.00, 0.88, 0.36]
CL_ROAD			= [0.08, 0.00, 1.00]
CL_NOROAD		= [1.00, 0.95, 0.00]
CL_ROAD_BKGND	= [0.65, 0.65, 0.65]
CL_AXES			= CL_GREY95
CL_MINOR_TICKS	= [0.50, 0.50, 0.50]
CL_MAJOR_TICKS	= [0.62, 0.62, 0.62]
CL_AXIS_NUMS	= CL_GREY95
CL_AXIS_LABELS	= CL_GREY95
#CL4_HISTO_GREEN	= [0.48, 0.86, 0.38, 0.76]
CL4_HISTO_GREEN	= [0.58, 0.74, 0.40, 0.85]
#CL4_HISTO_GREEN	= [0.53, 0.90, 0.30, 0.75]
#CL4_HISTO_ACT	= [0.57, 0.92, 0.48, 0.87]
CL4_HISTO_ACT	= [0.48, 0.86, 0.38, 0.95]
CL4_HISTO_TEMP	= [1.00, 0.93, 0.40, 0.82]
CL4_HISTO_RED	= [1.00, 0.17, 0.30, 0.80]

CL4_EDGE_NOROAD = [1.00, 0.95, 0.00, 1.00]
CL4_EDGE_NOROAD2= [1.00, 0.95, 0.00, 0.37]
CL4_EDGE_ROAD	= [0.08, 0.00, 1.00, 1.00]
CL4_EDGE_ROAD2	= [0.08, 0.00, 1.00, 0.30]
CL4_MATCH_GRP	= [0.00, 0.95, 0.00, 1.00]
CL4_MATCH_GRP2	= [0.00, 0.80, 0.00, 0.40]
CL4_UNTAGGED	= [1.00, 0.10, 0.20, 0.80]
CL4_SLOP		= [0.34, 1.00, 0.42, 0.45]
CL4_ANTISLOP	= [1.00, 0.85, 0.42, 0.51]
CL4_IGNORE		= [0.37, 1.00, 0.97, 0.67]

LIGHTS_TXT_DIR	= ''
LIGHTS_TXT		= None
LIGHTS_TEXTURE_NAME = '1000_lights_close.dds'
LIGHTS_TEXTURE_DIR  = ''
LIGHT_EDITOR_CONFIG	= None
g_ignore_g_lighteditor = 0
g_lighteditor 	= None
g_AGB_editor  	= None

TIM = [None, 'Start']



def elapsed(labelString):
	global TIM
	if TIM[0] == None:
		TIM[0] = Blender.sys.time()
	print "Time from", TIM[1], "to", labelString, "was", Blender.sys.time() - TIM[0], '\n'
	TIM[1] = labelString
	TIM[0] = Blender.sys.time()

def isnum(x):
	try:
		y = float(x)
		return True
	except:
		return False

def negate(sequence):
	return [-i for i in sequence] if isinstance(sequence, list) else tuple([-i for i in sequence])

def flipView(xyz):
	x, y, z = xyz
	return [x, -y, -z] if isinstance(xyz, list) else (x, -y, -z)

def clamp(x, xmin=0.0, xmax=1.0,):
	return max(min(x, xmax), xmin)

def superClamp(x, limit1=0.0, limit2=1.0, order=3):
	"""
	This is a swiss-army knife clamp function.
	x can be any of int, float, list or tuple types.
	By default all values are clamped between 0.0 and 1.0, but custom limits may be set with args 2 and 3.
	The function returns an object of the same  type as that which is passed to it.
	If a list or tuple is passed in:
		- if at least one member is a float, ALL members of the result will be floats.
		- By default, the result will be extended or truncated if necessary so as to be of length (order) 3.
		- If the result is shorter than 'order', it is padded with zeroes.
		- If the result is longer than 'order', only the first 'order' terms are returned
		- The value of 'order' may be customized with arg 4: if arg 4 is set to 0, no truncating/extending will be performed.
	"""
	if isinstance(x, float) or isinstance(x, int):
		return min(max(x, limit1), limit2)
	else:
		result = []
		try:
			if any([isinstance(i, float) for i in x ]):
				for i in x:
					result.append(float(min(max(i, limit1), limit2)))
				if order and (len(result) < order):
					for z in range(order - len(result)):
						result.append(0.0)
			else:
				for i in x:
					result.append(int(min(max(i, limit1), limit2)))
				if order and (len(result) < order):
					for z in range(order - len(result)):
						result.append(0)
			if order and (len(result) > order):
				result = result[:order]
		except:
			return None
		if isinstance(x, list):
			return result
		elif isinstance(x, tuple):
			return tuple(result)

def fallsIn(limit1, limit2, x):
	return limit1 <= x <= limit2 if limit2 > limit1 else limit1 >= x >= limit2

def fromEulers(pitch, heading):
	# Blender axes
	z = sin(radians(pitch))
	x = sin(radians(heading)) * cos(radians(pitch))
	y = cos(radians(heading)) * cos(radians(pitch))
	return [round(x,8), round(y, 8), round(z, 8)]

def eulersToMatrix(p,r,h):
	e = Euler(p, r, -h)
	return e.toMatrix()

def mix(seq1, seq2, ratio):
	m = []
	ratio = clamp(ratio)
	for a, b in zip(seq1, seq2):
		m.append(a*(1.0-ratio) + b*ratio)
	return tuple(m) if isinstance(seq2, tuple) else m

def mlt(j, k):
	return (pow(j[0] * k[0],1), pow(j[1] * k[1], 1), pow(j[2] * k[2], 1), j[3] * k[3])

def toEulers(xyz):
	""" Note - this funcion always returns a value of zero for roll, since we don't need yaw for lamps!
	"""
	temp = Vector(xyz)
	temp.normalize()
	x, y, z = temp
	return [degrees(asin(z)), 0.0,	degrees(atan2(x,y))]

def cleanString(inValue):

	"""
	This strips whitespace and commas, and adds two-space separators between tokens
	"""
	return "  ".join((" ".join(str(inValue).split(','))).split())

def sigfigs(num, sig_figs=4):
    """Round to specified number of sigfigs.
    """
    if num != 0:
        return round(num, -int(floor(log10(abs(num))) - (sig_figs - 1)))
    else:
        return 0

def fractionize(x, denom=1, up=0):
	if x == None or denom == 0:
		return x
	if not up:
		return floor(x * denom)/denom
	else:
		return ceil(x * denom)/denom

def snapToAxes(xyz, snapAngle=0.0):
	if snapAngle:
		t = abs(math.sin(math.radians(snapAngle)))
		x, y, z = xyz
		if abs(x) < t:
			x = 0.0
			m = math.hypot(y, z)
			y /= m
			z /= m
		elif abs(y) < t:
			y = 0.0
			m = math.hypot(x, z)
			x /= m
			z /= m
		elif abs(z) < t:
			z = 0.0
			m = math.hypot(x, y)
			x /= m
			y /= m
		return (x, y, z) if isinstance(xyz, tuple) else [x, y, z]
	return xyz

def HueToRGB(f1,  f2,  hue):
	if (hue < 0.0):
		hue += 1.0
	elif (hue > 1.0):
		hue -= 1.0

	if (6.0 * hue) < 1.0:
		res = f1 + (f2 - f1) * 6.0 * hue
	elif (2.0 * hue) < 1.0:
		res = f2
	elif (3.0 * hue) < 2.0:
		res = f1 + (f2 - f1) * ((2.0 / 3.0) - hue) * 6.0
	else:
		res = f1
	return res

def HSLToRGB(hsl):

	if hsl[1] == 0.0:
		rgb = [hsl[2], hsl[2], hsl[2]] # Luminance
	else:
		if hsl[2] < 0.5:
			f2 = hsl[2] * (1.0 + hsl[1])
		else:
			f2 = (hsl[2] + hsl[1]) - (hsl[1] * hsl[2])
		f1 = 2.0 * hsl[2] - f2
		rgb = [0.0, 0.0, 0.0]
		rgb[0] = HueToRGB(f1, f2, hsl[0] + (1.0/3.0))
		rgb[1] = HueToRGB(f1, f2, hsl[0])
		rgb[2] = HueToRGB(f1, f2, hsl[0] - (1.0/3.0))
	return rgb

def HSBToRGB(hsb):

	h, s, b = hsb
	if  s == 0:
		return [b, b, b]

	else:
		var_h = h * 6.0
		if ( var_h == 6.0 ):
			var_h = 0.0      #H must be < 1
		var_i = int( var_h )             # Or ... var_i = floor( var_h )
		var_1 = b * ( 1.0 - s )
		var_2 = b * ( 1.0 - s * ( var_h - var_i ) )
		var_3 = b * ( 1.0 - s * ( 1.0 - ( var_h - var_i ) ) )

		if var_i == 0 :
			var_r = b
			var_g = var_3
			var_b = var_1
		elif var_i == 1:
			var_r = var_2
			var_g = b
			var_b = var_1
		elif var_i == 2:
			var_r = var_1
			var_g = b
			var_b = var_3
		elif var_i == 3:
			var_r = var_1
			var_g = var_2
			var_b = b
		elif var_i == 4:
			var_r = var_3
			var_g = var_1
			var_b = b
		else:
			var_r = b
			var_g = var_1
			var_b = var_2
	return [var_r, var_g, var_b]

def RGBToHSL(RGB):

	hsl = [0.0, 0.0, 0.0]

	fmin = min(min(RGB[0], RGB[1]), RGB[2]) 			# Min. value of RGB
	fmax = max(max(RGB[0], RGB[1]), RGB[2])   			 # Max. value of RGB
	delta = fmax - fmin				           			 # Delta RGB value

	hsl[2] = (fmax + fmin) / 2.0 						#  Luminance

	if delta == 0.0:									# This is a gray, no chroma..
		hsl[0] = 0.0									#  Hue
		hsl[1] = 0.0									#  Saturation
	else:                                   		# Chromatic data...
		if hsl[2] < 0.5:
			hsl[1] = delta / (fmax + fmin)			#  Saturation
		else:
			hsl[1] = delta / (2.0 - fmax - fmin) 	#  Saturation

		deltaR = (((fmax - RGB[0]) / 6.0) + (delta / 2.0)) / delta
		deltaG = (((fmax - RGB[1]) / 6.0) + (delta / 2.0)) / delta
		deltaB = (((fmax - RGB[2]) / 6.0) + (delta / 2.0)) / delta

		if RGB[0] == fmax:
			hsl[0] = deltaB - deltaG				#  Hue
		elif RGB[1] == fmax:
			hsl[0] = (1.0 / 3.0) + deltaR - deltaB #  Hue
		elif RGB[2] == fmax:
			hsl[0] = (2.0 / 3.0) + deltaG - deltaR #  Hue

		if hsl[0] < 0.0:
			hsl[0] += 1.0 			#  Hue
		elif hsl[0] > 1.0:
			hsl[0] -= 1.0			#  Hue
	return hsl

def RGBToHSB(RGB):

	var_R, var_G, var_B = RGB

	var_Min = min( var_R, var_G, var_B )
	var_Max = max( var_R, var_G, var_B )
	del_Max = var_Max - var_Min

	V = var_Max

	if ( del_Max == 0 ):
		H = 0.0
		S = 0.0
	else:
		S = del_Max / var_Max
		del_R = ( ( ( var_Max - var_R ) / 6.0 ) + ( del_Max / 2.0 ) ) / del_Max
		del_G = ( ( ( var_Max - var_G ) / 6.0 ) + ( del_Max / 2.0 ) ) / del_Max
		del_B = ( ( ( var_Max - var_B ) / 6.0 ) + ( del_Max / 2.0 ) ) / del_Max

		if  var_R == var_Max:
			H = del_B - del_G
		elif var_G == var_Max:
			H = ( 1.0 / 3.0 ) + del_R - del_B
		elif var_B == var_Max:
			H = ( 2.0 / 3.0 ) + del_G - del_R

		if H < 0:
			H += 1.0
		if H > 1:
			H -= 1.0
	return [H, S, V]

def sRGBToLinear(inVals):
	if isinstance(inVals, float):
		return (inVals/12.92) if inVals <= 0.04045 else (math.pow((inVals + 0.055)/1.055, 2.4))
	rgb = []
	for x in inVals:
		rgb.append((x/12.92) if x <= 0.04045 else (math.pow((x + 0.055)/1.055, 2.4)))
	return rgb

def linearToSRGB(inVals):
	if isinstance(inVals, float) or isinstance(inVals, int):
		return (inVals * 12.92) if inVals <= 0.003131 else math.pow(inVals, 0.41666666) * 1.055 - 0.055
	sRGB = []
	for x in inVals:
		sRGB.append( (x * 12.92) if x <= 0.003131 else (math.pow(x, 0.41666666) * 1.055 - 0.055))
	return sRGB

def fltToPS(rgb):
	ints = []
	for c in rgb:
		ints.append(int(round(c * 255.0)))
	return ints

def PSToFlt(rgb):
	flts = []
	for c in rgb:
		flts.append(float(c)/255.0)
	return flts

def fltToPSHSB(hsb):
	h, s, b = hsb
	return [int(h*359.0), int(s * 100.0), int(b * 100.0)]

def stringToFloats(inStr):
	if not isinstance(inStr, str):
		return [0.5, 0.5, 0.5]
	toks = inStr.split()
	if len(toks) != 3:
		return [0.5, 0.5, 0.5]
	flts = []
	for t in toks:
		try:
			float(t)
			flts.append(float(t))
		except:
			PupMenu("Expected three, whitespace-separated floating-point numbers!")
			return [0.500,  0.500,  0.500]

	return flts

def normalizeVector(xyz):
	x,y,z = xyz
	m = math.sqrt(x*x + y*y +z*z)
	if m:
		x /= m
		y /= m
		z /= m
	return [x, y, z] if isinstance(xyz, list) else (x, y, z)

def stringToInts(inStr):
	if not isinstance(inStr, str):
		return [128, 128, 128]
	toks = inStr.split()
	if len(toks) != 3:
		return [128, 128, 128]
	ints = []
	for t in toks:
		if t.strip('0123456789+-'):
			return [128, 128, 128]
		else:
			ints.append(int(t))
	return ints

def numsToString(numList):
	s = ""
	for n in numList:
		if n == 0:
			s += "0  "
		elif isinstance(n, int):
			s += "%d  " % n
# 		else:
# 			s += str(sigfigs(n)) + "  "
		elif abs(n) >= 0.0001:
			s += "%.4g  " % n
		else:
			tok = ("%.3f" % round(n,3)).rstrip('0')
			if len(tok) > 1:
				if tok[-1] == '.':
					tok += '0  '
				else:
					tok += '  '
			s += tok
	return s.strip()

def drawOneLine(x1, y1, x2, y2, rgb):
    BGL.glColor3f(*rgb)
    BGL.glBegin(BGL.GL_LINES)
    BGL.glVertex2i(x1,y1)
    BGL.glVertex2i(x2,y2)
    BGL.glEnd()

def drawRectangle(x, y, w, h, colour):
	if isinstance(colour, float):
		BGL.glColor3f(colour, colour, colour)
	elif len(colour) == 3:
		BGL.glColor3f(*colour)
	elif len(colour) == 4:
		BGL.glColor4f(*colour)
	BGL.glRecti(x, y, x + w, y + h)

def drawButtonFrame(x, y, w, h, colour=0.663):
	""" This draws a rectangular, filled frame, the same colour as a blender number/string button.
		The frame is slightly lighter than a real button, to indicate that it is passive.
	"""

	BGL.glColor3f(0.4, 0.4, 0.4)
	BGL.glRecti(x, y, x + w + 1, y + h + 1)
	if isinstance(colour, float):
		BGL.glColor3f(colour, colour, colour)
	else:
		BGL.glColor3f(*colour)
	BGL.glRecti(x + 1, y + 1, x + w , y + h )

def splitStringByWidth(inStr, width, fontSize):
		"""
			The string value associated with textKey is split into a list of strings.
			The length of each string is defined as the longest that can be fitted into the
			specified width, given the specified font size.
		"""
		lines = []
		line  = ""
		for t in inStr.split():
			if GetStringWidth(" ".join([line, t]), fontSize) <= width:
				line += ' ' + t
			else:
				lines.append(line)
				line = t
		lines.append(line)
		return lines

def drawString(inStr, x, y, colour, mode='center', size='normal'):
	""" 'mode' can be one of 'ljust', 'rjust' and 'center'.
		In the case of 'ljust', x, y refer to the lower LH corner.
		In the case of 'rjust', x, y refer to the lower RH corner.
		In the case of 'center', x, y refer to the lower centre of the string.
		Size may be one of 'large', 'normal', 'normalfix', 'small' or 'tiny' """
	if mode == 'center':
		dx = GetStringWidth(inStr, size) / 2
	elif mode == 'rjust':
		dx = GetStringWidth(inStr, size)
	else:
		dx = 0
	if isinstance(colour, float):
		BGL.glColor3f(colour, colour, colour)
	else:
		BGL.glColor3f(*colour)
	BGL.glRasterPos2i(x - dx, y)
	Text(inStr, size)

def drawParagraph(inStr, x, y, w, lineSpacing, colour, mode, size):
	"""
		This draws a paragraph of text at the given location, paragraph width and line spacing.
		inStr is a single string.
		'mode' must be one of 'center', 'ljust' and 'rjust'.
	"""
	lines = splitStringByWidth(inStr, w, size)
	for s in lines:
		drawString(s, x, y, colour, mode, size)
		y -= lineSpacing

def drawAnnunciator(inValue, x, y, w, h, colour='green'):
	"""
		Draws a small rectangular LED of the specified colour, which relects the boolean state of inValue
	"""

	cols = {'green':[[0.10, 0.80, 0.10], [0.10, 0.20, 0.10]], 'red':[[1.00, 0.30, 0.20], [0.30, 0.00, 0.00]], 'amber':[[1.00, 0.69, 0.25], [0.36, 0.28, 0.17]]}

	fw = 2 	# frame width of the 'bezel' in pixels
	semiw = w/2
	if colour not in cols:
		print "Unrecognised colour passed to drawAnnunciator()\n"
		return
	if inValue:
		BGL.glColor3f(*cols[colour][1])
		BGL.glRecti(x - semiw, 		y, 		x + semiw + 1,		y + h + 1)
		BGL.glColor3f(*cols[colour][0])
		BGL.glRecti(x - semiw + fw, y + fw, x + semiw + 1 - fw, y + h + 1 - fw)
	else:
		BGL.glColor3f(0.25, 0.25, 0.25)
		BGL.glRecti(x - semiw, 		y, 		x + semiw + 1, 		y + h + 1)
		BGL.glColor3f(*cols[colour][1])
		BGL.glRecti(x - semiw + fw, y + fw, x + semiw + 1 - fw, y + h + 1 - fw)

def drawAnnunciatorRedGreen(inValue, x, y, w, h):
	"""
		Draws a small rectangular LED, shows red when inValue = 0, and green when inValue = 1
	"""

	cols = {'green':[[0.40, 0.70, 0.40], [0.40, 0.40, 0.40]], 'red':[[0.90, 0.40, 0.40], [0.40, 0.40, 0.40]]}

	fw = 1 	# frame width of the 'bezel' in pixels
	semiw = w/2
	if inValue:
		BGL.glColor3f(*cols['green'][1])
		BGL.glRecti(x - semiw, 		y, 		x + semiw + 1,		y + h + 1)
		BGL.glColor3f(*cols['green'][0])
		BGL.glRecti(x - semiw + fw, y + fw, x + semiw + 1 - fw, y + h + 1 - fw)
	else:
		BGL.glColor3f(*cols['red'][1])
		BGL.glRecti(x - semiw, 		y, 		x + semiw + 1, 		y + h + 1)
		BGL.glColor3f(*cols['red'][0])
		BGL.glRecti(x - semiw + fw, y + fw, x + semiw + 1 - fw, y + h + 1 - fw)

def str2Ints(spelling):
	sp = []
	for t in spelling.split():
		try:
			sp.append(int(t))
		except (TypeError, ValueError):
			PupMenu("Please enter one or more integers, separated by spaces.%t")
			return []
	return sp

def str2Nums(inStr):
	n = []
	for t in inStr.split():
		if not isnum(t):
			PupMenu("ERROR: %s is not a number!%t", t)
			n.append(0)
		else:
			n.append(float(t) if "." in t else int(t))
	return n

def ints2string(intList):
		s = ""
		for i in intList:
			s += str(i) + " "
		return s.rstrip()

def intlists2string(intLists):
	st = ""
	for lst in intLists:
		st += ints2string(lst) + ", "
	return st[:-2]

def str2intlists(s):
	if s == "":
		return []
	run_list = []
	for r in s.split(","):
		run = []
		for t in r.split():
			try:
				run.append(int(t))
			except (TypeError, ValueError):
				PupMenu("Please enter comma-separated lists of space-separated integers.%t")
				return []
		run_list.append(run)
	return run_list

def incrementName(nameString, existingNames=None, blenderStyle=False):
	"""
		Attempts to come up with something sensible when asked to create a unique name when passed a previous name and optionally a list of existing names with which to not conflict.
		If existingNames != None, and doesn't already include nameString, nameString is returned unchanged.
	"""
	if nameString is None:
		return None
	if existingNames and nameString not in existingNames:
		return nameString
	if not blenderStyle:
		# In this mode, we add or increment a pretty, whitespace separated numeric suffix.
		a, b, c = nameString.rstrip().rpartition(" ")
		n = a + b + str(int(c) + 1) if c.isdigit() else (nameString.rstrip() + ' 2').lstrip()
	else:
		# do the Blender '.nnn' thing.
		if '.' in nameString:
			a, b, c = nameString.rpartition('.')
			if c.isdigit():
				lngth = len(c)
				n = a + b + str(int(c) + 1).rjust(lngth,'0')
			else:
				n = nameString[:17] + '.001'
		else:
			n = nameString[:17] + '.001'
	if existingNames is None:
		return n
	while n in existingNames:
		n = incrementName(n, existingNames, blenderStyle)
	return n

# # # _______________________________________________________________________________________________________________________________________________________________________________________
#
#     LIGHT EDITOR
# # # _______________________________________________________________________________________________________________________________________________________________________________________
class PatchConfig:
	def __init__(self, id,	namesList, chnlList, modifierFunction=None):
		self.id				= id
		self.names			= namesList[:]
		self.chnls  		= chnlList[:]
		self.info			= ""
		def doNothing(self, channel, inValue):
			return inValue
		self.process_f		= doNothing if not modifierFunction else modifierFunction
PatchConfig(0,[],[])
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class SpillLightData:
	def __init__(self):
		self.r			= 1.0
		self.g			= 1.0
		self.b			= 1.0
		self.bright		= 1.0
		self.size		= 20.0
		self.x			= 0.0
		self.y			= 0.0
		self.z			= 0.0
		self.width		= 1.0
		self.freq		= 1.0
		self.phase		= 0.0
		self.amp		= 1.0
		self.day		= 0
		self.dataref	= ""

	def __getitem__(self, var):
		if not hasattr(self, var):
			return None
		return getattr(self, var)

	def __setitem__(self, var, val):
		if not hasattr(self, var):
			return None
		setattr(self, var, val)

	def copyFrom(self, other):
		self.r			= other.r
		self.g			= other.g
		self.b			= other.b
		self.bright		= other.bright
		self.size		= other.size
		self.x			= other.x
		self.y			= other.y
		self.z			= other.z
		self.width		= other.width
		self.freq		= other.freq
		self.phase		= other.phase
		self.amp		= other.amp
		self.day		= other.day
		self.dataref	= other.dataref
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class SpillLightParams:
	def __init__(self):
		self.r			= None
		self.g			= None
		self.b			= None
		self.bright		= None
		self.size		= None
		self.x			= None
		self.y			= None
		self.z			= None
		self.width		= None
		self.freq		= None
		self.phase		= None

	def __getitem__(self, var):
		if not hasattr(self, var):
			return None
		return getattr(self, var)

	def __setitem__(self, var, val):
		if not hasattr(self, var):
			return None
		setattr(self, var, val)

	def copyFrom(self, other):
		self.r			= other.r
		self.g			= other.g
		self.b			= other.b
		self.bright		= other.bright
		self.size		= other.size
		self.x			= other.x
		self.y			= other.y
		self.z			= other.z
		self.width		= other.width
		self.freq		= other.freq
		self.phase		= other.phase

	def reset(self):
		for k, v in vars(self).iteritems():
			vars(self)[k] = None
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class SpillLight:

	types 				 = ["SPILL_HW_DIR",
							"SPILL_HW_FLA",
							"SPILL_SW"]

	type_tok_count 		 = {'SPILL_HW_DIR': 12, 'SPILL_HW_FLA': 11, 'SPILL_SW': 12}

	names     			 = {'SPILL_HW_DIR' 	: [],
							'SPILL_HW_FLA'	: [],
							'SPILL_SW'		: []}

	name_groups			 = {	'SPILL_HW_DIR' 	: {
												'AIRPLANE'		: [],
												'SCENERY'		: [],
												'RGB'			: [],
												'BRIGHTNESS'	: [],
												'SIZE'			: [],
												'FLASH'			: [],
												'GEOMETRY'		: [],
												'OTHER'			: []
												},
							'SPILL_HW_FLA' 	: {
												'AIRPLANE'		: [],
												'SCENERY'		: [],
												'RGB'			: [],
												'BRIGHTNESS'	: [],
												'SIZE'			: [],
												'FLASH'			: [],
												'GEOMETRY'		: [],
												'OTHER'			: []
												},
							'SPILL_SW' 		: {
												'AIRPLANE'		: [],
												'SCENERY'		: [],
												'RGB'			: [],
												'BRIGHTNESS'	: [],
												'SIZE'			: [],
												'GEOMETRY'		: [],
												'FLASH'			: [],
												'OTHER'			: []
												}
												}
	parameter_groups = {
						'RGB'		: [2, 3, 4],
						'BRIGHTNESS': [5],
						'SIZE'		: [6],
						'GEOMETRY'	: [7, 8, 9, 10],
						'FLASH'		: [7, 8]
						}

	filter_cats = ['AIRPLANE',	'SCENERY',	'RGB',	'BRIGHTNESS', 'SIZE', 'GEOMETRY', 'FLASH', 'OTHER']

	# The following are SPILL_SW lights whose RGBA channels have been tricked into doing something different.

	names_no_rgba	= ['airplane_beacon_rotate_sp',  'airplane_beacon_strobe_sp', 'airplane_beacon_sp', 'appch_strobe_o', 'appch_strobe_u',
							'inset_appch_rabbit_o_sp', 'inset_appch_rabbit_u_sp', 'inset_appch_strobe_o_sp', 'inset_appch_strobe_u_sp',
							'wigwag_y1', 'wigwag_y2', 'hold_short_y1', 'hold_short_y2']
	lights_data = []
	names_set			 = 0

	def __init__(self):

		global LIGHTS_TXT
		if not SpillLight.names_set:
			hw_dir = []
			hw_fla = []
			sw	   = []

			SpillLight.lights_data = LIGHTS_TXT.readlines()
			for l in SpillLight.lights_data:
				if not l: continue

				tks = l.split()
				if not tks: continue
				if len(tks) == 12 and tks[0] == 'SPILL_HW_DIR':
					hw_dir.append(tks[1])

				elif len(tks) == 11 and tks[0] == 'SPILL_HW_FLA':
					hw_fla.append(tks[1])
				elif len(tks) == 12 and tks[0] == 'SPILL_SW':
					sw.append(tks[1])

			hw_dir_s = set(hw_dir)
			hw_dir	 = list(hw_dir_s)
			SpillLight.names['SPILL_HW_DIR'] = sorted(hw_dir)

			hw_fla_s = set(hw_fla)
			hw_fla = list(hw_fla_s)
			SpillLight.names['SPILL_HW_FLA'] = sorted(hw_fla)

			sw_s = set(sw)
			sw = list(sw_s)
			SpillLight.names['SPILL_SW']     = sorted(sw)

		# 	Now build lists of name indices, grouped by available parameters and aircraft/scenery use:

			for l in SpillLight.lights_data:
				if not l: continue
				tks = l.split()
				if not tks: continue

				for t in SpillLight.types:
					if len(tks) == SpillLight.type_tok_count[t] and tks[0] == t:
						for cat, value in SpillLight.parameter_groups.iteritems():
							if SpillLight.name_groups[t].has_key(cat):
								for param_index in SpillLight.parameter_groups[cat]:
									if not isnum(tks[param_index]):
										SpillLight.name_groups[t][cat].append(SpillLight.names[t].index(tks[1]))
						if tks[1].startswith('airplane'):
							SpillLight.name_groups[t]['AIRPLANE'].append(SpillLight.names[t].index(tks[1]))
						else:
							SpillLight.name_groups[t]['SCENERY'].append(SpillLight.names[t].index(tks[1]))


		#	These lists must have dupes removed and then be sorted so as to allow the up/down buttons to stay in sync with the menu items:

			for ng in SpillLight.name_groups.itervalues():
				for l in ng.itervalues():
					l_set = set(l)
					l[:] = []
					l.extend(list(l_set))
					l.sort()

		#	A few special cases: the 'flash' and 'geometry' buckets overlap because SPILL_HW_DIR has no flash params and SPILL_HW_FLA has no geometry params
			SpillLight.name_groups['SPILL_HW_DIR']['FLASH'][:] = []
			SpillLight.name_groups['SPILL_HW_FLA']['GEOMETRY'][:] = []


			SpillLight.names_set = 1


		self.type			= "SPILL_HW_DIR"
		self.name_enum		= 0
		self.name 			= SpillLight.names[self.type][self.name_enum]
		self.data_index 	= None
		self.sub_def_index	= None
		self.has_params 	= False
		self.def_string 	= ""
		self.arg_list   	= []
		self.data			= SpillLightData()
		self.params			= SpillLightParams()		# holds the parameter names from lights.txt

	def copy(self):
		"""
		Make a new SpillLight instance and copy our data to it.
		"""
		spl 				= SpillLight()
		spl.data 			= copy.deepcopy(self.data)
		spl.params 			= copy.deepcopy(self.params)
		spl.type			= self.type
		spl.name_enum		= self.name_enum
		spl.name 			= self.name
		spl.sub_def_index	= self.sub_def_index
		spl.data_index		= self.data_index
		spl.has_params 		= self.has_params
		spl.def_string 		= self.def_string
		spl.arg_list   		= self.arg_list

		return spl

	def copyFrom(self, other):

		self.type			= other.type
		self.name_enum		= other.name_enum
		self.name 			= other.name
		self.sub_def_index	= other.sub_def_index
		self.data_index		= other.data_index
		self.has_params 	= other.has_params
		self.def_string 	= other.def_string
		self.arg_list[:]  	= other.arg_list
		self.data.copyFrom(other.data)
		self.params.copyFrom(other.params)

	def hasName(self, inName):
		'''
		Return true if a light of name 'inName' exists.
		'''
		for tp in self.types:
			if inName in self.names[tp]:
				return True
		return False

	def refreshParam(self, paramName, paramValue):
		for k in vars(self.params).keys():
			if self.params[k] == paramName:
				self.data[k] = paramValue

	def refreshData(self, var, value):
		self.data[var] = value
		# Return true if we've modified a _non-parameterized_ variable.
 		# In other words, we've modified the actual light DEF, not just it's runtime args, so we'll need to export it.
		return True if self.params[var]  == None else False

	def checkForParams(self):
		self.has_params = any(vars(self.params).values())
		return self.has_params

	def makeNewDefString(self):
		'''
		Create a new def string based on latest edited values and parameter names.
		'''
		allvars = { 'SPILL_HW_DIR'	: ['r', 'g', 'b', 'bright',	'size', 'x', 'y', 'z',	'width', 'day'],
					'SPILL_HW_FLA' 	: ['r', 'g', 'b', 'bright',	'size', 'freq', 'phase', 'amp', 'day'],
					'SPILL_SW'		: ['r', 'g', 'b', 'bright',	'size', 'x', 'y', 'z',	'width', 'dataref']
					}
		new_def_string	= ""
		new_def_string += self.type
		new_def_string += '\t\t' if len(self.type) > 7 else '\t\t\t'
		new_def_string += self.name
		new_def_string += '\t' * ((39 - len(self.name))/8 + 1)
		for k in allvars[self.type]:
			if k in vars(self.params):
				if self.params[k]:
					v = self.params[k]
				else:
					v = str(self.data[k])
			else:
				v = str(self.data[k])

			if k != 'dataref':
				if k == 'freq':
					tok = "\t\t\t\t%s\t" % v[:7]
				elif k == 'day' and self.type == 'SPILL_HW_DIR':
					tok = "\t\t\t%s" % v[:7]
				else:
					tok = "%s\t" % v[:7]
				if k == 'size':
					tok += "\t"
			else:
				tok = v
			new_def_string += tok
		return new_def_string

	def	checkLength(self, length, type, nominalLength):
		if length > nominalLength:
			PupMenu("ERROR: Too many parameters for %s!" % type)
			return False
		elif length < nominalLength:
			PupMenu("ERROR: Too few parameters for %s!" % type)
			return False
		else:
			return True

	def getParamDef(self):
		tks = []
		self.arg_list[:] = []
		for line in SpillLight.lights_data:
			if line.startswith('LIGHT_PARAM_DEF'):
				tks = line.split()
				if len(tks) > 3 and tks[1] == self.name:
					self.arg_list[:] = tks[3:]
					break
		if self.arg_list and int(tks[2]) == len(self.arg_list):		# this should be zero if we've got our def.
			return False
		return True

	def getValuesFromProp(self, prop):
		if prop.type != 'STRING' or prop.name.upper() != 'PARAMS':
			PupMenu("ERROR: Cannot read property")
			return 0
		values = prop.data.split()
		if len(values) > len(self.arg_list):
			PupMenu("ERROR: Too many values in 'params' property!")
			return 0

		else:
			cnt = 0
			for v, name in zip(values, self.arg_list):
				if not isnum(v):
					PupMenu("ERROR: parameters must be numbers!")
					return 0
				for k in vars(self.params).keys():
					if self.params[k] == name:
						self.data[k] = float(v)
						cnt +=1
						break
			return cnt

	def setDataFromFile(self, byIndex=None):

		self.arg_list	= []
		self.params.reset()
		self.data  = SpillLightData()
		self.has_params	= False
		try:
			tks = []
			if byIndex:
				self.data_index = byIndex
			 	self.def_string = SpillLight.lights_data[byIndex]
				tks 			= self.def_string.split()
				self.type		= tks[0]
				self.name		= tks[1]
				self.name_enum	= SpillLight.names[self.type].index(self.name)
			#	print 'tks: ', tks, '\n'
			else:
				for j, l in enumerate(SpillLight.lights_data):
					tks = l.split()
					if (len(tks) == SpillLight.type_tok_count[self.type]) and (tks[0] == self.type) and (tks[1] == self.name):
						self.data_index 	= j
					#	self.sub_def_index 	= 0
						#print "position: %d" % j
						self.def_string 	= l
					#	print 'tks: ', tks, '\n'
						break

			for i, t in enumerate(tks):
				if i == 2:
					if isnum(t):
						self.data.r			= float(t)
					else:
						self.data.r			= 0.5
						self.params.r		= t
						self.has_params		= True
				elif i == 3:
					if isnum(t):
						self.data.g			= float(t)
					else:
						self.data.g			= 0.5
						self.params.g		= t
						self.has_params		= True
				elif i == 4:
					if isnum(t):
						self.data.b			= float(t)
					else:
						self.data.b			= 0.5
						self.params.b		= t
						self.has_params		= True
				elif i == 5:
					if isnum(t):
						self.data.bright	= float(t)
					else:
						self.data.bright	= int(0) if t == 'INDEX' else 1.0
						self.params.bright	= t
						self.has_params		= True
				elif i == 6:
					if isnum(t):
						self.data.size		= float(t)
					else:
						self.data.size		= 20.0
						self.params.size	= t
						self.has_params		= True

				if tks[0] in ['SPILL_HW_DIR', 'SPILL_SW']:

					if i == 7:
						if isnum(t):
							self.data.x			= float(t)
						else:
							self.data.x			= 0.0
							self.params.x		= t
							self.has_params		= True
					elif i == 8:
						if isnum(t):
							self.data.y			= float(t)
						else:
							self.data.y			= 0.0
							self.params.y		= t
							self.has_params		= True
					elif i == 9:
						if isnum(t):
							self.data.z			= float(t)
						else:
							self.data.z			= -1.0
							self.params.z		= t
							self.has_params		= True
					elif i == 10:
						if isnum(t):
							self.data.width		= float(t)
						else:
							self.data.width		= 0.866
							self.params.width	= t
							self.has_params		= True
					elif i == 11:
						if tks[0] == 'SPILL_HW_DIR':
							self.data.day		= int(t)		# non parameterisable

						elif tks[0] == 'SPILL_SW':
							self.data.dataref	= t				# non parameterisable

				elif tks[0] == 'SPILL_HW_FLA':

					if i == 7:
						if isnum(t):
							self.data.freq		= float(t)
						else:
							self.data.freq		= 1.0
							self.params.freq	= t
							self.has_params		= True
					elif i == 8:
						if isnum(t):
							self.data.phase		= float(t)
						else:
							self.data.phase		= 0.0
							self.params.phase	= t
							self.has_params		= True
					elif i == 9:
						self.data.amp			= float(t)					# non parameterisable
					elif i == 10:
						self.data.day			= int(t)					# non parameterisable

			# If we've collected any parameter names, get the params definition:

			if self.has_params:
				if self.getParamDef():
					PupMenu("ERROR: Couldn't load LIGHT_PARAM_DEF")
		except:
			print "Couldn't find def string"
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class BillboardData:
	def __init__(self):
		self.r			= 1.0
		self.g			= 1.0
		self.b			= 1.0
		self.bright		= 1.0
		self.size		= 1.0
		self.cell_size	= 2
		self.cell_x		= 5
		self.cell_y		= 3
		self.x			= 0.0
		self.y			= 0.0
		self.z			= -0.5
		self.width		= 0.5
		self.freq		= 1.0
		self.phase		= 0.0
		self.amp		= 0.0
		self.day		= 0
		self.dataref	= ""

	def __getitem__(self, var):
		if not hasattr(self, var):
			return None
		return getattr(self, var)

	def __setitem__(self, var, val):
		if not hasattr(self, var):
			return None
		setattr(self, var, val)

	def copy(self):
		cpy 			= BillboardData()
		cpy.r			= self.r
		cpy.g			= self.g
		cpy.b			= self.b
		cpy.bright		= self.bright
		cpy.size		= self.size
		cpy.cell_size	= self.cell_size
		cpy.cell_x		= self.cell_x
		cpy.cell_y		= self.cell_y
		cpy.x			= self.x
		cpy.y			= self.y
		cpy.z			= self.z
		cpy.width		= self.width
		cpy.freq		= self.freq
		cpy.phase		= self.phase
		cpy.amp			= self.amp
		cpy.day			= self.day
		cpy.dataref		= self.dataref
		return cpy

	def copyFrom(self, other):
		self.r			= other.r
		self.g			= other.g
		self.b			= other.b
		self.bright		= other.bright
		self.size		= other.size
		self.cell_size	= other.cell_size
		self.cell_x		= other.cell_x
		self.cell_y		= other.cell_y
		self.x			= other.x
		self.y			= other.y
		self.z			= other.z
		self.width		= other.width
		self.freq		= other.freq
		self.phase		= other.phase
		self.amp		= other.amp
		self.day		= other.day
		self.dataref	= other.dataref
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class BillboardParams:
	def __init__(self):
		self.r			= None
		self.g			= None
		self.b			= None
		self.bright		= None
		self.size		= None
		self.x			= None
		self.y			= None
		self.z			= None
		self.width		= None
		self.freq		= None
		self.phase		= None

	def copy(self):
		cpy 			= BillboardParams()
		cpy.r			= self.r
		cpy.g			= self.g
		cpy.b			= self.b
		cpy.bright		= self.bright
		cpy.size		= self.size
		cpy.x			= self.x
		cpy.y			= self.y
		cpy.z			= self.z
		cpy.width		= self.width
		cpy.freq		= self.freq
		cpy.phase		= self.phase
		return cpy

	def copyFrom(self, other):
		self.r			= other.r
		self.g			= other.g
		self.b			= other.b
		self.bright		= other.bright
		self.size		= other.size
		self.x			= other.x
		self.y			= other.y
		self.z			= other.z
		self.width		= other.width
		self.freq		= other.freq
		self.phase		= other.phase

	def reset(self):
		for k, v in vars(self).iteritems():
			vars(self)[k] = None

	def __getitem__(self, var):
		if not hasattr(self, var):
			return None
		return getattr(self, var)

	def __setitem__(self, var, val):
		if not hasattr(self, var):
			return None
		setattr(self, var, val)
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class Billboard:

	types				 = ["BILLBOARD_HW",
							"BILLBOARD_SW"]

	type_tok_count 		 = {'BILLBOARD_HW': 18, 'BILLBOARD_SW': 15}

	names     			 = {'BILLBOARD_HW' 	: [],
							'BILLBOARD_SW'	: []}

	name_groups			 = {'BILLBOARD_HW' 	: {
												'AIRPLANE'		: [],
												'SCENERY'		: [],
												'RGB'			: [],
												'SIZE'			: [],
												'BRIGHTNESS'	: [],
												'FLASH'			: [],
												'GEOMETRY'		: [],
												'OTHER'			: []
												},
							'BILLBOARD_SW' 	: {
												'AIRPLANE'		: [],
												'SCENERY'		: [],
												'RGB'			: [],
												'BRIGHTNESS'	: [],
												'SIZE'			: [],
												'GEOMETRY'		: [],
												'FLASH'			: [],
												'OTHER'			: []
												}
												}

	parameter_groups = {
						'RGB'		: [2, 3, 4],
						'BRIGHTNESS': [5],
						'SIZE'		: [6],
						'GEOMETRY'	: [10, 11, 12, 13]
						}

	filter_cats 		= ['AIRPLANE',	'SCENERY',	'RGB',	'BRIGHTNESS', 'SIZE', 'GEOMETRY', 'FLASH', 'OTHER']

	# The following are BILLBOARD_HW lights whose RGBA channels have been tricked into doing something different.

	names_no_rgba		= ['airplane_landing', 'airplane_landing0', 'airplane_landing1', 'airplane_landing2', 'airplane_landing3', 'airplane_landing_size', 'airplane_landing_flash',
							'airplane_taxi', 'airplane_taxi_size', 'airplane_taxi_flash', 'airplane_generic_size', 'airplane_generic_flash',
							'airplane_beacon_rotate', 'airplane_beacon_strobe', 'airplane_beacon',  'airplane_beacon_size', 'airplane_strobe_size', 'airplane_strobe',
							'airplane_nav_tail_size', 'airplane_nav_left_size', 'airplane_nav_right_size', 'airplane_nav_tail', 'airplane_nav_left', 'airplane_nav_right',
							'VASI', 'VASI3', 'appch_rabbit_o', 'appch_rabbit_u', 'appch_strobe_o', 'appch_strobe_u',
							'inset_appch_rabbit_o', 'inset_appch_rabbit_u', 'inset_appch_strobe_o',	'inset_appch_strobe_u',
							'wigwag_y1', 'wigwag_y2', 'hold_short_y1', 'hold_short_y2',
							'carrier_waveoff', 'carrier_meatball1', 'carrier_meatball2', 'carrier_meatball3', 'carrier_meatball4', 'carrier_meatball5',
							'frigate_SGSI_lo', 'frigate_SGSI_on', 'frigate_SGSI_hi']
	patches				= [] #PatchConfig(0,	['airplane_landing_core', 'airplane_taxi_core', 'airplane_spot_core', 'airplane_generic_core'],






	lights_data 		= []
	names_set			= 0
	bb_tex				= None

	def __init__(self):

		if not Billboard.bb_tex:
			try:
				Billboard.bb_tex = Load(LIGHTS_TEXTURE_DIR)
				Billboard.bb_tex.name = 'XPTBBTex'
	#			Billboard.bb_tex = Load('/Applications/X-Plane 10/Resources/bitmaps/world/lites/1000_lights_close.dds')
			except:
				print "couldn't get texture"


		global LIGHTS_TXT

		if not Billboard.names_set:
			hw_bb  = []
			sw_bb  = []
			Billboard.lights_data = SpillLight.lights_data
			for l in Billboard.lights_data:
				if not l: continue

				tks = l.split()
				if not tks: continue
				if len(tks)		== 18 and tks[0] == 'BILLBOARD_HW':
					hw_bb.append(tks[1])
				elif len(tks) 	== 15 and tks[0] == 'BILLBOARD_SW':
					sw_bb.append(tks[1])

			hw_bb_s = set(hw_bb)
			hw_bb[:]	 = []
			hw_bb.extend(list(hw_bb_s))

			Billboard.names['BILLBOARD_HW'] = sorted(hw_bb)
			sw_bb_s = set(sw_bb)
			sw_bb[:] = []
			sw_bb.extend( list(sw_bb_s))
			Billboard.names['BILLBOARD_SW']  = sorted(sw_bb)

		# 	Now build lists of name indices, grouped by available parameters and aircraft/scenery use:

			for l in Billboard.lights_data:
				if not l: continue
				tks = l.split()
				if not tks: continue
				for t in Billboard.types:
					if len(tks) == Billboard.type_tok_count[t] and tks[0] == t:
						for cat, value in Billboard.parameter_groups.iteritems():
							if Billboard.name_groups[t].has_key(cat):
								for param_index in Billboard.parameter_groups[cat]:
									if not isnum(tks[param_index]):
										Billboard.name_groups[t][cat].append(Billboard.names[t].index(tks[1]))
						if tks[1].startswith('airplane'):
							Billboard.name_groups[t]['AIRPLANE'].append(Billboard.names[t].index(tks[1]))
						else:
							Billboard.name_groups[t]['SCENERY'].append(Billboard.names[t].index(tks[1]))

		#	These lists must have dupes removed and then be sorted so as to allow the up/down buttons to stay in sync with the menu items:

			for ng in Billboard.name_groups.itervalues():
				for l in ng.itervalues():
					l_set = set(l)
					l[:] = []
					l.extend(list(l_set))
					l.sort()

		#	A few special cases: =
#			Billboard.name_groups['BILLBOARD_HW']['FLASH'][:] = []

			Billboard.names_set = 1

		self.type			= "BILLBOARD_HW"
		self.name_enum		= 0
		self.name 			= Billboard.names[self.type][self.name_enum]
		self.sub_def_index	= 0						# If this light has multiple bbs, this is our position in the list. e.g. if we're the second of three bbs, our sub_def_index is '1'
		self.data_index		= 0						# index of our def_string in the big array.
		self.has_params 	= False
		self.def_string 	= ""
		self.arg_list   	= []
		self.data			= BillboardData()
		self.params			= BillboardParams()		# holds the parameter names from lights.txt

	def copy(self):

		bb 					= Billboard()
		bb.data 			= copy.deepcopy(self.data)
		bb.params 			= copy.deepcopy(self.params)
		bb.type				= self.type
		bb.name_enum		= self.name_enum
		bb.name 			= self.name
		bb.sub_def_index	= self.sub_def_index
		bb.data_index		= self.data_index
		bb.has_params 		= self.has_params
		bb.def_string 		= self.def_string
		bb.arg_list   		= self.arg_list
		return bb

	def hasName(self, inName):
		'''
		Return true if a light of name 'inName' exists.
		'''
		for tp in self.types:
			if inName in self.names[tp]:
				return True
		return False

	def copyFrom(self, other):

		self.type			= other.type
		self.name_enum		= other.name_enum
		self.name 			= other.name
		self.sub_def_index	= other.sub_def_index
		self.data_index		= other.data_index
		self.has_params 	= other.has_params
		self.def_string 	= other.def_string
		self.arg_list[:]   	= other.arg_list
		self.data.copyFrom(other.data)
		self.params.copyFrom(other.params)

	def refreshParam(self, paramName, paramValue):

		for k in vars(self.params).keys():
			if self.params[k] == paramName:
				self.data[k] = paramValue

	def refreshData(self, var, value):

		self.data[var] = value
		# Return true if we've modified a _non-parameterized_ variable.
		# In other words, we've modified the actual light DEF, not just it's runtime args, so we'll need to export it.
		return True if (self.params[var] == None) else False

	def checkForParams(self):
		self.has_params = any(vars(self.params).values())
		return self.has_params

	def makeNewDefString(self):
		'''
		Create a new def string based on latest edited values and parameter names.
		'''
		allvars = { 'BILLBOARD_HW'	: ['r', 'g', 'b', 'bright',	'size',	'cell_size', 'cell_x', 'cell_y', 'x', 'y', 'z',	'width', 'freq', 'phase', 'amp', 'day'],
					'BILLBOARD_SW' 	: ['r', 'g', 'b', 'bright',	'size',	'cell_size', 'cell_x', 'cell_y', 'x', 'y', 'z',	'width', 'dataref']
					}
		new_def_string	= ""
		new_def_string += self.type
		new_def_string += '\t\t' if len(self.type) > 7 else '\t\t\t'
		new_def_string += self.name
		new_def_string += '\t' * ((39 - len(self.name))/8 + 1)
		for k in allvars[self.type]:
			if k in vars(self.params):
				if self.params[k]:
					v = self.params[k]
				else:
					v = str(self.data[k])
			else:
				v = str(self.data[k])
			if k not in ['cell_size', 'cell_x', 'dataref']:
				tok = v[:7]
				tok += '\t'
			elif k in ['cell_size', 'cell_x']:
				tok = "%s " % v
			else:
				tok = "\t\t\t\t\t%s" % v
			new_def_string += tok
		return new_def_string

	def getUVs(self, vals_only=0):
		x1 = self.data.cell_x * self.data.cell_size / 16.0
		x2 = (self.data.cell_x + 1) * self.data.cell_size / 16.0
		y1 = self.data.cell_y * self.data.cell_size / 8.0
		y2 = (self.data.cell_y + 1) * self.data.cell_size / 8.0
		if vals_only:
			return [x1, y1, x2, y2]
		else:
			return (Vector(x2,y2), Vector(x1,y2), Vector(x1,y1), Vector(x2,y1))

	def brightnessMustEqualOne(self):
		'''
		Returns True if circumstances dictate that brightness must equal one.
		HW billboards have the restriction that if any of variables R, G, or B are parameterized, the alpha channel ('bright') must have value 1.0.
		'''
		return self.type == 'BILLBOARD_HW' and any([self.params.r, self.params.g, self.params.b])

	def getParamDef(self):
		tks = []
		self.arg_list[:] = []
		for line in Billboard.lights_data:
			if line.startswith('LIGHT_PARAM_DEF'):
				tks = line.split()
				if len(tks) > 3 and tks[1] == self.name:
					self.arg_list[:] = tks[3:]
					break
		return False if self.arg_list and (int(tks[2]) == len(self.arg_list)) else True

	def getValuesFromProp(self, prop):

		if prop.type != 'STRING' or prop.name.upper() != 'PARAMS':
			PupMenu("ERROR: Cannot read property")
			return 0
		values = prop.data.split()
#		if len(values) > len(self.arg_list):
#			PupMenu("ERROR: Too many values in 'params' property!")
#			return 0
#		else:
		cnt = 0
		for v, name in zip(values, self.arg_list):
			if not isnum(v):
				PupMenu("ERROR: parameters must be numbers!")
				return 0
			for k in vars(self.params).keys():
				if self.params[k] == name:
					self.data[k] = float(v)
					cnt +=1
					break
			return cnt

	def remapIn(self, destIndex):
		'''
		Software light variables are frequently remapped, which is a bit of a PITA.
		Since our GUI elements are fairly variable-specific (e.g. X, Y and Z are accessed through a vector button), it seems a good idea to reorder the applicable vars
		on the way in to the editor and again on the way out, rather than maintaining their 'official' channels for the sake of 'correctness'. It's all a big mess anyway, so
		it seems smarter to keep the mess on the edges rather than letting it into our finely-tuned mechanism...
		Therefore, we have remapIn() to repatch SW light vars to HW light channels, and remapOut() for the inverse operation
		'''
		pass


	def setDataFromFile(self, byIndex=None):

		self.arg_list	= []
		self.params.reset()
		self.data  		= BillboardData()
		self.has_params	= False
		try:
			tks = []
			if byIndex:
				self.data_index = byIndex
			 	self.def_string = Billboard.lights_data[byIndex]
				tks 			= self.def_string.split()
				self.type		= tks[0]
				self.name		= tks[1]
				self.name_enum	= Billboard.names[self.type].index(self.name)
#				print 'tks: ', tks, '\n'
			else:
				for ind, l in enumerate(Billboard.lights_data):
					tks = l.split()
					if (len(tks) == Billboard.type_tok_count[self.type]) and (tks[0] == self.type) and (tks[1] == self.name):
						self.data_index 	= ind
						self.sub_def_index 	= 0
#						print "position: %d" % ind
						self.def_string 	= l
#						print 'tks: ', tks, '\n'
						break
			for i, t in enumerate(tks):
				if i == 2:
					if isnum(t):
						self.data.r			= float(t)
					else:
						self.data.r			= 1.0
						self.params.r		= t
						self.has_params		= True
				elif i == 3:
					if isnum(t):
						self.data.g			= float(t)
					else:
						self.data.g			= 1.0
						self.params.g		= t
						self.has_params		= True
				elif i == 4:
					if isnum(t):
						self.data.b			= float(t)
					else:
						self.data.b			= 1.0
						self.params.b		= t
						self.has_params		= True
				elif i == 5:
					if isnum(t):
						self.data.bright	= float(t)
					else:
						self.data.bright	= int(0) if t == 'INDEX' else 1.0
						self.params.bright	= t
						self.has_params		= True
				elif i == 6:
					if isnum(t):
						self.data.size		= float(t)
					else:
						self.data.size		= 1.0
						self.params.size	= t
						self.has_params		= True
				elif i == 7:
					self.data.cell_size	= int(t)
				elif i == 8:
					self.data.cell_x	= int(t)
				elif i == 9:
					self.data.cell_y	= int(t)
				elif i == 10:
					if isnum(t):
						self.data.x			= float(t)
					else:
						self.data.x			= 0.0
						self.params.x		= t
						self.has_params		= True
				elif i == 11:
					if isnum(t):
						self.data.y			= float(t)
					else:
						self.data.y			= 0.0
						self.params.y		= t
						self.has_params		= True
				elif i == 12:
					if isnum(t):
						self.data.z			= float(t)
					else:
						self.data.z			= -0.5
						self.params.z		= t
						self.has_params		= True
				elif i == 13:
					if isnum(t):
						self.data.width		= float(t)
					else:
						self.data.width		= 0.5
						self.params.width	= t
						self.has_params		= True
				if tks[0] == 'BILLBOARD_HW':
					if i == 14:
						if isnum(t):
							self.data.freq		= float(t)
						else:
							self.data.freq		= 1.0
							self.params.freq	= t
							self.has_params		= True
					elif i == 15:
						if isnum(t):
							self.data.phase		= float(t)
						else:
							self.data.phase		= 0.0
							self.params.phase	= t
							self.has_params		= True
					elif i == 16:
						if isnum(t):
							self.data.amp		= float(t)			# non parameterisable
					elif i == 17:
						if isnum(t):
							self.data.day		= int(t)			# non parameterisable

				elif tks[0] == 'BILLBOARD_SW':
					if i == 14:
						self.data.dataref		= t					# non parameterisable

			# If we've collected any parameter names, get the params definition:
			if self.has_params:
				if self.getParamDef():
					PupMenu("ERROR: Couldn't load LIGHT_PARAM_DEF")
		except:
			print "Couldn't find def string"
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class VisLampDef:
	def __init__(self, name, rotX, rotZ, dataBlock):
		self.name		= name
		self.rot_x		= rotX
		self.rot_z		= rotZ
		self.data		= dataBlock
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class XPlaneLightEditor:

	#
	# Macro to generate text object for scriptlink to draw openGL billboards
	#

	text_bb_scriptlink = """# script link
import bpy
import Blender
from Blender import Scene, Object, Window, Registry, BGL
from Blender.Mathutils import *
from math import floor

DISTANCE_EXPONENT 	= 0.40
OFFSET_NEAR			= 0.2
OFFSET_FAR			= 1.0
DISTANCE_FAR        = 1000.0
preview_size		= 100
light_obs = []
draw_bbs = 0

rtime			   = max(1.0, Blender.sys.time())
conf = Registry.GetKey("XPLightEditorRuntime", False)
if conf:
	try:
		preview_size 	= conf['preview_size']
		light_obs 		= conf['light_obs']
		draw_bbs		= conf['draw_bbs']
	except KeyError:
		print "ERROR: KeyError"


def fract(x):
	return x - floor(x)

def step(edge, x):
	return 0 if x < edge else 1

def mix(x, y, a):
	return x * (1.0 - a) + y * a

def clamp(x, lim1=0.0, lim2=1.0):
	return max(min(x, lim2), lim1)

def parseParams(inString):
	toks =	inString.split()
	return map(float, toks)

def assembleBillboards():

	global verts, uvs, col

	scn = Scene.GetCurrent()
	scene_layers = scn.layers

	# Bin any lights which aren't in visible layers:
	obs = [ob for ob in light_obs if any([item in scene_layers for item in ob.layers])]

	viewmatrot = viewmat.rotationPart()
	up = viewmatrot * Vector(0, 1.0, 0.0)
	up.normalize()
	rt = viewmatrot * Vector(1.0, 0, 0)
	rt.normalize()

	for ob in obs:
		# First get all the stuff which is common to all bbs in this light definition - there may be more than one:
		# We need to _ignore_ the local space rotation of the actual light ob (since rotating a light in any way other than by means of an animation has no affect on its vector in X-Plane),
		# but take into account the worldspace rotation of its immediate _parent_ , if it has one.

		obmat 			= ob.matrix.copy()
		if ob.parent:
			prntrot = ob.parent.matrix.rotationPart()
			obmat[0][0:3] = prntrot[0][:]
			obmat[1][0:3] = prntrot[1][:]
			obmat[2][0:3] = prntrot[2][:]
		else:
			obmat[0][0:3] = [1.0, 0.0, 0.0]
			obmat[1][0:3] = [0.0, 1.0, 0.0]
			obmat[2][0:3] = [0.0, 0.0, 1.0]

		obloc 			= Vector(obmat[3])
		modelviewmat 	= obmat * viewmat
		poseye 			= Vector(modelviewmat[3])		# position of the bb centre in eye-space
		veceye			= poseye.xyz.normalize()

		# shift the centre of the bb a little nearer the camera

		offset 			= OFFSET_NEAR + OFFSET_FAR * poseye.xyz.length/ DISTANCE_FAR
		poseye_closer   = poseye.copy()
		poseye_closer.xyz	-= veceye * offset
		poseye_closer_ws=  viewmatinverse * poseye_closer
		gl_normal_matrix = modelviewmat.rotationPart().invert()
		dist_atten =   pow( max(- poseye.z, 0),  DISTANCE_EXPONENT)

		for p in ob.getAllProperties():
			if "XPToolsBBPrms" in p.name:

				# unpack the bb data from our special prop:
				r, g, b, a, h, u1, v1, u2, v2, dx, dy, dz, dw, freq, phase, amp, day = parseParams(p.data)

				polar_atten = clamp(DotVecs(-veceye, gl_normal_matrix * Vector(dx, -dz, dy)) + dw)

				# Don't bother drawing it at all if it's too small to see:
				if polar_atten < 0.01:
					continue
				else:
					atten = polar_atten * dist_atten

					# flash attenuation is as per the the X-Plane shader:
					flash_attenuation = mix(1.0, abs(fract(rtime *  freq + phase) * 2.0 - 1.0), amp)  if amp < 1000.0 else   step(0.0, (amp - 1000.0) * 0.001 - fract(rtime * freq - phase))

					# scale the bb semi-length by the polar and distance atten and the global preview size multiplier.
					h *= (atten * 0.01 * preview_size)

					tr = poseye_closer_ws.xyz + (up + rt) * h
					tl = poseye_closer_ws.xyz + (up - rt) * h
					bl = poseye_closer_ws.xyz - (up + rt) * h
					br = poseye_closer_ws.xyz - (up - rt) * h

					verts.append(tl)
					verts.append(tr)
					verts.append(br)
					verts.append(bl)
					uvs.append([u1, v2])
					uvs.append([u2, v2])
					uvs.append([u2, v1])
					uvs.append([u1, v1])

					# Use flash attenuation to scale alpha.
					flash_attenuation *= a
					col.append([r * flash_attenuation, g * flash_attenuation, b * flash_attenuation, flash_attenuation ])

def setupOGL():

	BGL.glDisable(BGL.GL_LIGHTING)
	BGL.glMatrixMode(BGL.GL_PROJECTION)
	BGL.glLoadIdentity()
	BGL.glMatrixMode(BGL.GL_MODELVIEW)
	BGL.glLoadIdentity()

def backupOGL():

	BGL.glMatrixMode(BGL.GL_MODELVIEW)
	BGL.glPushMatrix()
	BGL.glMatrixMode(BGL.GL_PROJECTION)
	BGL.glPushMatrix()

def restoreOGL():

	BGL.glMatrixMode(BGL.GL_PROJECTION)
	BGL.glPopMatrix()
	BGL.glMatrixMode(BGL.GL_MODELVIEW)
	BGL.glPopMatrix()
def drawBillboards():

	current_col = BGL.Buffer(BGL.GL_FLOAT, 4)
	BGL.glGetFloatv(BGL.GL_CURRENT_COLOR, current_col)

	BGL.glEnable(BGL.GL_BLEND)
	BGL.glBlendFunc(BGL.GL_ONE, BGL.GL_ONE)
	tex_ID = BGL.Buffer(BGL.GL_INT, 1)
	BGL.glGenTextures(1, tex_ID)
	BGL.glBindTexture(BGL.GL_TEXTURE_2D, bb_tex.glLoad());
	BGL.glEnable(BGL.GL_TEXTURE_2D)

	viewMatrix = Window.GetPerspMatrix()
	viewBuff = [viewMatrix[i][j] for i in xrange(4) for j in xrange(4)]
	viewBuff = BGL.Buffer(BGL.GL_FLOAT, 16, viewBuff)

	BGL.glLoadIdentity()
	BGL.glMatrixMode(BGL.GL_PROJECTION)
	BGL.glPushMatrix()
	BGL.glLoadMatrixf(viewBuff)
	BGL.glEnable(BGL.GL_DEPTH_TEST)
	BGL.glDepthMask(BGL.GL_FALSE)

	BGL.glBegin(BGL.GL_QUADS)

	for i, cl in enumerate(col):
		BGL.glColor4f(*cl)

		for j in range(4):
			BGL.glTexCoord2d(*uvs[i*4 + j])
			BGL.glVertex3d(*verts[i*4 + j])

	BGL.glEnd()
	BGL.glDisable(BGL.GL_BLEND)
	BGL.glDisable(BGL.GL_TEXTURE_2D)
	BGL.glDepthMask(BGL.GL_TRUE)
	BGL.glDisable(BGL.GL_DEPTH_TEST)
	BGL.glColor4f(*current_col)

if draw_bbs:

	try:
		bb_tex = Blender.Image.Get('XPTBBTex')
	except:
		bb_tex = None
		print "Scriptlink can't get the billboard texture!"

	verts 	= []
	uvs		= []
	col		= []
	viewmat = Window.GetViewMatrix()
	viewmatinverse = viewmat.copy().transpose().invert()

	backupOGL()
	setupOGL()
	assembleBillboards()
	drawBillboards()
	restoreOGL()
"""

	#
	# Main class definition
	#

	mode = 'SPILL'

	SPILL_BRIGHTNESS_GAIN	=	2.0										# Used to match the rendered brightness of blender spill preview lights to that of the same light in X-Plane, for visualization purposes.

	allparams = { 'BILLBOARD_HW'	: ['r', 'g', 'b', 'size', 'x', 'y', 'z', 'width', 'freq', 'phase'],
					'BILLBOARD_SW'	: ['r', 'g', 'b', 'bright', 'size', 'x', 'y', 'z', 'width'],
					'SPILL_HW_DIR'	: ['r', 'g', 'b', 'bright', 'size', 'x', 'y', 'z', 'width'],
					'SPILL_HW_FLA'	: ['r', 'g', 'b', 'bright', 'size', 'freq', 'phase'],
					'SPILL_SW'		: ['r', 'g', 'b', 'bright', 'size', 'x', 'y', 'z', 'width']
					}

	def __init__(self, eventLowest, lightsFileDir=None):

		self.vis_lamp					= None
		self.ob_act_prev				= None							# We save the current active object before doing a test render, so we can restore it afterwards.
		self.vis_lamp_dict				= {}							# Dict of unique blender lamp data blocks for spill visualizations. Keys are tuples containing the data, values are names of the corresponding blender datablock
		self.vis_lamp_data_free			= []							# A list of any available unlinked blender lamp datablocks with name 'XPToolsVisLamp.xxx'
		self.vis_lamp_ob_defs			= {}							# A dict of VisLampDef instances, containing a blender lamp datablock, two object rotation values, and a key string built from the concatenation of the X-Plane light name and the params string, if any.
		self.vis_lamp_obs				= []							# blender lamp objects for visualisations, with name 'XPToolsVisLpOb.xxx'
		self.vis_lamp_obs_free			= []							# list of vis_lamp_obs with name 'XPToolsVisLpOb.xxx' which have previously been unlinked ('deleted') from the scene. We reuse these in preference to creating new ones.
		self.vis_loc_obs				= []							# X-Plane light objects selected as locations to draw vis lamps/obs.
		self.vis_loc_obs_selected		= []							# Keep the list so as to restore the original selection when we nuke the vis lamps/bbs
		self.context_backup				= {}
		self.context_settings 			= dict(radiosityRender = 0, aspectY = 100.0, aspectX = 100.0, mode = 5, renderer = 0,
										borderRender = 0, oversampling = 1,  singleLayer = 1, freeImages = 1, renderwinSize = 100)
		self.render_layer_settings		= dict(enable = True, name = "XPToolsRenLayer", lightGroup = None, zLayerMask = 0, passIndex = False, passAOXOR = False, passRadiosity = False,
											enableZMask = False, passZ = False, layerMask = 1048575, passCombined = True, enableSky = False, enableSolid = True, enableZMaskAll = True, enableStrand = False, enableZTra = True)
		self.render_layer				= None
		self.render_cam_data			= None
		self.render_cam					= None
		self.blender_world_mode			= 0								# State of Blender AO etc before we messed with the render settings.
		self.context_configured			= 0
		self.event0						= eventLowest					# Starting point for event enums
		self.spill 						= SpillLight()
		self.billboard 					= Billboard()
		self.light_parts				= []
		self.multiple_defs				= []							# List of [type, line indices] in Billboard.lights_data for the above
		self.params_list				= []							# list of all parameterized vars for the current light def (may include multiple light parts)
		self.params_values				= []							# list of values for the current light definition to be passed to XPlane in a PARAMS property for parameterized lights
		self.has_multi					= 0
		self.sub_def_index				= 0
		self.snapping					= 0
		self.invert_view				= 0
		self.light_ob_layer				= DEF_LIGHT_LAYER
		self.type_menu_string			= "Light type %%t|%s %%x0|%s %%x1|%s %%x2" % ("hardware - directionsl", "hardware - flashing", "software")
		self.filter_menu_string			= ""
		self.multi_menu_string			= ""
		self.params_menu_string			= ""
		self.name						= ""
		self.light_prev_spill			= []							# list of four ints: mn_type_menu.val, mn_filter_menu.val, mn_name_menu.val, sub_def_index
		self.light_prev_billboard		= []
		self.mode_prev					= ''
		self.dev_mode					= 0
		self.mouse_xy					= [0,0]
		self.win_id						= 0

		self.evt_spill					= self.event0
		self.evt_billboards				= self.event0 + 1
		self.evt_type_menu				= self.event0 + 2
		self.evt_name_menu				= self.event0 + 3
		self.evt_filter_menu			= self.event0 + 4
		self.evt_colourpicker			= self.event0 + 5
		self.evt_sR_int					= self.event0 + 6
		self.evt_sG_int					= self.event0 + 7
		self.evt_sB_int					= self.event0 + 8
		self.evt_rgb_normalize			= self.event0 + 9
		self.evt_rgb_copy				= self.event0 + 10
		self.evt_rgb_paste				= self.event0 + 11
		self.evt_bright					= self.event0 + 12
		self.evt_size					= self.event0 + 13
		self.evt_vector					= self.event0 + 14
		self.evt_snapping				= self.event0 + 15
		self.evt_width_deg				= self.event0 + 16
		self.evt_omni					= self.event0 + 17
		self.evt_freq					= self.event0 + 18
		self.evt_phase					= self.event0 + 19
		self.evt_data_to_selected		= self.event0 + 20
		self.evt_create_new_at_cursor	= self.event0 + 21
		self.evt_read_light_ob			= self.event0 + 22
		self.evt_visualize_at_cursor	= self.event0 + 23
		self.evt_visualize_selected		= self.event0 + 24
		self.evt_visualize_at_all_lamps	= self.event0 + 25
		self.evt_clear_visualizations	= self.event0 + 26
		self.evt_next_name				= self.event0 + 27
		self.evt_prev_name				= self.event0 + 28
		self.evt_invert_view			= self.event0 + 29
		self.evt_edge					= self.event0 + 30
		self.evt_pitch					= self.event0 + 31
		self.evt_heading				= self.event0 + 32
		self.evt_light_ob_layer			= self.event0 + 33
		self.evt_distance_to_surface	= self.event0 + 34
		self.evt_file_selector			= self.event0 + 35
		self.evt_xray					= self.event0 + 36
		self.evt_preview_size			= self.event0 + 37
		self.evt_show_icons				= self.event0 + 38
		self.evt_multi_menu				= self.event0 + 39
		self.evt_cell_size				= self.event0 + 40
		self.evt_cell_x					= self.event0 + 41
		self.evt_cell_y					= self.event0 + 42
		self.evt_2D_preview_size		= self.event0 + 43
		self.evt_params_menu			= self.event0 + 44
		self.evt_amp					= self.event0 + 45
		self.evt_day_flag				= self.event0 + 46
		self.evt_make_new_def			= self.event0 + 47
		self.evt_cancel_new_def			= self.event0 + 48
		self.evt_rename_light			= self.event0 + 49
		self.evt_remove_part			= self.event0 + 50
		self.evt_options				= self.event0 + 51
		self.evt_nuke_render_mat		= self.event0 + 52
		self.evt_do_render				= self.event0 + 53
		self.evt_reset_render			= self.event0 + 54

		self.tg_spill					= Create(1)
		self.tg_billboards				= Create(0)
		self.mn_type_menu				= Create(0)
		self.mn_name_menu				= Create(0)
		self.mn_filter_menu				= Create(0)
		self.mn_multi_menu				= Create(0)
		self.mn_params_menu				= Create(0)
		self.cp_colourpicker			= Create(0.5,0.5,0.5)
		self.nb_sR_int					= Create(128)
		self.nb_sG_int					= Create(128)
		self.nb_sB_int					= Create(128)
		self.bt_rgb_normalize			= Create(0)
		self.bt_rgb_copy				= Create(0)
		self.bt_rgb_paste				= Create(0)
		self.nb_bright					= Create(0.5)
		self.nb_size					= Create(16.0)
		self.nm_vector					= Create(0,1,0)					# axes transformed from xp to blender
		self.tg_snapping				= Create(1)
		self.nb_width_deg				= Create(120)
		self.nb_edge					= Create(0)
		self.nb_freq					= Create(1.0)
		self.nb_phase					= Create(0.0)
		self.nb_amp						= Create(0.0)
		self.tg_day_flag				= Create(0)
		self.bt_data_to_selected		= Create(0)
		self.bt_create_new_at_cursor	= Create(0)
		self.bt_read_light_ob			= Create(0)
		self.bt_visualize_at_cursor		= Create(0)
		self.bt_visualize_selected		= Create(0)
		self.bt_visualize_at_all_lamps	= Create(0)
		self.bt_clear_visualizations	= Create(0)
		self.bt_next_name				= Create(0)
		self.bt_prev_name				= Create(0)
		self.tg_invert_view				= Create(0)
		self.tg_omni					= Create(1)
		self.nb_pitch					= Create(0)
		self.nb_heading					= Create(0)
		self.nb_light_ob_layer			= Create(DEF_LIGHT_LAYER)
		self.sl_distance_to_surface		= Create(0.0)
		self.tg_xray					= Create(0)
		self.bt_make_new_def			= Create(0)
		self.bt_cancel_new_def			= Create(0)
		self.bt_remove_part				= Create(0)
		self.st_new_light_name			= Create(self.name)
		self.nb_preview_size			= Create(100)
		self.nb_2D_preview_size			= Create(100)
		self.tg_show_icons				= Create(0)
		self.mn_cell_size				= Create(1)
		self.nb_cell_x					= Create(0)
		self.nb_cell_y					= Create(0)
		self.bt_options					= Create(0)
		self.bt_nuke_render_mat			= Create(0)
		self.bt_do_render				= Create(0)
		self.bt_reset_render			= Create(0)

		self.preview_size				= 100
		self.sRGB						= [0.5, 0.5, 0.5]
		self.sRGB_for_cp				= tuple(self.sRGB)
		self.sRGB_int					= fltToPS(self.sRGB)
		self.sRGB_clipboard				= [0.0, 0.0, 0.0]
		self.xyz_blender				= [0.0, 1.0, 0.0]
		self.spot_radius				= 0.0
		self.edge_type					= ""
		self.width_deg					= 120
		self.omni						= 0
		self.dimmed						= 0
		self.live_updates				= 0
		self.spill.type					= 'SPILL_HW_DIR'
		self.billboard.type				= 'BILLBOARD_HW'
		self.mn_type_menu.val			= 0
		self.spill.name 				= 'full_custom_halo_night'
		self.billboard.name				= 'spot_params_bb'
		self.mn_filter_menu.val			= 1
		self.buildFilterMenuString()
		self.name_menu_list				= []
		self.buildNameMenuString()
		self.mn_type_menu.val			= SpillLight.types.index(self.spill.type)
		self.spill.name_enum			= SpillLight.names[self.spill.type].index(self.spill.name)
		self.mn_name_menu.val			= self.spill.name_enum
		self.spill.setDataFromFile()
		self.refreshPanelValues()
		self.draw_bbs					= 0
		self.mouse_over_CP				= False
		self.light_obs 					= []
		self.view3D_back_colour			= [155, 155, 155, 255]
		self.txt_output					= None
		self.unsaved_def				= False
		self.defs_out					= None

		#
		# Set up a registry dict to pass runtime vals to the openGL scriptlink
		#
		self.setRuntime()

		#
		# Create a Blender text object for the scriptlink
		#
		try:
			self.scriptlink = Blender.Text.Get("XPToolsSL.py")
			Blender.Text.unlink(self.scriptlink)
			self.scriptlink = Blender.Text.New("XPToolsSL.py")
			self.scriptlink.write(self.text_bb_scriptlink)
		except:
			self.scriptlink = Blender.Text.New("XPToolsSL.py")
			self.scriptlink.write(self.text_bb_scriptlink)

		scene = Blender.Scene.GetCurrent()
		scene.addScriptLink("XPToolsSL.py", 'Redraw')

		#
		# Unpack persistent config data from the registry
		#

		rep = self.unpackConfig()

		if rep:
			if self.mode_prev:
				XPlaneLightEditor.mode = self.mode_prev
				self.setMode()
				self.unpackMenuListAndLoadLight()

	def string2Values(self, inString):
		"""
			Converts a single string of values into a list of floats and/or ints.
		"""
		outlist = []
		if not isinstance(inString, str):
			return []
		if inString.strip('0123456789-+. \t'):
			print "%s - Params must be numbers!" % inString
			return []
		toks = inString.split()
		for t in toks:
			if '.' not in t:
				outlist.append(int(t))
			else:
				outlist.append(float(t))
		return outlist

	def unpackConfig(self):

		if LIGHT_EDITOR_CONFIG:
			try:
				if 'light_prev_spill' in LIGHT_EDITOR_CONFIG:
					self.light_prev_spill  		= LIGHT_EDITOR_CONFIG['light_prev_spill']
				else: return False
				if 'light_prev_billboard' in LIGHT_EDITOR_CONFIG:
					self.light_prev_billboard 	= LIGHT_EDITOR_CONFIG['light_prev_billboard']
				else: return False
				if 'mode_prev' in LIGHT_EDITOR_CONFIG:
					self.mode_prev				= LIGHT_EDITOR_CONFIG['mode_prev']
				else: return False
				if 'preview_size' in LIGHT_EDITOR_CONFIG:
					self.preview_size 			= LIGHT_EDITOR_CONFIG['preview_size']
				return True
			except:
				print "ERROR. Unable to unpack LIGHT_EDITOR_CONFIG"

	def packConfig(self):

		if XPlaneLightEditor.mode == 'SPILL':
			self.light_prev_spill = [self.mn_type_menu.val, self.mn_filter_menu.val, self.mn_name_menu.val, self.sub_def_index]
		elif XPlaneLightEditor.mode == 'BILLBOARD':
			self.light_prev_billboard = [self.mn_type_menu.val, self.mn_filter_menu.val, self.mn_name_menu.val, self.sub_def_index]
		try:
			global LIGHT_EDITOR_CONFIG
			LIGHT_EDITOR_CONFIG = {}
			LIGHT_EDITOR_CONFIG['light_prev_spill']			= self.light_prev_spill
			LIGHT_EDITOR_CONFIG['light_prev_billboard']		= self.light_prev_billboard
			LIGHT_EDITOR_CONFIG['mode_prev']				= XPlaneLightEditor.mode
			LIGHT_EDITOR_CONFIG['preview_size']				= self.preview_size
		except:
			print "ERROR. Unable to pack LIGHT_EDITOR_CONFIG"

	def unpackMenuListAndLoadLight(self):
		'''
		Initialize the editor with menu settings from the registry cache
		'''

		if XPlaneLightEditor.mode == 'SPILL' and self.light_prev_spill:
			self.mn_type_menu.val 	= self.light_prev_spill[0]
			self.mn_filter_menu.val	= self.light_prev_spill[1]
			self.mn_name_menu.val	= self.light_prev_spill[2]
			self.loadLight(SpillLight.names[SpillLight.types[self.mn_type_menu.val]][self.mn_name_menu.val], None)
			if self.has_multi:
				self.mn_multi_menu.val = self.light_prev_spill[3]
				self.swapParts(self.light_prev_spill[3])

		elif XPlaneLightEditor.mode == 'BILLBOARD' and self.light_prev_billboard:
			self.mn_type_menu.val 	= self.light_prev_billboard[0]
			self.mn_filter_menu.val	= self.light_prev_billboard[1]
			self.mn_name_menu.val	= self.light_prev_billboard[2]
			self.loadLight(Billboard.names[Billboard.types[self.mn_type_menu.val]][self.mn_name_menu.val], None)
			if self.has_multi:
				self.mn_multi_menu.val = self.light_prev_billboard[3]
				self.swapParts(self.light_prev_billboard[3])

	def setRuntime(self):
		d = {}
		d['draw_bbs']					= self.draw_bbs
		d['preview_size']				= self.preview_size
		d['light_obs']					= self.light_obs
		Registry.SetKey('XPLightEditorRuntime', d, False)

	def getRuntime(self):
		temp = Registry.GetKey('XPLightEditorRuntime', False)
		if temp:
			try:
				self.draw_bbs			= temp['draw_bbs']
				self.preview_size		= temp['preview_size']
				self.light_obs			= temp['light_obs']
			except:
				self.setRuntime()

	def showBillboards(self):

		scn = Blender.Scene.GetCurrent()
		self.light_obs[:] = [ob for ob in scn.objects for prop in ob.getAllProperties() if ob.type == 'Lamp' and prop.name =='XPToolsBBPrms']
		self.draw_bbs	=	1
		self.setRuntime()

		# The blender lamp icons tend to obscure the billboard images, so we hide them while previews are turned on.
		# We appear to have to select them and then use a keystroke to hide them.
		if not self.tg_show_icons.val:
			scn.objects.selected = self.light_obs
			v3d = Window.GetScreenInfo(Window.Types.VIEW3D)
			if not v3d:
				print "No open 3D window!"
				return
			id = v3d[0]['id']
			Window.SetKeyQualifiers(0)
			Window.QAdd(id, HKEY, 1)
			Window.QAdd(id, HKEY, 0)
			Window.QHandle(id)
		# If the current VIEW3D is ortho, we warn the user to switch to perspective view.
		# Test for ortho projection - if the fourth value of each row = 0, we're ortho
		pmat = Blender.Window.GetPerspMatrix()
		if pmat[0][3] == pmat[1][3] == pmat[2][3] == 0:
			is_ortho = True
			PupMenu('Halos may look wrong in ortho view - suggest switching to perspective view (numpad 5)')

		else:
			is_ortho = False

# 		if is_ortho:
# 			Window.QAdd(id, PAD5, 1)
# 			Window.QAdd(id, PAD5, 0)

		self.dimBlenderScene(True)
		Window.Redraw(-1)

	def hideBillboards(self):
		scn = Blender.Scene.GetCurrent()
		self.draw_bbs	=	0
		self.light_obs[:] = []
		self.setRuntime()

		if not self.tg_show_icons.val:
			v3d = Window.GetScreenInfo(Window.Types.VIEW3D)
			if not v3d:
				print "ERROR: No open 3D window!"
				return

			id = v3d[0]['id']
			cds = v3d[0]['vertices']
			mousexy = Window.GetMouseCoords()
			centre = [int((cds[0] + cds[2])/2), int((cds[1] + cds[3])/2)]
			Window.SetMouseCoords(centre)
			Window.QHandle(id)
			Window.SetKeyQualifiers(Window.Qual.ALT)
			Window.QAdd(id, LEFTALTKEY, 1)
			Window.QAdd(id, HKEY, 1)
			Window.QAdd(id, HKEY, 0)
			Window.QAdd(id, LEFTALTKEY, 0)
			Window.QHandle(id)
			Window.SetKeyQualifiers(0)
			Window.SetMouseCoords(mousexy)
			scn.objects.selected = []

		self.dimBlenderScene(False)
		Window.Redraw(-1)

	def dimBlenderScene(self, dim=1):
		"""
		# Here we set the 3D view background colour to black, for previewing billboards.
		"""
		if self.tg_xray.val:		# this is a recycled button name. It now means "disable auto-dim"
			if not self.dimmed:
				return
		# case 1 - we want dimming:
		if dim:
			theme = Theme.Get()[0]
			temp = theme.get('view3d').back
		#	if not any(temp):		# the background colour is already black, so we're probably out of sync and already have the default colour cached - do nothing
			if self.dimmed:
				return
			self.view3D_back_colour[:]	= []
			self.view3D_back_colour.extend(theme.get('view3d').back)
			theme.get('view3d').back = [0, 0, 0, 0]
			self.dimmed = 1
		# case 0 - we want to restore the original background colour:
		else:
			if not self.dimmed:
				return
			theme = Theme.Get()[0]
			theme.get('view3d').back = self.view3D_back_colour
			self.dimmed = 0
		Window.RedrawAll()

	def buildTypeMenuString(self):
		if XPlaneLightEditor.mode == 'SPILL':
			self.type_menu_string			= "light type %%t|%s %%x0|%s %%x1|%s %%x2" % ("hardware lights - directional", "hardware lights - flashing", "software lights")
		elif XPlaneLightEditor.mode == 'BILLBOARD':
			self.type_menu_string			= "light type %%t|%s %%x0|%s %%x1" % ("hardware lights", "software lights")

	def buildFilterMenuString(self):

		items = ['aircraft lights', 'scenery lights', 'colour', 'brightness',
				'size',	'direction / width',	'frequency / phase', 'other']
		self.filter_menu_string		= "filter list %t"
		if XPlaneLightEditor.mode == 'SPILL':
			for cat, item in zip(SpillLight.filter_cats, items):
				if SpillLight.name_groups[SpillLight.types[self.mn_type_menu.val]][cat]:
					self.filter_menu_string += "|%s%s %%x%d" % (str(len(SpillLight.name_groups[SpillLight.types[self.mn_type_menu.val]][cat])).ljust(4),  item, SpillLight.filter_cats.index(cat))
		elif XPlaneLightEditor.mode == 'BILLBOARD':
			for cat, item in zip(Billboard.filter_cats, items):
				if Billboard.name_groups[Billboard.types[self.mn_type_menu.val]][cat]:
					self.filter_menu_string += "|%s%s %%x%d" % (str(len(Billboard.name_groups[Billboard.types[self.mn_type_menu.val]][cat])).ljust(4),  item, Billboard.filter_cats.index(cat))

	def buildParamsMenuString(self):
		'''
		This menu displays a list of the vars that may legitimately be parameterized for the current light.
		Those currently parameterized are indicated by '+'
		'''

		self.params_menu_string	= "parameters%t"
		if XPlaneLightEditor.mode == 'BILLBOARD':
			for i, v in enumerate(XPlaneLightEditor.allparams[self.billboard.type]):
				self.params_menu_string += "|%s %s%%x%d" % ('+' if self.billboard.params[v] else '  ', v.upper(), i+1)
		elif XPlaneLightEditor.mode == 'SPILL':
			for i, v in enumerate(XPlaneLightEditor.allparams[self.spill.type]):
				self.params_menu_string += "|%s %s%%x%d" % ('+' if self.spill.params[v] else '  ', v.upper(), i+1)

	def updateParameterList(self, menuIndex):

		#  The menu indices vals = list index + 1
		i = menuIndex - 1

		if XPlaneLightEditor.mode == 'BILLBOARD':
			if not 0 <= i < len(XPlaneLightEditor.allparams[self.billboard.type]):
				print "ERROR: updateParameterList() index overrun"
				return
			param = self.billboard.params[XPlaneLightEditor.allparams[self.billboard.type][i]]

			if param != None:  # we're already parameterized, so now remove it
				rep = PupMenu("REMOVE THIS PARAM?:%t|    NO|  YES")
				if rep == 2:
					if param in self.billboard.arg_list:
						self.billboard.arg_list.remove(param)
					self.billboard.params[XPlaneLightEditor.allparams[self.billboard.type][i]] = None
				else:
					return
			else:
				## Check to see if adding this param would require that brightness = 1.0
				if XPlaneLightEditor.allparams[self.billboard.type][i] in ['r', 'g', 'b'] \
								and self.billboard.type == 'BILLBOARD_HW' \
								and self.billboard.data.bright != 1.0 \
								and not any([self.billboard.params.r, self.billboard.params.g, self.billboard.params.b]):

					rep = PupMenu("Parameterizing this var requires that brightness = 1. Continue anyway?%t|    NO|  YES")
					if rep == 2:
						rep2 = PupMenu("Scale existing RGB values to preserve current appearance?%t|    NO|  YES")
						if rep2 == 2:
							self.billboard.data.r *= self.billboard.data.bright
							self.billboard.data.g *= self.billboard.data.bright
							self.billboard.data.b *= self.billboard.data.bright
						self.billboard.data.bright = 1.0
						for v in ['r', 'g', 'b', 'bright']:
							self.refreshMultis(v)

				# prompt user for name for a param
				newname = Create(XPlaneLightEditor.allparams[self.billboard.type][i].upper())
				block = []
				block.append(("Name: ", newname, 0, 7, "Enter a name for the new parameter you've requested. Entering an existing name will cause the corresponding value to be shared between the like-named params."))
				retval = PupBlock('New parameter', block)
				if retval == 1:
					self.billboard.params[XPlaneLightEditor.allparams[self.billboard.type][i]] = newname.val.upper()
					self.billboard.arg_list.append(newname.val.upper())
				else:
					return

			self.light_parts[self.sub_def_index].params.copyFrom(self.billboard.params)
			self.light_parts[self.sub_def_index].arg_list[:] = self.billboard.arg_list
			self.billboard.checkForParams()
			self.refreshPanelValues()
		elif XPlaneLightEditor.mode == 'SPILL':
			if not 0 <= i < len(XPlaneLightEditor.allparams[self.spill.type]):
				print "ERROR: updateParameterList() index overrun"
				return
			param = self.spill.params[XPlaneLightEditor.allparams[self.spill.type][i]]
			if param != None:
				rep = PupMenu("REMOVE THIS PARAM?:%t|    NO|  YES")
				if rep == 2:
					if param in self.spill.arg_list:
						self.spill.arg_list.remove(param)
					self.spill.params[XPlaneLightEditor.allparams[self.spill.type][i]] = None
				else:
					return
			else:
				newname = Create(XPlaneLightEditor.allparams[self.spill.type][i].upper())
				block = []
				block.append(("Name: ", newname, 0, 7, "Enter a name for the new parameter you've requested. Entering an existing name will cause the corresponding value to be shared between the like-named params."))
				retval = PupBlock('New parameter', block)
				if retval == 1:
					self.spill.params[XPlaneLightEditor.allparams[self.spill.type][i]] = newname.val.upper()
					self.spill.arg_list.append(newname.val.upper())
				else:
					return
			self.light_parts[self.sub_def_index].params.copyFrom(self.spill.params)
			self.light_parts[self.sub_def_index].arg_list[:] = self.spill.arg_list
			self.spill.checkForParams()

		self.light_parts[self.sub_def_index].checkForParams()
		self.buildParamsMenuString()
		self.unsaved_def = True

	def getFirstAvailFilterCat(self):
		""" This is cheesy - read the filter menu name string to find the index of the first available filter category (one which contains lights in other words)
		"""
		ind = self.filter_menu_string.find("%x")
		ind += 2
		f = ""
		while len(self.filter_menu_string) - 1 > ind and self.filter_menu_string[ind].isdigit():
		 	f += self.filter_menu_string[ind]
		 	ind += 1

		return int(f)

	def buildNameMenuString(self):

		self.name_menu_string			= "light name %t"
		self.name_menu_list[:]			= []		# list of pointers into the list of names.
		if XPlaneLightEditor.mode == 'SPILL':
			for n in SpillLight.name_groups[SpillLight.types[self.mn_type_menu.val]][SpillLight.filter_cats[self.mn_filter_menu.val]]:
				self.name_menu_string += ("|%s %%x%d" % (SpillLight.names[SpillLight.types[self.mn_type_menu.val]][n], n))
				self.name_menu_list.append(n)
		elif XPlaneLightEditor.mode == 'BILLBOARD':
			for n in Billboard.name_groups[Billboard.types[self.mn_type_menu.val]][Billboard.filter_cats[self.mn_filter_menu.val]]:
				self.name_menu_string += ("|%s %%x%d" % (Billboard.names[Billboard.types[self.mn_type_menu.val]][n], n))
				self.name_menu_list.append(n)

	def getNextNameMenuEnum(self, direction):
		"""
			When the user changes the name filter menu, or increments/decrements the name list via the up/down buttons,
			we need to find the enumeration of the next available item on the list.
			<direction> == -1 means 'previous'; 1 means 'next', and zero means 'first'
		"""
#		print "called once"
		pick = self.mn_name_menu.val
		if not self.name_menu_list:		#  if the list is empty, do nothing
			return pick
		if pick in self.name_menu_list:
			item = self.name_menu_list.index(pick)
		else:
			item = 0

		if not direction:				#  we want the first name on the list
			return self.name_menu_list[0]

		elif direction < 0:				# 	we want the previous name, if any, else loop around
			return self.name_menu_list[(item - 1)%len(self.name_menu_list)]

		elif direction > 0:				# 	we want the next name, if any, else loop around
			return self.name_menu_list[(item + 1)%len(self.name_menu_list)]

	def snapToAxes(self, xyz):

		if self.snapping:
			t = 0.25
			x, y, z = xyz
			if abs(x) < t:
				x = 0.0
				m = math.hypot(y, z)
				y /= m
				z /= m
			if abs(y) < t:
				y = 0.0
				m = math.hypot(x, z)
				x /= m
				z /= m
			if abs(z) < t:
				z = 0.0
				m = math.hypot(x, y)
				x /= m
				y /= m
			return (x, y, z) if isinstance(xyz, tuple) else [x, y, z]
		return xyz

	def sRGBCopy(self):
		'''
		copy/paste functionality for RGB values between lights and/or light parts
		'''
		self.sRGB_clipboard	= self.sRGB

	def sRGBPaste(self):
		'''
		copy/paste functionality for RGB values between lights and/or light parts
		'''
		self.sRGB = self.sRGB_clipboard
		self.sRGB_for_cp = tuple(superClamp(self.sRGB))
		self.sRGB_int					= fltToPS(self.sRGB)
		self.nb_sR_int.val 				= self.sRGB_int[0]
		self.nb_sG_int.val				= self.sRGB_int[1]
		self.nb_sB_int.val 				= self.sRGB_int[2]

		if XPlaneLightEditor.mode == 'SPILL':
			self.spill.data.r, self.spill.data.g, self.spill.data.b = sRGBToLinear(self.sRGB)
		elif XPlaneLightEditor.mode == 'BILLBOARD':
			self.billboard.data.r, self.billboard.data.g, self.billboard.data.b = self.sRGB
		Draw()

	def calculateSpotRadius(self):
		return
		# We have to substitute the value of self.spill.data.width in the omni case, else the math comes out wrong.
		# width = 1 for omni is just a convention, not a mathematical fact!

		w = self.spill.data.width if not self.omni else 0.0
		dts = clamp(self.sl_distance_to_surface.val, 0, self.spill.data.size)

		self.spot_radius	= (math.tan(math.acos(w)) * dts)\
							if (dts < self.spill.data.size * w)\
							else (math.sqrt(math.pow(self.spill.data.size, 2.0) - math.pow(dts, 2.0)))
		dist1 = self.spill.data.size * w						# distance at which spot size is largest
		dist2 = (self.spill.data.size + dist1) * 0.5			# distance at which the spot edge is getting very soft
		self.edge_type					= 'hard' if dts < dist1 else ('med' if dts < dist2 else 'soft')

	def normalizeRGBValues(self):

		if not any(self.sRGB):
			PupMenu("ERROR: at least one non-zero value required to normalize!")
			return
		rat = 1.0/max(self.sRGB)
		n = []
		for	c in self.sRGB:
			n.append(c * rat)

		self.sRGB[:] = []
		self.sRGB.extend(n)
		self.sRGB_for_cp 		= tuple(superClamp(n))
		self.sRGB_int			= fltToPS(self.sRGB)
		self.nb_sR_int.val 		= self.sRGB_int[0]
		self.nb_sG_int.val		= self.sRGB_int[1]
		self.nb_sB_int.val 		= self.sRGB_int[2]

		if XPlaneLightEditor.mode == 'SPILL':
			self.spill.data.r, self.spill.data.g, self.spill.data.b = sRGBToLinear(self.sRGB)
		elif XPlaneLightEditor.mode == 'BILLBOARD':
			self.billboard.data.r, self.billboard.data.g, self.billboard.data.b = self.sRGB
		Draw()

	def decodeBBGeomParams(self):
		xyz = Vector(self.billboard.data.x, self.billboard.data.y, self.billboard.data.z)
		if not any([self.billboard.data.x, self.billboard.data.y, self.billboard.data.z]):
			self.omni = 1
			mag = 0
			self.width_deg  = 360
			self.nb_edge.val = 0
		else:
			self.omni = 0
			mag = xyz.length
			w = 0.0
			k	= mag + self.billboard.data.width - 1.0
			# k -> edge 'hardness' factor. k = 0 is exisiting x-plane behaviour, where the light reaches maximum brightness exactly on axis.
			# Higher values of k cause max brightness to be reached sooner.

			try:
				w = self.billboard.data.width/(k + 1.0)
				if w >= 0.5:
					w = 0.5
				self.width_deg = 2.0 * degrees(acos(w/(w - 1.0)))
			except ZeroDivisionError:
				print "div zero"
				print "'w' is too large! Guessing a cone angle of 2 degrees"
				self.width_deg = 2.0

			if abs(k) > 0.0000001:
				try:
					self.nb_edge.val = int(clamp(round(log(k / (2.0 * ( 1.0 - cos(radians(self.width_deg * 0.5))))) + 0.006737946) + 5, 0, 10))
#					print 'BP13 nb_edge.val: ', self.nb_edge.val
				except ZeroDivisionError:
					print "div zero at bp 2 in decodeBBparams"
					self.nb_edge.val = 0
			else:
				self.nb_edge.val = 0

			xyz.normalize()
			self.xyz_blender				= [self.billboard.data.x, -self.billboard.data.z, self.billboard.data.y]

	def encodeBBGeomParams(self):
		"""
		By tweaking the relative values of x,y,z and w, it's possible to get quite a wide range of edge falloff slopes
		"""
		if self.omni:
			self.billboard.data.x = 0.0
			self.billboard.data.y = 0.0
			self.billboard.data.z = 0.0
			self.xyz_blender	  = [0.0, 0.0, 0.0]
			self.billboard.data.width = 1.0
			self.width_deg = 360.0
			self.nb_edge.val = 0.0
		else:
			# this is in case we've just been 'de-omnied' and consequently have been left with a zero-length XYZ vector.
			# Set to default 'south-facing' vector
			if not any([self.billboard.data.x, self.billboard.data.y, self.billboard.data.z]):
				self.billboard.data.z = 1.0
				self.width_deg = 360.0
				self.nb_width_deg.val = 360
				self.nb_edge.val = 0.0
			if self.width_deg < 2.0:
				self.width_deg = 2.0
				self.nb_width_deg.val = 2
			cos_a 	= cos(radians(self.width_deg * 0.5))
			k = 2.0 * (1.0 - cos(radians(self.width_deg * 0.5))) * (exp(self.nb_edge.val - 5) - 0.006737946)
			self.billboard.data.width = (1.0 + k) * cos_a /(cos_a - 1.0)
			mag	= 1.0 + k - self.billboard.data.width

			# rescale the direction vector:
			xyz = Vector(self.billboard.data.x, self.billboard.data.y, self.billboard.data.z)
			xyz.normalize()
			self.billboard.data.x, self.billboard.data.y, self.billboard.data.z = xyz * mag
			self.xyz_blender			= [self.billboard.data.x, -self.billboard.data.z, self.billboard.data.y]
		for var in ['x', 'y', 'z', 'width']:
			self.refreshMultis(var)

	def refreshMultis(self, var):

		if XPlaneLightEditor.mode == 'SPILL':
			if not self.dev_mode:
				param_name = self.spill.params[var]
				if param_name:
					for part in self.light_parts:
						part.refreshParam(param_name, self.spill.data[var])
			else:
				# In dev mode, we update any specified var, regardless of whether or not it's 'parameterized'.
				if self.sub_def_index < len(self.light_parts):

					# self.light_parts.refreshData() returns True if we've modified a non-parameterized var.
					# The following slightly weird expression updates self.unsaved_def with the result of the latest data update

					self.unsaved_def =  self.light_parts[self.sub_def_index].refreshData(var, self.spill.data[var]) or self.unsaved_def

		elif XPlaneLightEditor.mode == 'BILLBOARD':
			if not self.dev_mode:
				param_name = self.billboard.params[var]
				if param_name:
					for part in self.light_parts:
						part.refreshParam(param_name, self.billboard.data[var])
			else:
				if self.sub_def_index < len(self.light_parts):
					self.unsaved_def = self.light_parts[self.sub_def_index].refreshData(var, self.billboard.data[var]) or self.unsaved_def
		Draw()

	def refreshPanelValues(self):

		"""
		Here we set up all the panel controls with values from a newly loaded light
		"""

		if self.dev_mode:
			self.buildParamsMenuString()

		if XPlaneLightEditor.mode == 'SPILL':
			self.tg_spill.val = 1
			self.tg_billboards.val = 0
			self.linear_rgb				= [self.spill.data.r, self.spill.data.g, self.spill.data.b]
			self.sRGB					= linearToSRGB(self.linear_rgb)
		elif XPlaneLightEditor.mode == 'BILLBOARD':
			self.tg_spill.val = 0
			self.tg_billboards.val = 1
			self.sRGB					= [self.billboard.data.r, self.billboard.data.g, self.billboard.data.b]

		self.sRGB_for_cp				= superClamp(tuple(self.sRGB), 0.0, 1.0)
		self.sRGB_int					= fltToPS(self.sRGB)
		self.nb_sR_int.val 				= self.sRGB_int[0]
		self.nb_sG_int.val 				= self.sRGB_int[1]
		self.nb_sB_int.val 				= self.sRGB_int[2]

		#print "bp0: ", self.xyz_blender
		if XPlaneLightEditor.mode == 'SPILL':
			self.nb_bright.val			= linearToSRGB(self.spill.data.bright)
		if XPlaneLightEditor.mode == 'BILLBOARD':
			self.nb_bright.val			= self.billboard.data.bright

		self.xyz_blender				= [self.spill.data.x, -self.spill.data.z, self.spill.data.y] if XPlaneLightEditor.mode == 'SPILL' else [self.billboard.data.x, -self.billboard.data.z, self.billboard.data.y]
		#print "bp1: ", self.xyz_blender, self.spill.data.width if XPlaneLightEditor.mode == 'SPILL' else  self.billboard.data.width
		if XPlaneLightEditor.mode == 'BILLBOARD':
			self.decodeBBGeomParams()
			self.omni						= 0 if self.billboard.data.width != 1.0 and any([self.billboard.data.x, -self.billboard.data.z, self.billboard.data.y]) else 1
		elif XPlaneLightEditor.mode == 'SPILL':
			self.omni						= 0 if self.spill.data.width != 1.0 and any([self.spill.data.x, -self.spill.data.z, self.spill.data.y]) else 1
		self.tg_omni.val				= self.omni
		if self.omni:
			self.invert_view			= 0
			self.nm_vector.val			= (0.0, 0.0, 1.0)
		else:
			self.invert_view			= self.tg_invert_view.val
			self.nm_vector.val			= tuple(self.xyz_blender if not self.invert_view else flipView(self.xyz_blender))
		#print 'bp2: ', self.nm_vector.val,  self.spill.data.width if XPlaneLightEditor.mode == 'SPILL' else  self.billboard.data.width
		self.nb_pitch.val, unused, self.nb_heading.val = toEulers(self.xyz_blender)
		if XPlaneLightEditor.mode == 'SPILL':
			self.nb_size.val				= self.spill.data.size
		#	self.sl_distance_to_surface.val	= self.spill.data.size * 0.75									# this is just an authoring aid to allow the user to see what the radius of the projected spot will be at arbitrary distances
			self.width_deg					= int(2.0 * math.degrees(math.acos(self.spill.data.width)))
		#	self.calculateSpotRadius()
			self.nb_freq.val 				= self.spill.data.freq
			self.nb_phase.val 				= self.spill.data.phase
			if self.dev_mode:
				self.nb_amp.val 			= self.spill.data.amp
				self.tg_day_flag.val		= self.spill.data.day

			#print 'bp3: ', self.xyz_blender, self.width_deg, self.spill.data.width
		if XPlaneLightEditor.mode == 'BILLBOARD':
			self.nb_size.val				= self.billboard.data.size
			self.nb_freq.val 				= self.billboard.data.freq
			self.nb_phase.val 				= self.billboard.data.phase

			if self.dev_mode:
				self.mn_cell_size.val		= self.billboard.data.cell_size >> 1
				self.nb_cell_x.val			= self.billboard.data.cell_x
				self.nb_cell_y.val			= self.billboard.data.cell_y
				self.nb_amp.val 			= self.billboard.data.amp
				self.tg_day_flag.val		= self.billboard.data.day
			#print 'bp4: ', self.xyz_blender, self.width_deg,  self.billboard.data.width
		self.nb_width_deg.val			= int(round(self.width_deg))
		#print 'bp5: ', self.xyz_blender, self.width_deg, self.spill.data.width if XPlaneLightEditor.mode == 'SPILL' else  self.billboard.data.width
		Draw()

	def generateParamsValues(self):
		"""
			This builds a liat of args for the PARAMS property based on the current editor values
		"""

		# clear the list of PARAMS values:
		self.params_values[:] = []

		if XPlaneLightEditor.mode == 'SPILL':

			if self.omni:
				self.spill.data.x = 0.0
				self.spill.data.y = 0.0
				self.spill.data.z = 0.0
				self.spill.data.width = 1.0

			if self.spill.has_params and not self.has_multi:

				for ar in self.spill.arg_list:
					for k in vars(self.spill.params).keys():
						if vars(self.spill.params)[k] == ar:
							self.params_values.append(vars(self.spill.data)[k])
							break

		elif XPlaneLightEditor.mode == 'BILLBOARD':

			# If we're omni, we need to update the XYZW params because we've been preserving the values that were extant before the 'omni' button was hit:
			if self.omni:
				self.billboard.data.x = 0.0
				self.billboard.data.y = 0.0
				self.billboard.data.z = 0.0
				self.billboard.data.width = 1.0

			if self.billboard.has_params and not self.has_multi:
				for ar in self.billboard.arg_list:
					for k in vars(self.billboard.params).keys():
						if vars(self.billboard.params)[k] == ar:
							self.params_values.append(vars(self.billboard.data)[k])
							break

		# Writing params is somewhat complicated in the case of multi-part lights by the fact that we may have to gather args from more than one source
		# and combine them into a single 'params' blender property.

		if self.has_multi:
			# make a list of keys and as yet empty values
			param_kvs = []
			for part in self.light_parts:

				if part.has_params:
					if part.sub_def_index == self.sub_def_index:
						part.data.copyFrom(self.billboard.data if isinstance(part, Billboard) else self.spill.data)
					for prm in part.arg_list:
						param_kvs.append([prm, None])
					break

			if param_kvs:
				for [prm, val] in param_kvs:
					got_val = False
					while not got_val:
						for part in self.light_parts:
							if got_val:
								break
							for k in vars(part.params).keys():
								if got_val:
									break
								if part.params[k] == prm:
									self.params_values.append(part.data[k])
									got_val = True
									break
						else:
							got_val = True
							break

	def addXPlaneLampToScene(self, addLocation):
		"""
			This creates a new  XPlane light object from the definition currently displayed in the GUI
			and adds it to the current scene at the location we pass in.
			If the current definition includes a billboard (halo), a property containing the visualization params is also added to the ob.
		"""

		scn = Blender.Scene.GetCurrent()
		if self.light_ob_layer not in scn.layers:
			layers = scn.layers
			layers.append(self.light_ob_layer)
			scn.layers = layers

		# If we already have an instance of our lamp data in the scene, re-use it, else make a new one:
		for lamp in Blender.Lamp.Get():
			if lamp.name == 'XPlane_Light':
				xp_lamp = lamp
				break
		else:
			xp_lamp = Blender.Lamp.New('Lamp')
			xp_lamp.name = 	'XPlane_Light'

		# Improve Blender perf by nuking the actual lighting behaviour of the lamp:
		xp_lamp.setMode('NoDiffuse', 'NoSpecular', 'Sphere', 'RayShadow')
		xp_lamp.dist		= 0.01
		xp_lamp.energy		= 0.0
		xp_lamp_ob 			= scn.objects.new(xp_lamp)
		xp_lamp_ob.loc 		= addLocation
		xp_lamp_ob.layers	= [self.light_ob_layer]

		if XPlaneLightEditor.mode == 'SPILL':
			xp_lamp_ob.name 	= self.spill.name
			xp_lamp_ob.addProperty('name', self.spill.name, 'STRING')

		elif XPlaneLightEditor.mode == 'BILLBOARD':
			xp_lamp_ob.name 	= self.billboard.name
			xp_lamp_ob.addProperty('name', self.billboard.name, 'STRING')

		# Generate values for the PARAMS property, if any.
		if self.params_list:
#			print "have params list"
			self.generateParamsValues()
#			print "arg values: ", self.params_values, "\n"
			if self.params_values:
				xp_lamp_ob.addProperty('PARAMS', numsToString(self.params_values), 'STRING')
#				print numsToString(self.params_values)
		# Now add a property/properties containing the values that are needed by the billboard visualization script:

		self.addPreviewProperties(xp_lamp_ob)

		scn.objects.active = xp_lamp_ob
		scn.objects.selected = [xp_lamp_ob]
		if self.draw_bbs:
			self.light_obs.append(xp_lamp_ob)

		return xp_lamp_ob

	def addPreviewProperties(self, XPLampOb):

		for prt in self.light_parts:
			if isinstance(prt, Billboard):
				x1, y1, x2, y2 = prt.getUVs(1)
				propnames = [p.name for p in XPLampOb.getAllProperties()]
				avail = 'XPToolsBBPrms'
				inc = 1
				while avail in propnames and inc < 10:
					avail = 'XPToolsBBPrms%d' % inc
					inc += 1
				XPLampOb.addProperty(avail, numsToString([prt.data.r, prt.data.g, prt.data.b, prt.data.bright, prt.data.size/2.0,
																	x1, y1, x2, y2,
																	prt.data.x, prt.data.y, prt.data.z, prt.data.width,
																	prt.data.freq, prt.data.phase, prt.data.amp, prt.data.day]), 'STRING')

	def updateSelectedXPLampsWithCurrent(self):

		"""
			Updates selected blender lamps to the definition currently in the editor
		"""

		scn = Blender.Scene.GetCurrent()
		obs = scn.objects.selected

		# Sanity checks:
		if not obs:
			PupMenu("ERROR: No objects are selected!")
			return
		rep = PupMenu("Overwrite all selected lights with the current settings?:%t|    NO|  YES")
		if rep != 2:
			return

		# Generate PARAMS values for the current definition (we do this only once for speed):
		self.generateParamsValues()

		for ob in obs:
			self.overwriteXPlaneLampWithCurrent(ob)

	def overwriteXPlaneLampWithCurrent(self, ob):
		"""
			Rename the passed-in ob to the current editor definition name, if different.
			Delete, add or modify PARAMS property, as required.
			Delete, add or modify XPToolsBBPrms property, as required.
		"""
		if ob.getData().type == Blender.Lamp.Types["Lamp"]:
			try:
				obname = strip_suffix(ob.name)
				props = ob.getAllProperties()
				for p in props:
					if p.name.upper() == 'NAME':
						obname = p.data
						break

				# Change the old name to the new one, if different:
				if self.name != obname:
					ob.name = self.name
					for p in props:
						if p.name.upper() == 'NAME':
							p.data = self.name

				# Is the new def parameterized?
				if self.params_values:
					for p in props:
						if p.name.upper() == 'PARAMS':
							p.data = numsToString(self.params_values)
							break
					else:
						ob.addProperty('PARAMS', numsToString(self.params_values), 'STRING')
				else:
					# remove any existing PARAMS property if our new def isn't parameterized:
					for p in props:
						if p.name.upper() == 'PARAMS':
							ob.removeProperty(p)

				# Clear any existing billboard preview properties and add new ones:
				for p in props:
					if 'XPToolsBBPrms' in p.name:
						ob.removeProperty(p)
				self.addPreviewProperties(ob)

			except:
				print 'failed to update light %s\n' % obname

	def	loadLight(self, lightName, event):

		""" This attempts to load a full XP light definition, including both spill and billboard types and multiples thereof.
			'lightName' is ... the name of the light which we are passed
			'event' is the button event that started the process. It tells us whether we need to go looking for existing param values or not.
		"""

		# Reading the basic def from lights
		#
		# if mode BILLBOARD - 	look for billboard defs, load first one we find into self.billboard
		#						look for any others that may exist, if we find any, copy first def to a new instance in self.light_parts
		#						then load the next one into a second instance in self.light_parts. repeat till exhausted
		#						then look for SPILL lights of the same name and repeat the process for spills

		self.multiple_defs[:]   = []
		self.light_parts[:]		= []
		self.has_multi			= 0
		self.name				= lightName

		# Find the indices of all occurrences of lightName in lights.txt which are either a spill def or a billboard def (but not a param def)

		for l in Billboard.lights_data:			# we can use this for both billboards and spill : it's just a copy of lights.txt
			tks = l.split()
			if tks:
				if len(tks) > 3:
					if tks[1] == lightName and tks[0] in Billboard.types:
						if len(tks) == Billboard.type_tok_count[tks[0]]:
							self.multiple_defs.append(['BILLBOARD', Billboard.lights_data.index(l)])
					elif tks[1] == lightName and tks[0] in SpillLight.types:
						if len(tks) == SpillLight.type_tok_count[tks[0]]:
							self.multiple_defs.append(['SPILL', Billboard.lights_data.index(l)])
					elif tks[1] == lightName and tks[0] == 'LIGHT_PARAM_DEF':
						self.params_list[:] = tks[3:]

		self.multiple_defs.sort()
		#print 'multiple defs: ', self.multiple_defs
		# Make a list of bb or spill objects, one per def

		self.has_multi = 0 if len(self.multiple_defs) <= 1 else 1

		for i, [cls, ind] in enumerate(self.multiple_defs):
			tks = Billboard.lights_data[ind].split()
			if cls == 'BILLBOARD':
				newbb = Billboard()
				newbb.setDataFromFile(ind)
				newbb.sub_def_index = i
				self.light_parts.append(newbb)
			elif cls == 'SPILL':
				newsp = SpillLight()
				newsp.sub_def_index = i
				newsp.setDataFromFile(ind)
				self.light_parts.append(newsp)

		# If we have multi parts, create a menu to allow the user to choose which part to edit:

		if self.has_multi:
			self.multi_menu_string = "light part%t"
		 	for j, [cls, ind] in enumerate(self.multiple_defs):
		 		if cls == 'BILLBOARD':
			 		self.multi_menu_string += "|part %d: %s%%x%d" % (j+1, 'halo', j)
		 		elif cls == 'SPILL':
			 		self.multi_menu_string += "|part %d: %s%%x%d" % (j+1, 'spill', j)

		# ... and copy the first sub-def which matches the current editor mode into the editor :

		found_match = FALSE
		if XPlaneLightEditor.mode == 'BILLBOARD':
			for i, [cls, ind] in enumerate(self.multiple_defs):
				for part in self.light_parts:
					if isinstance(part, Billboard):
						found_match = TRUE
						self.billboard.copyFrom(part)
						self.sub_def_index  = part.sub_def_index
						self.mn_multi_menu.val = part.sub_def_index
						break
		elif XPlaneLightEditor.mode == 'SPILL':
			for part in self.light_parts:
				if isinstance(part, SpillLight):
					found_match = TRUE
					self.spill.copyFrom(part)
					self.sub_def_index  = part.sub_def_index
					self.mn_multi_menu.val = part.sub_def_index
					break

		# ... or switch modes if one doesn't exist:
		if not found_match:
			if isinstance(self.light_parts[0], Billboard):
				XPlaneLightEditor.mode = 'BILLBOARD'
				self.billboard.copyFrom(self.light_parts[0])
				self.mn_multi_menu.val = self.light_parts[0].sub_def_index
				self.sub_def_index	= self.light_parts[0].sub_def_index
			elif isinstance(part, SpillLight):
				XPlaneLightEditor.mode = 'SPILL'
				self.spill.copyFrom(part)
				self.mn_multi_menu.val = self.light_parts[0].sub_def_index
				self.sub_def_index	= self.light_parts[0].sub_def_index

		self.buildTypeMenuString()
		self.buildFilterMenuString()
		self.buildNameMenuString()

		if XPlaneLightEditor.mode   == 'BILLBOARD':
			self.mn_type_menu.val 	= Billboard.types.index(self.billboard.type)
			self.mn_name_menu.val	= self.billboard.name_enum

		elif XPlaneLightEditor.mode == 'SPILL':
			self.mn_type_menu.val 	= SpillLight.types.index(self.spill.type)
			self.mn_name_menu.val	= self.spill.name_enum

		self.refreshPanelValues()

	def swapParts(self, subDefIndex):

		if subDefIndex >= len(self.light_parts):
			print 'list index out of range, swapParts(%d)', subDefIndex
			return
		if len(self.light_parts) <= 1:
			print "ERROR: swapParts() - only one part in list: can't swap"
			return

		# Copy the current editor settings to the appropriate item of self.light_parts:

		if isinstance(self.light_parts[self.sub_def_index], Billboard):
			self.light_parts[self.sub_def_index].copyFrom(self.billboard)

		elif isinstance(self.light_parts[self.sub_def_index], SpillLight):
			self.light_parts[self.sub_def_index].copyFrom(self.spill)

		# Now update the editor sub_def_index to reflect the menu selection:

		self.sub_def_index = subDefIndex

		if isinstance(self.light_parts[self.sub_def_index], Billboard):

			self.tg_spill.val = 0
			self.tg_billboards.val = 1
			XPlaneLightEditor.mode = 'BILLBOARD'

			#print 'XPlaneLightEditor.mode: ',  XPlaneLightEditor.mode
			self.billboard.copyFrom(self.light_parts[self.sub_def_index])

			if not self.unsaved_def:
				try:
					self.billboard.name_enum 	= Billboard.names[self.billboard.type].index(self.name)
					self.buildTypeMenuString()
					self.mn_type_menu.val 		= Billboard.types.index(self.billboard.type)
					self.buildFilterMenuString()
					for cat, name_enums in Billboard.name_groups[self.billboard.type].iteritems():
						if self.billboard.name_enum in name_enums:
							self.mn_filter_menu.val = Billboard.filter_cats.index(cat)
							break
					self.buildNameMenuString()
					self.mn_name_menu.val		= self.billboard.name_enum
				except:
					print "Cannot find name_enum - new def must be added to light.txt!"
			else:
				if self.billboard.type in Billboard.types:
					self.mn_type_menu.val 		= Billboard.types.index(self.billboard.type)

		elif isinstance(self.light_parts[subDefIndex], SpillLight):

			self.tg_spill.val = 1
			self.tg_billboards.val = 0
			XPlaneLightEditor.mode = 'SPILL'

			self.spill.copyFrom(self.light_parts[self.sub_def_index])

			if not self.unsaved_def:
				try:
					self.spill.name_enum 	= SpillLight.names[self.spill.type].index(self.name)
					self.buildTypeMenuString()
					self.mn_type_menu.val 		= SpillLight.types.index(self.spill.type)
					self.buildFilterMenuString()
					for cat, name_enums in SpillLight.name_groups[self.spill.type].iteritems():
						if self.spill.name_enum in name_enums:
							self.mn_filter_menu.val = SpillLight.filter_cats.index(cat)
							break
					self.buildNameMenuString()
					self.mn_name_menu.val		= self.spill.name_enum
				except:
					print "Cannot find name_enum - new def must be added to light.txt!"
			else:
				if self.spill.type in SpillLight.types:
					self.mn_type_menu.val 		= SpillLight.types.index(self.spill.type)

		self.refreshPanelValues()
		Draw()

	def makeLightDef(self):

		"""
			Output the current light settings in a format which can be copy-pasted into lights.txt.
		"""
		#
		# Create/locate a text object for new light definitions output
		#

		if not self.defs_out:
			try:
				self.defs_out = Blender.Text.Get("LightDefsOut.txt")

			except:
				self.defs_out = Blender.Text.New("LightDefsOut.txt")
			self.defs_out.write("\n#" + "_" * 170 + "\n\n")
		# Make a LIGHT_PARAM_DEF if we need one. Note that in the case of a multi-part light, any given var may conceivably have more than one keyname -
		# e.g. data.size may be parameterized as 'SIZE' for the billboard part, and 'THROW' for the spill part, so we need to build our list from params.values(), not params.keys()
		# However, we don't want DUPLICATES of VALUES, so we need to check for those as we go along.

		outstr = []
		paramkeys = []

		for part in self.light_parts:
			for p in XPlaneLightEditor.allparams[part.type]:
				paramkey = part.params[p]
				if paramkey != None and paramkey not in paramkeys:
					paramkeys.append(paramkey)

		if paramkeys:
			paramdef = "LIGHT_PARAM_DEF\t\t%s" % self.name
			paramdef += "\t" * ((39 - len(self.name))/8)
			paramdef += "%d\t" % len(paramkeys)
			for k in paramkeys:
				paramdef += "%s\t" % k
			outstr.append(paramdef)

		# Now get a def string from each of the light's parts

		for part in self.light_parts:
			outstr.append(part.makeNewDefString())
		for o in outstr:
			self.defs_out.write(o + '\n')
		self.defs_out.write('\n\n')

		# Now that we've exported the new def, we can turn off the 'unsaved def' flag.
		self.unsaved_def = False

	def renameLight(self):
		'''
		Called when the self.st_new_light_name button is touched.

		Check if new value is different from current name (it might not have actually been changed) and if yes, copy the new value to all light parts.
		'''

		newname = Create(self.name)
		block = []
		block.append(("Name: ", newname, 0, 32, "Enter a new name for this light"))
		retval = PupBlock('New light name', block)
		if retval == 1:
			if newname.val == self.name:
				return False
			else:
				self.name = newname.val
				self.billboard.name = self.name
				self.spill.name = self.name
				for part in self.light_parts:
					part.name = self.name
				Draw()
				return True
		return False

	def newPartDialogue(self, event):
		'''
		A pop-up dialog to allow the user enter the type for a new light part
		'''
		types = ['BILLBOARD_HW', 'BILLBOARD_SW',  'SPILL_HW_DIR', 'SPILL_HW_FLA', 'SPILL_SW']
		if event == self.evt_billboards:
			name = "select type:%%t|%s%%x0|%s%%x1" % (types[0], types[1])
		elif event == self.evt_spill:
			name = "select type:%%t|%s%%x2|%s%%x3|%s%%x4" % (types[2], types[3], types[4])

		selection =	PupMenu(name)
		return None if selection == -1 else types[selection]

	def addNewPart(self, newType):
		'''
		In dev mode, shift-clicking on "SPILL' or 'HALOS' creates a new part, with the same name as that currently in the editor.

		If one exists, the new part will be a copy of the first existing part of the same class as that requested.
		Otherwise, the new part is set to default settings.
		'''

		for part in self.light_parts:
			if part.type == newType:
				newpart = part.copy()
				break
		else:
			if newType in Billboard.types:
				newpart = Billboard()
			elif newType in SpillLight.types:
				newpart = SpillLight()
			newpart.name = self.name
			newpart.type = newType
		newpart.sub_def_index = len(self.light_parts)

		self.multiple_defs.append(['BILLBOARD' if isinstance(newpart, Billboard) else 'SPILL', newpart.data_index])
		self.has_multi	=	True
		self.light_parts.append(newpart)

		# update the multi-part menu string:
		self.multi_menu_string = "Multi-Part%t"

		for j, [cls, ind] in enumerate(self.multiple_defs):
			if cls == 'BILLBOARD':
				self.multi_menu_string += "|part %d: %s%%x%d" % (j+1, 'halo', j)
			elif cls == 'SPILL':
				self.multi_menu_string += "|part %d: %s%%x%d" % (j+1, 'spill', j)
		self.unsaved_def = True
		self.refreshPanelValues()

	def getParamsFromBlenderProp(self, obName, prop):

		tks = []
		arg_list = []			# arg_list will contain the XP parameterized light parameter key strings.

		for line in Billboard.lights_data:
			if line.startswith('LIGHT_PARAM_DEF'):
				tks = line.split()
				if len(tks) > 3 and tks[1] == self.name:
					arg_list = tks[3:]
					break

		# It seems that there are a few LIGHT_PARAM_DEFS in lights.txt with unused params -
		# i.e. the param name appears in LIGHT_PARAM_DEF but not in any of BILLBOARD_HW, SPILL_HW_DIR etc
		# So run a check on the param identifiers that we actually USE.

		arg_list_copy = arg_list[:]
		for part in [part for part in self.light_parts if part.has_params]:
			for k in vars(part.params).keys():
				if part.params[k] in arg_list_copy:
					arg_list_copy.remove(part.params[k])

		if arg_list_copy:
			for item in arg_list_copy:
				arg_list.remove(item)

		# Now get the actual values from the Blender 'params' prop
		if prop.type != 'STRING' or prop.name.upper() != 'PARAMS':
			PupMenu("ERROR: %s Cannot read property" % obName)
			print "ERROR: %s Cannot read property" % obName
			return 0
		values = self.string2Values(prop.data)
		#print 'values:', values
		if not values:
			return 0
		if len(values) != len(arg_list):
			PupMenu("ERROR: %s Wrong number of args in 'params' property!" % obName)
			print "ERROR: %s Wrong number of args in 'params' property!\n\n" % obName
			return 0

 		# Now loop round the (potenitally) multiple sub defs and try to match our param key strings to the sub-def param strings.
 		# If we succeed, we copy over the respective value

		for i, arg in enumerate(arg_list):
			for part in [part for part in self.light_parts if part.has_params]:
				for k in vars(part.params).keys():
					if part.params[k] == arg:
						part.data[k]  = values[i]

		# Then copy the data from whichever part is currently also in the editor.

		if isinstance(self.light_parts[self.sub_def_index], Billboard):
			self.billboard.data.copyFrom(self.light_parts[self.sub_def_index].data)

		elif isinstance(self.light_parts[self.sub_def_index], SpillLight):
			self.spill.data.copyFrom(self.light_parts[self.sub_def_index].data)
		self.refreshPanelValues()

	def updateExistingLightsForPreviews(self):
		'''
		Iteratively check all blender lamp objects in the current scene and if necessary update them to be compatible with XPTools 3D previews.

		1. Check if "Lamp' type lamp
		2. if yes, check for presence of XPToolsBBPrms prop(s) on ob
		3. - if yes, we're good: move on to next lamp
		2. - if no, check if it's a known XPlane billboard named light.
		4. -- if yes, load all sub_defs and write necessary props
		'''
		scn = bpy.data.scenes.active
		rejected = []

		for ob in [ob for ob in scn.objects if ob.type == 'Lamp' and strip_suffix(ob.name) not in ['PIN', 'PINSTOP_PT', 'PINANCHOR_PT', 'PINPIVOT_LEAD_PT', 'PINPIVOT_TRAIL_PT', 'PINGROUND_PT']]:
			if ob.getData().type == Blender.Lamp.Types["Lamp"]:
				try:
					done = 0
					obname = strip_suffix(ob.name)
					props = ob.getAllProperties()
					if props:
						for p in props:
							if 'XPToolsBBPrms' in p.name:
								done = 1
							# if the lamp has a 'name' property, use it!
							if p.name.upper() == 'NAME':
								obname = p.data

					if not done:
						# check for validity of name. Keep a list of rejects for debug.
						if self.billboard.hasName(obname):
							self.readXPlaneLightFromBlender(ob)
							self.addPreviewProperties(ob)
						else:
							#print 'bp10'
							rejected.append(obname)
				except:
					print 'failed to update light %s\n' % obname

	def readXPlaneLightFromBlender(self, lampOb=None):
		'''
		Attempts to load the passed-in blender object into the editor. If called with no arg, it defaults to the active scene object.
		'''
		if not lampOb:
			ob = Blender.Scene.GetCurrent().objects.active
			if not ob:
				PupMenu("ERROR: No active object!")
				return None
			if ob.type != 'Lamp':
				PupMenu("ERROR: The active object is not a 'Lamp'!")
				return None
		else:
			ob = lampOb
		try:
			# if the lamp has a 'name' property, use it! Else strip any .00n goo from the OB name and hope for the best:
			props = ob.getAllProperties()
			obname = strip_suffix(ob.name)
			for p in props:
				if p.name.upper() == 'NAME':
					obname = p.data
					break

			self.name = obname
			self.loadLight(obname, None)
			for p in props:
				if p.name.upper() == 'PARAMS':
					self.getParamsFromBlenderProp(ob.name, p)
					break
		except:
			print "ERROR: %s Couldn't get values from prop!" % ob.name

	def	createVisLampObDef(self, lightname, obparams, ob, datablock):
		"""
			Build a vis_lamp_ob_def from the passed-in args:
			name: name of the X-Plane light definition
			params: x-plane 'params' args string
			ob:    the vis_loc_ob to load
			datablock: the lamp datablock to wrap with the def
		"""
		if not self.spill.hasName(lightname):
			return False
		defname = (lightname + ' ' + obparams).rstrip()

		self.readXPlaneLightFromBlender(ob)

		datablock.type 		= Lamp.Types['Lamp'] if self.omni else Lamp.Types['Spot']
		datablock.col 		= self.sRGB_for_cp
		datablock.dist		= self.spill.data.size
		datablock.energy	= linearToSRGB(self.spill.data.bright) * XPlaneLightEditor.SPILL_BRIGHTNESS_GAIN
		datablock.spotBlend = 1.0
		datablock.spotSize	= self.width_deg
		datablock.clipEnd	= self.spill.data.size
		datablock.mode		|= Lamp.Modes["RayShadow"]
		datablock.mode		|= Lamp.Modes["Shadows"]
		datablock.mode		|= Lamp.Modes["Sphere"]

		rotx = radians(self.nb_pitch.val + 90.0)
		rotz = radians(-self.nb_heading.val)

		# Create a vis_lamp_ob_def wrapper and add it to the list:
		self.vis_lamp_ob_defs[defname] = VisLampDef(defname, rotx, rotz, datablock)
		return True

	def liveUpdate(self, buttonEvent):
		"""
			Updates the settings of the active object if it is a lamp with name 'XPToolsVisLpObRT'
			When the user hits 'preview at cursor', an 'orphan' vis_lamp_ob based on the current editor settings is added at the cursor location.
			Subsequent tweaks of the GUI are written to the lamp by this function in realtime.
		"""
		#scn = Blender.Scene.GetCurrent()

		try:
			ob_act = Object.Get('XPToolsVisLpObRT')
		except:
			return

		dat = ob_act.getData()

		if buttonEvent == self.evt_omni:
			dat.type = Lamp.Types['Spot'] if not self.omni else Lamp.Types['Lamp']

		elif buttonEvent in [self.evt_colourpicker, self.evt_sR_int, self.evt_sG_int, self.evt_sB_int]:
			dat.col	= self.sRGB_for_cp

		elif buttonEvent == self.evt_size:
			dat.dist = self.spill.data.size
			dat.clipEnd = dat.dist

		elif buttonEvent == self.evt_bright:
			dat.energy = linearToSRGB(self.spill.data.bright) * XPlaneLightEditor.SPILL_BRIGHTNESS_GAIN

		elif buttonEvent == self.evt_width_deg:
			dat.spotSize = self.width_deg

		elif buttonEvent in [self.evt_vector, self.evt_pitch, self.evt_heading]:
			ob_act.RotX	= radians(self.nb_pitch.val + 90.0)
			ob_act.RotZ	= radians(-self.nb_heading.val)

		Window.RedrawAll()

	def nukeVisLamps(self):

		scn = Blender.Scene.GetCurrent()
		try:
			vis_lamp_group = Group.Get('XPToolsVisLampObs')
			vis_lamp_group.objects = []
		except (NameError, ValueError):
			print "No group 'XPToolsVisLampObs' - doing nothing"

		while self.vis_lamp_obs:
			unlinked = self.vis_lamp_obs.pop()
			print "Unlinking:", unlinked.name
			scn.objects.unlink(unlinked)

		if self.vis_loc_obs_selected:
			scn.objects.selected = self.vis_loc_obs_selected
			self.vis_loc_obs_selected[:] = []
		self.vis_loc_obs[:] = []

		for ob in [ob for ob in scn.objects if "XPToolsVisLpOb" in ob.name]:
			scn.objects.unlink(ob)

		self.vis_lamp_ob_defs.clear()

		try:
			vis_lamp_group = Group.Get('XPToolsVisLampObs')
			vis_lamp_group.objects = []
		except (NameError, ValueError):
			print "No group 'XPToolsVisLampObs' - doing nothing"

		Window.RedrawAll()

	def cleanUpVisLampsFromScene(self):
		"""
		Remove any stray vislamps before quitting
		"""
		scn = Blender.Scene.GetCurrent()
		for ob in scn.objects:
			if ob.type == "Lamp" and 'XPToolsVisLpOb' in ob.name:
				#self.vis_lamp_obs_free.append(ob)
				scn.objects.unlink(ob)
		#scn.update()
		Window.RedrawAll()

	def stripBBProps(self):
		'''
		Strips properties with name 'XPToolsBBPrms' from all lights in scene
		'''
		total_removed = 0
		obs_cleaned = 0
		for ob in [ob for ob in bpy.data.scenes.active.objects if ob.type == 'Lamp']:
			props_removed = 0
			for p in ob.getAllProperties():
				if 'XPToolsBBPrms' in p.name:
					ob.removeProperty(p)
					props_removed += 1
			if props_removed:
				total_removed += props_removed
				obs_cleaned += 1
		print "A total of %d properties were stripped from %d lamp objects.\n" % (total_removed, obs_cleaned)

	def cleanUpBBGooFromScene(self):
		scn = bpy.data.scenes.active
		try:
			self.dimBlenderScene(False)
			scn.clearScriptLinks(['XPToolsSL.py'])
			Blender.Text.unlink(self.scriptlink)
		except:
			pass
		Window.RedrawAll()

	def inBoth(self, listA, listB):
		"""
			Returns True if listA and listB have at least one common element, else False.
		"""
		for a in listA:
			if a in listB:
				return True
		return False

	def makeSuffixList(self, objList):
		suffix_list = []
		for ob in objList:
			e = ob.name.rfind('.')
			if e == -1:
				suffix_list.append(0)
			else:
				suffix_list.append(int(ob.name[e:]))
		suffix_list.sort()
		return suffix_list

	def nextAvailableSuffix(self, suffixList, previous):

		previous += 1
		while previous in suffixList:
			previous += 1
		return previous

	def addMultipleVisLamps(self, locType):

		scn = Blender.Scene.GetCurrent()

		if self.light_ob_layer not in scn.layers:
			layers = scn.layers
			layers.append(self.light_ob_layer)
			scn.layers = layers
		layers_set = set(scn.layers)
		Window.SetActiveLayer(1<<(self.light_ob_layer - 1))

		# find any previously deleted vis_lamp_obs for recycling:
		all_vis_lamp_obs = set([ob for ob in Blender.Object.Get() if ob.name.startswith('XPToolsVisLpOb')])
		scn_vis_lamp_obs = set([ob for ob in scn.objects if ob.name.startswith('XPToolsVisLpOb')])
		free_vis_lamp_obs = list(all_vis_lamp_obs - scn_vis_lamp_obs)
		free_vis_lamp_obs.sort()
		# find any previously deleted vis_lamp datablocks for recycling:
		free_vis_lamp_datablocks = [blk for blk in Blender.Lamp.Get() if blk.name.startswith('XPToolsVisLamp') and not blk.users]

		# These are the 'real' X-Plane lights that we're going to simulate by adding renderable, wireframe vis_lamp_obs at each of their locations:
		self.vis_loc_obs[:] = []
		self.vis_loc_obs_selected = list(scn.objects.context)

		print '\n'
		print 'all_vis_lamp_obs: ', len(all_vis_lamp_obs)
		print 'scn_vis_lamp_obs: ', len(scn_vis_lamp_obs)
		print 'free_vis_lamp_obs:', len(free_vis_lamp_obs)
		print 'free_vis_lamp_datablocks:' , len(free_vis_lamp_datablocks)
		all_vis_lamp_obs.clear()
		scn_vis_lamp_obs.clear()

		if locType == 'cursor':

			vis_lamp_ob = lampdat = None
			try:
				vis_lamp_ob = Object.Get('XPToolsVisLpObRT')
				lampdat = vis_lamp_ob.data
				try:
					scn.objects.link(vis_lamp_ob)
				except:
					print "%s already in scene" % vis_lamp_ob.name
			except:
				if free_vis_lamp_datablocks:
					lampdat = free_vis_lamp_datablocks.pop()
				else:
					lampdat = Blender.Lamp.New('Lamp', 'XPToolsVisLamp')
				if free_vis_lamp_obs:
					vis_lamp_ob = free_vis_lamp_obs.pop()
					vis_lamp_ob.link(lampdat)
					try:
						scn.objects.link(vis_lamp_ob)
					except:
						print "%s already in scene" % vis_lamp_ob.name
				else:
					vis_lamp_ob 	= scn.objects.new(lampdat)
				vis_lamp_ob.name = 'XPToolsVisLpObRT'

			lampdat.type 		= Lamp.Types['Lamp'] if self.omni else Lamp.Types['Spot']
			lampdat.col 		= self.sRGB_for_cp
			lampdat.dist		= self.spill.data.size
			lampdat.energy		= linearToSRGB(self.spill.data.bright) * XPlaneLightEditor.SPILL_BRIGHTNESS_GAIN
			lampdat.spotBlend	= 1.0
			lampdat.spotSize	= self.width_deg
			lampdat.clipEnd		= self.spill.data.size
			lampdat.mode		|= Lamp.Modes["RayShadow"]
			lampdat.mode		|= Lamp.Modes["Shadows"]
			lampdat.mode		|= Lamp.Modes["Sphere"]

			vis_lamp_ob.layers	= [self.light_ob_layer]
			vis_lamp_ob.loc 	= tuple(Window.GetCursorPos())
			vis_lamp_ob.RotX	= radians(self.nb_pitch.val + 90.0)
			vis_lamp_ob.RotZ	= radians(-self.nb_heading.val)
			scn.objects.active 	= vis_lamp_ob
			vis_lamp_ob.sel	= 1
			self.vis_lamp_obs.append(vis_lamp_ob)

		else:
			# Clear the vis lamp objects list:
			self.vis_lamp_obs[:] = []
			allobs = []
			if locType == 'selected':
				allobs = scn.objects.context
			elif locType == 'all':
				allobs = scn.objects

			scn.objects.selected = []

			for ob in allobs:
				if ob.type != 'Lamp' or ob.name.startswith('PIN') or (not set(ob.layers) & layers_set):
					continue
				else:
					self.vis_loc_obs.append(ob)
#			print 'Allobs:', len(allobs)
#			print 'self.vis_loc_obs:', len(self.vis_loc_obs)
#			print '\n'

			for ob in self.vis_loc_obs:
			#	try:
				obname = strip_suffix(ob.name)
				obparams = ""
				prps = ob.getAllProperties()
				for prp in prps:
					if prp.name.upper() == 'NAME':
						obname = prp.data
					elif prp.name.upper() == 'PARAMS':
						obparams = prp.data

				defname = (obname + ' ' + obparams).rstrip()

				if defname in self.vis_lamp_ob_defs:
					print defname, 'in self.vis_lamp_ob_defs'
				else:
					if self.spill.hasName(obname):
						if free_vis_lamp_datablocks:
							dat = free_vis_lamp_datablocks.pop()
						else:
							dat = Blender.Lamp.New('Lamp', 'XPToolsVisLamp')
						ret = self.createVisLampObDef(obname, obparams, ob, dat)
						if not ret:
							print "ERROR: createVisLampObDef() returned False for %s" % obname
							continue
						else:
							print "created new def for %s" % defname
					else:
						print 'This is a billboard def - ignoring'
						continue

				if free_vis_lamp_obs:
					vis_lamp_ob = free_vis_lamp_obs.pop()
					vis_lamp_ob.link(self.vis_lamp_ob_defs[defname].data)
					try:
						scn.objects.link(vis_lamp_ob)
					except:
						print "%s already in scene" % vis_lamp_ob.name

				else:
					vis_lamp_ob = scn.objects.new(self.vis_lamp_ob_defs[defname].data)
					vis_lamp_ob.name	= 'XPToolsVisLpOb'

				# For some reason, writing to object.layers takes about 0.2 seconds, so we test first and only update if necessary:
				if vis_lamp_ob.layers != [self.light_ob_layer]:
					vis_lamp_ob.layers	= [self.light_ob_layer]
				# give it the original untransformed rotations from the light def:
				vis_lamp_ob.RotX	= self.vis_lamp_ob_defs[defname].rot_x
				vis_lamp_ob.RotZ	= self.vis_lamp_ob_defs[defname].rot_z

				# Now transform by the matrix of the vis_loc_ob's parent, if any:
				if ob.parent:
					vis_lamp_ob.setMatrix(vis_lamp_ob.getMatrix() * ob.parent.getMatrix().rotationPart().resize4x4())

				vis_lamp_ob.loc = ob.getLocation('worldspace')	# this takes into account any offset courtesy of a possible parent.

				self.vis_lamp_obs.append(vis_lamp_ob)
#				print 'added ob:', vis_lamp_ob.name, '\n'
				ob.select(False)
 			#	except:
			#		print 'Failed to load spill light %s, ignoring.\n' % obname

		# Add the vis_lamp_obs to a group. This group is then specified in the rendering context, which means that ONLY this lamps will take part in the render.
		# Otherwise, we get random amounts of light coming from all the other lamps, pins etc in the scene.
		vlo = []
		for ob in self.vis_lamp_obs:
			vlo.append(ob.name)
		vlo.sort()
# 		for ob in vlo:
# 			print ob, Blender.Object.Get(ob).data.name

		try:
			vis_lamp_group = Group.Get('XPToolsVisLampObs')
		except (NameError, ValueError):
			vis_lamp_group = Group.New('XPToolsVisLampObs')
		vis_lamp_group.objects = self.vis_lamp_obs

# 			print "scn.objects:\n"
# 			for ob in scn.objects:
# 				if ob.type == 'Lamp':
# 					print ob.name, ob.loc, ob.layers
# 			print '\nAll objects:\n'
# 			for ob in Blender.Object.Get():
# 				if ob.type == 'Lamp':
# 					print ob.name
#			print '\nAll lamp datablocks:\n'
# 			for lam in Blender.Lamp.Get():
# 				print lam.name, lam.type

		Window.RedrawAll()

	def addRenderMatToObs(self, selected=False):
		"""
			Adds a temporary basic material to selected/all mesh objects to enable the user
			 to perform a test render to check the appearance of spill lights.
		"""

		# Check if the material already exists - if not, create a new one:
		mats = Material.Get()
		for m in mats:
			if m.name == "XPToolsBasicMat":
				break
		else:
			m = Material.New('XPToolsBasicMat')
			m.mode |= Material.Modes.TEXFACE_ALPHA
			m.mode |= Material.Modes.TEXFACE
			m.mode |= Material.Modes.SHADOW
			m.mode |= Material.Modes.TRACEABLE
			m.mode |= Material.Modes.RAYBIAS
			m.mode |= Material.Modes.ZTRANSP
			m.mode &= ~Material.Modes.SHADOWBUF
			m.spec = 0.0
			m.hard = 1
			m.ref  = 1.0
			m.amb  = 0.0

		scn = Blender.Scene.GetCurrent()
		mat = Material.Get('XPToolsBasicMat')

		# set which objects to operate on, based on the optional parameter 'selected'
		# By default, operate on ALL mesh objs in the scene.

		if selected:
			obs = scn.objects.selected
		else:
			obs = scn.objects

		for ob in [ob for ob in obs if ob.type == 'Mesh']:
			if mat not in ob.getData(mesh=1).materials and len(ob.getData(mesh=1).materials) < 16:
				ob.getData(mesh=1).materials += [mat]

		Window.RedrawAll()

	def nukeRenderMat(self):
		"""
			Removes the temporary render material from all objects in the scene.
		"""
		scn = Blender.Scene.GetCurrent()

		# Does the material exist? If not, there's no need to remove it!
		mats = Material.Get()
		for m in mats:
			if m.name == "XPToolsBasicMat":
				break
		else:
			return

		mat = Material.Get('XPToolsBasicMat')

		for ob in [ob for ob in scn.objects if ob.type == 'Mesh']:
			try:
				mat_list = ob.getData(mesh=1).materials
				if mat in mat_list:
					mat_list.remove(mat)
					#print "new mat_list:", mat_list
					ob.getData(mesh=1).materials = mat_list
			except:
				print "Couldn't get material from object %s" % ob.name

		Window.RedrawAll()

	def copyRenderContext(self):
		"""
			Build a dict containing all the writable vars and values of the current rendering context.
		"""

		scn = Blender.Scene.GetCurrent()
		ren_con = scn.getRenderingContext()

		if self.context_backup == {}:
			for k, v in self.context_settings.iteritems():
				self.context_backup[k] = v

		# copy the active values to the backup dict:
		for k, v in self.context_backup.iteritems():
			try: v = ren_con.__getattribute__(k)
			except:
				print "copyRenderContext can't copy <%s %s> to backup" % (k, repr(v))

	def configureRenderContext(self, sourceDict):
		"""
			Configure the blender render context.
		"""

		if not self.context_backup:
			PupMenu("Context has not been backed-up.%t")
			return
		scn = Blender.Scene.GetCurrent()

		# Get the current context
		context = scn.getRenderingContext()

		# Write new settings to it:
		for k, v in sourceDict.iteritems():
			try:
				context.__setattr__(k, v)
			except:
				print "********* config can't set %s %s" % (k, repr(v))

	def setupVisLampRender(self):
		"""
		Setup the rendering context for test render of spill lights
		"""
		# Find the ID of the largest 3d window:
		v3d = Window.GetScreenInfo(Window.Types.VIEW3D)
		win_area = []
		for v in v3d:
			x1, y1, x2, y2 = v['vertices']
			win_area.append((x2 - x1) * (y2 - y1))
		idx = win_area.index(max(win_area))
		self.win_id = v3d[idx]['id']

		# Check for an existing XPToolsCamera, if none, add one:
		scn = Blender.Scene.GetCurrent()
		try:
			self.render_cam_data = Camera.Get('XPToolsCamData')
		except (NameError, ValueError):
			self.render_cam_data = Camera.New('persp', 'XPToolsCamData')
			self.render_cam_data.lens = 50.0
			self.render_cam_data.clipEnd = 5000.0
			self.render_cam_data.clipStart = 1.0
			self.render_cam_data.drawName = True
		try:
			self.render_cam = Object.Get('XPToolsCamera')
		except:
			self.render_cam = Object.New('Camera', 'XPToolsCamera')
			self.render_cam.link(self.render_cam_data)
		try:
			scn.link(self.render_cam)
		except RuntimeError:
			print "Camera already in scene"

		layer_mask = Window.GetActiveLayer()
		act_layer = 0
		while layer_mask > 0:
			layer_mask = layer_mask>>1
			act_layer += 1
		self.render_cam.layers = [act_layer]

		# Get the existing rendering context and save a copy for restore.
		self.copyRenderContext()

		# Reconfigure the render context to our preferred settings:
		self.configureRenderContext(self.context_settings)

		# Set the aspect ratio of the render image to that of the current 3d view window:
		x1, y1, x2, y2 = v3d[idx]['vertices']
		scn.render.sizeY = int(scn.render.sizeX * float(y2 - y1)/float(x2 - x1))
		Render.EnableDispWin()

		#Add a new render layer to the context if we don't already have one
		for i, lyr in enumerate(scn.render.renderLayers):
			if lyr.name == "XPToolsRenLayer":
				self.render_layer = scn.render.renderLayers[i]
				scn.render.activeLayer = i
				break
		else:
			self.render_layer = scn.render.addRenderLayer()
			for k, v in self.render_layer_settings.iteritems():
				try:
					exec "self.render_layer.%s = v" % k
				except:
					print "setup new layer can't find key %s" % k
			for i, lyr in enumerate(scn.render.renderLayers):
				if lyr.name == "XPToolsRenLayer":
					scn.render.activeLayer = i
					break
			try:
				self.render_layer.lightGroup = Group.Get("XPToolsVisLampObs")
			except (NameError, ValueError):
				self.render_layer.lightGroup = Group.New("XPToolsVisLampObs")

		# Save the current AO state for restore, and turn it off:
		world = Blender.World.GetCurrent()
		if world:
			self.blender_world_mode = world.mode
			world.mode &= ~16

		self.context_configured = True

	def cleanupVisLampRender(self):
		"""
		Return rendering context to original state
		"""

		if not self.context_configured:
			return
		# Remove the XPTools render layer:
		scn = Blender.Scene.GetCurrent()
		for lyr in scn.render.renderLayers:
			if "XPToolsRenLayer" in lyr.name and len(scn.render.renderLayers) > 1:
				scn.render.removeRenderLayer(lyr)

		# Restore the blender world mode (AO status etc).
		if Blender.World.GetCurrent() and self.blender_world_mode:
			Blender.World.GetCurrent().mode = self.blender_world_mode

		# Restore the original render context settings:
		self.configureRenderContext(self.context_backup)

		# Clear our backup copy of the user's original context settings:
		self.context_backup.clear()

		# Close the preview window:
		Render.CloseRenderWindow()

		# Remove the camera:
		if self.render_cam:
			scn.objects.unlink(self.render_cam)

		# Strip the material:
		self.nukeRenderMat()

		self.context_configured = False

		Window.RedrawAll()

	def doVisLampRender(self):
		"""
			Set up the rendering context if it isn't already setup, add a material to all meshes, add a screen-aligned camera and perform a render.
		"""
		scn = Blender.Scene.GetCurrent()

		if not self.context_configured:
			self.setupVisLampRender()
			self.addRenderMatToObs()

		# Save the
		self.ob_act_prev = scn.objects.active

		# Set the camera to the current 3d view position, by clicking the mouse in the largest 3d view pane (if there are more than one).
		try:
			scn.objects.camera = self.render_cam
			layer_mask = Window.GetActiveLayer()
			act_layer = 0
			while layer_mask > 0:
				layer_mask = layer_mask>>1
			act_layer += 1
			self.render_cam.layers = [act_layer]
		except:
			print "No render camera!"
			return
		Window.CameraView(True)
		Window.RedrawAll()
		# do the render:
		context = scn.getRenderingContext()
		context.render()
		scn.objects.active = self.render_cam
		Window.CameraView(False)

		Window.QAdd(self.win_id, PADASTERKEY,1)
		Window.QAdd(self.win_id, PADASTERKEY,0)
		Window.QHandle(self.win_id)
		self.render_cam.layers = []

		Window.QAdd(self.win_id, LEFTMOUSE,0)
		Window.QAdd(self.win_id, RIGHTMOUSE,0)
		Window.QAdd(self.win_id, PAD0,1)
		Window.QAdd(self.win_id, PAD0,0)
		Window.QHandle(self.win_id)

		scn.objects.active = self.ob_act_prev
		Window.RedrawAll()

	def brightnessCallBack(self, evt, val):
		""" We seem to have to have these callbacks in order to pass optional precision-specifying args to Draw.Number()
			At least, I can't figure out how to pass them without getting errors...
		"""
		pass

	def sizeCallBack(self, evt, val):
		""" see above
		"""
		pass

	def dummyCallBack(self, evt, val):
		pass

	def setMode(self):
		self.buildTypeMenuString()
		self.mn_type_menu.val = 0
		if XPlaneLightEditor.mode == 'SPILL':
			self.spill.type = SpillLight.types[self.mn_type_menu.val]
			self.buildFilterMenuString()
			self.mn_filter_menu.val = self.getFirstAvailFilterCat()
			self.buildNameMenuString()
			self.mn_name_menu.val = self.getNextNameMenuEnum(0)
			self.spill.name_enum = self.mn_name_menu.val
			self.spill.name = SpillLight.names[self.spill.type][self.mn_name_menu.val]
			self.name = self.spill.name

		elif XPlaneLightEditor.mode == 'BILLBOARD':
			self.billboard.type = Billboard.types[self.mn_type_menu.val]
			self.buildFilterMenuString()
			self.mn_filter_menu.val = self.getFirstAvailFilterCat()
			self.buildNameMenuString()
			self.mn_name_menu.val = self.getNextNameMenuEnum(0)
			self.billboard.name_enum = self.mn_name_menu.val
			self.billboard.name = Billboard.names[self.billboard.type][self.mn_name_menu.val]
			self.name = self.billboard.name

		self.loadLight(self.name, None)
		self.refreshPanelValues()

	def exitLightEditor(self):

		self.packConfig()
		self.cleanUpVisLampsFromScene()
		self.cleanUpBBGooFromScene()
		self.cleanupVisLampRender()

	def eventHandler(self, evt, val):
		global g_qualifs
		if g_qualifs & Window.Qual.SHIFT and g_qualifs & Window.Qual.CTRL and evt == DKEY and val == 1:
			self.dev_mode = 1 - self.dev_mode
			if self.dev_mode:
				self.buildParamsMenuString()
				self.refreshPanelValues()

		if evt in [LEFTMOUSE, RIGHTMOUSE] and val == 1:
			self.mouse_xy = Window.GetMouseCoords()
			if XPlaneLightEditor.mode == 'BILLBOARD':
				if self.mouseInRect(210,	60,   	144,	110):
					self.mouse_over_CP = 1 - self.mouse_over_CP
					Draw()

	def buttonEvent(self, evt):

		global g_qualifs, DEF_LIGHT_LAYER
		if evt in [self.evt_spill, self.evt_billboards]:

			if self.dev_mode and g_qualifs & Window.Qual.SHIFT:
				want_new_type = self.newPartDialogue(evt)
				if want_new_type:
					self.addNewPart(want_new_type)
				Window.SetKeyQualifiers(0)
				g_qualifs = 0
				Draw()
			elif not self.unsaved_def:

				# Before we change mode, cache the menu vals of the light currently in the editor so we can snap back to it when next we change mode:

				self.packConfig()

				# Now change mode

				if evt == self.evt_spill:
					XPlaneLightEditor.mode = 'SPILL' if self.tg_spill.val else 'BILLBOARD'
					self.tg_billboards.val = 1 - self.tg_spill.val
				elif evt == self.evt_billboards:
					XPlaneLightEditor.mode = 'BILLBOARD' if self.tg_billboards.val else 'SPILL'
					self.tg_spill.val = 1 - self.tg_billboards.val

				self.setMode()

				# Restore any cached name we may have, now that we've switched mode:

				self.unpackMenuListAndLoadLight()

		elif evt == self.evt_type_menu:

			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.type = SpillLight.types[self.mn_type_menu.val]
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.type = Billboard.types[self.mn_type_menu.val]
			self.buildFilterMenuString()
			self.mn_filter_menu.val = self.getFirstAvailFilterCat()
			self.buildNameMenuString()
			self.mn_name_menu.val = self.getNextNameMenuEnum(0)

			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.name_enum = self.mn_name_menu.val
				self.spill.name = SpillLight.names[self.spill.type][self.mn_name_menu.val]
				self.name = self.spill.name

			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.name_enum = self.mn_name_menu.val
				self.billboard.name = Billboard.names[self.billboard.type][self.mn_name_menu.val]
				self.name = self.billboard.name

			self.refreshPanelValues()
			self.loadLight(self.name, evt)
			self.packConfig()

		elif evt in [self.evt_next_name, self.evt_prev_name, self.evt_name_menu]:

			if evt == self.evt_next_name:
				self.mn_name_menu.val = self.getNextNameMenuEnum(1)

			elif evt == self.evt_prev_name:
				self.mn_name_menu.val = self.getNextNameMenuEnum(-1)

			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.name_enum = self.mn_name_menu.val
				self.spill.name = SpillLight.names[self.spill.type][self.mn_name_menu.val]
				self.name = self.spill.name

			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.name_enum = self.mn_name_menu.val
				self.billboard.name = Billboard.names[self.billboard.type][self.mn_name_menu.val]
				self.name = self.billboard.name

			self.loadLight(self.name, evt)
			self.packConfig()

		elif evt == self.evt_filter_menu:

			self.buildNameMenuString()
			self.mn_name_menu.val = self.getNextNameMenuEnum(0)

			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.name_enum = self.mn_name_menu.val
				self.spill.name = SpillLight.names[self.spill.type][self.mn_name_menu.val]
				self.name = self.spill.name
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.name_enum = self.mn_name_menu.val
				self.billboard.name = Billboard.names[self.billboard.type][self.mn_name_menu.val]
				self.name = self.billboard.name
			self.loadLight(self.name, evt)

		elif evt == self.evt_multi_menu:
			self.swapParts(self.mn_multi_menu.val)
			self.packConfig()

		elif evt == self.evt_params_menu:
			self.updateParameterList(self.mn_params_menu.val)

		elif evt == self.evt_colourpicker:

			if XPlaneLightEditor.mode == 'SPILL':
				if all([self.spill.params.r, self.spill.params.g, self.spill.params.b]) or self.dev_mode:
					self.sRGB_for_cp 			= self.cp_colourpicker.val
					self.spill.data.r, self.spill.data.g, self.spill.data.b = sRGBToLinear(self.sRGB_for_cp)
					self.sRGB 					= list(self.sRGB_for_cp)
					self.sRGB_int 				= fltToPS(self.sRGB)
					self.nb_sR_int.val 			= self.sRGB_int[0]
					self.nb_sG_int.val			= self.sRGB_int[1]
					self.nb_sB_int.val 			= self.sRGB_int[2]
					for v in ['r', 'g', 'b']:
						self.refreshMultis(v)

			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.sRGB_for_cp  = self.cp_colourpicker.val
				if all([self.billboard.params.r, self.billboard.params.g, self.billboard.params.b]) or self.dev_mode:
					self.billboard.data.r, self.billboard.data.g, self.billboard.data.b = self.sRGB_for_cp
					self.sRGB 					= list(self.sRGB_for_cp)
					self.sRGB_int 				= fltToPS(self.sRGB)
					self.nb_sR_int.val 			= self.sRGB_int[0]
					self.nb_sG_int.val			= self.sRGB_int[1]
					self.nb_sB_int.val 			= self.sRGB_int[2]
					for v in ['r', 'g', 'b']:
						self.refreshMultis(v)

		#	g_qualifs = Window.GetKeyQualifiers()
			Draw()

		elif evt == self.evt_rgb_normalize:
			self.normalizeRGBValues()

		elif evt == self.evt_rgb_copy:
			self.sRGBCopy()

		elif evt == self.evt_rgb_paste:
			self.sRGBPaste()

		elif evt in [self.evt_sR_int, self.evt_sG_int, self.evt_sB_int]:
			if XPlaneLightEditor.mode == 'SPILL':
				if evt == self.evt_sR_int:
					self.sRGB_int[0] 	= self.nb_sR_int.val
					self.sRGB[0]		= round(float(self.sRGB_int[0]) / 255.0, 4)
					self.spill.data.r 	= sRGBToLinear(self.sRGB[0])
				elif evt == self.evt_sG_int:
					self.sRGB_int[1] 	= self.nb_sG_int.val
					self.sRGB[1]		= round(float(self.sRGB_int[1]) / 255.0, 4)
					self.spill.data.g 	= sRGBToLinear(self.sRGB[1])
				elif evt == self.evt_sB_int:
					self.sRGB_int[2] 	= self.nb_sB_int.val
					self.sRGB[2]		= round(float(self.sRGB_int[2]) / 255.0, 4)
					self.spill.data.b 	= sRGBToLinear(self.sRGB[2])
				self.sRGB_for_cp 		= tuple(self.sRGB)
				self.cp_colourpicker.val = self.sRGB_for_cp
			else:
				if evt == self.evt_sR_int:
					self.sRGB_int[0] 		= self.nb_sR_int.val
					self.sRGB[0]			= round(float(self.sRGB_int[0]) / 255.0, 4)
					self.billboard.data.r 	= self.sRGB[0]
				elif evt == self.evt_sG_int:
					self.sRGB_int[1] 		= self.nb_sG_int.val
					self.sRGB[1]			= round(float(self.sRGB_int[1]) / 255.0, 4)
					self.billboard.data.g 	= self.sRGB[1]
				elif evt == self.evt_sB_int:
					self.sRGB_int[2] 		= self.nb_sB_int.val
					self.sRGB[2]			= round(float(self.sRGB_int[2]) / 255.0, 4)
					self.billboard.data.b 	= self.sRGB[2]
				self.sRGB_for_cp 		= tuple(self.sRGB)
				self.cp_colourpicker.val = self.sRGB_for_cp
			for v in ['r', 'g', 'b']:
				self.refreshMultis(v)

		elif evt == self.evt_bright:
			if XPlaneLightEditor.mode == 'SPILL':
				if self.spill.params.bright == 'INDEX':
					self.spill.data.bright = int(self.nb_bright.val)
					self.nb_bright.val	= int(self.nb_bright.val)
				else:
					self.spill.data.bright = sRGBToLinear(self.nb_bright.val)
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				if self.billboard.params.bright == 'INDEX':
					self.billboard.data.bright = int(self.nb_bright.val)
					self.nb_bright.val	= int(self.nb_bright.val)
				else:
					self.nb_bright.val  = clamp(self.nb_bright.val)
					self.billboard.data.bright = self.nb_bright.val
			self.refreshMultis('bright')

		elif evt == self.evt_vector:
			if XPlaneLightEditor.mode == 'SPILL':
				if self.omni:
					self.nm_vector.val				=  0.0, 0.0, 1.0
				elif  self.spill.params.x or self.spill.params.y or self.spill.params.z or self.dev_mode:
					self.xyz_blender 				=  self.snapToAxes(self.nm_vector.val if not self.invert_view else tuple(flipView(self.nm_vector.val)))
					self.spill.data.x				=  self.xyz_blender[0]
					self.spill.data.y				=  self.xyz_blender[2]
					self.spill.data.z				=  -self.xyz_blender[1]
					self.nm_vector.val				=  tuple(self.xyz_blender) if not self.invert_view else tuple(flipView(self.xyz_blender))
					self.nb_pitch.val, unused, self.nb_heading.val = toEulers(self.xyz_blender)
				else:
					self.nm_vector.val				=   tuple(self.xyz_blender) if not self.invert_view else tuple(flipView(self.xyz_blender))

			elif XPlaneLightEditor.mode == 'BILLBOARD':
				if self.omni:
					self.nm_vector.val				=  0.0, 0.0, 1.0
					self.xyz_blender 				= [0, 0, 0]
				elif self.billboard.params.x or self.billboard.params.y or self.billboard.params.z or self.dev_mode:
					self.xyz_blender 				=  self.snapToAxes(self.nm_vector.val if not self.invert_view else tuple(flipView(self.nm_vector.val)))
					self.billboard.data.x			=  self.xyz_blender[0]
					self.billboard.data.y			=  self.xyz_blender[2]
					self.billboard.data.z			=  -self.xyz_blender[1]
					self.nm_vector.val				=  tuple(self.xyz_blender) if not self.invert_view else tuple(flipView(self.xyz_blender))
					self.nb_pitch.val, unused, self.nb_heading.val = toEulers(self.xyz_blender)
					self.encodeBBGeomParams()
				else:
					self.nm_vector.val				=   tuple(self.xyz_blender) if not self.invert_view else tuple(flipView(self.xyz_blender))
			for v in ['x', 'y', 'z']:
				self.refreshMultis(v)

		elif evt in [self.evt_pitch, self.evt_heading]:
			self.xyz_blender = fromEulers(round(self.nb_pitch.val,0), round(self.nb_heading.val,0))
			temp = Vector(self.xyz_blender)
			temp.normalize()
			self.nm_vector.val				=  tuple(temp) if not self.invert_view else tuple(flipView(temp))
			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.data.x				= self.xyz_blender[0]
				self.spill.data.y				= self.xyz_blender[2]
				self.spill.data.z				= -self.xyz_blender[1]
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.data.x			= self.xyz_blender[0]
				self.billboard.data.y			= self.xyz_blender[2]
				self.billboard.data.z			= -self.xyz_blender[1]
				self.encodeBBGeomParams()
			for v in ['x', 'y', 'z']:
				self.refreshMultis(v)

		elif evt == self.evt_size:
			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.data.size			= round(self.nb_size.val, 2)
				self.sl_distance_to_surface.val = min(self.sl_distance_to_surface.val, self.nb_size.val)
				self.calculateSpotRadius()
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.data.size		= round(self.nb_size.val, 2)
			self.refreshMultis('size')

		elif evt == self.evt_width_deg:
			self.width_deg = self.nb_width_deg.val
			if XPlaneLightEditor.mode == 'SPILL':
			#	if self.spill.params.width or self.dev_mode:
				self.spill.data.width		= cos(radians(0.5 * float(self.width_deg)))

				self.calculateSpotRadius()
			elif XPlaneLightEditor.mode == 'BILLBOARD':
#				if self.billboard.params.width or self.dev_mode:
				self.encodeBBGeomParams()
			self.refreshMultis('width')

		elif evt == self.evt_edge:
			self.encodeBBGeomParams()

		elif evt == self.evt_invert_view:
			self.invert_view = self.tg_invert_view.val if not self.omni else 0
			if not self.omni:
				self.nm_vector.val			=  tuple(self.xyz_blender) if not self.invert_view else tuple(flipView(self.xyz_blender))

		elif evt == self.evt_omni:

			self.omni = self.tg_omni.val
			if XPlaneLightEditor.mode == 'SPILL':
				if self.omni:
					self.spill.data.width = 1.0
					self.invert_view = 0
					self.nm_vector.val			= (0.0, 0.0, 1.0)
					self.calculateSpotRadius()
				else:
					self.invert_view = self.tg_invert_view.val
					# reset to 'default' heading and width
					self.xyz_blender			= [0.0, -1.0, 0.0]
					self.spill.data.x			= 0.0
					self.spill.data.y			= 0.0
					self.spill.data.z			= 1.0
					self.spill.data.width		= 0.0
					self.width_deg				= 180
					self.nb_width_deg.val		= 180
					self.nb_pitch.val			= 0
					self.nb_heading.val			= 180
					self.nm_vector.val			= tuple(self.xyz_blender) if not self.invert_view else tuple(flipView(self.xyz_blender))
					self.calculateSpotRadius()

			elif XPlaneLightEditor.mode == 'BILLBOARD':
				if self.omni:
					self.billboard.data.width = 1.0
					self.invert_view = 0
					self.nm_vector.val			= (0.0, 0.0, 1.0)

				else:
					self.invert_view = self.tg_invert_view.val
					self.nm_vector.val			= tuple(self.xyz_blender) if not self.invert_view else tuple(flipView(self.xyz_blender))
					if self.width_deg < 2.0:
						self.width_deg = 2.0
						self.nb_width_deg.val = 2
				self.encodeBBGeomParams()

		elif evt == self.evt_distance_to_surface:
			self.calculateSpotRadius()

		elif evt ==	self.evt_snapping:
			self.snapping = self.tg_snapping.val

		elif evt == self.evt_light_ob_layer:
			self.light_ob_layer = DEF_LIGHT_LAYER = int(self.nb_light_ob_layer.val)

		elif evt == self.evt_visualize_at_cursor:
			if XPlaneLightEditor.mode == 'SPILL':
				#self.nukeVisLamps()
				self.addMultipleVisLamps('cursor')
				self.live_updates = True

		elif evt == self.evt_visualize_selected:
			if XPlaneLightEditor.mode == 'SPILL':
				self.nukeVisLamps()
				self.addMultipleVisLamps('selected')
				self.live_updates = False

		elif evt == self.evt_visualize_at_all_lamps:
			if XPlaneLightEditor.mode == 'SPILL':
				self.nukeVisLamps()
				self.addMultipleVisLamps('all')
				self.live_updates = False

			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.updateExistingLightsForPreviews()
				self.showBillboards()

		elif evt == self.evt_do_render:
				self.doVisLampRender()

		elif evt == self.evt_reset_render:
			self.cleanupVisLampRender()

		elif evt == self.evt_clear_visualizations:
			if XPlaneLightEditor.mode == 'SPILL':
				self.nukeVisLamps()
				self.live_updates = False
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				if g_qualifs & Window.Qual.SHIFT:
					self.stripBBProps()
				else:
					self.hideBillboards()

		elif evt == self.evt_create_new_at_cursor:
			loc = tuple(Window.GetCursorPos())
			self.addXPlaneLampToScene(loc)
			self.live_updates = True
			Window.Redraw()

		elif evt == self.evt_read_light_ob:
			self.readXPlaneLightFromBlender()

		elif evt == self.evt_data_to_selected:
			self.updateSelectedXPLampsWithCurrent()

		elif evt == self.evt_freq:
			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.data.freq = self.nb_freq.val
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.data.freq = self.nb_freq.val
			self.refreshMultis('freq')

		elif evt == self.evt_phase:
			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.data.phase = self.nb_phase.val
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.data.phase = self.nb_phase.val
			self.refreshMultis('phase')

		elif evt == self.evt_amp:
			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.data.amp = self.nb_amp.val
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.data.amp = self.nb_amp.val
			self.refreshMultis('amp')

		elif evt == self.evt_day_flag:
			if XPlaneLightEditor.mode == 'SPILL':
				self.spill.data.day = self.tg_day_flag.val
			elif XPlaneLightEditor.mode == 'BILLBOARD':
				self.billboard.data.day = self.tg_day_flag.val
			self.refreshMultis('day')

		elif evt == self.evt_xray:
			if self.dimmed and self.tg_xray.val:
				self.dimBlenderScene(dim=False)
			elif not self.dimmed and not self.tg_xray.val:
				if self.draw_bbs:
					self.dimBlenderScene()

		elif evt == self.evt_preview_size:
			self.preview_size = self.nb_preview_size.val
			self.setRuntime()
			Window.Redraw(-1)

		elif evt == self.evt_cell_size:
			self.billboard.data.cell_size = clamp(self.mn_cell_size.val, 1, 4)
			self.light_parts[self.sub_def_index].data.cell_size = self.billboard.data.cell_size
			self.refreshMultis('cell_size')

		elif evt == self.evt_cell_x:
			self.billboard.data.cell_x 	= clamp(self.nb_cell_x.val, 0, 16/self.billboard.data.cell_size - 1)
			self.light_parts[self.sub_def_index].data.cell_x = self.billboard.data.cell_x
			self.refreshMultis('cell_x')

		elif evt == self.evt_cell_y:
			self.billboard.data.cell_y 	= clamp(self.nb_cell_y.val, 0, int(8/self.billboard.data.cell_size) - 1)
			self.light_parts[self.sub_def_index].data.cell_y = self.billboard.data.cell_y
			self.refreshMultis('cell_y')

		elif evt == self.evt_make_new_def:
			self.makeLightDef()
			Draw()

		elif evt == self.evt_cancel_new_def:
			self.unsaved_def = False
			Draw()

		elif evt == self.evt_rename_light:
			self.renameLight()
			Draw()

		if self.live_updates:
			self.liveUpdate(evt)

	def mouseInRect(self, x, y, w, h):
		'''
		Returns True if the mouse is inside the specified rectangle, else False
		'''
		id = Window.GetAreaID()
		areas = Window.GetScreenInfo(Window.Types.SCRIPT)
		area_cds = []
		for area in areas:
			if area['id'] == id:
				area_cds = area['vertices']
				break
		if not area_cds:
			return False
		x1 = area_cds[0] + x
		y1 = area_cds[1] + y

		return True if x1 <= self.mouse_xy[0] <= x1 + w and y1 <= self.mouse_xy[1] <= y1 + h else False

	def draw2DBBPreview(self, x1, y1, x2, y2):

		u1, v1, u2, v2 = self.billboard.getUVs(1)

		# scale the UVs to zoom the 2d preview image:

		zm = 1.0 - (100.0 / self.nb_2D_preview_size.val)
		uvw = 0.5 * (u2 - u1)
		uvh = 0.5 * (v2 - v1)

		u1 += zm * uvw
		u2 -= zm * uvw
		v1 += zm * uvh
		v2 -= zm * uvh

		# The 2D preview window is not exactly square (it's cropped at the bottom) so we must adjust the bottom uvs accordingly

		win_w = x2 - x1
		win_h = y2 - y1
		aspect_ratio = float(win_h)/float(win_w)
		v1 += (1.0 - aspect_ratio) * (v2 - v1)

		uvs   = [[u1, v2], [u1, v1], [u2, v1], [u2, v2]]
		verts = [[x1, y2], [x1, y1], [x2, y1], [x2, y2]]

		current_col = Buffer(GL_FLOAT, 4)
		glGetFloatv(GL_CURRENT_COLOR, current_col)
		glEnable(GL_BLEND)
		glBlendFunc(GL_ONE, GL_ONE)
		tex_ID = Buffer(GL_INT, 1)
		glGenTextures(1, tex_ID)
		glBindTexture(GL_TEXTURE_2D, Billboard.bb_tex.glLoad());
		glEnable(GL_TEXTURE_2D)

		# A few BILLBOARD_SW edge cases where R G B and/or A parameters serve a different purpose than normal - we ignore RGBA in such cases:

		brt = self.billboard.data.bright if self.billboard.params.bright != 'INDEX' else 1.0
		if self.billboard.type == 'BILLBOARD_SW' and self.billboard.name in Billboard.names_no_rgba:
			glColor4f(1.0, 1.0, 1.0, 1.0)
		else:
			glColor4f(self.billboard.data.r * brt, self.billboard.data.g * brt, self.billboard.data.b * brt, self.billboard.data.bright)

		glBegin(GL_POLYGON)
		glEnable(GL_TEXTURE_2D)
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL)
		glBindTexture(GL_TEXTURE_2D, tex_ID[0])

		for i in range(4):
			glTexCoord2d(*uvs[i])
			glVertex2i(*verts[i])

		glEnd()
		glDisable(GL_BLEND)
		glDisable(GL_TEXTURE_2D)
		glColor4f(*current_col)

	def drawEditor(self):

	#	MENUS

		spillname = 'Add new spill' if self.dev_mode and g_qualifs & Window.Qual.SHIFT else 'SPILL'
		billboardname = 'Add new halo' if self.dev_mode and g_qualifs & Window.Qual.SHIFT else 'HALOS'
		self.tg_spill						= 	Toggle(spillname,						self.evt_spill,				10,		148,	95,		22,		self.tg_spill.val,													"Manage spill ('global') lights.")
		self.tg_billboards					= 	Toggle(billboardname,					self.evt_billboards,		105,	148,	95,		22,		self.tg_billboards.val,												"Manage halo ('billboard') lights.")
		if not self.unsaved_def:
			self.mn_type_menu				=	Menu(	self.type_menu_string,			self.evt_type_menu,			10,		126,    190,	22,		self.mn_type_menu.val,												"")
			self.mn_name_menu				= 	Menu(	self.name_menu_string,			self.evt_name_menu,			10,		82,		190,	22,		self.mn_name_menu.val,												"")
			self.bt_next_name				= 	PushButton(">",							self.evt_next_name,			152,	60,		48,		22,																			"Load the next light definition.")
			self.bt_prev_name				= 	PushButton("<",							self.evt_prev_name,			105,	60,		47,		22,																			"Load the previous light definition.")
			self.mn_filter_menu				= 	Menu(	self.filter_menu_string,		self.evt_filter_menu,		10,		104,	190,	22,		self.mn_filter_menu.val,											"")
			self.bt_read_light_ob			= 	PushButton("load active",				self.evt_read_light_ob,		10,		60,		95,		22,																			"Load the active blender lamp's definition into the editor (must be a legitimate X-Plane light).")
		else:
			drawString(							"UNSAVED DEFINITION",												105, 	67, 					CL_WHITE,			'center',	'normal')

			self.bt_make_new_def			=	PushButton("EXPORT",					self.evt_make_new_def,		105,	104,	95,		22,																			"")
			self.bt_rename_light			=	PushButton("RENAME LIGHT DEF",			self.evt_rename_light,		10,		126,	190,	22,																			"")
			self.bt_cancel_new_def			=	PushButton("CANCEL",					self.evt_cancel_new_def,	10,		104,	95,		22,																			"")
			drawButtonFrame(																						10,		82,		190,	22,  	0.80)
			drawString(									self.name ,													14, 	89, 					0.1,			'ljust',	'normal')
		#	self.st_new_light_name			=   String(		"",							self.evt_new_light_name,	10,	    104,	190,	22,		self.name,								 32, 						"")
		self.nb_light_ob_layer				=  	Number(	"LIGHT LAYER: ",				self.evt_light_ob_layer,	10,		10,		190,	22,		self.nb_light_ob_layer.val,		1,		20,							"Blender layer in which to add new lights.")

		if self.dev_mode:
			self.mn_params_menu				= 	Menu(	self.params_menu_string,		self.evt_params_menu,		720,    10,		95,		22,		0,												"")
			if self.has_multi and g_qualifs & Window.Qual.SHIFT:
				self.bt_remove_part			=	PushButton("delete part >",				self.evt_remove_part,		10,		38,		95,		22,																			"")
		if self.has_multi:
			self.mn_multi_menu				= 	Menu(	self.multi_menu_string,			self.evt_multi_menu,		105,	38,		95,		22,		self.mn_multi_menu.val,												"Select which part of a multipart light to edit. Some light definitions include both a halo part AND a spill part, or even multiples thereof.")
			drawButtonFrame(																						10,		38,		95,		22,  	CL_GUI_GREEN)
			drawString(									"multi-part light",												57, 	45, 					0.9,			'center',	'normal')

	#  	COLOUR
		self.bt_rgb_copy					= PushButton(" copy ",						self.evt_rgb_copy,			210,	10,		48,		22,																				"Copy the currently displayed colour to the clipboard.")

		if XPlaneLightEditor.mode == 'SPILL':

			drawRectangle(																							210,	60,		145,	111, 	0.2)
			if self.spill.type == 'SPILL_SW' and self.spill.name in SpillLight.names_no_rgba:
				drawRectangle(																						211,	61,		143,	109, 	1.0)
			else:
				drawRectangle(																						211,	61,		143,	109, 	self.sRGB if self.spill.params.bright == 'INDEX' else mix(CL_BLACK, self.sRGB, clamp(linearToSRGB(self.spill.data.bright))))
				self.cp_colourpicker  		= ColorPicker(								self.evt_colourpicker,		210,	60,   	144,	22,		self.sRGB_for_cp,														"Set the colour of the spill light, in sRGB colour space.")

			want_rgb_buts = 1 if all([self.spill.params.r, self.spill.params.g, self.spill.params.b]) or self.dev_mode else 0

			if want_rgb_buts:
				self.nb_sR_int				= Number("", 								self.evt_sR_int, 			210, 	38,  	48, 	22, 	self.sRGB_int[0], 	0, 			255, 									"Set the sRGB level of the spill light red channel, in the range 0-255.")
				self.nb_sG_int				= Number("", 								self.evt_sG_int, 			258, 	38,  	48, 	22, 	self.sRGB_int[1], 	0, 			255, 									"Set the sRGB level of the spill light green channel, in the range 0-255.")
				self.nb_sB_int				= Number("", 								self.evt_sB_int, 			306, 	38,  	48, 	22, 	self.sRGB_int[2], 	0, 			255, 									"Set the sRGB level of the spill light blue channel, in the range 0-255.")
				self.bt_rgb_paste			= PushButton(" paste ",						self.evt_rgb_paste,			258,	10,		48,		22,																				"Paste the colour from the clipboard into the colour picker.")
				self.bt_rgb_normalize		= PushButton(" norm ",						self.evt_rgb_normalize,		306,	10,		48,		22,																				"Normalize the currently displayed colour to maximum brightness before clipping.")
			else:
				drawButtonFrame(																					210, 	38,  	48, 	22)
				drawString(							"%d" % self.sRGB_int[0],										234, 	45, 					CL_TEXT_LT,			'center',	'normal')
				drawButtonFrame(																					258, 	38,  	48, 	22)
				drawString(							"%d" % self.sRGB_int[1],										282, 	45, 					CL_TEXT_LT,			'center',	'normal')
				drawButtonFrame(																					306, 	38,  	48, 	22)
				drawString(							"%d" % self.sRGB_int[2],										330, 	45, 					CL_TEXT_LT,			'center',	'normal')
				drawButtonFrame(																					258,	10,		48,		22,		CL_ACT_BUT)
				drawString(							'paste',														283, 	17, 					CL_GREY40,			'center',	'normal')
				drawButtonFrame(																					306,	10,		48,		22,		CL_ACT_BUT)
				drawString(							'norm',															331, 	17, 					CL_GREY40,			'center',	'normal')

			if self.spill.params.bright  or self.dev_mode:
				if self.spill.params.bright == 'INDEX':
					self.nb_bright			= Number("indx: ", 							self.evt_bright, 			364, 	38,  	73, 	22, 	int(self.nb_bright.val), 0, 			63, 	"Set the XPlane light index for this spill light.", self.brightnessCallBack, 0.015625)
				else:
					self.nb_bright			= Number("brt: ", 							self.evt_bright, 			364, 	38,  	73, 	22, 	float(self.nb_bright.val), 0.0, 		2.5, 	"Set the brightness of this spill light.", self.brightnessCallBack, 10.0, 2.0)
			else:
				drawButtonFrame(																					364, 	38,  	73, 	22)
				drawString(							"brt:" +'%5.2f' % self.spill.data.bright,						400, 	45, 					CL_TEXT_LT,			'center',	'normal')

		elif XPlaneLightEditor.mode == 'BILLBOARD':
			drawRectangle(																							210,	38,		145,	133, 	0.0)
	#		if self.mouse_over_CP:
			want_rgb_buts = 1 if all([self.billboard.params.r, self.billboard.params.g, self.billboard.params.b]) or self.dev_mode else 0
			if g_qualifs & Window.Qual.SHIFT:
				cpwy = 38 + 22 * want_rgb_buts
				self.cp_colourpicker  		= ColorPicker(								self.evt_colourpicker,		210,	cpwy,  	144,	22,		self.sRGB_for_cp,													"")
			if want_rgb_buts:
				self.nb_sR_int				= Number("", 								self.evt_sR_int, 			210, 	38,  	48, 	22, 	self.sRGB_int[0], 	0, 			255, 									"")
				self.nb_sG_int				= Number("", 								self.evt_sG_int, 			258, 	38,  	48, 	22, 	self.sRGB_int[1], 	0, 			255, 									"")
				self.nb_sB_int				= Number("", 								self.evt_sB_int, 			306, 	38,  	48, 	22, 	self.sRGB_int[2], 	0, 			255, 									"")
				self.bt_rgb_paste			= PushButton(" paste ",						self.evt_rgb_paste,			258,	10,		48,		22,																				"")
				self.bt_rgb_normalize		= PushButton(" norm ",						self.evt_rgb_normalize,		306,	10,		48,		22,																				"")
			else:
				drawButtonFrame(																					258,	10,		48,		22,		CL_ACT_BUT)
				drawString(							'paste',														283, 	17, 					CL_GREY40,			'center',	'normal')
				drawButtonFrame(																					306,	10,		48,		22,		CL_ACT_BUT)
				drawString(							'norm',															331, 	17, 					CL_GREY40,			'center',	'normal')

			self.draw2DBBPreview(210, 38, 355, 171)

			if self.billboard.params.bright  or self.dev_mode and not self.billboard.brightnessMustEqualOne():
				if self.billboard.params.bright == 'INDEX':
					self.nb_bright			= Number("indx: ", 							self.evt_bright, 			364, 	38,  	73, 	22, 	int(self.nb_bright.val), 0, 			63, 	"", self.brightnessCallBack, 0.015625)
				else:
					self.nb_bright			= Number("brt: ", 							self.evt_bright, 			364, 	38,  	73, 	22, 	float(self.nb_bright.val), 0.0, 		1.0, 	"", self.brightnessCallBack, 10.0, 2.0)
			else:
				drawButtonFrame(																					364, 	38,  	144, 	22)
				drawString(							"brt:"	 +'%5.2f' % self.billboard.data.bright,					400, 	45, 					CL_TEXT_LT,			'center',	'normal')
			if self.dev_mode:
				self.mn_cell_size			= Menu( "cell size%t|1 %x1|2 %x2|4 %x4",	self.evt_cell_size,			364,	60,		49,		22,		self.billboard.data.cell_size,					"")
				self.nb_cell_x				= Number("x",								self.evt_cell_x,			413,	60,		48,		22,		self.nb_cell_x.val,			0, 16/self.billboard.data.cell_size - 1,	"")
				self.nb_cell_y				= Number("y",								self.evt_cell_y,			461,	60,		49,		22,		self.nb_cell_y.val,			0, 8/self.billboard.data.cell_size - 1,	"")

		#GEOMETRY

		if XPlaneLightEditor.mode == 'SPILL':
			if not self.spill.type == 'SPILL_HW_FLA':

				drawRectangle(																						520,	82,		191,	89, 	0.2)
				drawRectangle(																						521,	83,		189,	87,		0.63)
				drawRectangle(																						521,	160,	189,	10,		0.5)
				self.nm_vector 				= Normal(									self.evt_vector,			579,	85,   	72,    72,		self.nm_vector.val,												"Set the spill light direction vector. The light shines FROM the bright side of the sphere, as if it were glowing from within.")

				if self.omni:
					drawString("OMNIDIRECTIONAL",																	615,	162, 				CL_VECT_ORANGE,			'center',	'tiny')
				else:
					xyz = Vector(self.spill.data.x, self.spill.data.y, self.spill.data.z)
					xyz.normalize()
					drawString("%.4f   %.4f   %.4f" % (xyz.x, xyz.y, xyz.z),										615,	162, 				CL_VECT_ORANGE,			'center',	'tiny')
				# mini-axes:
				drawRectangle(														 									536,	123,	14,		2, 		[0.65, 0.0, 0.0])
				drawString(								'x',							  								556, 	121, 					[0.55, 0.0, 0.0],	'center',	'small')
				if self.invert_view:
					drawString(							'bottom',														673, 	90, 					CL_TEXT_LT,			'ljust',	'small')
					drawRectangle(																						531,	104,	2,		14, 	[0.0, 0.5, 0])
					drawString(							'y',															532, 	95, 					[0.0, 0.44, 0],		'center',	'small')
					drawString(							'-',															526, 	121, 					[0.2, 0.2, 1.0],	'center',	'small')
				else:
					drawString(							'top',															682, 	90, 					CL_TEXT_LT,			'ljust',	'small')
					drawRectangle(																						531,	130,	2,		14, 	[0.0, 0.5, 0])
					drawString(							'y',															532, 	148, 					[0.0, 0.44, 0],		'center',	'small')
				drawString(								'z',															532, 	121, 					[0.2, 0.2, 1.0],	'center',	'small')

				self.tg_snapping		= Toggle(	"snap", 							self.evt_snapping,	 		672,   126,   	 35, 	22, 	self.snapping,													"When enabled, the vector selector will snap to axes when within a few degrees.")
				self.tg_invert_view		= Toggle(	"flip", 							self.evt_invert_view,	 	672,   104,   	 35, 	22, 	self.tg_invert_view.val,										"Inverts the view of the vector selector for convenience, if required. Does not effect the behaviour of the light itself.")
			else:
				if self.spill.params.freq  or self.dev_mode:
					self.nb_freq		= Number(	"flash frequency: ",				self.evt_freq,				520,   148,		190,	22,		self.nb_freq.val,		0,		10.0,		"",		self.dummyCallBack,	10.0,	3.0)
				else:
					drawButtonFrame				(																	520,   148,  	190, 	22)
					drawString					(	'flash frequency: %3.5r' % self.spill.data.freq,				615,   155, 					CL_TEXT_LT,			'center',	'normal')
				if self.spill.params.phase  or self.dev_mode:
					self.nb_phase		= Number(	"flash phase offset: ",				self.evt_phase,				520,   126,		190,	22,		self.nb_phase.val,		0,		10.0,		"",		self.dummyCallBack,	10.0,	3.0)
				else:
					drawButtonFrame				(																	520,   126,  	190, 	22)
					drawString					(	'flash phase offset: %3.6r' % self.spill.data.phase,			615,   133, 					CL_TEXT_LT,			'center',	'normal')
				if self.dev_mode:
					self.nb_amp			= Number(	"flash amplitude: ", 				self.evt_amp, 				520,   104,  	190, 	22, 	self.nb_amp.val,	 0.0, 		9999.0, 					"")
				else:
					drawButtonFrame				(																	520,   104,  	190, 	22)
					drawString					(	'flash amplitude: %3.6r' % self.spill.data.amp,					615,   111, 					CL_TEXT_LT,			'center',	'normal')
			if self.spill.type != 'SPILL_HW_FLA':
				if any([self.spill.params.x, self.spill.params.y, self.spill.params.z, self.dev_mode]) and not self.omni:
					self.nb_pitch		= Number(	"pitch: ", 							self.evt_pitch, 			520, 	60,  	95, 	22, 	self.nb_pitch.val,	 -90.0, 		90.0, 					"")
					self.nb_heading		= Number(	"hdng: ", 							self.evt_heading, 			615, 	60,  	95, 	22, 	self.nb_heading.val,-180.0, 		180.0, 					"")
				else:
					drawButtonFrame(																				520, 	60,  	95, 	22)
					drawButtonFrame(																				615, 	60,  	95, 	22)
					drawString(						'pitch: %d' % self.nb_pitch.val if not self.omni else "---",	567, 	67, 					CL_TEXT_LT,			'center',	'normal')
					drawString(						'hdng: %d' % self.nb_heading.val if not self.omni else "---",	662, 	67, 					CL_TEXT_LT,			'center',	'normal')
			else:
					drawButtonFrame(																				520, 	38,  	190, 	44)
					drawString(						'ALL FLASHING HW SPILL',										615, 	67, 					CL_TEXT_LT,			'center',	'normal')
					drawString(						'IS OMNIDIRECTIONAL',											615, 	45, 					CL_TEXT_LT,			'center',	'normal')

			if (self.spill.params.width  or self.dev_mode) and self.spill.type != 'SPILL_HW_FLA':
				self.tg_omni			= Toggle(	"omni",			 					self.evt_omni,	 			720,   38,   	95, 	22, 	self.tg_omni.val,														"Set the overall width of the spill light cone, in degrees.")
			else:
				if self.omni or self.spill.type == 'SPILL_HW_FLA':
					drawButtonFrame(																				720, 	38,  	95, 	22,		CL_GUI_GREEN)
					drawString(						'omni',															767, 	45, 					0.8,				'center',	'normal')
				else:
					drawButtonFrame(																				720, 	38,  	95, 	22, 	CL_GUI_LT_BLUE)
					drawString(						'omni',															767, 	45, 					CL_TEXT_LT,			'center',	'normal')

			if self.spill.params.size  or self.dev_mode:
				self.nb_size			= Number(	"size: ", 							self.evt_size, 				437, 	38,  	73, 	22, 	self.nb_size.val, 	0, 			256.0, 					"Set the size of the spill light, in terms of its maximum throw, in metres.", self.sizeCallBack, 10.0, 1.0)
			else:
				drawButtonFrame(																					437, 	38,  	73, 	22)
				drawString(							"size: " + str(self.spill.data.size),							473, 	45, 					CL_TEXT_LT,			'center',	'normal')
			if self.dev_mode:
				self.tg_day_flag		= Toggle(	'24 hours',							self.evt_day_flag,			720,	60,		95,		22,		self.tg_day_flag.val,									"")
			else:
				drawButtonFrame					(																	720, 	60,  	95, 	22)
				drawString						(	"24 hrs" if self.spill.data.day else "night time only",			767, 	67, 					CL_TEXT_LT,			'center',	'normal')

		elif XPlaneLightEditor.mode == 'BILLBOARD':
			drawRectangle(																							520,	82,		191,	89, 	0.2)
			drawRectangle(																							521,	83,		189,	87,		0.63)
			drawRectangle(																							521,	160,	189,	10,		0.5)
			self.nm_vector 				= Normal(									self.evt_vector,				579,	85,   	72,    72,		self.nm_vector.val,												"")
			if self.omni:
				drawString("OMNIDIRECTIONAL",																		615,	162, 				CL_VECT_ORANGE,			'center',	'tiny')
			else:
				xyz = Vector(self.billboard.data.x, self.billboard.data.y, self.billboard.data.z)
				xyz.normalize()
				drawString("%.4f   %.4f   %.4f" % (xyz.x, xyz.y, xyz.z),											615,	162, 				CL_VECT_ORANGE,			'center',	'tiny')
			# mini-axes:
			drawRectangle(														 									536,	123,	14,		2, 		[0.65, 0.0, 0.0])
			drawString(								'x',							  								556, 	121, 					[0.55, 0.0, 0.0],	'center',	'small')
			if self.invert_view:
				drawString(							'bottom',														673, 	90, 					CL_TEXT_LT,			'ljust',	'small')
				drawRectangle(																						531,	104,	2,		14, 	[0.0, 0.5, 0])
				drawString(							'y',															532, 	95, 					[0.0, 0.44, 0],		'center',	'small')
				drawString(							'-',															526, 	121, 					[0.2, 0.2, 1.0],	'center',	'small')
			else:
				drawString(							'top',															682, 	90, 					CL_TEXT_LT,			'ljust',	'small')
				drawRectangle(																						531,	130,	2,		14, 	[0.0, 0.5, 0])
				drawString(							'y',															532, 	148, 					[0.0, 0.44, 0],		'center',	'small')
			drawString(								'z',															532, 	121, 					[0.2, 0.2, 1.0],	'center',	'small')

			self.tg_snapping			= Toggle(	"snap", 							self.evt_snapping,	 		672,   126,   	 35, 	22, 	self.snapping,													"")
			self.tg_invert_view			= Toggle(	"flip", 							self.evt_invert_view,	 	672,   104,   	 35, 	22, 	self.tg_invert_view.val,										"")

			if self.billboard.params.freq  or self.dev_mode:
				self.nb_freq		= Number(		"freq: ",							self.evt_freq,				720,   148,		95,		22,		self.nb_freq.val,	0,			10.0,		"",		self.dummyCallBack,	10.0,	3.0)
			else:
				drawButtonFrame				(																		720,   148,  	95, 	22)
				drawString					(		'freq: %3.5r' % self.billboard.data.freq,						767,   155, 					CL_TEXT_LT,			'center',	'normal')
			if self.billboard.params.phase  or self.dev_mode:
				self.nb_phase		= Number(		"phase: ",							self.evt_phase,				720,   126,		95,		22,		self.nb_phase.val,	0,			10.0,		"",		self.dummyCallBack,	10.0,	3.0)
			else:
				drawButtonFrame				(																		720,   126,  	95, 	22)
				drawString					(		'phase: %3.6r' % self.billboard.data.phase,						767,   133, 					CL_TEXT_LT,			'center',	'normal')
			if self.dev_mode:
				self.nb_amp			= Number(		"amp: ", 							self.evt_amp, 				720, 	104,  	95, 	22, 	self.nb_amp.val,	 0.0, 		9999.0, 					"")
			else:
				drawButtonFrame				(																		720,   104,  	95, 	22)
				drawString					(		'amp: %3.6r' % self.billboard.data.amp,							767,   111, 					CL_TEXT_LT,			'center',	'normal')
			if any([self.billboard.params.x, self.billboard.params.y, self.billboard.params.z, self.dev_mode]) and not self.omni:
				self.nb_pitch			= Number(	"pitch: ", 							self.evt_pitch, 			520, 	60,  	95, 	22, 	self.nb_pitch.val,	 -90.0, 		90.0, 					"")
				self.nb_heading			= Number(	"hdng: ", 							self.evt_heading, 			615, 	60,  	95, 	22, 	self.nb_heading.val,-180.0, 		180.0, 					"")
			else:
				drawButtonFrame(																					520, 	60,  	95, 	22)
				drawButtonFrame(																					615, 	60,  	95, 	22)
				drawString(							'pitch: %d' % self.nb_pitch.val if not self.omni else "---",	567, 	67, 					CL_TEXT_LT,			'center',	'normal')
				drawString(							'hdng: %d' % self.nb_heading.val if not self.omni else "---",	662, 	67, 					CL_TEXT_LT,			'center',	'normal')

			if self.billboard.params.width or self.dev_mode:
				self.tg_omni				= Toggle("omni",			 			self.evt_omni,	 				720,    38,     95, 	22, 	self.tg_omni.val,														"")
				if not self.omni:
					self.nb_width_deg		= Number("width: ", 					self.evt_width_deg, 			520, 	38,  	95, 	22,  	self.nb_width_deg.val, 	2, 		360, 							"")
					self.nb_edge			= Number("edge: ", 						self.evt_edge, 					615, 	38,  	95, 	22,  	self.nb_edge.val, 		0, 		10, 										"")
				else:
					drawButtonFrame(																				520, 	38,  	95, 	22)
					drawString(				"---",																	567, 	45, 					CL_TEXT_LT,			'center',	'normal')
					drawButtonFrame(																				615, 	38,  	95, 	22)
					drawString(				"---",																	662, 	45, 					CL_TEXT_LT,			'center',	'normal')
			else:
				drawButtonFrame(																					520, 	38,  	95, 	22)
				drawString(					"width: " + str(self.nb_width_deg.val) if not self.omni else "---",		567, 	45, 					CL_TEXT_LT,			'center',	'normal')
				drawButtonFrame(																					615, 	38,  	95, 	22)
				drawString(					"edge: " + str(self.nb_edge.val) if not self.omni else "---",			662, 	45, 					CL_TEXT_LT,			'center',	'normal')
				if self.omni:
					drawButtonFrame(																				720, 	38,  	95, 	22,		CL_GUI_GREEN)
					drawString(						'omni',															767, 	45, 					0.8,				'center',	'normal')
				else:
					drawButtonFrame(																				720, 	38,  	95, 	22, 	CL_GUI_LT_BLUE)
					drawString(						'omni',															767, 	45, 					CL_TEXT_LT,			'center',	'normal')

			if self.billboard.params.size or self.dev_mode:
				self.nb_size			= Number(	"size ", 							self.evt_size, 				437, 	38,  	73, 	22, 	self.nb_size.val, 	0, 			256.0, 					"", self.sizeCallBack, 10.0, 1.0)
			else:
				drawButtonFrame(																					437, 	38,  	73, 	22)
				drawString(							"size " + str(self.billboard.data.size),						473, 	45, 					CL_TEXT_LT,			'center',	'normal')

				#drawRectangle(																							364,	60,		147,	63, 	CL_GREY70)
				#drawButtonFrame(																						364, 	60,  	146, 	63,		CL_GREY70)
			if self.dev_mode:
				self.tg_day_flag		= Toggle(	'24 hours',							self.evt_day_flag,			720,	60,		95,		22,		self.tg_day_flag.val,									"")
			else:
				drawButtonFrame				(																		720, 	60,  	95, 	22)
				drawString					(		"24 hours" if self.billboard.data.day else "night time only",	767, 	67, 					CL_TEXT_LT,			'center',	'normal')

		if XPlaneLightEditor.mode == 'SPILL':

			if not self.spill.type == 'SPILL_HW_FLA':
				if (self.spill.params.width or self.dev_mode) and not self.omni:
					self.nb_width_deg		= Number("width: ", 					self.evt_width_deg, 			520, 	38,  	190, 	22,  	self.nb_width_deg.val, 	2, 		180, 							"")

				else:
					drawButtonFrame(																				520, 	38,  	190, 	22)
					drawString(				"width: " + str(self.nb_width_deg.val) if not self.omni else "---",		615, 	45, 					CL_TEXT_LT,			'center',	'normal')

	#	ACTIONS

		self.bt_data_to_selected		= PushButton("APPLY TO SELECTED",		self.evt_data_to_selected,		364,	10,		146,	22,																			"Apply the current editor settings to selected Blender lamps. Updates both names and parameter values, if present.")
		self.bt_create_new_at_cursor	= PushButton("CREATE NEW AT CURSOR",	self.evt_create_new_at_cursor,	520,	10,		190,	22,																			"Create a new XPlane light at the current cursor position, using the current editor settings.")
	#	drawButtonFrame(																						364, 	148,  	146, 	22,			0.63)

		if XPlaneLightEditor.mode == 'SPILL':
			#namevisbutton = "preview at selected" if g_qualifs & Window.Qual.SHIFT else "preview at cursor"
			self.bt_visualize_at_cursor	= PushButton("at cursor",				self.evt_visualize_at_cursor,	364,	126,	73,		22,																			"Add a renderable, correctly-sized, wireframe preview spill light at the cursor, using the current editor settings.")
			self.bt_visualize_selected	= PushButton("selected",				self.evt_visualize_selected,	437,	126,	73,		22,																			"Add renderable, correctly-sized, wireframe preview spill lights for all valid X-Plane lights in the current selection.")
			self.bt_visualize_at_all_lamps	= PushButton("all visible",			self.evt_visualize_at_all_lamps,364,	104,	73,		22,																			"Add renderable, correctly-sized, wireframe preview spill lights for all valid X-Plane lights in the current scene.")
			self.bt_clear_visualizations= PushButton("remove",					self.evt_clear_visualizations,	437,	104,	73,		22,																			"Remove all preview spill lights from the scene.")
			drawString(									"TEST RENDER",											767, 	155, 						CL_TEXT_LT,			'center',	'small')
			drawString(									"PREVIEW SPILL LIGHTS",										437, 	155, 					CL_TEXT_LT,			'center',	'small')
			namerendermat = "add to selected" if g_qualifs & Window.Qual.SHIFT else "add to all"

			self.bt_do_render			= PushButton("render 3d view",			self.evt_do_render,				720,	126,	95,		22,																			"Perform a quick test render of the current 3d view.")
			self.bt_reset_render		= PushButton("clean up",				self.evt_reset_render,			720,	104,	95,		22,																			"Clean up after the render and restore previous rendering and materials configurations.")

		elif XPlaneLightEditor.mode == 'BILLBOARD':
			drawString(									"PREVIEW HALOS",										437, 	155, 						CL_TEXT_LT,			'center',	'small')
			self.bt_clear_visualizations= PushButton('hide',					self.evt_clear_visualizations,	437,	126,	73,		22,		"Remove halo previews.")
			self.bt_visualize_at_all_lamps	= PushButton('show',				self.evt_visualize_at_all_lamps,364,	126,	73,		22,		"Show halo previews in the 3-D view for all X-Plane lights in the scene. Note: 'Script Links' must be enabled in the Blender 'Script' buttons panel.")
			self.tg_xray				= Toggle("no autodim",			 		self.evt_xray,					364,    104,   	73, 	22, 	self.tg_xray.val,													"By default, the 3d window background colour is changed to black when previewing halos. Disable with this button, if desired.")
			if not self.draw_bbs:
				self.tg_show_icons		= Toggle("draw icons",		 			self.evt_show_icons,			437,    104,   	73, 	22, 	self.tg_show_icons.val,												"By default, relevant Blender lamp icons are hidden when halo previews are drawn, to prevent them partially obscuring the halos. Disable hiding with this button, e.g. to be able to select a previewed lamp.")
			else:
				if self.tg_show_icons.val:
					drawButtonFrame(																			437, 	104,  	73, 	22,		CL_GUI_GREEN)
					drawString(					'draw icons',													474, 	111, 	0.85,	'center',	'normal')
				else:
					drawButtonFrame(																			437, 	104,  	73, 	22,		CL_GUI_LT_BLUE)
					drawString(					'draw icons',													474, 	111, 	CL_TEXT_LT,	'center',	'normal')
#			self.nb_2D_preview_size		= Slider("2d: ",		 				self.evt_2D_preview_size,		364, 	82,  	98, 	15, 	self.nb_2D_preview_size.val, 100, 	400, 	1,	"Adjusts the size of the 2D preview - does not affect their size in X-Plane.")
			self.nb_preview_size		= Slider("size: ",		 				self.evt_preview_size,			364, 	82,  	146, 	22, 	self.preview_size, 			 50, 	200, 	1,	"Global size adjustment of halo previews - tweak to taste. Has no affect on the actual size of lights in X-Plane.")
# # # _______________________________________________________________________________________________________________________________________________________________________________________

def   fileSelectorCallback(xplanefolder):

	global g_ignore_g_lighteditor, LIGHTS_TXT_DIR, LIGHTS_TXT, LIGHTS_TEXTURE_DIR, g_lighteditor
	try:
		spr = sys.sep
		if len(xplanefolder) > 0:
			if xplanefolder[-1] != spr:
				retval = PupMenu('Please select a folder, not a file (the second line in the selector window must be empty) Try again?%t|NO|YES')
				if retval == 2:
					Window.FileSelector(fileSelectorCallback, 'Select X-plane Folder')
					return 0
				else:
					PupMenu('Lights functionality disabled. Exit and re-run the script to try again.')
					g_ignore_g_lighteditor = 1
					return 0
			else:
				lites_dir = xplanefolder + "Resources%sbitmaps%sworld%slites%s" % (spr, spr, spr, spr)
				if sys.exists(lites_dir + 'lights.txt') == 1:
					LIGHTS_TXT_DIR = lites_dir + 'lights.txt'
					#print "bp1", LIGHTS_TXT_DIR
					LIGHTS_TXT = open(LIGHTS_TXT_DIR, 'r')
					#print LIGHTS_TXT_DIR
				else:
					PupMenu("Path to lights.txt is invalid")
					return 0
				if sys.exists(lites_dir + LIGHTS_TEXTURE_NAME):
					LIGHTS_TEXTURE_DIR = lites_dir + LIGHTS_TEXTURE_NAME
					#print LIGHTS_TEXTURE_DIR
				else:
					PupMenu("Path to %s is invalid" % LIGHTS_TEXTURE_NAME)
					return 0

				initLightEditor()
				return 1
	except:
		pass

def   initLightEditor():
	'''
	Instantiates the various components of the light editor, in the correct sequence.
	'''
	global g_lighteditor, g_ignore_g_lighteditor
	SpillLightData()
	SpillLightParams()
	SpillLight()
	BillboardData()
	BillboardParams()
	Billboard()
	g_lighteditor = XPlaneLightEditor(400)
	g_ignore_g_lighteditor = 1

def   lightEditorStart():
	'''
	Make sure we have the paths to lights.txt and the billboard texture.

	if yes, open lights.txt and instantiate the editor objects.
	if no, open the file browser.
	'''
	global LIGHTS_TXT

	if not g_lighteditor:
		# Do we already have the paths from the registry?
		if LIGHTS_TXT_DIR and LIGHTS_TEXTURE_DIR:
			if sys.exists(LIGHTS_TXT_DIR) == 1:
				LIGHTS_TXT = open(LIGHTS_TXT_DIR, 'r')
			else:
				PupMenu("Path to lights.txt is invalid")
			initLightEditor()
		else:
			Window.FileSelector(fileSelectorCallback, 'Select X-plane Folder')
# # # _______________________________________________________________________________________________________________________________________________________________________________________

XPTButtonNumber 			= {}
for i in range(64):
	XPTButtonNumber[str(i)] = Create(0)

XPTButtonString				= {}
for i in range(4):
	XPTButtonString[str(i)]	= Create("")

XPTButtonMenu				= {}
for i in range(64):
	XPTButtonMenu[str(i)] = Create(0)
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class XPButton(object):
	global BUTTONS, EVENTS
	counter = 0

	def __init__(self, name, initVal):

		self.name 	= name
		XPButton.counter += 1
		self.index 	= XPButton.counter
		self.__button = BUTTONS[str(self.index)]
		self.setVal(initVal)
		self.evt	= EVENTS[str(self.index)]

	def setVal(self, value):
		self.__button.val = value

	def getVal(self):
		return 	self.__button.val
# # # _______________________________________________________________________________________________________________________________________________________________________________________
#
#	  AGB EDITOR
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBObject(object):

	height_buckets = {'1': 0, '2': 10, '3': 16, '4': 24, '5': 32, '6': 40, '7': 80, '8': 120, '9': 999}
	depth_buckets  = {'1': [30,60], '2': [60,120], '3': [90,180]}

	def __init__(self, wrapper):

		self.name						= wrapper.name
		self.wrapper					= wrapper
		self.grid_obs					= []
		self.spelling_sets				= []
		self.depth_min					= 60
		self.depth_max					= 150
		self.half_block					= 0
		self.height_min					= 0
		self.height_max					= 999
		self.height_menu_index			= None
		self.depth_menu_index			= None

		for p in self.wrapper.getAllProperties():
			if p.name == 'XPToolsAGBProps' and isinstance(p.data, str):
				toks = p.data.split()
				if len(toks) == 5:
					self.depth_min  = int(toks[0])
					self.depth_max  = int(toks[1])
					self.height_min = int(toks[2])
					self.height_max = int(toks[3])
					self.half_block = int(toks[4])
				break

		if self.height_max == 999 and self.height_min == 0:
			self.height_menu_index = 1		# 'None' bucket
		else:
			for k, v in AGBObject.height_buckets.iteritems():
				if v == self.height_max:
					self.height_menu_index = int(k)

		for k, v in AGBObject.depth_buckets.iteritems():
			if v[0] == self.depth_min:
				self.depth_menu_index = int(k)

		for t in Blender.Text.Get():
			if t.name == self.name:
				self.text				= t						# Associated blender textfile
				self.parseText()
				break
		else:
			self.text 					= Blender.Text.New(self.name, 1)
			for s in self.spelling_sets:
				s.printSet()

	def numCuts(self, tileline, seq):
		if seq != None and len(seq) > 1:
			t = 0
			for i in range(1, len(seq)):
				if seq[i] - seq[i-1] != 1 or tileline[seq[i-1]].is_slop:
					t +=1
			return t
		return 0

	def updateSpellingSetMetrics(self):
		if self.spelling_sets:
			print "we have spelling sets"
			for ss in self.spelling_sets:
				print ss.comment

				if len(self.grid_obs) == 1:
					grid_ob = self.grid_obs[0]
					ss.grid_name = self.grid_obs[0].name
				else:
					for gr in self.grid_obs:
						if gr.name == ss.grid_name:
							grid_ob = gr
							break
					else:
						print "No grid ob!"
						continue
				print grid_ob.name
				for s in ss.spellings_s:
					s.setMetaData(self.numCuts(grid_ob.columns, s.spelling), grid_ob.computeSpellingMetrics(0, s.spelling))
		#			print s.spelling, s.min, s.max
				for t in ss.spellings_t:
					t.setMetaData(self.numCuts(grid_ob.rows, t.spelling), grid_ob.computeSpellingMetrics(1, t.spelling))

	def saveActiveConstraintsIDProps(self, activeSpellingSetIndex):
		if len(self.spelling_sets) > 0 and activeSpellingSetIndex != None:
			setIDProp(self.wrapper, 'XPToolsIDProps/AGB/SpellingSets/' + self.spelling_sets[activeSpellingSetIndex].comment, self.spelling_sets[activeSpellingSetIndex].getConstraintsConfig(self))
			print "saveActiveConstraintsIDProps"

	def loadActiveConstraintsIDProps(self, activeSpellingSetIndex):
		if len(self.spelling_sets) > 0 and activeSpellingSetIndex != None:
			retval = getIDProp(self.wrapper, 'XPToolsIDProps/AGB/SpellingSets/' + self.spelling_sets[activeSpellingSetIndex].comment, None)
			if retval != None:
				self.spelling_sets[activeSpellingSetIndex].setConstraintsConfig(self, retval)
				print "loadActiveConstraintsIDProps"

	def parseText(self):

		if self.text:
			got_set_header = override = False
			temp = ""
			for line in self.text.asLines():
			#print line
				if not line:
					continue
				toks = line.split()
				if not toks:
					continue
				if toks[0] == '#' and len(toks) > 2 and toks[1] == 'XPToolsSpellingSet':
					temp = " ".join(toks[2:])[:32]
				if toks[0] == '#' and len(toks) == 2 and toks[1] == 'op':
					override = True
				if toks[0] == "SPELLING_SET":
					got_set_header = True
					self.spelling_sets.append(AGBSpellingSet())
					if len(toks) > 1:
						for t in toks[1:]:
							self.spelling_sets[-1].param_list.append(int(t))
						self.spelling_sets[-1].decodeEdges()
					if override:
						self.spelling_sets[-1].override_params = True
						override = False
					self.spelling_sets[-1].comment = temp if temp else str(len(self.spelling_sets) -1)
					temp = ""

				elif got_set_header == False and toks[0] in ["SPELLING_S", "SPELLING_T"] :   # Old -style AGB text file
					self.spelling_sets.append(AGBSpellingSet())
					self.spelling_sets[-1].param_list = "0 1 2 3 4 5 6 7"
					got_set_header = True

				if self.spelling_sets:

					if toks[0] == "HEIGHT_RANGE" and len(toks) > 2:
						self.spelling_sets[-1].height_min = int(toks[1])
						self.spelling_sets[-1].height_max = int(toks[2])
						self.spelling_sets[-1].has_height_range = True

					elif toks[0] == "USE_GRID" and len(toks) > 1:
						self.spelling_sets[-1].grid_name = toks[1]

					elif toks[0] == "FREQUENCY" and len(toks) > 1:
						self.spelling_sets[-1].frequency = int(toks[1])

					elif toks[0] == "SPELLING_S" and len(toks) > 1:
						self.spelling_sets[-1].setData(0, tuple(map(int, line.split("#", 1)[0].split()[1:])))

					elif toks[0] == "SPELLING_T" and len(toks) > 1:
						self.spelling_sets[-1].setData(1, tuple(map(int, line.split("#", 1)[0].split()[1:])))

	def renameAGB(self, newName):

		scn = bpy.data.scenes.active
		new_name = incrementName(newName, (ob.name for ob in Object.Get()), blenderStyle=True)
		if self.text != None:
			self.text.name = new_name
			AGBold_real_name = self.wrapper.name[3:]+".agb"
			AGBnew_real_name = new_name[3:]+".agb"
			for i, line in enumerate(self.text.asLines()):
				if line[:2] == "# " and line[2:].split()[0] == AGBold_real_name:
					line = "# " + AGBnew_real_name + "\n"
					self.text.setCursorPos(i, 0)
					self.text.setSelectPos(i+1, 0)
					self.text.insert(line)

		if self.wrapper != None:
			self.wrapper.name = new_name
			setIDProp(self.wrapper, 'XPToolsIDProps/AGB/name', new_name)
		self.name = new_name

		Window.RedrawAll()
		return self.name

	def saveAGB(self, dialogs=True):

		if self.text and self.spelling_sets:
			do_backup = True

			if dialogs:
				retval = PupMenu("Save AGB. This will replace the current AGB text file with the contents of the editor. Proceed?%t|Cancel|Save, backing up previous version|Save, overwriting previous version")
				if retval <2:
					return
				if retval == 3:
					do_backup = False
			if do_backup:
				text_backup = Blender.Text.New("_" + self.text.name)
				for line in self.text.asLines():
					text_backup.write(line + "\n")
			self.text.clear()
			self.text.write("#\n# %s\n" % (self.name[3:] + ".agb"))
			self.text.write("# " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
			self.text.write("\n#\n")
			for s in self.spelling_sets:
				s.printSet(self.text)

		# Note that we don't save spelling set constraint properties here - they are specific to the panel control settings for the ACTIVE spelling set, and only exist as a dict of properties on the wrapper.
		# Only one is loaded into the editor at any given time. This allows one grid object to masquerade in multiple guises suited to specific spelling set configurations - the constraints are set on (and read from) the grid ob
		# but not actually stored there - they are updated whenever the identity of the active spelling set changes.
		# If we were to try to save them now, spelling sets which share a common grid would end up with identical constraint settings.

		# Save AGB params to a conventional Blender prop on the  wrapper:
		propstring = "%d %d %d %d %d" % (int(self.depth_min), int(self.depth_max), int(self.height_min), int(self.height_max), int(self.half_block))
		setProp(self.wrapper, 'XPToolsAGBProps', propstring)
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBSpelling(object):

	def __init__(self, spelling, numCuts=None, lengths=[0.0, 0.0], startRoad=1, endRoad=1):
		self.spelling 	= spelling
		self.min		= lengths[0]
		self.max		= lengths[1]
		self.cuts		= numCuts
		self.start_type	= startRoad
		self.end_type	= endRoad
		self.protected	= False

	def copy(self):
		return AGBSpelling(self.spelling, self.cuts, [self.min, self.max], self.start_type, self.end_type)

	def setMetaData(self, numCuts, lengths, startRoad=1, endRoad=1):
		self.min		= lengths[0]
		self.max		= lengths[1]
		self.cuts		= numCuts
		self.start_type	= startRoad
		self.end_type	= endRoad
	def overlap(self, spelling):
		return spelling.min <= self.max and self.min <= spelling.max
	def containedBy(self, minMax):
		return minMax[0] <= self.min and minMax[1] >= self.max
	def excludedBy(self, minMax):
		return self.min > minMax[1] or self.max < minMax[0]
	def __str__(self):
		return reduce(lambda x, y: x + " " + y, [str(s) for s in self.spelling]) if len(self.spelling) > 0 else ""
	def __lt__(self, minMax):
		return self.max < minMax[0]
	def __le__(self, minMax):
		return self.max <= minMax[0]
	def __gt__(self, minMax):
		return self.min > minMax[1]
	def __ge__(self, minMax):
		return self.min >= minMax[1]
	def __eq__(self, spellingOb):
		if not isinstance(spellingOb, AGBSpelling):
			return False
		return self.spelling == spellingOb.spelling
	def __ne__(self, spellingOb):
		if not isinstance(spellingOb, AGBSpelling):
			return True
		return self.spelling != spellingOb.spelling

	def overlapsAny(self, minMax):
		return self.min <= minMax[1] and minMax[0] <= self.max

	def overlapsMin(self, minMax):
		return self.min < minMax[0] <= self.max <= minMax[1]

	def overlapsMax(self, minMax):
		return  minMax[0] < self.min <= minMax[1] <= self.max

	def straddleLengthMax(self, spelling):
		return self.max - spelling.min if spelling.min <= self.max < spelling.max else -1

	def straddleLengthMin(self, spelling):
		return spelling.max - self.min if spelling.min < self.min <= spelling.max else -1
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBSpellingSet(object):

	def __init__(self):
		self.comment			= ""					# Human readable name
		self.param_list			= []					# List of X-Plane's edge match configs
		self.road_edges			= []					# List of decoded versions of above, in the form of lists of 4 boolean values, one for each side of the block.
		self.override_params	= False					# True if we ignore all edge param info, and define the spelling set as being eligible for all edge type combinations
		self.height_min			= 0
		self.height_max			= 999
		self.has_height_range	= False
		self.grid_name			= None
		self.frequency			= 0
		self.spellings_s		= []					# List of AGBSpelling() obs [   ( tuple spelling, int number_of_cuts, float min_length, float  max_length, int start_is_road_facing, int end_is_road_facing), ...  ]
		self.spellings_t		= []
		self.rejected_s			= set()					# set of spelling _tuples_ only, NOT AGBSpellings
		self.rejected_t			= set()
		self.keep_protected		= True

	def setData(self, axis, spelling, numCuts=-1, lengths=[0.0, 0.0], startRoad=1, endRoad=1):
		if isinstance(spelling, str):
			spelling = tuple(map(lambda x: int(x), spelling.split()))
		if not axis:
			self.spellings_s.append( AGBSpelling(spelling, numCuts, lengths, startRoad, endRoad) )
		else:
			self.spellings_t.append( AGBSpelling(spelling, numCuts, lengths, startRoad, endRoad) )

	def copy(self, newName=None):
		newset = AGBSpellingSet()
		if newName == None:
			newset.comment		= incrementName(self.comment)
		else:
			newset.comment		= newName
		newset.param_list		= self.param_list[:]
		newset.road_edges		= self.road_edges[:]
		newset.height_min		= self.height_min
		newset.height_max		= self.height_max
		newset.grid_name		= self.grid_name
		newset.frequency		= self.frequency
		newset.spellings_s		= self.spellings_s[:]
		newset.spellings_t		= self.spellings_t[:]
		return newset

	def getConstraintsConfig(self, AGBOb):
		config = {}
		if AGBOb != None and AGBOb.grid_obs:
			for g in AGBOb.grid_obs:
				if g.name == self.grid_name:
					config['constraints'] = g.getTilelineConstraints()
					break
			else:
				print "ERROR. Couldn't find grid object %s" % self.grid_name
				return
			config['spelling_set_name'] = self.comment
			config['grid_name']			= self.grid_name
			config['time_stamp'] 		= datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
			config['common_constraints']= g_AGB_editor.getCommonConstraintsSettings()
		return config

	def setConstraintsConfig(self, AGBOb, data):
		global g_AGB_editor
		if data == None:
			print "No Constraints data"
			return
		try:
			if data['grid_name'] != self.grid_name:
				print "ERROR: Spelling set grid_name different from config grid_name."
				return
			if AGBOb != None and AGBOb.grid_obs:
				if data['grid_name'] == self.grid_name and data['spelling_set_name'] == self.comment:
					for g in AGBOb.grid_obs:
						if g.name == self.grid_name:
							g.setTilelineConstraints(data['constraints'])
							break
					else:
						return self.grid_name

					g_AGB_editor.setCommonConstraintsSettings(data['common_constraints'])

		except (KeyError, ValueError), detail:
			print "ERROR: ", detail
		else:
			return 'CONSTRAINTS_OK'

	def hasSpellings(self, axis=None):
		if axis != 1 and len(self.spellings_s) > 0:
			return True
		if axis != 0 and len(self.spellings_t) > 0:
			return True
		return False

	def clearSpellings(self, axis=None):
		if axis != 1:
			if not self.keep_protected:
				self.spellings_s[:] = []
			else:
				self.spellings_s = filter(lambda x: x.protected, self.spellings_s)
		if axis != 0:
			if not self.keep_protected:
				self.spellings_t[:] = []
			else:
				self.spellings_t = filter(lambda x: x.protected, self.spellings_t)

	def replaceSpellings(self, axis, spellingSet, lengthRange=False):

		my_spellings = self.spellings_t if axis else self.spellings_s

		if lengthRange:		# Delete all old spellings lying entirely within the ranges of the new spellings, and insert new.
			ranges = spellingSet.getLengthRanges(axis)
			for s in my_spellings[:]:
				for r in ranges:
					if s.containedBy(r):
						if not self.keep_protected or not s.protected:
							my_spellings.remove(s)
							break
			self.extendSpellings(axis, spellingSet.spellings_t if axis else spellingSet.spellings_s)
		else:
			if not self.keep_protected:
				my_spellings[:] = []
			else:
				my_spellings[:] = filter(lambda x: x.protected, my_spellings)

			self.extendSpellings(axis, spellingSet.spellings_t if axis else spellingSet.spellings_s)

	def extendSpellings(self, axis, spellings):
		if axis == 0:
			for s in spellings:
				if s not in self.spellings_s and s.spelling not in self.rejected_s:
					self.spellings_s.append(s.copy())
		elif axis == 1:
			for s in spellings:
				if s not in self.spellings_t and s.spelling not in self.rejected_t:
					self.spellings_t.append(s.copy())
		self.sortSpellings(axis, 'min')

	def defineProtected(self, axis, protectedStrings):
		if not protectedStrings:
			return
		ptuples = [tuple(map(lambda x: int(x), ps.split())) for ps in protectedStrings]

		if axis == 0:
			for s in self.spellings_s:
				if s.spelling in ptuples:
					s.protected = True
		elif axis == 1:
			for s in self.spellings_t:
				if s.spelling in ptuples:
					s.protected = True

	def unprotectProtected(self, axis):
		for s in [self.spellings_s, self.spellings_t][axis]:
			s.protected = False

	def extendRejected(self, axis, rejectedStrings):
		"""
			rejectedStrings is a list of strings, read from selected preview grid ob properties.
			Each string contains a single spelling.
		"""
		if not rejectedStrings:
			return
		for r in (tuple(map(lambda x: int(x), rs.split())) for rs in rejectedStrings):
			if axis == 0:
				self.rejected_s.add(r)
			elif axis == 1:
				self.rejected_t.add(r)

		if axis == 0 and self.rejected_s:
			for s in self.spellings_s:
				if s.spelling in self.rejected_s:
					self.spellings_s.remove(s)

		elif axis == 1 and self.rejected_t:
			for s in self.spellings_t:
				if s.spelling in self.rejected_t:
					self.spellings_t.remove(s)

	def resetRejected(self, axis):
		"""
		Clears the set of rejected spellings for the specified axis
		"""
		if axis == 0:
			self.rejected_s.clear()
		elif axis == 1:
			self.rejected_t.clear()

	def sortSpellings(self, axis=None, mode='min'):
		"""
			If no axis is given, both lists are sorted
		"""
		modes = {'cuts'   : lambda x: x.cuts,
				 'min'	  : lambda x: x.min,
				 'max'    : lambda x: x.max,
				 'tiles'  : lambda x: len(x.spelling)}
		if axis != 1 and self.spellings_s:
			self.spellings_s.sort(key = modes[mode])
		if axis != 0 and self.spellings_t:
			self.spellings_t.sort(key = modes[mode])

	def getLengthRanges(self, axis):
		for r in self.collapseRanges([[s.min, s.max] for s in [self.spellings_s, self.spellings_t][axis]]):
			yield r

	def collapseRanges(self, minMax):
		if len(minMax) == 0:
			yield r
		r = minMax[0][:]
		for low, high in sorted( minMax):
			if r[1] >= low:
				r[1] = max(r[1], high)
			else:
				yield r
				r[0] = low
				r[1] = high
		yield r

# 	def getPreviewCount(self, sMin, sMax, tMin, tMax, perSpelling=None, perMetre=None):
# 		"""
# 			Calculate how many previews are likely to be made from this spelling set within the specified ranges and at the specified pitch.
# 		"""
# 		if not self.hasSpellings(0):
# 			PupMenu("Can't make previews - there are no horizontal spellings in the requested source!%t")
# 			return False
# 		if not self.hasSpellings(1):
# 			PupMenu("Can't make previews - there are no vertical spellings in the requested source!%t")
# 			return False
#
# 		s_range = [sMin, sMax]
# 		t_range = [tMin, tMax]
#
# 		s_num = len([s for s in self.spellings_s if s.overlapsAny(s_range)])
#
#

	def	decodeEdges(self):
		self.road_edges = []
		for p in self.param_list:
			a = [1]
			for i in range(3):
				a.append(int(bool(p&(1<<i))))
			self.road_edges.append(a)

	def encodeEdges(self):
		"""
			Iterate through both lists of spellings, noting which combinations of edges are present in each.
			Combine the two results to find the possible combinations. Note that this doesn't necessarily imply full coverage of those combinations!
		"""

		if self.override_params:
			self.param_list = [0, 1, 2, 3, 4, 5, 6, 7]
			self.decodeEdges()
		else:
			edges_s = []
			edges_t = []

			for s in self.spellings_s:
				if [s.start_type, s.end_type] not in edges_s:
					edges_s.append([s.start_type, s.end_type])
			for s in self.spellings_t:
				if [s.start_type, s.end_type] not in edges_t:
					edges_t.append([s.start_type, s.end_type])
			self.road_edges	= []
			for s in edges_s:
				for t in edges_t:
					self.road_edges.append([1, s[1], t[1], s[0]])
			# t[0] (the front road) is ignored because it is always assumed to be road-facing:
			self.param_list = []
			for c in self.road_edges:
				self.param_list.append(c[1]*4 + c[2]*2 + c[3])
			self.param_list.sort()

	def printSet(self, destination=None, debug=True):

		if destination == None:
			if not AGBEditor.temp_text:
				self.getTempText()
			txt = AGBEditor.temp_text
		else:
			txt = destination
			if not txt:
				PupMenu("ERROR: Can't find requested text object!%t")
				return

		def frmt(x):
			return str(x).rjust(3)

		rh_col = 0
		if self.spellings_s:
			rh_col = max(len(x.spelling) for x in self.spellings_s)
		if self.spellings_t:
			rh_col = max(rh_col, max(len(x.spelling) for x in self.spellings_t))
		rh_col *= 4

		outtext = []
		outtext.append('\n')

		if self.comment:
			outtext.append("#\n# XPToolsSpellingSet %s\n" % self.comment)

		outtext.append("# op\n\n" if self.override_params else "#\n\n")
		outtext.append("SPELLING_SET  " + "  ".join(map(str, self.param_list)) + "\n")
		if self.has_height_range:
			outtext.append("HEIGHT_RANGE  " + str(self.height_min) + " " + str(self.height_max) + "\n")
		if self.grid_name:
			outtext.append("USE_GRID      %s\n" % self.grid_name)
		if self.frequency:
			outtext.append("FREQUENCY     %d\n" % self.frequency)
		if self.spellings_s:
			outtext.append( '\n')
			for s in self.spellings_s:
				sp_str = " ".join(map(frmt, s.spelling))
				if debug:
					outtext.append("SPELLING_S " + sp_str + (rh_col - len(sp_str)) * " " + "#%s %d %d %.3f  %.3f\n" % (frmt(s.cuts),  s.start_type, s.end_type, s.min, s.max))
				else:
					outtext.append("SPELLING_S " + sp_str)
		if self.spellings_t:
			outtext.append( '\n')
			for t in self.spellings_t:
				sp_str = " ".join(map(frmt, t.spelling))
				if debug:
					outtext.append("SPELLING_T " + sp_str + (rh_col - len(sp_str)) * " " + "#%s %d %d %.3f  %.3f\n" % (frmt(t.cuts), t.start_type, t.end_type, t.min, t.max))
				else:
					outtext.append("SPELLING_S " + sp_str)

		outtext.append("\n")

	#	txt.clear()
		for t in outtext:
			txt.write(t)
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBMatchGroup(object):
	def __init__(self, menuIndex, name, matches):
		self.index		= menuIndex						# int
		self.name		= name							# string
		self.matches	= matches						# list of ints

	def setIndex(self, newIndex):
		self.index		= newIndex
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBGrid(object):

	VERT_LIMIT = 0.01
	SPELLING_ROUNDING_DENOMINATOR = 4
	def __init__(self, initOb):

		# We are passed a Blender object which must be of type 'Mesh'.
		self.init_ob		= initOb
		self.wrapper		= None
		prnt = initOb.getParent()
		if prnt != None:
			if prnt.type == 'Empty':
				self.wrapper	= prnt
				self.name 		= prnt.name							# If the grid has a wrapper already, we use its name
		elif initOb.name.startswith('TILE'):
			self.name = initOb.name[4:]
		else:
			self.name = initOb.name										# else use the mesh's name sans the TILE prefix

		self.loc 				= initOb.getLocation('worldspace')	# initOb is a blender mesh object consisting of individual vertices located along the x and y axes at the positions of cuts.
		self.init_mesh 			= initOb.getData(mesh=1)
		self.is_pukka			= False
		self.origin_uv			= [0.0, 0.0]								# UV coords of the mesh origin, as a ratio in the range 0 ... 1.0
		self.tex_scaleX			= None										# texture scale in tex_widths/m
		self.tex_scaleY			= None										# texture scale in tex_heights/m
		self.image				= None
		self.image_size_pix		= []
		self.pxpm_x				= 2.0
		self.pxpm_y				= 2.0

		self.def_mesh_ob		= None
		self.def_mesh			= None										# Master, undeformed grid mesh

		self.offsets_x			= []										# list of  all x-axis tile edge offsets (metres)
		self.offsets_y			= []										# list of  all y-axis tile edge offsets (metres)

		self.rows				= []										# list of AGBTileLines
		self.columns			= []										# list of AGBTileLines
		self.num_rows			= 0											# len(self.rows)
		self.num_columns		= 0											# len(self.columns)

		self.tiles				= []
		self.num_tiles			= 0

		self.cuts_h				= []
		self.cuts_v				= []

		self.match_groups_h 	= []										# list of AGBMatchGroups
		self.match_groups_v 	= []										# list of AGBMatchGroups
		self.match_group_active	= None										# string
		self.edge_groups_h		= [AGBMatchGroup(0, 'HR', []), AGBMatchGroup(1, 'HN', [])]
		self.edge_groups_v		= [AGBMatchGroup(0, 'VR', []), AGBMatchGroup(1, 'VN', [])]
		self.overlap_h			= []										# This is the same as OVERLAP_H etc, in METRES
		self.overlap_v			= []										#
		self.antislop_h			= []										#
		self.antislop_v			= []										#

		self.metarect_data		= {'edge_min': [0,0,0,0], 'edge_best': [0,0,0,0], 'edge_max': [0,0,0,0]}

		self.got_geom			= False
		self.got_meta_rects		= False
		self.is_grid			= False
		self.has_uvs			= False
		self.has_image			= False
		self.has_antislop		= False
		self.has_ignore			= False
		self.need_slop			= True
		self.need_edge_type		= True
		self.need_match_grp		= True

		self.GUI_rects			= {'match_group':[], 'match_group_active':[], 'edge_road':[], 'edge_road_active':[], 'edge_no_road':[], 'edge_no_road_active':[], 'untagged':[]}
		self.GUI_zone_rects		= []
		self.GUI_chequerboard	= []
		self.GUI_lines			= []
		self.GUI_label_locs_x  	= []
		self.GUI_label_locs_y  	= []
		self.GUI_bugs_x			= []
		self.GUI_bugs_y			= []

		self.annotations		= []
		self.dgroups			= []
		self.text3d_ob			= None

	def m2p(self, inMetres, axis):
		""" takes either a float, int or list of floats or ints, and multiplies them by the appropriate pxpm scale
		"""
		if isinstance(inMetres, list):
			return map(lambda x: x *(self.pxpm_y if axis == 'v' else self.pxpm_x), inMetres)
		else:
			return inMetres * (self.pxpm_y if axis == 'v' else self.pxpm_x)

	def p2m(self, inPixels, axis):
		if isinstance(inPixels, list):
			return map(lambda x: x /(self.pxpm_y if axis == 'v' else self.pxpm_x), inPixels)
		else:
			return inPixels / (self.pxpm_y if axis == 'v' else self.pxpm_x)

	def testFace(self, face, faceMode, negate=0):
		if negate:
			return not (face.mode & Blender.Mesh.FaceModes[faceMode])
		else:
			return face.mode & Blender.Mesh.FaceModes[faceMode]

	def realCloseFloat(self, x, y, limit):
		return abs(x - y) < limit

	def realClose(self, vec1, vec2, limit):
		for i in range(len(vec1)):
			if abs(vec1[i] - vec2[i]) > limit:
				return False
		return True

	def checkIsPukka(self):

		if self.init_ob == None:
			return False
		if 'AGBProps' in self.init_ob.properties.keys():
			props = self.init_ob.properties['AGBProps']
		else:
			return False
		try:
			ispukka	 = props['IsPukka'] if 'IsPukka' in props else False
			numverts = props['NumVerts'] if 'NumVerts' in props else 0
			numverts_now 	= len(self.init_mesh.verts)
			if self.init_mesh.faceUV and ispukka and numverts == numverts_now:
				self.is_pukka = True
				return True
			else:
				self.is_pukka = False
				return False
		except KeyError:
			return False

	def sameCardinality(self, gridOb):
		if gridOb != None:
			if self.num_columns == gridOb.num_columns and self.num_rows == gridOb.num_rows:
				return True
		return False

	def getBoundingRect(self):
		if self.def_mesh_ob != None:
			bb = self.def_mesh_ob.boundingBox
			return (bb[0][0], bb[0][1], bb[6][0], bb[6][1])
		else:
			return None

	def readBasicGeomFromMesh(self):
		"""
			Get the vertex row and column locations of the init mesh, and build our 'virtual grid' geometry.
			If succesful, check if the init mesh is already a grid, set self.is_grid = True.
		"""
		if self.init_mesh == None:
			PupMenu("ERROR: The selected object does appear to contain a mesh!%t")
			return 0
		self.offsets_x		= list(set([round(v.co[0],3) for v in self.init_mesh.verts]))		# list of  all x-axis tile edge offsets (metres)
		self.offsets_x.sort()
		self.offsets_y		= list(set([round(v.co[1],3) for v in self.init_mesh.verts]))		# list of  all y-axis tile edge offsets (metres)
		self.offsets_y.sort()

		self.num_rows		= len(self.offsets_y) - 1
		self.num_columns	= len(self.offsets_x) - 1
		self.num_tiles		= self.num_rows * self.num_columns

		self.rows = []
		for i in range(self.num_rows):
			self.rows.append(AGBTileLine('row', i,  self.num_columns, self.num_rows, self.offsets_y[i], self.offsets_y[i+1]))

		self.columns = []
		for i in range(self.num_columns):
			self.columns.append(AGBTileLine('column', i, self.num_columns, self.num_rows,  self.offsets_x[i], self.offsets_x[i+1]))

		# The cuts lists are lists of integer blender edge indices, which will be set when we build our new master mesh.
		self.cuts_h	= []
		self.cuts_v	= []

		i = 0
		for c in range(self.num_columns + 1):
			cut = []
			for r in range(self.num_rows):
				cut.append(i)
				i += 1
			self.cuts_h.append(cut)
		for r in range(self.num_rows + 1):
			cut = []
			for c in range(self.num_columns):
				cut.append(i)
				i += 1
			self.cuts_v.append(cut)

		# build a 2-dimensional list of AGBTiles objects:
		vert_index = 0
		self.tiles = []
		for y in range(self.num_rows):
			self.tiles.append([])
			for x in range(self.num_columns):
				t = AGBTile(x, y, vert_index, self.offsets_x[x], self.offsets_y[y], self.offsets_x[x+1], self.offsets_y[y+1])
				self.tiles[y].append(t)
				vert_index += 4

		self.got_geom = True

		# Is this mesh already a grid perhaps?
		self.is_grid = True if len(self.init_mesh.faces) == self.num_rows * self.num_columns else False

		# If it's a grid, maybe it has UV coords and an assigned image?
		# If so, calculate texture scale and origin and get the texture image data.
		# If not, prompt the user to set them.
		if self.is_grid:
			self.has_uvs = self.init_mesh.faceUV
		return self.has_uvs

	def getTextureParamsFromMesh(self):

		all_uvs = []
		for f in self.init_mesh.faces:
			for uv in f.uv:
				all_uvs.append(uv)
		u_min = min([c[0] for c in all_uvs])
		u_max = max([c[0] for c in all_uvs])
		v_min = min([c[1] for c in all_uvs])
		v_max = max([c[1] for c in all_uvs])

		self.tex_scaleX = (u_max - u_min)/(self.offsets_x[-1] - self.offsets_x[0])
		self.tex_scaleY = (v_max - v_min)/(self.offsets_y[-1] - self.offsets_y[0])
		# Don't assume that the grid origin is at the lower left corner of the mesh:
		self.origin_uv = [u_min - self.offsets_x[0] * self.tex_scaleX, v_min - self.offsets_y[0] * self.tex_scaleY]

		# Get a face's image - we assume it will be the same image for the others too.
		if self.init_mesh.faces[0].image:
			self.image 			= self.init_mesh.faces[0].image
			self.image_size_pix = self.init_mesh.faces[0].image.getSize()
			self.pxpm_x 		= self.image_size_pix[0] * self.tex_scaleX
			self.pxpm_y			= self.image_size_pix[1] * self.tex_scaleY
			self.has_image = True
		return True

	def setTextureData(self, imageName, sizeX, sizeY, originX, originY):
		"""
			called by the editor to update our texture vars with the GUI settings.
		"""
		try:
			self.image 	= Blender.Image.Get(imageName)
		except:
			PupMenu("ERROR: Can't find image %s %%t" % imageName)

		if self.image:
			self.has_image 	= True
			self.image_size_pix = self.image.size
		if sizeX >= 1.0:
			self.tex_scaleX = 1.0/sizeX
		if sizeY >= 1.0:
			self.tex_scaleY = 1.0/sizeY

		self.origin_uv 	= [originX * self.tex_scaleX, originY * self.tex_scaleY]
		self.pxpm_x 	= self.image_size_pix[0] * self.tex_scaleX
		self.pxpm_y		= self.image_size_pix[1] * self.tex_scaleY
		print "imageName, sizeX, sizeY, originX, originY", imageName, sizeX, sizeY, originX, originY
		print "self.tex_scaleX, self.tex_scaleY, self.origin_uv, self.pxpm_x, self.pxpm_y", self.tex_scaleX, self.tex_scaleY, self.origin_uv, self.pxpm_x, self.pxpm_y
		if not self.init_mesh.faceUV:
			self.init_mesh.addUVLayer("UV")

	def getOverlapValuesFromWrapper(self):

		# If grid already has a parent wrapper, search for overlap and antislop properties, and if found, update the AGBTileLine values:
		if self.wrapper != None:
			self.overlap_h  = self.p2m(str2Nums(getProp(self.wrapper, 'OVERLAP_H', 	"")), 'h')
			self.overlap_v  = self.p2m(str2Nums(getProp(self.wrapper, 'OVERLAP_V', 	"")), 'v')
			self.antislop_h = self.p2m(str2Nums(getProp(self.wrapper, 'ANTISLOP_H', "")), 'h')
			self.antislop_v = self.p2m(str2Nums(getProp(self.wrapper, 'ANTISLOP_V', "")), 'v')

			if self.overlap_h :
				slopzones_h = filter(lambda x: x.is_slop, self.columns)
				if len(self.overlap_h) > 1:
					if len(slopzones_h) > len(self.overlap_h): PupMenu("ERROR: There are fewer overlap_h args than slop zones!%t")
					else:
						for z, o in zip(slopzones_h, self.overlap_h):
							z.min_overlap = o
				elif len(self.overlap_h) == 1:
					for z in slopzones_h:
						z.min_overlap = self.overlap_h[0]

			if self.overlap_v:
				slopzones_v = filter(lambda x: x.is_slop, self.rows)
				if len(self.overlap_v) > 1:
					if len(slopzones_v) > len(self.overlap_v): PupMenu("ERROR: There are fewer overlap_v args than slop zones!%t")
					else:
						for z, o in zip(slopzones_v, self.overlap_v):
							z.min_overlap = o
				elif len(self.overlap_v) == 1:
					for z in slopzones_v:
						z.min_overlap = self.overlap_v[0]

			if self.antislop_h:
				antisloptiles_h = filter(lambda x: x.is_antislop, self.columns)
				if len(self.antislop_h) > 1:
					if len(antisloptiles_h) > len(self.antislop_h): PupMenu("ERROR: There are fewer antislop_v args than antislop zones!%t")
					else:
						for z, o in zip(antisloptiles_h, self.antislop_h):
							z.max_antislop = o
				elif len(self.antislop_h) == 1:
					for z in antisloptiles_h:
						z.max_antislop = self.antislop_h[0]
			if self.antislop_v:
				antisloptiles_v = filter(lambda x: x.is_antislop, self.rows)
				if len(self.antislop_v) > 1:
					if len(antisloptiles_v) > len(self.antislop_v): PupMenu("ERROR: There are fewer antislop_v args than antislop zones!%t")
					else:
						for z, o in zip(antisloptiles_v, self.antislop_v):
							z.max_antislop = o
				elif len(self.antislop_v) == 1:
					for z in antisloptiles_v:
						z.max_antislop = self.antislop_v[0]

			for tilelines in [self.rows, self.columns]:
				for t in tilelines:
					t.updateMetrics()

	# 	print '\nCOLUMNS'
# 		for c in self.columns:
# 			c.printData()
# 		print '\nROWS'
# 		for c in self.rows:
# 			c.printData()
		return [self.has_image, self.has_uvs]

	def wrapGrid(self, wrapperLayer=None):
		scn = bpy.data.scenes.active
		if self.wrapper != None and self.wrapper in scn.objects:
			return
		self.wrapper = scn.objects.new('Empty', self.name)
		self.wrapper.loc = self.loc
		if wrapperLayer != None:
			self.wrapper.layers = [wrapperLayer]
		else:
			self.wrapper.layers = self.def_mesh_ob.layers
		self.wrapper.emptyShape = Blender.Object.EmptyShapes['CUBE']
		self.wrapper.drawSize  = 3.0
		self.wrapper.nameMode = True
		self.wrapper.makeParent([self.def_mesh_ob])
		scn.objects.active = self.wrapper

	def updateWrapperProps(self, panelMode=None):

		if self.wrapper == None:
			self.wrapGrid()

		if panelMode == 'slop':
			if self.overlap_h:
				setProp(self.wrapper, 'OVERLAP_H', numsToString(self.m2p(self.overlap_h, 'h')))
			if self.overlap_v:
				setProp(self.wrapper, 'OVERLAP_V', numsToString(self.m2p(self.overlap_v, 'v')))

		elif panelMode == 'antislop':
			if self.antislop_h:
				setProp(self.wrapper, 'ANTISLOP_H', numsToString(self.m2p(self.antislop_h, 'h')))
			if self.antislop_v:
				setProp(self.wrapper, 'ANTISLOP_V', numsToString(self.m2p(self.antislop_v, 'v')))

	def setOverlapValuesFromGUI(self, panelMode, bugIndex, bug):

		if not isnum(bug.val):
			return

		lenbugx = len(self.GUI_bugs_x)
		if bugIndex < lenbugx:
			if panelMode == 'slop':
				self.columns[self.GUI_bugs_x[bugIndex][0]].min_overlap = eval(bug.val)
				self.overlap_h[:] = []
				for c in self.columns:
					if c.is_slop:
						self.overlap_h.append(c.min_overlap)
						c.updateMetrics()

			elif panelMode == 'antislop':
				self.columns[self.GUI_bugs_x[bugIndex][0]].max_antislop = eval(bug.val)
				self.antislop_h[:] = []
				for c in self.columns:
					if c.is_antislop:
						self.antislop_h.append(c.max_antislop)
						c.updateMetrics()

		elif lenbugx <= bugIndex < (lenbugx + len(self.GUI_bugs_y)):
			if panelMode == 'slop':
				self.rows[self.GUI_bugs_y[bugIndex - lenbugx][0]].min_overlap = eval(bug.val)
				self.overlap_v[:] = []
				for r in self.rows:
					if r.is_slop:
						self.overlap_v.append(r.min_overlap)
						r.updateMetrics()

			elif panelMode == 'antislop':
				self.rows[self.GUI_bugs_y[bugIndex - lenbugx][0]].max_antislop = eval(bug.val)
				self.antislop_v[:] = []
				for r in self.rows:
					if r.is_antislop:
						self.antislop_v.append(r.max_antislop)
						r.updateMetrics()

		# Update wrapper blender props if we have one
		self.updateWrapperProps(panelMode)

	def createUVCoords(self):
		"""
			Calculate UV coords and update the AGBTile array
		"""
		for y in self.tiles:
			for x in y:
				x.uvs[:] = []
				for co in x.coords:
					x.uvs.append(Vector(co[0]*self.tex_scaleX + self.origin_uv[0],  co[1]*self.tex_scaleY + self.origin_uv[1]))
		self.has_uvs = True

	def getMatchGroupsFromMesh(self):
		"""
			Search for vert groups with a legitimate name and build two lists of AGBMatchGroup objects.
			Update the TileLines
		"""
		if self.def_mesh == None:
			return
		vgs = self.def_mesh.getVertGroupNames()

		gcuts = []
		for vg in vgs:
			if len(vg) == 2 and vg[1].isdigit():
				if vg[0] == 'H':
					verts = self.def_mesh.getVertsFromGroup(vg)
					if verts:
						gcuts = []
						for v in [v for v in verts if v < len(self.cuts_h)]:
							gcuts.append(v)
						mg = AGBMatchGroup(0, vg, gcuts)
						self.match_groups_h.append(mg)

				elif vg[0] == 'V':
					verts = self.def_mesh.getVertsFromGroup(vg)
					if verts:
						gcuts = []
						for v in [v for v in verts if not v % len(self.cuts_h)]:
							gcuts.append(v / len(self.cuts_h))
						mg = AGBMatchGroup(0, vg, gcuts)
						self.match_groups_v.append(mg)

		for grp in [self.match_groups_h, self.match_groups_v]:
			grp.sort(key=lambda x: x.name)
			i = 0
			for g in grp:
				g.index = i
				i += 1

		for cuts, tl, mg  in zip([self.cuts_h, self.cuts_v], [self.columns, self.rows], [self.match_groups_h, self.match_groups_v]):
			for i, c in enumerate(cuts):
				for g in mg:
					if i in g.matches:
						if i < len(tl):
							tl[i].match_trail = g.name
						if i > 0:
							tl[i-1].match_lead = g.name

	def getEdgeGroupsFromMesh(self):
		"""
			Search for vert groups with a legitimate name and build two lists of AGBMatchGroup objects.
			Update the TileLines
		"""

		if self.def_mesh == None:
			return
		vgs = self.def_mesh.getVertGroupNames()
		if not vgs:
			return
		gcuts = []
		verts = []
		for vg in vgs:
			if vg in ['HR', 'HN']:
				verts = self.def_mesh.getVertsFromGroup(vg)
				if verts != None:
					gcuts = []
					for v in [v for v in verts if v < len(self.cuts_h)]:
						gcuts.append(v)
					for grp in self.edge_groups_h:
						if grp.name == vg:
							grp.matches = gcuts
			elif vg in ['VR', 'VN']:
				verts = self.def_mesh.getVertsFromGroup(vg)
				if verts != None:
					gcuts = []
					for v in [v for v in verts if not v % len(self.cuts_h)]:
						gcuts.append(v / len(self.cuts_h))
					for grp in self.edge_groups_v:
						if grp.name == vg:
							grp.matches = gcuts

		for cuts, tl, mg  in zip([self.cuts_h, self.cuts_v], [self.columns, self.rows], [self.edge_groups_h, self.edge_groups_v]):
			for i, c in enumerate(cuts):
				for g in mg:
					if i in g.matches:
						if i < len(tl):
							tl[i].match_trail = g.name
						if i > 0:
							tl[i-1].match_lead = g.name

	def getMatchGroup(self, name):
		for g in [g for g in self.match_groups_h + self.match_groups_v if g.name == name]:
			return g
		return None

	def selectMatchGroup(self, axis, index):

		def selectCut(c):
			for edge in c:
				self.def_mesh.edges[edge].sel = 1

		editmode = Window.EditMode()
		if editmode:
			Window.EditMode(0)
		self.deselectAllEdges()

		if axis == 'h':
			for g in self.match_groups_h:
				if g.index == index:
					self.match_group_active = g.name
					for m in g.matches:
						selectCut(self.cuts_h[m])
					break

		elif  axis == 'v':
			for g in self.match_groups_v:
				if g.index == index:
					self.match_group_active = g.name
					for m in g.matches:
						selectCut(self.cuts_v[m])
					break

		if editmode:
			Window.EditMode(1)

	def deselectAllEdges(self):
		for e in self.def_mesh.edges:
			e.sel = 0

	def setMatchGroups(self, evtString, groupIndex=0):

		if evtString == 'evt_open_pnl_match_grp':
			Mesh.Mode(Mesh.SelectModes['EDGE'])

			Window.EditMode(1)
			return None
		elif evtString in ['evt_add_match_group_h', 'evt_add_match_group_v']:
			Window.EditMode(0)
			edges_sel = set()
			for e in self.def_mesh.edges:
				if e.sel:
					edges_sel.add(e.index)

			if not edges_sel:
				PupMenu('Error: No edges are selected!%t')
				Window.EditMode(1)
				return None

			# Find the lowest available name in the list (some may have been deleted and made available)
			newgroupname = None
			gindex = None
			if evtString == 'evt_add_match_group_h':
				for n in range(1, 20):
					for g in self.match_groups_h:
						if str(n) in g.name:
							break
					else:
						newgroupname = 'H' + str(n)
						gindex = n
						break
			elif evtString == 'evt_add_match_group_v':
				for n in range(1, 20):
					for g in self.match_groups_v:
						if str(n) in g.name:
							break
					else:
						newgroupname = 'V' + str(n)
						gindex = n
						break

			# find which edgesa have been selected and then find single verts on the x axis which share the same cut line as each edge.
			match_group_verts = []
			match_group_cuts = []

			if evtString == 'evt_add_match_group_h':
				for e in edges_sel:
					for c in self.columns:
						if e in c.edges[0]:
							if c.match_trail != None:
								PupMenu("WARNING: One or more of the selected vertical edges cannot be assigned because they are already allocated to an existing group. Doing nothing.%t")
								Window.EditMode(1)
								return None
							c.match_trail = newgroupname
							match_group_verts.append(c.verts[0])
							match_group_cuts.append(c.index)
							break
						elif c.index == len(self.columns) - 1 and e in c.edges[1] and  c.match_lead == None:
							c.match_lead = newgroupname
							match_group_verts.append(c.verts[1])
							match_group_cuts.append(c.index + 1)
							break

				if not match_group_cuts:
					Window.EditMode(1)
					return None
				else:
					for i in range(1, self.num_columns):
						self.columns[i-1].match_lead = self.columns[i].match_trail

				self.match_groups_h.append(AGBMatchGroup(gindex, newgroupname, match_group_cuts))
				self.match_groups_h.sort(key=lambda x: x.name)

			elif evtString == 'evt_add_match_group_v':
				for e in edges_sel:
					for r in self.rows:
						if e in r.edges[0]:
							if r.match_trail != None:
								PupMenu("WARNING: One or more of the selected edges cannot be assigned because they are already allocated to an existing group. Doing nothing.%t")
								Window.EditMode(1)
								return None
							r.match_trail = newgroupname
							match_group_verts.append(r.verts[0])
							match_group_cuts.append(r.index)
							break
						elif r.index == len(self.rows) - 1 and e in r.edges[1] and r.match_lead == None:
							r.match_lead = newgroupname
							match_group_verts.append(r.verts[1])
							match_group_cuts.append(r.index + 1)
							break

				if not match_group_cuts:
					Window.EditMode(1)
					return None
				else:
					for i in range(1, self.num_rows):
						self.rows[i-1].match_lead = self.rows[i].match_trail

				self.match_groups_v.append(AGBMatchGroup(gindex, newgroupname, match_group_cuts))
				self.match_groups_v.sort(key=lambda x: x.name)


			# Create a new empty vert group named from the groupIndex int we've been passed:
			if newgroupname not in self.def_mesh.getVertGroupNames():
				self.def_mesh.addVertGroup(newgroupname)

			# assign the verts to the vert group
			self.def_mesh.assignVertsToGroup(newgroupname, match_group_verts, 1.0, Blender.Mesh.AssignModes.REPLACE)
			self.match_group_active = newgroupname
			Window.EditMode(1)
			return gindex

	def deleteMatchGroup(self, evtString, groupIndex):

		Window.EditMode(0)
		ret = None
		if evtString == 'evt_del_match_group_h':
			for g in self.match_groups_h:
				if g.index == groupIndex:
					for c in self.columns:
						if c.match_trail == g.name:
							c.match_trail = None
						if c.match_lead == g.name:
							c.match_lead = None
					self.def_mesh.removeVertGroup(g.name)
					self.match_groups_h.remove(g)
					self.match_group_active = None

		elif evtString == 'evt_del_match_group_v':
			for g in self.match_groups_v:
				if g.index == groupIndex:
					for r in self.rows:
						if r.match_trail == g.name:
							r.match_trail = None
						if r.match_lead == g.name:
							r.match_lead = None
					self.def_mesh.removeVertGroup(g.name)
					self.match_groups_v.remove(g)
					self.match_group_active = None

		self.deselectAllEdges()

		Window.EditMode(1)
		return None

	def setEdgeGroups(self, roadFacing, evtString):

		if evtString == 'evt_open_pnl_edge_type':
			Mesh.Mode(Mesh.SelectModes['EDGE'])
			Window.EditMode(1)

		elif evtString == 'evt_add_edge_group':
			Window.EditMode(0)
			edges_sel = set()
			for e in self.def_mesh.edges:
				if e.sel:
					edges_sel.add(e.index)

			if not edges_sel:
				PupMenu('Error: No edges are selected!%t')
				Window.EditMode(1)
				return None

	#		self.clearEdgeGroup(roadFacing)
			edge_group_verts = []
			edge_group_cuts  = []
			for edge_groups, cuts, lines, num_lines, groupnames in zip([self.edge_groups_h, self.edge_groups_v],
																	[self.cuts_h, self.cuts_v],
																	[self.columns, self.rows],
																	[self.num_columns, self.num_rows],
																	[["HR" , "HN"], ["VR", "VN"]] ):
				groupname 			= groupnames[0] if roadFacing else 	groupnames[1]
				edge_group_verts = []
				edge_group_cuts  = []

				for e in edges_sel:
					for c in lines:
						if e in c.edges[0]:
							if c.match_trail != None:
								PupMenu("WARNING: One or more of the selected edges cannot be assigned because they are already allocated to an existing group. Doing nothing.%t")
								Window.EditMode(1)
								return None
							c.match_trail = groupname
							if c.index > 0:
								lines[c.index - 1].match_lead = groupname
							edge_group_verts.append(c.verts[0])
							edge_group_cuts.append(c.index)
							break
						elif c.index == num_lines - 1 and e in c.edges[1] and c.match_lead == None:
							c.match_lead = groupname
							edge_group_verts.append(c.verts[1])
							edge_group_cuts.append(c.index + 1)
							break

				if not edge_group_cuts:
					continue
				else:
					for eg in edge_groups:
						if eg.name == groupname:
							eg.matches = edge_group_cuts
# 					print "edge_group.name, matches:", eg.name, eg.matches
					if groupname not in self.def_mesh.getVertGroupNames():
						self.def_mesh.addVertGroup(groupname)
					self.def_mesh.assignVertsToGroup(groupname, edge_group_verts,1.0, Blender.Mesh.AssignModes.REPLACE)

# 			print 'calling printTileLineData() from bp 1'
# 			self.printTileLineData(self.columns)
# 			self.printTileLineData(self.rows)
#self.selectEdgeGroup(roadFacing)

		Window.EditMode(1)

	def selectEdgeGroup(self, roadFacing):

		editmode = Window.EditMode()
		if editmode:
			Window.EditMode(0)

		def selectCut(c):
			for edge in c:
				self.def_mesh.edges[edge].sel = 1

		self.deselectAllEdges()
# 		print 'calling printTileLineData() from bp 2'
# 		self.printTileLineData(self.columns)
# 		self.printTileLineData(self.rows)
		for groups, cuts in zip([self.edge_groups_h, self.edge_groups_v], [self.cuts_h, self.cuts_v]):
			for g in groups:
				if g.index == (0 if roadFacing else 1):
# 					print 'len(cuts)', len(cuts)
# 					print 'g.name', g.name, 'g.index', g.index,'matches', g.matches
					for m in g.matches:
# 						print 'm', m
						selectCut(cuts[m])

		if editmode:
			Window.EditMode(1)

	def clearEdgeGroup(self, roadFacing):

		Window.EditMode(0)
		vgs = self.def_mesh.getVertGroupNames()

		index = 0 if roadFacing else 1
		for lines, group in zip([self.columns, self.rows], [self.edge_groups_h, self.edge_groups_v]):
			g = group[index]
			g.matches = []
			if g.name in vgs:
				verts = self.def_mesh.getVertsFromGroup(g.name)
				if verts:
					self.def_mesh.removeVertsFromGroup(g.name, verts)
			for line in lines:
				if line.match_trail == g.name:
					line.match_trail = None
				if line.match_lead == g.name:
					line.match_lead = None
		self.deselectAllEdges()
		Window.EditMode(1)

	def printTileLineData(self, tileLineList):
		for t in tileLineList:
			print 'row' if t.row else 'col', t.index, '\tm_trail:', t.match_trail, '\tm_lead:', t.match_lead, '\tslop %.3f' % (t.min_overlap if t.is_slop else '            '), 'antislop %.3f' % (t.max_antislop if t.is_antislop else '           ')

	def getTilelineConstraints(self):
		"""
			Compresses the various tileline constraint attributes into a dict of dicts of dicts ([axis][index][attrs])
		"""
		constraints = {'grid_name':self.name}
		for k, tilelines in zip(['columns', 'rows'], [self.columns, self.rows]):
			cs = {}
			for i, t in enumerate(tilelines):
				cs[str(i)] = t.packConstraints()
			constraints[k] = cs
		return constraints

	def setTilelineConstraints(self, data):

		if not data:
			print "ERROR: no data passed to setTilelineConstraints()%t"
			return False
		if data['grid_name'] != self.name:
			print "ERROR: Spelling constraint configuration doesn't match the grid object%t"
			return False

		for k, tilelines in zip(['columns', 'rows'], [self.columns, self.rows]):
			try:
				if len(tilelines) != len(data[k]):
					print "ERROR: tileline/constraint_data length mismatch.%t"
					return False
				for t in tilelines:
					t.unpackConstraints(data[k][str(t.index)])
			except StandardError, detail:
				print 'ERROR', detail
				return False
		return True

	def createDefaultMesh(self, meshLayer):

		""" Build a grid mesh in its default 'rest postiion' and parent it to a new wrapper empty
		"""

		editmode = Window.EditMode()
		if editmode:
			Window.EditMode(0)

		all_coords = []
		all_edges  = []
		all_faces  = []

		#
		# Build a manifold grid, taking care to keep verts, edges and faces in a predictable order.
		#

		# Add coords in order L->R, F->B
		i = 0
		for v, y in enumerate(self.offsets_y):
			for h, x in enumerate(self.offsets_x):
				all_coords.append([x, y, 0])
				i += 1

		# Add vertical edges in order F->B, L->R
		j, k = 0, self.num_columns + 1
		index = 0
		for x in range(k):
			j = x
			for y in range(self.num_rows):
				all_edges.append((j, j+k))
				j += k

		# Add horizontal edges in order L->R, F->B
		# DON'T zero 'index' - the edge INDICES are cumulative!
		i, k = 0, self.num_rows + 1
		for y in range(k):
			for x in range(self.num_columns):
				all_edges.append( (i, i + 1))
				index += 1
				i += 1
			i += 1

		# Add faces in order L->R, F->B
		i = 0
		j = self.num_columns + 1
		for y in range(self.num_rows):
			for x in range(self.num_columns):
				all_faces.append([i, i+1, j+1, j])
				i += 1
				j += 1
			i += 1
			j += 1

		#
		# Test the newly created mesh data against the original - if the face, edge and vert index orders are the same, use the original.
		# If not, build a new one.
		#

		self.is_pukka = True
		# verts
		for vo, vn in zip(self.init_mesh.verts, all_coords):
#			print 'vo.co', vo.co, 'vn', vn
			if not self.realClose(vo.co, vn, AGBGrid.VERT_LIMIT):
				self.is_pukka = False
				break
		else:
			#edges
			for eo, en in zip(self.init_mesh.edges, all_edges):
				if not eo.key == en:
					self.is_pukka = False
					break

			else:
				#faces
				for fo, fn in zip(self.init_mesh.faces, all_faces):
					for vo, vn in zip(fo.verts, fn):
						if not vo.index == vn:
							self.is_pukka = False
					if not self.is_pukka:
						break

		scn = bpy.data.scenes.active

		if not self.is_pukka:
			print "NOTE: Grid mesh is not pukka. Building a replacement now.\n"
# 			p = self.init_ob.parent
# 			if p != None:
# 				if self.AGB_ob and p.parent == self.AGB_ob.wrapper:
#
			if not self.def_mesh:
				self.def_mesh = bpy.data.meshes.new('XPToolsAGBGrid')

			self.def_mesh.verts = None
			self.def_mesh.verts.extend(all_coords)
			self.def_mesh.edges.extend(all_edges)
			self.def_mesh.faces.extend(all_faces)

			if self.has_uvs:
				i = 0
				while i < len(self.def_mesh.faces):
					for y in self.tiles:
						for x in y:
							self.def_mesh.faces[i].uv = tuple(x.uvs)
							#print "coords", x.coords, "uvs", x.uvs
							i += 1
				for new in self.def_mesh.faces:
					for old in self.init_mesh.faces:
						if self.realClose(old.cent, new.cent, AGBGrid.VERT_LIMIT):
							new.mode = old.mode
							break

			if self.has_image:
				for f in self.def_mesh.faces:
					f.image = self.image
					f.transp = Mesh.FaceTranspModes.ALPHA

			# Create a new blender ob and link it to the new mesh:
			# We want to keep the name of the original ob, so to avoid getting blender .00x suffix
			newname = self.init_ob.name if self.init_ob.name.startswith('TILE') else 'TILE' + self.init_ob.name[:13]
			self.def_mesh_ob = scn.objects.new(self.def_mesh, newname)
			if self.def_mesh_ob not in scn.objects:
				scn.link(self.def_mesh_ob)
			self.def_mesh_ob.setMatrix(self.init_ob.matrix)
			self.def_mesh_ob.layers = [meshLayer]

			# Unparent the initOb and substitute the new one
			if self.wrapper != None:
				self.init_ob.clrParent(2)
				self.wrapper.makeParent([self.def_mesh_ob])

			# Test for possible group membership:
			for grp in bpy.data.groups:
				if self.init_ob in grp.objects:
					grp.objects.link(self.def_mesh_ob)

			self.init_ob.select(0)
			self.def_mesh_ob.select(1)
			if AGBEditor.DELETE_INIT_OB_ON_LOAD:
				self.deleteInitOb()
			self.is_pukka = True
		else:
			# If pukka, keep the initOb AND mesh:
			self.def_mesh_ob =  self.init_ob
			self.def_mesh = self.init_mesh
			self.def_mesh_ob.select(1)
			self.def_mesh_ob.makeDisplayList()
		#if 'AGBProps' not in self.def_mesh_ob.properties:
		self.def_mesh_ob.properties['AGBProps'] = {'IsPukka': True, 'NumVerts': len(self.def_mesh.verts)}

		if editmode:
			Window.EditMode(1)

	def deleteInitOb(self):
		if self.init_ob != None and self.init_ob != self.def_mesh_ob:
			scn = bpy.data.scenes.active

			scn.objects.unlink(self.init_ob)
			self.init_ob 	= None
			self.init_ob 	= self.def_mesh_ob
			self.init_mesh 	= self.def_mesh

	def selectConstraintFaces(self, spellingsAxis, constraintMode):

		modes = ['exclude', 'require', 'center', 'xor']
		Mesh.Mode(Mesh.SelectModes['FACE'])
		editmode = Window.EditMode()
		if editmode:
			Window.EditMode(0)

		all_faces = []
		if spellingsAxis:
			for t in (t for t in self.rows if getattr(t, modes[constraintMode])):
				all_faces.extend(t.faces)
		else:
			for t in (t for t in self.columns if getattr(t, modes[constraintMode])):
				all_faces.extend(t.faces)

		self.def_mesh.sel = False

		for f in all_faces:
			self.def_mesh.faces[f].sel = 1

		if editmode:
			Window.EditMode(1)

	def getMeshFaceData(self):
		"""
		Get the mesh's facemodes and vertex groups and update our AGBTileLine data
		"""
		# facemodes first:
		attrs = [('DYNAMIC', 'is_antislop', False, 'max_antislop'),   ('TWOSIDE', 'is_slop', True, 'min_overlap'),   ('INVISIBLE', 'ignore', True, None)]

		for c in self.columns:
#			print "COLUMN", c.index, "\nc.verts:", c.verts, "\nc.edges:", c.edges, "\nc.faces:", c.faces
			c.is_slop = c.is_antislop = c.ignore = True
			for a in attrs:
				for i in c.faces:
					if self.testFace(self.def_mesh.faces[i], a[0], a[2]):
						setattr(c, a[1], False)
						if a[3] != None:
							setattr(c, a[3], 0)
						break

		for r in self.rows:
#			print "ROW", r.index, "\nr.verts:", r.verts, "\nr.edges:", r.edges, "\nr.faces:", r.faces
			r.is_slop = r.is_antislop = r.ignore = True
			for a in attrs:
				for i in r.faces:
					if self.testFace(self.def_mesh.faces[i], a[0], a[2]):
						setattr(r, a[1], False)
						if a[3] != None:
							setattr(r, a[3], 0)
						break

		# Do we have at least one slop zone in EACH axis?
		self.need_slop = False if any(filter(lambda x: x.is_slop, self.columns)) and any(filter(lambda x: x.is_slop, self.rows)) else True

		# Do we have at least one antislop row or column?
		self.has_antislop = True if any(filter(lambda x: x.is_antislop, self.columns)) or any(filter(lambda x: x.is_antislop, self.rows)) else False

		# Do we have at least one ignored row or column?
		self.has_ignore = True if any(filter(lambda x: x.ignore, self.columns)) or any(filter(lambda x: x.ignore, self.rows)) else False

	def getMetaRectData(self):

		if not self.wrapper:
			return False
		scn = bpy.data.scenes.active
		num_rects = 0
		for ob in getChildren(self.wrapper, scn.objects):
			if ob.type == 'Mesh' and ob.name.upper().startswith('META'):
				me = ob.getData(mesh=1)
				if len(me.faces) != 1:
					PupMenu("ERROR: Meta rects must be single quads!%t")
					return False
				f = me.faces[0]
				if len(f.verts) != 4:
					PupMenu("ERROR: Meta rects must be single quads!%t")
					return False
				m_verts = []
				ob_mat = ob.getMatrix('localspace')
				for v in f.verts:
					vt = xform(v,ob_mat)
					m_verts.append(vt)
				grid_mat = self.def_mesh_ob.getMatrix('localspace')
				self.def_mesh.verts
				gv0 =  xform(self.def_mesh.verts[0], grid_mat)
				gv1 =  xform(self.def_mesh.verts[-1], grid_mat)

				offsets = [ min(m_verts, key = lambda x: x[0])[0] - gv0[0], min(m_verts, key = lambda x: x[1])[1] - gv0[1],
						    max(m_verts, key = lambda x: x[0])[0] - gv1[0], max(m_verts, key = lambda x: x[1])[1] - gv1[1] ]
				if ob.name.upper().startswith('METAEDGE_BEST'):
					self.metarect_data['edge_best'] = offsets
					num_rects += 1
				elif ob.name.upper().startswith('METAEDGE_MIN'):
					self.metarect_data['edge_min'] = offsets
					num_rects += 1
				elif ob.name.upper().startswith('METABLOCK_EDGE'):
					self.metarect_data['edge_max'] = [offsets[0] - 4.25, offsets[1] - 4.25, offsets[2] + 4.25, offsets[3] + 4.25]
					num_rects += 1
		if num_rects == 3:
			self.got_meta_rects = True
# 			for k,v in self.metarect_data.iteritems():
# 				print k , v
			return True
		self.got_meta_rects = False
		return False

	def addMetaRects(self, offsetBest, offsetMin, blockEdge, layer):
		"""
			Automatically add three metarects to the grid, at plausible provisional locations
		"""
		scn = bpy.data.scenes.active

		x0, x1, y0, y1 = self.offsets_x[0], self.offsets_x[-1], self.offsets_y[0], self.offsets_y[-1]
		rect_names = ['METAEDGE_BEST', 'METAEDGE_MIN', 'METABLOCK_EDGE']
		insets	   = [offsetBest, offsetMin, blockEdge]
		rects = []
		for rectname, n in zip(rect_names, insets):
			newmesh = bpy.data.meshes.new('XPToolsMetaRect')
			newmesh.verts.extend([[x0+n, y0+n, 0.0],[x1-n, y0+n, 0.0], [x1-n, y1-n, 0.0], [x0+n, y1-n, 0.0]])
			newmesh.faces.extend([0, 1, 2, 3])
			ob = scn.objects.new(newmesh, rectname)
			ob.loc = self.loc
			ob.drawType = Object.DrawTypes['WIRE']
			rects.append(ob)
		for r in rects:
			r.layers = [layer]
		if self.wrapper:
			self.wrapper.makeParent(rects)
		Window.RedrawAll()

	def computeSpellingMetrics(self, axis, spelling):
		"""
			Computes the minimum and maximum length in metres of a list of tile indices,
			taking into account the available range of slop.
		"""

		l_min, l_max = 0.0, 0.0
		try:
			if axis == 0:
				for t in spelling:
					l_min += self.columns[t].len_min
					l_max += self.columns[t].len_max
				return (fractionize(l_min + self.metarect_data['edge_best'][2] - self.metarect_data['edge_best'][0],  AGBGrid.SPELLING_ROUNDING_DENOMINATOR,  1),
							fractionize(l_max + self.metarect_data['edge_best'][2] - self.metarect_data['edge_best'][0],  AGBGrid.SPELLING_ROUNDING_DENOMINATOR))

			elif axis == 1:
				for t in spelling:
					l_min += self.rows[t].len_min
					l_max += self.rows[t].len_max
				return (fractionize(l_min + self.metarect_data['edge_best'][3] - self.metarect_data['edge_best'][1],  AGBGrid.SPELLING_ROUNDING_DENOMINATOR,  1),
							fractionize(l_max + self.metarect_data['edge_best'][3] - self.metarect_data['edge_best'][1],  AGBGrid.SPELLING_ROUNDING_DENOMINATOR))
		except IndexError:
			PupMenu("ERROR: list index overrun in AGBGrid.computeSpellingMetrics()%t")
			return (0.0, 0.0)

	def getLengthAt(self, axis, spelling, index):
		"""
			Returns the min and maximum lengths between the start of the spelling and the centre of the tile at position index of the spelling.
			Takes into account the inset fo the initial EDGEBEST metarect
		"""
		try:
			if axis == 0:
				return (reduce(lambda x, y: x+y, (self.columns[t].len_min for t in spelling[:index]), 0.0) + self.columns[spelling[index]].len_min*0.5 - self.metarect_data['edge_best'][0],
						reduce(lambda x, y: x+y, (self.columns[t].len_max for t in spelling[:index]), 0.0) + self.columns[spelling[index]].len_max*0.5 - self.metarect_data['edge_best'][0])

			elif axis == 1:
				return (reduce(lambda x, y: x+y, (self.rows[t].len_min for t in spelling[:index]), 0.0) + self.rows[spelling[index]].len_min*0.5 - self.metarect_data['edge_best'][1],
						reduce(lambda x, y: x+y, (self.rows[t].len_max for t in spelling[:index]), 0.0) + self.rows[spelling[index]].len_max*0.5 - self.metarect_data['edge_best'][1])

		except IndexError:
			print "ERROR: list index overrun in AGBGrid.computeSpellingMetrics(), spelling: %s, index: %d" % (str(spelling), index)
			return (0.0, 0.0)

	def getDuplis(self, glist, indexS, indexT, loc):
		cellname = "XPToolsDupGrp%02d%02d" % (indexS, indexT)
		for g in self.dgroups:
			if cellname in g.name:
				glist.append([g, loc])

	def createPreviewMeshFromSpellings(self, loc, spellS, spellT, targetS=None, targetT=None, previewObs=None, doAnnotations=True, dupliObs=None, source='temp'):
		"""
		We are passed a worldspace location at which to build a new mesh, a pair of spellings, and optionally a pair of target lengths.
		"""
		Z_OFFSET = -0.05		# Ugly hack to avoid z-buffer thrash in the blender 3d view: we draw slop tiles with a small negative z-axis offset, i.e. slightly below 'regular' tiles.

		# Calculate size adjustments due to slop usage to fit the target lengths
		# It turns out that antislop is invariable i.e. it is always applied, regardless of the compression of the block.

	#	print "targetS, targetT" , targetS, targetT
		percent_slop  = [0.0, 0.0]

		if targetS != None and targetT != None:
			for axis, tileline, spelling, target, pc_slop in zip([0, 1], [self.columns, self.rows], [spellS, spellT], [targetS, targetT], percent_slop):
				spell_min, spell_max =  self.computeSpellingMetrics(axis, spelling)
	#			print "spell_min, spell_max", spell_min, spell_max
				available_slop = reduce(lambda x, y: x+y, [t.len_max for t in tileline if t.is_slop], 0.0)

				slop_required = spell_max - target
				if available_slop < slop_required:
					PupMenu("ERROR: Insufficient slop in the current configuration for %s spelling %s.%%t" % ("vertical" if axis else "horizontal", str(spelling)) )
					continue
				pc_slop = 0.0
				try:
					percent_slop[axis] = slop_required/available_slop
				except ZeroDivisionError:
					print "ERROR: ZeroDivisionError in createPreviewMeshFromSpellings()"

		scn = bpy.data.scenes.active
		scn.objects.selected = []

		all_verts = []
		all_faces = []
		all_uvs	  = []
		dupli_ob_data  = []

		v_index = 0

		running_s = running_t = 0
		for r in spellT:
			if self.rows[r].is_antislop:
				running_t -= self.rows[r].max_antislop
			len_t_adjusted = self.rows[r].len if not self.rows[r].is_slop else self.rows[r].min_overlap + self.rows[r].len_max * (1.0 - percent_slop[1])

			running_s = 0
			for c in spellS:
				if self.columns[c].is_antislop:
					running_s -= self.columns[c].max_antislop
				len_s_adjusted = self.columns[c].len if not self.columns[c].is_slop else (self.columns[c].min_overlap + self.columns[c].len_max * (1.0 - percent_slop[0]))

				all_verts += self.tiles[r][c].getCoordsAt(running_s, running_t, len_s_adjusted, len_t_adjusted, Z_OFFSET if self.columns[c].is_slop or self.rows[r].is_slop else 0.0)
				all_faces.append([v_index, v_index + 1, v_index + 2, v_index + 3])
				all_uvs.append(self.tiles[r][c].getUVs(len_s_adjusted, len_t_adjusted))
				v_index += 4

				if doAnnotations:
					self.getDuplis(dupli_ob_data, c, r, Vector(running_s, running_t, 0.0))

				running_s += len_s_adjusted
				running_s -= self.columns[c].min_overlap
			running_t += len_t_adjusted
			running_t -= self.rows[r].min_overlap

		if all_verts and all_faces and all_uvs:
			try:
				ob = previewObs.pop()
				if ob not in scn.objects:
					scn.objects.link(ob)

			except:
				newmesh = bpy.data.meshes.new('XPTAGBPrevMesh')
				ob = scn.objects.new(newmesh, 'XPTAGBPreview')

			me = ob.getData(mesh=1)
			me.verts = None
			me.verts.extend(all_verts)
			me.faces.extend(all_faces)
			for f in me.faces:
				f.uv = tuple(all_uvs[f.index])
				f.image = self.image
				f.transp = Mesh.FaceTranspModes.ALPHA

	#		me.remDoubles(0.01)
			ob.loc = loc
			setProp(ob, "XPToolsAGBSp_S", " ".join(map(str, spellS)))
			setProp(ob, "XPToolsAGBSp_T", " ".join(map(str, spellT)))
			setProp(ob, "XPToolsAGBSrc" , source)
			if doAnnotations:

				# We've built a list of dupligroups and corresponding localspace locations.
				# Now convert the locations to worldspace and add empty objects, enabling dupGroups and specifying the corresponding group

				mat = ob.getMatrix('worldspace')
				for d in dupli_ob_data:
					try:
						dup = dupliObs.pop()
						if dup not in scn.objects:
							scn.objects.link(dup)

					except:
						dup = scn.objects.new('Empty', 'XPToolsAGBDupl')
						dup.drawSize = 0.01
						dup.emptyShape = Blender.Object.EmptyShapes['AXES']
						dup.enableDupGroup = True

					dup.DupGroup = d[0]
					dup.loc = mat.rotationPart() * d[1] + mat.translationPart()

		return [running_s, running_t]

	def loosenPack(self, inList, minSpacing):
		"""
			This tries to gracefully limit the minimum difference in values between consecutive numbers in a sequence.
		"""
		seq = inList
# 		print 'seq', seq
		length = len(seq)
		diffs = []
		for i in range(1, length):
			diffs.append(seq[i][1] - seq[i-1][1])
		m = min(diffs)
		if m >= minSpacing:
			return seq
		else:
			delta = (minSpacing - m) * 0.5
			lower = diffs.index(m)
			upper = diffs.index(m) + 1
			seq[upper][1] += int(delta)
			seq[lower][1] -= int(delta)
			while upper < (length - 1):
				seq[upper+1][1] = max(seq[upper+1][1], seq[upper][1] + minSpacing)
				upper += 1
			while lower > 0:
				seq[lower-1][1] = min(seq[lower-1][1], seq[lower][1] - minSpacing)
				lower -= 1
			return seq

	def setGUITileData(self, x, y, w, h, bugWidth, bugHeight, panelMode, index=0):
		"""
			We're passed the location and size of the grid (in pixels) as drawn in the GUI, and we calculate the sizes of zone rectangles to draw over it,
			and the locations of the data-entry bugs (blender string-type buttons).
		"""
		BUGW  = bugWidth
		BUGH  = bugHeight
		FNTH  = 6						# font height
		ppmx  = w / (self.offsets_x[-1] - self.offsets_x[0])
		ppmy  = h / (self.offsets_y[-1] - self.offsets_y[0])
		self.GUI_zone_rects = []
		self.GUI_chequerboard = []
		bugx  = []
		bugy  = []

		self.GUI_label_locs_x = []
		self.GUI_label_locs_y = []
		nudge = 0						# This is a flag to nudge the label up/along a bit when labels are very close together
		for c in self.columns:
			if not c.ignore:
				offset = int(x + (c.loc + c.len * 0.5)*ppmx)
				self.GUI_label_locs_x.append([offset, nudge, str(c.index)])
				if len(self.GUI_label_locs_x)> 1:
					if (GetStringWidth(self.GUI_label_locs_x[-1][2], 'tiny') + GetStringWidth(self.GUI_label_locs_x[-2][2], 'tiny')) * 0.5 >=  (self.GUI_label_locs_x[-1][0] - self.GUI_label_locs_x[-2][0]):
						nudge = 1 - nudge
						self.GUI_label_locs_x[-1][1] = nudge
					else:
						self.GUI_label_locs_x[-1][1] = 0
		nudge = 0
		for c in self.rows:
			if not c.ignore:
				offset = int(y + (c.loc + c.len * 0.5)*ppmy - FNTH/2)
				self.GUI_label_locs_y.append([offset, nudge, str(c.index)])
				if len(self.GUI_label_locs_y)> 1:
					if FNTH >=  (self.GUI_label_locs_y[-1][0] - self.GUI_label_locs_y[-2][0]):
						nudge = 1 - nudge
						self.GUI_label_locs_y[-1][1] = nudge
					else:
						self.GUI_label_locs_y[-1][1] = 0

		cbx = []
		for c in self.columns:
			cbx.append( x + int(c.loc * ppmx))
		cbx.append(cbx[-1] + int(self.columns[-1].len * ppmx))
		cby = []
		for r in self.rows:
			cby.append( y + int(r.loc * ppmy))
		cby.append(cby[-1] + int(self.rows[-1].len * ppmy))

		for j in range(self.num_rows):
			for	i in range(j%2, self.num_columns, 2):
				self.GUI_chequerboard.append( [cbx[i], cby[j], cbx[i+1] - cbx[i],  cby[j+1]- cby[j]] )

		if panelMode == 'slop':
			cls = [int(x), int(x+w)]
			# What's that? Where alpha transparent row and column rects cross over each other, their alpha values combine and it looks lame.
			# Workaround is to split the rows into multiple, shorter rects, so that they don't overlap the columns. 'cls' is an accumulation of 'start' and 'end' points for these rects.
			bugx = []
			for c in self.columns:
				if c.is_slop:
					left  = x + c.loc*ppmx
					width = max(c.len*ppmx, 1)
					self.GUI_zone_rects.append( [int(left), 	int(y), 		int(width), 	int(h)]	)
					cls[-1:-1] =  [int(left), int(left) + int(width)]
					bugx.append( [c.index, int(left + width*0.5), niceFround(c.min_overlap, 4)] )
			if len(bugx) > 1:
				self.GUI_bugs_x[:] = self.loosenPack(bugx, BUGW)
			else:
				self.GUI_bugs_x[:] = bugx
			bugy = []
			for r in self.rows:
				if r.is_slop:
					bottom  = y + r.loc*ppmy
					height  =  max(r.len*ppmy, 1)
					for i in range(0, len(cls), 2):
						self.GUI_zone_rects.append( [cls[i], 		int(bottom),	cls[i+1] - cls[i],   		int(height)])
					bugy.append( [r.index, int(bottom + height*0.5), niceFround(r.min_overlap, 4)] )
			if len(bugy) > 1:
				self.GUI_bugs_y[:] = self.loosenPack(bugy, BUGH)
			else:
				self.GUI_bugs_y[:] = bugy

		elif panelMode == 'antislop':
			cls = [int(x), int(x+w)]
			bugx = []
			for c in self.columns:
				if c.is_antislop:
					left  = x + c.loc*ppmx
					width = max(c.len*ppmx, 1)
					self.GUI_zone_rects.append( [int(left), 	int(y), 		int(width), 	int(h)]	)
					cls[-1:-1] =  [int(left), int(left) + int(width)]
					bugx.append( [c.index, int(left + width*0.5), niceFround(c.max_antislop, 4)] )
			if len(bugx) > 1:
				self.GUI_bugs_x[:] = self.loosenPack(bugx, BUGW)
			else:
				self.GUI_bugs_x[:] = bugx
			bugy = []
			for r in self.rows:
				if r.is_antislop:
					bottom  = y + r.loc*ppmy
					height  =  max(r.len*ppmy, 1)
					for i in range(0, len(cls), 2):
						self.GUI_zone_rects.append( [cls[i], 		int(bottom),	cls[i+1] - cls[i],   		int(height)])
					bugy.append( [r.index, int(bottom + height*0.5), niceFround(r.max_antislop, 4) ] )
			if len(bugy) > 1:
				self.GUI_bugs_y[:] = self.loosenPack(bugy, BUGH)
			else:
				self.GUI_bugs_y[:] = bugy

		elif panelMode == 'ignore':
			cls = [int(x), int(x+w)]
			for c in self.columns:
				if c.ignore:
					left  = x + c.loc*ppmx
					width = max(c.len*ppmx, 1)
					self.GUI_zone_rects.append( [int(left), 	int(y), 		int(width), 	int(h)]	)
					cls[-1:-1] =  [int(left), int(left) + int(width)]
			for r in self.rows:
				if r.ignore:
					bottom  = y + r.loc*ppmy
					height  =  max(r.len*ppmy, 1)
					for i in range(0, len(cls), 2):
						self.GUI_zone_rects.append( [cls[i], 		int(bottom),	cls[i+1] - cls[i],   		int(height)])

		elif panelMode in ['match_grp', 'edge_type']:

			for k in self.GUI_rects.iterkeys():
				self.GUI_rects[k] = []

			for groups, lines in zip([self.edge_groups_h, self.edge_groups_v],  [self.columns, self.rows]):

				for line in lines:
					rect_trail_c = [int(x + line.loc*ppmx),			  	int(y), 	1, 			int(h)]
					rect_trail_r = [int(x), 		int(y + line.loc*ppmy),			int(w),   	1	  ]

					if line.match_trail == None:
						if line.column:
							self.GUI_rects['untagged'].append(rect_trail_c)
						elif line.row:
							self.GUI_rects['untagged'].append(rect_trail_r)

					elif panelMode == 'match_grp':
						if line.match_trail == self.match_group_active:
							if line.column:
								self.GUI_rects['match_group_active'].append(rect_trail_c)
							elif line.row:
								self.GUI_rects['match_group_active'].append(rect_trail_r)
						elif line.match_trail[-1].isdigit():
							if line.column:
								self.GUI_rects['match_group'].append(rect_trail_c)
							elif line.row:
								self.GUI_rects['match_group'].append(rect_trail_r)
						elif line.match_trail in ['HR','VR']:
							if line.column:
								self.GUI_rects['edge_road'].append(rect_trail_c)
							elif line.row:
								self.GUI_rects['edge_road'].append(rect_trail_r)
						elif line.match_trail in ['HN','VN']:
							if line.column:
								self.GUI_rects['edge_no_road'].append(rect_trail_c)
							elif line.row:
								self.GUI_rects['edge_no_road'].append(rect_trail_r)

					elif panelMode == 'edge_type':
						if line.match_trail in ['HR','VR']:
							if line.column:
								self.GUI_rects['edge_road_active' if not index else 'edge_road'].append(rect_trail_c)
							elif line.row:
								self.GUI_rects['edge_road_active' if not index else 'edge_road'].append(rect_trail_r)
						elif line.match_trail in ['HN','VN']:
							if line.column:
								self.GUI_rects['edge_no_road_active' if index else 'edge_no_road'].append(rect_trail_c)
							elif line.row:
								self.GUI_rects['edge_no_road_active' if index else 'edge_no_road'].append(rect_trail_r)
						elif line.match_trail[-1].isdigit():
							if line.column:
								self.GUI_rects['match_group'].append(rect_trail_c)
							elif line.row:
								self.GUI_rects['match_group'].append(rect_trail_r)

				if len(lines) > 0:
					rect_lead_c  = [int(x + (lines[-1].loc + lines[-1].len)*ppmx), int(y), 	1, 			int(h)]
					rect_lead_r  = [int(x), int(y + (lines[-1].loc + lines[-1].len)*ppmy),	int(w),   	1     ]

					if lines[-1].match_lead == None:
						if lines[-1].column:
							self.GUI_rects['untagged'].append(rect_lead_c)
						elif lines[-1].row:
							self.GUI_rects['untagged'].append(rect_lead_r)

					elif panelMode == 'match_grp':
						if lines[-1].match_lead == self.match_group_active:
							if lines[-1].column:
								self.GUI_rects['match_group_active'].append(rect_lead_c)
							elif lines[-1].row:
								self.GUI_rects['match_group_active'].append(rect_lead_r)
						elif lines[-1].match_lead[-1].isdigit():
							if lines[-1].column:
								self.GUI_rects['match_group'].append(rect_lead_c)
							elif lines[-1].row:
								self.GUI_rects['match_group'].append(rect_lead_r)
						elif lines[-1].match_lead in ['HR','VR']:
							if line.column:
								self.GUI_rects['edge_road'].append(rect_lead_c)
							elif line.row:
								self.GUI_rects['edge_road'].append(rect_lead_r)
						elif lines[-1].match_lead in ['HN','VN']:
							if line.column:
								self.GUI_rects['edge_no_road'].append(rect_lead_c)
							elif line.row:
								self.GUI_rects['edge_no_road'].append(rect_lead_r)

					elif panelMode == 'edge_type':
						if lines[-1].match_lead in ['HR','VR']:
							if line.column:
								self.GUI_rects['edge_road_active' if not index else 'edge_road'].append(rect_lead_c)
							elif line.row:
								self.GUI_rects['edge_road_active' if not index else 'edge_road'].append(rect_lead_r)
						elif lines[-1].match_lead in ['HN','VN']:
							if line.column:
								self.GUI_rects['edge_no_road_active' if index else 'edge_no_road'].append(rect_lead_c)
							elif line.row:
								self.GUI_rects['edge_no_road_active' if index else 'edge_no_road'].append(rect_lead_r)
						elif lines[-1].match_lead[-1].isdigit():
							if line.column:
								self.GUI_rects['match_group'].append(rect_lead_c)
							elif line.row:
								self.GUI_rects['match_group'].append(rect_lead_r)

	def doAnnotations(self):
		"""
			Determine which annotations belong to which tiles; create and assign duplgroups; set dupligroup offsets.
		"""
		annotations = []
		self.dgroups = []

		scn = bpy.data.scenes.active
		dg_prefix = "XPToolsDupGrp"

		# Clear any existing XPT dupligroups of objects:
		for d in [d for d in Blender.Group.Get() if dg_prefix in d.name]:
			d.objects = []

		prefixes = ['OBJ', 'TREE', 'FAC']

		for ob in getChildren(self.wrapper, scn.objects):
			for pre in prefixes:
				if ob.name.startswith(pre):
					annotations.append([ob, ob.getMatrix('worldspace')[3][0:3]])

		mm = self.def_mesh_ob.getMatrix("worldspace")

		offsets_x_ws = []
		offsets_y_ws = []

		for x in self.offsets_x:
			offsets_x_ws.append(x + self.loc[0])
		for y in self.offsets_y:
			offsets_y_ws.append(y + self.loc[1])

		for j in range(len(offsets_y_ws)-1):
			for i in range(len(offsets_x_ws)-1):
				for a in annotations:
					if offsets_x_ws[i] <= a[1][0] < offsets_x_ws[i+1]:
						if offsets_y_ws[j] <= a[1][1] < offsets_y_ws[j+1]:
							if a[0].type == "Empty":
								aa = getChildren(a[0], scn.objects)
								aa.append(a[0])
							else:
								aa = [a[0]]
							if dg_prefix + "%02d%02d" % (i, j) in [dg.name for dg in self.dgroups]:
								myg = dg
							else:
								myg = Blender.Group.New(dg_prefix + "%02d%02d" % (i, j))
								myg.dupliOffset = xform(self.def_mesh.verts[j*(self.num_columns + 1) + i], mm)
								self.dgroups.append(myg)
							for k in aa:
								myg.objects.link(k)
		#for g in self.dgroups:
		#	print g.name
		#	for o in g.objects:
		#		print o.name, o.loc

	def add3DLabels(self):
		""" Adds a 3-d text mesh ob at the tile grid origin, denoting the row and column indices
		"""

		scn = bpy.data.scenes.active
		if self.text3d_ob != None:
			return
		TEXT3D_SIZE = 2.0


		t3d = Text3d.New('t3d')

		for ob in Object.Get():
			if ob.name.startswith('XPToolsText3d') and ob.users == 0:
				ob_t = ob
				break
		else:
			ob_t = scn.objects.new(t3d, 'XPToolsText3d')

		if ob_t not in scn.objects:
			scn.link(ob_t)

		me = bpy.data.meshes.new('temp')
		ob = scn.objects.new(me)
		mesh = bpy.data.meshes.new('XPToolsAGBLabels')
		text3dob = scn.objects.new(mesh)
		text3dob.loc = self.loc

		for i, axis in enumerate([self.columns, self.rows]):
			for line in axis:

				t3d.setText(str(line.index))
				t3d.setSize(TEXT3D_SIZE)
				t3d.setAlignment(Text3d.MIDDLE)
				t3d.setDefaultResolution(4)

				me.getFromObject(ob_t)

				if i == 0:
					ob.loc = Vector(self.loc) + Vector(line.loc + line.len*0.5, -TEXT3D_SIZE, 0.0)
				elif i == 1:
					ob.loc = Vector(self.loc) + Vector(- TEXT3D_SIZE * 0.875, line.loc + line.len * 0.5, 0.0)

				text3dob.join([ob])

		#text3dob.makeDisplayList
	#	elapsed("bp5 add3dLabels()")
		scn.unlink(ob)
		scn.unlink(ob_t)
		self.text3d_ob = text3dob
		self.text3d_ob.sel = 0
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBTile(object):

	def __init__(self, indX, indY, vertIndex, x0, y0, x1, y1):

		self.ind_x			= indX
		self.ind_y			= indY
		self.x0				= x0
		self.x1 			= x1
		self.y0				= y0
		self.y1 			= y1
		self.coords			= [[x0, y0, 0], [x1, y0, 0], [x1, y1, 0], [x0, y1, 0]]
		self.face			= [vertIndex, vertIndex + 1, vertIndex + 2, vertIndex + 3]
		self.uvs 			= []
		self.dupligroups	= []

	def getCoordsAt(self, x, y, lengthReq=None, heightReq=None, zOffset=0.0):
		leng = (self.x1 - self.x0) if lengthReq == None else lengthReq
		hgt	= (self.y1 - self.y0) if heightReq == None else heightReq
		return [[x, y, zOffset], [x+leng, y, zOffset], [x+leng, y+hgt, zOffset], [x, y+hgt, zOffset]]

	def getUVs(self, lengthReq=None, heightReq=None):
		"""
			lengthRatio and heightRatio specify the desired tile size relative to the default values, for slop cropping.
		"""
		if lengthReq == None and heightReq == None:
			return self.uvs
		u0 = self.uvs[0][0]
		v0 = self.uvs[0][1]
		u1 = self.uvs[2][0]
		v1 = self.uvs[2][1]
		lengthRatio = lengthReq / (self.x1 - self.x0)
		heightRatio = heightReq / (self.y1 - self.y0)
		return [ Vector(u0, v0), 	Vector(u0+(u1-u0)*lengthRatio, v0), 	Vector(u0+(u1-u0)*lengthRatio, v0+(v1-v0)*heightRatio), 	Vector(u0, v0+(v1-v0)*heightRatio)]

	def setUVs(self, UV_origin, scaleX, scaleY):
		self.uvs[:] = []
		for v in self.coords:
			self.uvs.append( Vector([v[0] * scaleX + UV_origin[0], v[1] * scaleY + UV_origin[1]]) )
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBTileLine(object):

	def __init__(self, lineType, index, numCols, numRows, offsetTrail, offsetLead):

		self.row			= True if lineType == 'row'	else False
		self.column			= True if lineType == 'column' else False
		self.index			= index							# zero-based spelling index
		if self.column:
			self.verts		= [index, index + 1]			# list of two vert indicies: [trailing-edge, leading edge], both on the edge of the grid

			self.edges		= [ range(index*numRows, (index+1)*numRows) , range((index+1)*numRows, (index+2)*numRows)]	# list of two lists of edge indices

			self.faces		= range(index, numCols*numRows, numCols)
		elif self.row:
			self.verts	 	= [index * (numCols + 1), (index + 1)*(numCols + 1)]
			num_col_edges	= numRows * (numCols+1)
			self.edges		= [ range(index*numCols+num_col_edges, (index+1)*numCols+num_col_edges),  range((index+1)*numCols+num_col_edges, (index+2)*numCols+num_col_edges)]
			self.faces		= range(index*numCols, (index+1)*numCols)

		self.loc			= offsetTrail					# localspace trailing edge offset from origin along x axis if self.index_y, else y-axis (metres).
		self.len			= offsetLead - offsetTrail		# overall _physical_ length (or depth) of this tile line in metres.

		self.is_slop		= False
		self.min_overlap	= 0.0							# same as OVERLAP_H/OVERLAP_V, only in metres, not pixels.
		self.is_antislop	= False
		self.max_antislop	= 0.0							# same as ANTISLOP_H/ANTISLOP_V, only in metres, not pixels.

		self.len_min		= offsetLead - offsetTrail		# 'spelling' length - used to add up spelling metrics. For a regular tile, min and max are the same
		self.len_max		= offsetLead - offsetTrail		#	"	"		"		"		"	 "

		self.match_lead		= None							# (string) match group of leading edge (the one furthest from origin).
		self.match_trail	= None							# (string) match group of trailing edge (the one nearest the origin).
		self.ignore			= False							# Exclude this tile line from all generated spellings (e.g. use for texture-bleed tiles)
		self.exclude		= False							# Exclude this tile line from this spelling set
		self.require		= False							# This tile line MUST be used in this spelling set
		self.center			= False							# Try to keep this tile in the metrical center of spellings
		self.xor			= False							# Only one xor tile may be included in a given spelling.

		self.dupes_min		= None							# If != None, sets the minimum number of duplicates in each spelling. Overrides the global setting. Range -1... 6
		self.dupes_max		= None							# If != None, sets the maximum number of duplicates in each spelling. Overrides the global setting
		self.only_before	= []							# If not empty, the current tile can only appear in a spelling if it precedes one of the specified tiles.
		self.only_after		= []							# if not empty, the current tile can only appear in a spelling if it follows one of the specified tiles.

	def packConstraints(self):
		p = Bitfield(0)
		for i, k in enumerate(['exclude', 'require', 'center', 'xor']):
			p[i] 	= int(getattr(self, k))
		if self.dupes_min == None:
			p[4:8] = 0
		else:
			p[4:8] = min(max(0, self.dupes_min + 1), 15)
		if self.dupes_max == None:
			p[8:12] = 0
		else:
			p[8:12] = min(max(0, self.dupes_max + 1), 15)
		return dict(only_after=intlists2string(self.only_after), only_before=intlists2string(self.only_before), flags=int(p))

	def unpackConstraints(self, data):
		try:
			p = Bitfield(data['flags'])
			for i, k in enumerate(['exclude', 'require', 'center', 'xor']):
				setattr(self, k, bool(p[i]))
			if p[4:8] > 0:
				self.dupes_min = int(p[4:8]) - 1
			elif p[4:8] == 0:
				self.dupes_min = None
			if p[8:12] > 0:
				self.dupes_max = int(p[8:12]) - 1
			elif p[8:12] == 0:
				self.dupes_max = None
			self.only_before = str2intlists(data['only_before'])
			self.only_after = str2intlists(data['only_after'])
		except StandardError, detail:
			print "ERROR", detail
			PupMenu("ERROR: UnpackConstraints failed. %s%%t" % detail)

	def updateMetrics(self):
		""" This just calulates the min and max lengths based on slop and antislop data which may have been set since instantiation
		"""
		if self.is_slop:
			self.len_min = 0.0
			self.len_max = self.len - self.min_overlap
		elif self.is_antislop:
			self.len_min = self.len - self.max_antislop
			self.len_max = self.len_min

	def printData(self):
		print self.index, 'loc', self.loc, 'len', self.len, 'slop' if self.is_slop else '    ', 'mino', self.min_overlap,\
				'anti' if self.is_antislop else '    ', 'maxa', self.max_antislop, 'excl' if self.exclude else '    ', 'reqr' if self.require else '    ', 'ignr' if self.ignore else '   ',\
				'cent' if self.center else '   ', 'xor ' if self.xor else '    ', self.match_trail, self.match_lead
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class GUIBug(object):
	pass
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class RangePlot(object):

	def __init__(self, name, x, w, xMin, xMax):

		self.name				= name
		self.lowest				= xMin
		self.highest			= xMax if xMax > xMin else xMin + 1
		self.x					= x
		self.w					= w
		self.m2pix 				= w/float(self.highest - self.lowest)
		self.minor_ticks_x		= []
		self.major_ticks_x		= []
		self.minor_tick_values	= []
		self.major_tick_values	= []

	def getAxesCoords(self, extensionHeight, yOffset):
		# Tick marks
		spacing_metres = [[100, 20], [50, 10], [20, 5], [10, 2], [5, 1], [2, 1]]
		i = 0
		str_width = GetStringWidth(str(int(self.highest)), 'tiny')*1.5
		while i < len(spacing_metres) and spacing_metres[i][0] * self.m2pix > str_width:
			major_tick_spacing = spacing_metres[i][0]
			minor_tick_spacing = spacing_metres[i][1]
			i += 1

		self.minor_tick_values = [t for t in range(0, int(self.highest)+1, minor_tick_spacing) if t >= self.lowest]
		self.major_tick_values = [t for t in range(0, int(self.highest)+1, major_tick_spacing) if t >= self.lowest]
		self.minor_ticks_x = []
		for t in self.minor_tick_values:
			self.minor_ticks_x.append(self.x+int((t - self.lowest)*self.m2pix))
		self.major_ticks_x = []
		for t in self.major_tick_values:
			self.major_ticks_x.append(self.x+int((t - self.lowest)*self.m2pix))

		out_axes = [[],[],[]]		# out_axes[0] -> regular tick marks, out_axes[1] -> extended minor ticks, out_axes[2] = extended major ticks
		out_axes[0].extend([self.x, yOffset, self.x + self.w, yOffset])	# x-axis line
		for mintv, mintx in zip(self.minor_tick_values, self.minor_ticks_x):
			if mintv not in self.major_tick_values:
				out_axes[0].extend([mintx, yOffset - 1, mintx, yOffset + 1])
				if extensionHeight:
					out_axes[1].extend([mintx, yOffset + 1, mintx, yOffset + 1 + extensionHeight])
		for majtv, majtx in zip(self.major_tick_values, self.major_ticks_x):
			out_axes[0].extend([majtx, yOffset - 2, majtx, yOffset+1])
			if extensionHeight:
				out_axes[2].extend([majtx, yOffset + 1, majtx, yOffset + 1 + extensionHeight])
		return out_axes

	def getNumberData(self, yOffset):
		string_data = []
		# Axis numbers:
		for x, v in zip(self.major_ticks_x, self.major_tick_values):
			string_data.append([str(v), x, yOffset - 9])
		# Axis name:
		string_data.append([self.name, self.x + self.w/2, yOffset - 18])
		return string_data

	def getFullRangeCoords(self, yOffset, h):
		return [self.x, yOffset, self.x + self.w, yOffset + h]

	def getAbsentRangesCoords(self, minMaxList, yOffset, h):
		"""
			Return the 2D vertex coords for range gaps rects
		"""
		return self.rangesToCoords(self.getMissingRanges(minMaxList), yOffset, h)

	def getPresentRangesCoords(self, minMaxList, yOffset, h):
		"""
			Return the 2D coords for covered ranges rects
		"""
		return self.rangesToCoords(self.getContiguousRanges(minMaxList), yOffset, h)

	def rangesToCoords(self, ranges, yOffset, h):
		out_coords = []
		for g in ranges:
			x0 = int(self.x + self.m2pix * (g[0] - self.lowest))
			x1 = max(int(self.x + self.m2pix * (g[1] - self.lowest)), x0 + 1)		# The max function is to make sure that even the tiniest range gaps are drawn.

			out_coords.append([x0, yOffset, x1, yOffset + h])
		return out_coords

	def collapseRanges(self, minMax):
		if len(minMax) == 0:
			yield []
		r = minMax[0][:]
		for low, high in sorted( minMax):
			if r[1] >= low:
				r[1] = max(r[1], high)
			else:
				yield r
				r[0] = low
				r[1] = high
		yield r

	def getContiguousRanges(self, minMaxList):
		"""
			Takes a list of [min, max] value lists, and returns a list of [min, max] ranges over which the input data is contiguous.
		"""
		if len(minMaxList) == 0:
			return []
		out_ranges = []
		for r in self.collapseRanges(minMaxList):
			if r[1] < self.lowest:
				continue
			elif r[0] <= self.highest:
				out_ranges.append([max(r[0], self.lowest), min(r[1], self.highest)])
		return out_ranges

	def getMissingRanges(self, minMaxList):
		"""
			Return pairs of values describing the inverse of getContiguousRanges between the limits of self.lowest and self.highest
		"""
		overlaps = self.getContiguousRanges(minMaxList)
		num_overlaps = len(overlaps)
		if num_overlaps == 0:
			return [[self.lowest, self.highest]]

		gaps = []
		if overlaps[0][0] > self.lowest:
			gaps.append([self.lowest, overlaps[0][0]])
		if num_overlaps > 1:
			for i in range(1, len(overlaps)):
				if overlaps[i][0] >= self.lowest and overlaps[i-1][1] <= self.highest:
					gaps.append([max(overlaps[i-1][1], self.lowest), min(overlaps[i][0], self.highest)])
		if overlaps[-1][1] < self.highest:
			gaps.append([overlaps[-1][1], self.highest])
		return gaps
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBHistogram(object):
	"""
		in drawAGBEditor:		drawHistograms(yOffset)			 yoffset is the scrolling offset
		in showHistograms:		process data, decide vertical splits, if any.

								Create two RanglePlot instances, s and t, passing (self, name, x, w, h, xMin, xMax) to each

								create on AGBHistogram instance per set of data, passing: x, yOffset, w, h, plot_s, plot_t sMinMaxList, tMinMaxList, modeString, evt
								mode can be one of:
								mode		purpose			icons	gaps  regions  background	axes	labels	height	button
								'split'		split_heights	Y		red		-	 	green		-		-		Y		-
								'single'	single_height	Y		red		-		green		-		-		-		-
								'spelling_set' spelling_sets Y		-		green	-			-		-		Y		Y
								'axes'		master_axes		-		-		-		-			Y		Y		-		-
	"""

	def __init__(self, name, x, yOffset, w, h, rangePlotOb1, rangePlotOb2, mode, hExtension=0, sMinMaxList=None, tMinMaxList=None, roadEdges=None, heightStr=None, evt=None, isActive=False, yWindowBottom=None, yWindowTop=None, halfBlock=0):

		self.name 				= name
		self.x					= x
		self.y_offset			= yOffset
		self.w					= w
		self.h					= h
		self.button_x			= rangePlotOb2.x + rangePlotOb2.w + 20
		self.road_edges			= roadEdges
		self.half_block			= halfBlock
		self.evt				= evt
		self.mode				= mode
		self.number_data		= []
		self.label_data			= []
		self.is_active			= isActive						# True if showing the active spelling set
		self.window_bottom		= yWindowBottom
		self.window_top			= yWindowTop

		if self.mode == 'axes':
			self.axes_verts = [[], [], []]
			for p in [rangePlotOb1, rangePlotOb2]:
				all_verts = p.getAxesCoords(hExtension, yOffset)									# three lists of verts: main axes, minor extended, major extended
				for i in range(3):
					self.axes_verts[i].extend(all_verts[i])
				self.number_data.extend(p.getNumberData(yOffset)[:-1])								# list of ['value', x, y] values. Last one in list is the axis label
				self.label_data.append(p.getNumberData(yOffset)[-1])								# axis labels

		elif self.mode in ['spelling_set', 'temp']:
			self.green_verts = []
			for p, mm in zip([rangePlotOb1, rangePlotOb2], [sMinMaxList, tMinMaxList]):
				self.green_verts.extend(p.getPresentRangesCoords(mm, yOffset, h))					# List of verts for drawing the _covered_ ranges for both axes

			if heightStr:																			# Height bucket upper and lower values, to be drawn midway between the two plots, one above the other
				toks = heightStr.split()
				self.number_data.append([toks[-1], rangePlotOb2.x - 13, yOffset + self.h - 4])
				if len(toks) == 2:
					self.number_data.append([toks[0], rangePlotOb2.x - 13, yOffset - 1])

			self.button = Create(0)

		elif self.mode in ['single', 'split']:
			self.green_verts = []
			self.red_verts	 = []
			for p, mm in zip([rangePlotOb1, rangePlotOb2], [sMinMaxList, tMinMaxList]):
				self.green_verts.append(p.getFullRangeCoords(yOffset, h))
				self.red_verts.extend(p.getAbsentRangesCoords(mm, yOffset, h))

			if self.mode == 'split' and heightStr:													# Height bucket value(s), to be drawn centered between the two plots, level with the top of the plots
				toks = heightStr.split()															# If the string contains TWO values (i.e. it's the first bucket), the first value is to be drawn level with the BOTTOM of the plots
				self.number_data.append([toks[-1], rangePlotOb2.x - 13, yOffset + h - 3])
				if len(toks) == 2:
					self.number_data.append([toks[0], rangePlotOb2.x - 13, yOffset - 3])

	def drawHistogram(self, scrollOffset=0, scrollLimit=0):

		global CL_ROAD, CL_NOROAD, CL_AXES, CL_AXIS_LABELS, CL_MAJOR_TICKS, CL_MINOR_TICKS

		if self.mode == 'axes':

			BGL.glColor3f(*CL_AXES)
			BGL.glBegin(BGL.GL_LINES)
			for i in range(0, len(self.axes_verts[0]), 2):
				BGL.glVertex2i(self.axes_verts[0][i], self.axes_verts[0][i+1])
			BGL.glEnd()

			BGL.glColor3f(*CL_MINOR_TICKS)
			BGL.glBegin(BGL.GL_LINES)
			for i in range(0, len(self.axes_verts[1]), 2):
				BGL.glVertex2i(self.axes_verts[1][i], self.axes_verts[1][i+1])
			BGL.glEnd()

			BGL.glColor3f(*CL_MAJOR_TICKS)
			BGL.glBegin(BGL.GL_LINES)
			for i in range(0, len(self.axes_verts[2]), 2):
				BGL.glVertex2i(self.axes_verts[2][i], self.axes_verts[2][i+1])
			BGL.glEnd()

			for n in self.number_data:
				drawString(n[0],	n[1], 	n[2],	CL_AXIS_NUMS,	'center', 'tiny')

			for n in self.label_data:
				drawString(n[0],	n[1], 	n[2],	CL_AXIS_LABELS,	'center', 'tiny')
# 			agb_name = self.name[3:] + '.agb'
# 			drawString(agb_name,	(self.button_x + self.x +  self.w )/2, 	self.y_offset+7,	CL_TEXT_DK,	'center', 'normal')

		elif self.mode == 'spelling_set':

			if self.window_bottom != None and self.window_top != None:
				if self.window_bottom < (self.y_offset + scrollOffset) <= self.window_top:

					BGL.glEnable( BGL.GL_BLEND )
					BGL.glBlendFunc(BGL.GL_SRC_ALPHA, BGL.GL_ONE_MINUS_SRC_ALPHA)
					if self.is_active:
						BGL.glColor4f(*CL4_HISTO_ACT)
					else:
						BGL.glColor4f(*CL4_HISTO_GREEN)
					for r in self.green_verts:
						BGL.glRecti(*self.offsetRect(r, scrollOffset))
					BGL.glDisable( BGL.GL_BLEND )

					for n in self.number_data:
						drawString(n[0],	n[1], 	n[2]+scrollOffset,	CL_WHITE,	'center', 'tiny')

					self.button = PushButton(self.name,		self.evt,	self.button_x, self.y_offset + scrollOffset - 3, 815 - self.button_x, self.h + 6,	"Click to edit this spelling set.")

					drawRectangle(10, self.y_offset + scrollOffset - 1, 115, self.h+2, CL_GREY56)
					if self.road_edges != None:
						self.drawEdgeTypeIcons(self.road_edges,		self.x + 3, 	self.y_offset + (self.h - 8)/2 + scrollOffset, 	11, 	8, 		1, 		14, 	CL_ROAD, CL_NOROAD, CL_ROAD_BKGND, self.half_block)

		elif self.mode == 'temp':

			BGL.glEnable( BGL.GL_BLEND )
			BGL.glBlendFunc(BGL.GL_SRC_ALPHA, BGL.GL_ONE_MINUS_SRC_ALPHA)

			BGL.glColor4f(*CL4_HISTO_TEMP)
			for r in self.green_verts:
				BGL.glRecti(*r)
			BGL.glDisable( BGL.GL_BLEND )

			if self.road_edges != None:
				drawRectangle(10, self.y_offset - 1, 115, self.h+2, CL_GREY60)
				self.drawEdgeTypeIcons(self.road_edges,		self.x + 3, 	self.y_offset + (self.h - 8)/2, 	11, 	8, 		1, 		14, 	CL_ROAD, CL_NOROAD, CL_ROAD_BKGND, self.half_block)

			drawString(self.name,	(self.button_x + self.x +  self.w )/2, 	self.y_offset+1,	CL4_HISTO_TEMP[:3],	'center', 'normal')

		elif self.mode in ['single', 'split']:

			BGL.glEnable( BGL.GL_BLEND )
			BGL.glBlendFunc(BGL.GL_SRC_ALPHA, BGL.GL_ONE_MINUS_SRC_ALPHA)

			BGL.glColor4f(*CL4_HISTO_GREEN)
			for g in self.green_verts:
				BGL.glRecti(*g)

			BGL.glColor4f(*CL4_HISTO_RED)
			for r in self.red_verts:
				BGL.glRecti(*r)

			BGL.glDisable( BGL.GL_BLEND )

			if self.road_edges != None:
				self.drawEdgeTypeIcons(self.road_edges,		self.x + 3, 	self.y_offset + (self.h - 8)/2, 	11, 	8, 		1, 		14, 	CL_ROAD, CL_NOROAD, CL_ROAD_BKGND, self.half_block)

			if self.mode == 'split':
				for n in self.number_data:
					drawString(n[0],	n[1], 	n[2],	CL_WHITE,	'center', 'tiny')

	def drawScrollArrow(self, x, y, w, h, direction, colour):
		"""
			Draws a chevron from a point midway along its base.
		"""
		y1 = y + h if  direction == 'up' else y - h

		BGL.glColor3f(*colour)
		BGL.glBegin(BGL.GL_TRIANGLES)
		BGL.glVertex2i(x - w/2, y)
		BGL.glVertex2i(x, 		y1)
		BGL.glVertex2i(x + w/2, y)
		BGL.glEnd()

	def offsetRect(self, coordList, yOffset):
		return [coordList[0], coordList[1] + yOffset, coordList[2], coordList[3] + yOffset]

	def drawEdgeTypeIcons(self, data, x, y, w, h, th, spacing, colourRoad, colourNoRoad, colourBackground, halfBlock=0):

		MUTING_RATIO = 0.25
		rects = []
		inner_rects = []
		edges = []
		rects_muted = []
		edges_muted = []

		if not halfBlock:
			all_icons = [[1, 1, 1, 1]]
		else:
			all_icons = [[1, 0, 0, 0], [1, 1, 0, 0], [1, 0, 1, 0], [1, 1, 1, 0], [1, 0, 0, 1], [1, 1, 0, 1], [1, 0, 1, 1], [1, 1, 1, 1]]

		for d in all_icons:

			flag = 1 if d in data else 0
			x1, y1 = x+w, y+h

			(rects if flag else rects_muted).append([x, y, x1, y1])
			inner_rects.append([x+th, y+th, x1-th, y1-th])

			if d[0]:
				(edges if flag else edges_muted).append([x, 	y, 		x1,		y+th])
			if d[1]:
				(edges if flag else edges_muted).append([x1-th,	y,		x1,		y1])
			if d[2]:
				(edges if flag else edges_muted).append([x, 	y1-th, 	x1,		y1])
			if d[3]:
				(edges if flag else edges_muted).append([x,		y,		x+th,	y1])

			x += spacing

		BGL.glColor3f(*colourNoRoad)
		for r in rects:
			BGL.glRecti(*r)

		BGL.glColor3f(*mix(colourBackground, colourNoRoad, MUTING_RATIO))
		for rm in rects_muted:
			BGL.glRecti(*rm)

		BGL.glColor3f(*colourBackground)
		for r in inner_rects:
			BGL.glRecti(*r)

		BGL.glColor3f(*colourRoad)
		for e in edges:
			BGL.glRecti(*e)

		BGL.glColor3f(*mix(colourBackground, colourRoad, MUTING_RATIO))
		for em in edges_muted:
			BGL.glRecti(*em)
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class AGBEditor(object):

	DELETE_INIT_OB_ON_LOAD			= True

	for t in Blender.Text.Get():
		if "XPTAGBSpellings" in t.name:
			temp_text  = t
			break

	else:		# Temporary text file for generated spelling output
		temp_text  = Blender.Text.New("XPTAGBSpellings", 1)

	class PanelState(object):
		def __init__(self):
			self.slop					= False
			self.antislop				= False
			self.edge_type				= False
			self.match_grp				= False
			self.ignore					= False

		def setMode(self, mode):
			for k, v in self.__dict__.iteritems():
				object.__setattr__(self, k, False)
			object.__setattr__(self, mode, True)

		def getMode(self):
			for k, v in self.__dict__.iteritems():
				if v:
					return k
			return None

		def anyPanel(self):
			return any(self.__dict__.itervalues())

		def closeAny(self):
			for k in self.__dict__.iterkeys():
				self.__dict__[k] = False

		def printStatus(self):
			for k, v in self.__dict__.iteritems():
				print k, 'ON ' if v else 'off',
			print '\n'

	def __init__(self, event):

		global DEF_TILE_WRAPPER_LAYER, DEF_OBJ_WRAPPER_LAYER, DEF_METARECT_LAYER, DEF_TILE_MESH_LAYER
		evt = event
		global XPTButtonNumber
		global XPTButtonString

		self.ButtonNumber				= {}
		self.evt_ButtonNumber			= {}
		self.ButtonNumber_val			= {}
		self.ButtonNumber.update(XPTButtonNumber)
		for i in range(len(self.ButtonNumber)):
			self.evt_ButtonNumber[str(i)] = evt
			self.ButtonNumber_val[str(i)] = 0
			evt += 1
		self.ButtonMenu					= {}
		self.evt_ButtonMenu				= {}
		self.ButtonMenu_val				= {}
		self.ButtonMenu.update(XPTButtonMenu)
		for i in range(len(self.ButtonMenu)):
			self.evt_ButtonMenu[str(i)] = evt
			self.ButtonMenu_val[str(i)] = 0
			evt += 1
		self.ButtonString				= {}
		self.evt_ButtonString			= {}
		self.ButtonString.update(XPTButtonString)
		for i in range(len(self.ButtonString)):
			self.evt_ButtonString[str(i)] = evt
			evt += 1

		self.AGB_ob						= None					# AGBObject in editor, if any
		self.spellings_file				= None					# text file for barfing spellings into
		self.delete_on_quit				= None
		self.unsaved_changes			= False
		self.grid_obs					= []					# List of AGBGrid objects associated with this AGB
		self.temp_grid_ob				= None					# A new AGBGrid object, not yet added to an AGB
		self.active_grid_index			= None				    # Index of the grid object in self.grid_obs currently in the editor
		self.active_sp_set_index		= None					# spelling set currently in the editor
		self.mesh_layer					= DEF_TILE_MESH_LAYER
		self.tex_panel					= False
		self.panels						= AGBEditor.PanelState()

		self.match_groups_h_mnstr 		= ""
		self.match_groups_v_mnstr 		= ""
		self.AGB_depth_menustring		= "AGB Depth Range (m)%t| 30 - 60%x1| 60 - 120%x2| 90 + %x3"
		self.AGB_height_menustring		= "AGB Height Range (m)%t|   none %x1| under 10%x2| 10 - 16%x3| 16 - 24%x4| 24 - 32%x5| 32 - 40%x6| 40 - 80%x7| 80 - 120%x8| over 120 %x9"
		self.mn_sp_set_act_menustring	= "Spelling Set %t"

		self.GUIimageWindow				= (0, 0, 2, 2, 32, 16)
		self.draw_image_data			= ()
		self.draw_grid_size				= []					# Size of the GUI grid image, w x h in pixels.
		self.preview_source 			= 'undefined'			# or 'temp' -> previews of temporary generated spellings, 'active' -> previews of active spelling set
		self.preview_obs				= []					# List of unlinked AGB grid preview mesh obs.
		self.dupli_obs					= []					# List of unlinked empty objects used as dupliobs for agb preview annotations.
		self.spellings_axis				= 0						# 0 -> horizontal, 1 -> vertical
		self.constraints_mode			= 1						# Display mode: 0 -> 'exclude', 1 -> 'require', 2 -> 'center', 3 -> XOR
		self.show_constraints			= False
		self.show_histograms			= False
		self.show_temp					= False
		self.histo_prefs				= {'source_is_AGB': True, 'custom_ranges': False, 'show_edge_splits': False}
		self.range_plots				= []
		self.histograms					= []
		self.histograms_scrolling		= []
		self.scroll_offset				= 0
		self.scroll_limit				= 0						# Prevents excessive scrolling of histograms
		self.SPELLING_SET_HT			= 18
		self.temp_sp_set				= AGBSpellingSet()
		self.temp_sp_set.comment		= "Temporary Generated Output"

		self.maxed_out_iterations		= False					# GUI messages
		self.last_iterations			= ""
		self.progress_message			= ""
		self.progress_message2			= ""

		self.pv_increment_metres		= 10.0
		self.pv_increment_spellings		= 2
		self.common_constraints			= { 'horizontal': dict( min_length=55, max_length=250,	max_dupes=0,  max_cuts=3, max_iterations=1000000, center_precision=4.0, sp_variety=1, xor_is_or=0),
											  'vertical':  dict( min_length=55, max_length=125,	max_dupes=0,  max_cuts=3, max_iterations=500000, center_precision=4.0, sp_variety=1, xor_is_or=0) }

		self.evt_AGB_name, 			evt = evt, evt + 1
		self.evt_load_AGB, 			evt = evt, evt + 1
		self.evt_new_AGB, 			evt = evt, evt + 1
		self.evt_save_AGB, 			evt = evt, evt + 1
		self.evt_duplicate_AGB,		evt = evt, evt + 1
		self.evt_add_grid_to_AGB,	evt = evt, evt + 1
		self.evt_duplicate_grid, 	evt = evt, evt + 1
		self.evt_delete_grid, 		evt = evt, evt + 1
		self.evt_active_grid_index,	evt = evt, evt + 1
		self.evt_texscale_x, 		evt = evt, evt + 1
		self.evt_texscale_y, 		evt = evt, evt + 1
		self.evt_texorig_x, 		evt = evt, evt + 1
		self.evt_texorig_y, 		evt = evt, evt + 1
		self.evt_AGB_tile_layer, 	evt = evt, evt + 1
		self.evt_analyse_active, 	evt = evt, evt + 1
		self.evt_build_grid, 		evt = evt, evt + 1
		self.evt_tex_image, 		evt = evt, evt + 1
		self.evt_get_tex_image, 	evt = evt, evt + 1
		self.evt_open_tex_panel, 	evt = evt, evt + 1
		self.evt_set_tex_data,		evt = evt, evt + 1
		self.evt_exit_tex_panel,	evt = evt, evt + 1
		self.evt_wrapper_layer, 	evt = evt, evt + 1
		self.evt_mesh_layer, 		evt = evt, evt + 1
		self.evt_meta_layer, 		evt = evt, evt + 1
		self.evt_new_grid_name, 	evt = evt, evt + 1
		self.evt_mesh_ob_name,	 	evt = evt, evt + 1
		self.evt_wrap_grid, 		evt = evt, evt + 1
		self.evt_make_prev_grids,	evt = evt, evt + 1
		self.evt_clear_prev_grids,	evt = evt, evt + 1
		self.evt_pv_annotations,	evt = evt, evt + 1

		self.evt_AGB_depth_range,	evt = evt, evt + 1
		self.evt_AGB_height_bucket,	evt = evt, evt + 1
		self.evt_AGB_full_half,		evt = evt, evt + 1

		self.evt_sp_set_act,		evt = evt, evt + 1
		self.evt_sp_set_next,		evt = evt, evt + 1
		self.evt_sp_set_previous,	evt = evt, evt + 1
		self.evt_override_params,	evt = evt, evt + 1
		self.evt_rename_sp_set,		evt = evt, evt + 1
		self.evt_new_sp_set,		evt = evt, evt + 1
		self.evt_delete_sp_set,		evt = evt, evt + 1
		self.evt_copy_sp_set,		evt = evt, evt + 1
		self.evt_sp_set_name,		evt = evt, evt + 1
		self.evt_define_heights,	evt = evt, evt + 1
		self.evt_sp_set_height_min,	evt = evt, evt + 1
		self.evt_sp_set_height_max,	evt = evt, evt + 1
		self.evt_sp_set_frequency,	evt = evt, evt + 1

		self.evt_sp_exclude,		evt = evt, evt + 1
		self.evt_sp_require,		evt = evt, evt + 1
		self.evt_sp_center,			evt = evt, evt + 1
		self.evt_sp_xor,			evt = evt, evt + 1
		self.evt_toggle_or_xor,		evt = evt, evt + 1
		self.evt_sp_sel_constraints,evt = evt, evt + 1
		self.evt_sp_clr_constraints,evt = evt, evt + 1
		self.evt_print_spellings,	evt = evt, evt + 1
		self.evt_clear_temp_text,	evt = evt, evt + 1
		self.evt_show_histograms,	evt = evt, evt + 1
		self.evt_reset_scale,		evt = evt, evt + 1
		self.evt_scroll_up,			evt = evt, evt + 1
		self.evt_scroll_down,		evt = evt, evt + 1
		self.evt_edge_type_splits,	evt = evt, evt + 1
		self.evt_show_temp,			evt = evt, evt + 1
		self.evt_sp_set_select 			= []
		for i in range(20):
			self.evt_sp_set_select.append(evt)
			evt += 1
		self.evt_axis_horizontal,	evt = evt, evt + 1
		self.evt_axis_vertical,		evt = evt, evt + 1
		self.evt_make_spellings,	evt = evt, evt + 1

		self.evt_max_length,		evt = evt, evt + 1
		self.evt_min_length,		evt = evt, evt + 1


		self.evt_max_repeats,		evt = evt, evt + 1
		self.evt_max_dupes,			evt = evt, evt + 1
		self.evt_max_cuts,			evt = evt, evt + 1
		self.evt_max_iterations,	evt = evt, evt + 1
		self.evt_center_precision,	evt = evt, evt + 1
		self.evt_sp_variety,		evt = evt, evt + 1
		self.evt_add_to_sp_set,		evt = evt, evt + 1
		self.evt_filter_spellings,	evt = evt, evt + 1
		self.evt_min_length_pv, 	evt = evt, evt + 1
		self.evt_max_length_pv,		evt = evt, evt + 1
		self.evt_min_depth_pv, 		evt = evt, evt + 1
		self.evt_max_depth_pv,		evt = evt, evt + 1
		self.evt_by_spelling_pv, 	evt = evt, evt + 1
		self.evt_by_size_pv, 		evt = evt, evt + 1
		self.evt_increment_pv,		evt = evt, evt + 1

		self.evt_match_groups_h,	evt = evt, evt + 1
		self.evt_match_groups_v,	evt = evt, evt + 1
		self.evt_add_match_group_h,	evt = evt, evt + 1
		self.evt_add_match_group_v,	evt = evt, evt + 1
		self.evt_del_match_group_h,	evt = evt, evt + 1
		self.evt_del_match_group_v,	evt = evt, evt + 1
		self.evt_sel_match_group_h,	evt = evt, evt + 1
		self.evt_sel_match_group_v,	evt = evt, evt + 1

		self.evt_add_edge_group,	evt = evt, evt + 1
		self.evt_del_edge_group,	evt = evt, evt + 1
		self.evt_road_facing,		evt = evt, evt + 1
		self.evt_non_road_facing,	evt = evt, evt + 1

		self.evt_open_pnl_slop,		evt = evt, evt + 1
		self.evt_open_pnl_antislop,	evt = evt, evt + 1
		self.evt_open_pnl_ignore,	evt = evt, evt + 1
		self.evt_open_pnl_edge_type,evt = evt, evt + 1
		self.evt_open_pnl_match_grp,evt = evt, evt + 1
		self.evt_panel_apply,		evt = evt, evt + 1
		self.evt_panel_clear,		evt = evt, evt + 1
		self.evt_panel_cancel,		evt = evt, evt + 1
		self.evt_panel_show_faces,	evt = evt, evt + 1

		self.evt_panel_data_bugs		= []
		for i in range(8):
			self.evt_panel_data_bugs.append(evt)
			evt += 1

		self.evt_bug,				evt = evt, evt + 1
		self.evt_panel_help,		evt = evt, evt + 1
		self.evt_test_colour,		evt = evt, evt + 1
		self.evt_test_alpha,		evt = evt, evt + 1

		self.st_AGB_name				= Create("")
		self.bt_load_AGB				= Create(0)
		self.bt_save_AGB				= Create(0)
		self.bt_new_AGB					= Create(0)
		self.bt_duplicate_AGB			= Create(0)
		self.bt_add_grid_to_AGB			= Create(0)
		self.bt_duplicate_grid			= Create(0)
		self.bt_delete_grid				= Create(0)

		self.mn_active_grid_index		= Create(0)
		self.nb_texscale_x				= Create(256.0)
		self.nb_texscale_y				= Create(256.0)
		self.nb_texorig_x				= Create(0.0)
		self.nb_texorig_y				= Create(0.0)
		self.nb_AGB_tile_layer			= Create(0)
		self.bt_analyse_active			= Create(0)
		self.bt_build_grid				= Create(0)
		self.mn_tex_image				= Create(0)
		self.bt_get_tex_image			= Create(0)
		self.bt_open_tex_panel			= Create(0)
		self.bt_set_tex_data			= Create(0)
		self.bt_exit_tex_panel			= Create(0)
		self.nb_wrapper_layer			= Create(DEF_TILE_WRAPPER_LAYER)
		self.nb_mesh_layer				= Create(DEF_TILE_MESH_LAYER)
		self.nb_meta_layer				= Create(DEF_METARECT_LAYER)
		self.st_new_grid_name			= Create("")
		self.st_mesh_ob_name			= Create("")
		self.bt_wrap_grid				= Create(0)
		self.bt_make_prev_grids			= Create(0)
		self.bt_clear_prev_grids		= Create(0)
		self.tg_pv_annotations			= Create(0)

		self.mn_AGB_depth_range			= Create(0)
		self.mn_AGB_height_bucket		= Create(0)
		self.tg_AGB_full_half			= Create(0)
		self.mn_sp_set_act				= Create(0)
		self.bt_sp_set_next				= Create(0)
		self.bt_sp_set_previous			= Create(0)
		self.tg_override_params			= Create(1)
		self.bt_rename_sp_set			= Create(0)
		self.bt_new_sp_set				= Create(0)
		self.bt_copy_sp_set				= Create(0)
		self.bt_delete_sp_set			= Create(0)
		self.st_sp_set_name				= Create("")
		self.bt_define_heights			= Create(0)
		self.nb_sp_set_height_min		= Create(0)
		self.nb_sp_set_height_max		= Create(999)
		self.nb_sp_set_frequency		= Create(5)

		self.tg_sp_exclude				= Create(0)		#
		self.tg_sp_require				= Create(1)		# Four interlocking buttons, selecting which of four spelling constraints is displayed/edited
		self.tg_sp_center				= Create(0)		#
		self.tg_sp_xor					= Create(0)		#
		self.bt_toggle_or_xor			= Create(0)
		self.bt_sp_sel_constraints		= Create(0)
		self.bt_sp_clr_constraints		= Create(0)
		self.bt_print_spellings			= Create(0)
		self.bt_clear_temp_text			= Create(0)
		self.bt_show_histograms			= Create(0)
		self.bt_scroll_up				= Create(0)
		self.bt_scroll_down				= Create(0)
		self.bt_reset_scale				= Create(0)
		self.tg_edge_type_splits		= Create(0)
		self.bt_show_temp				= Create(0)
		self.tg_axis_horizontal			= Create(1)
		self.tg_axis_vertical			= Create(0)
		self.bt_make_spellings			= Create(0)
		self.bt_add_to_sp_set			= Create(0)
		self.bt_filter_spellings		= Create(0)
		self.nb_max_length				= Create(180.0)	# max desired length for generated spellings
		self.nb_min_length				= Create(55.0)	# min desired length for generated spellings

		self.nb_max_dupes				= Create(0)
		self.nb_max_cuts				= Create(3)
		self.nb_max_iterations			= Create(50000)
		self.nb_center_precision		= Create(5.0)
		self.sl_sp_variety				= Create(1)

		self.nb_min_length_pv			= Create(60)
		self.nb_max_length_pv			= Create(180)
		self.nb_min_depth_pv			= Create(60)
		self.nb_max_depth_pv			= Create(120)
		self.tg_by_spelling_pv			= Create(0)
		self.tg_by_size_pv				= Create(1)
		self.nb_increment_pv			= Create(10)

		self.mn_match_groups_h			= Create(0)
		self.mn_match_groups_v			= Create(0)
		self.bt_add_match_group_h		= Create(0)
		self.bt_add_match_group_v		= Create(0)
		self.bt_del_match_group_h		= Create(0)
		self.bt_del_match_group_v		= Create(0)
		self.bt_sel_match_group_h		= Create(0)
		self.bt_sel_match_group_v		= Create(0)

		self.bt_add_edge_group			= Create(0)
		self.bt_del_edge_group			= Create(0)
		self.tg_road_facing				= Create(1)
		self.tg_non_road_facing			= Create(0)

		self.bt_open_pnl_slop			= Create(0)
		self.bt_open_pnl_antislop		= Create(0)
		self.bt_open_pnl_ignore			= Create(0)
		self.bt_open_pnl_edge_type		= Create(0)
		self.bt_open_pnl_match_grp		= Create(0)
		self.bt_panel_apply				= Create(0)
		self.bt_panel_clear				= Create(0)
		self.bt_panel_cancel			= Create(0)
		self.bt_panel_show_faces		= Create(0)
		self.tg_panel_help				= Create(0)
		self.st_panel_data_bugs			= []
		for i in range(8):
			self.st_panel_data_bugs.append(Create(""))

		self.st_only_after 				= Create("")
		self.st_only_before				= Create("")
		self.nb_dupes_min				= Create(0)
		self.nb_dupes_max				= Create(0)
		self.bt_dupes_min_on_off		= Create(0)
		self.bt_dupes_max_on_off		= Create(0)
		self.bt_close_pup  				= Create(0)
		self.bt_reset_pup_vals			= Create(0)
		self.bt_cancel_pup				= Create(0)

		self.mn_test_run_menu			= Create(0)
		self.tg_bug						= Create(0)
		self.bug_value					= 0.0
		self.cp_test_colour				= Create(0.5, 0.5, 0.5)
		self.sl_test_alpha				= Create(1.0)
		self.test_colour				= [0.0, 0.0, 0.0, 1.0]

		self.buildGridImageMenu()
		self.buildGridObsMenuString()
		self.buildMatchGroupMenuString('h')
		self.buildMatchGroupMenuString('v')
		self.buildSpellingSetMenuString()
		# self.saveQuitText()
		self.queryTempTextSize(10000)

	def saveQuitText(self):

		if self.delete_on_quit == None:
			for t in Blender.Text.Get():
				if t.name == "XPTSafety":
					self.delete_on_quit = t
					retval = PupMenu("It appears that XPlaneTools crashed last time it ran. Do you want to try to restore to its most recent state?%t| NO | YES")
					break
			else:
				self.delete_on_quit = Blender.Text.New("XPTSafety")
		scn = bpy.data.scenes.active
		if self.AGB_ob != None:
			self.delete_on_quit.clear()
			self.delete_on_quit.write("This is an XPlaneTools auto-backup file - do not attempt to use it with an art-asset!\n")
			self.delete_on_quit.write("SCENE   			%s\n" % scn.name)
			self.delete_on_quit.write("AGB				%s\n" % self.AGB_ob.wrapper.name)
			if self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
				self.delete_on_quit.write("ACTIVE_SP_SET	%s\n" % self.AGB_ob.spelling_sets[self.active_sp_set_index].comment)
			self.delete_on_quit.write("ACTIVE_GRID		%s\n" % self.temp_grid_ob.name)
			self.delete_on_quit.write("\n")
			self.delete_on_quit.write("#\n# %s\n" % (self.AGB_ob.name[3:] + ".agb"))
			self.delete_on_quit.write("# " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
			self.delete_on_quit.write("\n#\n")
			for s in self.AGB_ob.spelling_sets:
				s.printSet(self.delete_on_quit)
			if self.temp_sp_set and self.temp_sp_set.hasSpellings():
				self.printDivider(self.delete_on_quit)
				self.temp_sp_set.printSet(self.delete_on_quit)

	def deleteQuitText(self):
		if self.delete_on_quit != None:
			Blender.Text.unlink(self.delete_on_quit)

	def getTempText(self):
		for t in Blender.Text.Get():
			if "XPTAGBSpellings" in t.name:
				AGBEditor.temp_text  = t
				break
		else:		# Temporary text file for generated spelling output
			AGBEditor.temp_text  = Blender.Text.New("XPTAGBSpellings")

	def queryTempTextSize(self, size):

		num_lines = AGBEditor.temp_text.getNLines()
		if num_lines > size:
			retval = PupMenu("The spelling generator output history text object is getting quite large - do you want to clear it?%t| NO | YES")
			if retval == 2:
				AGBEditor.temp_text.clear()
				Window.RedrawAll()

	def clearTempText(self):
		retval = PupMenu("You are about to clear the spelling generator output history - proceed?%t| NO | YES")
		if retval == 2:
			AGBEditor.temp_text.clear()

	def showHistograms(self):

		self.SPELLING_SET_HT	= 19
		PLOT_HEIGHT_SP_SET		= 12
		PLOT_HEIGHT_SINGLE_HT	= 24
		PLOT_HEIGHT_SPLIT_HT	= 12
		SPACE_ABOVE_MASTER		= 8
		SPACE_ABOVE_TEMP		= 7
		TOTAL_PANEL_HEIGHT		= 149	# vertical space available above the x-axis, for the actual plots.
		AGB_SOURCE_Y			= 28
		PLOT_X					= 137	# leftmost location of s-plot
		INTER_PLOT_SPACE		= 25

		if not self.AGB_ob:
			PupMenu("No AGB - can't show histograms.%t")
			return False

		# 	layout:
		#	margin	edge_icons	space	plot_s	space	plot_t	space	button
		#
		#	13			109		 15				 25				 20		 145		(widths)
		#			13		 122	 137<-	available_width	-> 650	670		815		(x locs)


		# First calculate the relative width ratio of the s and t axes:

		avail_width	=  513 - INTER_PLOT_SPACE

		# Choose either default or custom ranges for the two plots:

		if self.histo_prefs['custom_ranges']:
			sMin = float(self.nb_min_length_pv.val)
			sMax = float(max(self.nb_max_length_pv.val, self.nb_min_length_pv.val + 1))
			tMin = float(self.nb_min_depth_pv.val)
			tMax = float(max(self.nb_max_depth_pv.val, self.nb_min_depth_pv.val + 1))

		else:																	# leave a margin of 5m at either end of the axes to assist suthors to cover the required ranges.
			sMin = self.nb_min_length_pv.val = self.AGB_ob.depth_min - 5
			sMax = self.nb_max_length_pv.val = self.AGB_ob.depth_max * 2.0 + 5
			tMin = self.nb_min_depth_pv.val  = self.AGB_ob.depth_min - 5
			tMax = self.nb_max_depth_pv.val  = self.AGB_ob.depth_max + 5

		w_depth   = int(avail_width *(tMax - tMin)/ float((sMax - sMin) + (tMax - tMin)))
		w_width  = int(avail_width *(sMax - sMin)/ float((sMax - sMin) + (tMax - tMin)))

		def overlap(minMax1, minMax2):
			return True if (minMax1[0] <= minMax2[1] and minMax2[0] <= minMax1[1]) else False

		# Next, work out how many plots we're going to need:
		total_height = 0
		height_splits = set()

		# This is now 'in addition to' the AGB histos, not 'instead of'
		num_sets = len(self.AGB_ob.spelling_sets)
		height_ranges = []

		for ss in self.AGB_ob.spelling_sets:
			if ss.has_height_range:
				hr = [ss.height_min, ss.height_max]
				if hr not in height_ranges:
					height_ranges.append(hr)

		total_height += (num_sets-1) * self.SPELLING_SET_HT + PLOT_HEIGHT_SP_SET

		if not height_ranges:
			total_height += PLOT_HEIGHT_SINGLE_HT + SPACE_ABOVE_MASTER
		else:
			# Find the combined set of ranges:
			height_splits = set([self.AGB_ob.height_min, self.AGB_ob.height_max])
			for hr in height_ranges:
				for h in hr:
					height_splits.add(h)

			height_splits = list(height_splits)		# a list of height splits starting at agb.height_min and ending at agb.height_max
			height_splits.sort()
			print "height_splits", height_splits
			n = len(height_splits) - 1

			total_height += (n * (PLOT_HEIGHT_SPLIT_HT + 1) - 1 + SPACE_ABOVE_MASTER)

		if self.show_temp:
			total_height += PLOT_HEIGHT_SP_SET + SPACE_ABOVE_TEMP

		# Now define our two plot obs. These take care of the x-axis calculations and axis lines etc
		self.range_plots = [RangePlot("W I D T H", PLOT_X, w_width, sMin, sMax), RangePlot("D E P T H", PLOT_X + w_width + INTER_PLOT_SPACE, w_depth, tMin, tMax)]

		#
		# Finally, create an AGBHistogram object for each vertical element of the display:
		#

		self.histograms[:] = []
		self.histograms_scrolling[:] = []

		# Axes:
		axes = AGBHistogram(self.AGB_ob.name, 10, AGB_SOURCE_Y, 805, 0, self.range_plots[0], self.range_plots[1], 'axes', hExtension=min(TOTAL_PANEL_HEIGHT, total_height))
		self.histograms.append(axes)


		running_y = AGB_SOURCE_Y + 1	# Cumulative y offset value. The '1' is to allow for the thickness of the axis line.

		self.AGB_ob.updateSpellingSetMetrics()

		# Combined AGB histo
		if not height_ranges:
			agb_minMax_s = []
			agb_minMax_t = []
			for ss in self.AGB_ob.spelling_sets:
				agb_minMax_s.extend([[s.min, s.max] for s in ss.spellings_s])
				agb_minMax_t.extend([[s.min, s.max] for s in ss.spellings_t])
			agb_minMax_s.sort()
			agb_minMax_t.sort()
			agb_histo = AGBHistogram(self.AGB_ob.name, 10, running_y,	805, PLOT_HEIGHT_SINGLE_HT,	self.range_plots[0], self.range_plots[1],  'single', 0, agb_minMax_s, agb_minMax_t)
			self.histograms.append(agb_histo)
			running_y += PLOT_HEIGHT_SINGLE_HT

		# Split AGB histo
		else:
			hst_index = 0
			height_ranges.sort(key = lambda x: x[1])

			# Find out which spelling sets fall into which height bucket(s)
			buckets = []
			min_maxs = []
			for ss in self.AGB_ob.spelling_sets:
				if ss.has_height_range:
					min_maxs.append([ss.height_min, ss.height_max])
				else:
					min_maxs.append([self.AGB_ob.height_min, self.AGB_ob.height_max])
			print "min_maxs", min_maxs
			for i in range(1, len(height_splits)):
				members = []
				for j, mm in enumerate(min_maxs):
					if height_splits[i-1] >= mm[0] and height_splits[i] <= mm[1]:
# 					if overlap([height_splits[i-1], height_splits[i]], min_maxs[j]):
						members.append(j)
				buckets.append(members)
			print "buckets", buckets

			for i in range(len(buckets)):
				b_minMax_s = []
				b_minMax_t = []
				for m in buckets[i]:
					b_minMax_s.extend([[s.min, s.max] for s in self.AGB_ob.spelling_sets[m].spellings_s])
					b_minMax_t.extend([[s.min, s.max] for s in self.AGB_ob.spelling_sets[m].spellings_t])
				b_minMax_s.sort()
				b_minMax_t.sort()
				if i == 0:
					height_string = "%d %d" % (height_splits[0],  height_splits[1])
				else:
					height_string = ("%d" % height_splits[i+1]) if i + 1 < len(height_splits) else ""

				b_histo = AGBHistogram("", 10, running_y,	805, PLOT_HEIGHT_SPLIT_HT,	self.range_plots[0], self.range_plots[1],  'split', 0, b_minMax_s, b_minMax_t, heightStr=height_string)
				self.histograms.append(b_histo)
				running_y += PLOT_HEIGHT_SPLIT_HT + 1
				hst_index += 1


		# Add a space between the master histogram and the scrolling spelling-set histograms:
		running_y += SPACE_ABOVE_MASTER

		# Optional Temporary output histo:
		if self.show_temp:

			histo = AGBHistogram("Generated spellings", 10, running_y, 805, PLOT_HEIGHT_SP_SET,	self.range_plots[0], self.range_plots[1],  'temp', 0, [[s.min, s.max] for s in self.temp_sp_set.spellings_s], [[s.min, s.max] for s in self.temp_sp_set.spellings_t], roadEdges=self.temp_sp_set.road_edges, halfBlock=self.AGB_ob.half_block)
			self.histograms.append(histo)
			running_y += PLOT_HEIGHT_SP_SET + SPACE_ABOVE_TEMP

		# Finally, individual spelling sets:

		lower_limit = running_y - 1
		upper_limit	= TOTAL_PANEL_HEIGHT + AGB_SOURCE_Y - 12  # Extra  to prevent overshoot at the top - we measure to the bottom of the plot, not the top


		for i, ss in enumerate (self.AGB_ob.spelling_sets):
			if i < len(self.evt_sp_set_select):		# max twenty evt enums and thus spelling sets - ought to be enough!
				histo = AGBHistogram(ss.comment, 10, running_y, 805, PLOT_HEIGHT_SP_SET, self.range_plots[0], self.range_plots[1],  'spelling_set', 0, [[s.min, s.max] for s in ss.spellings_s], [[s.min, s.max] for s in ss.spellings_t],
								roadEdges=ss.road_edges, heightStr="%d %d" % (ss.height_min, ss.height_max) if ss.has_height_range else None, evt=self.evt_sp_set_select[i], isActive=True if i == self.active_sp_set_index else False, yWindowBottom = lower_limit, yWindowTop = upper_limit, halfBlock=self.AGB_ob.half_block)
				self.histograms_scrolling.append(histo)
				running_y += self.SPELLING_SET_HT

		self.scroll_limit = max((running_y - upper_limit)/self.SPELLING_SET_HT, 0)
		self.scroll_offset = min(self.scroll_offset, self.scroll_limit)

		# Scroll button y-locations.
		if self.scroll_limit:
			num_vis_ss = (TOTAL_PANEL_HEIGHT + AGB_SOURCE_Y + 6 - lower_limit)/self.SPELLING_SET_HT
			self.scroll_up_bt_y = lower_limit + 1 + (num_vis_ss - 1) * self.SPELLING_SET_HT - 3	# This is the bottom of the top spelling set button. The -3 is because the button stack is 3 pixels lower than the plot stack, because each buttons is 6 pixels taller than its respective plot.
			self.scroll_down_bt_y = lower_limit + 1 - 3
			self.scroll_bt_h	= PLOT_HEIGHT_SP_SET + 6

		return True

	def setDrawImageData(self, x, y, w, h, bugw, bugh):
		"""
			Calculate zoom and cropping vals for drawing the grid tex in the GUI.
			'bugw' and 'bugh' are ignored by this function.
		"""
		self.draw_image_data = ()

		if self.temp_grid_ob == None or self.temp_grid_ob.def_mesh_ob == None or not self.temp_grid_ob.is_pukka: return

		width = self.temp_grid_ob.offsets_x[-1] - self.temp_grid_ob.offsets_x[0]
		depth = self.temp_grid_ob.offsets_y[-1] - self.temp_grid_ob.offsets_y[0]

		grid_width_px  = self.temp_grid_ob.pxpm_x * width
		grid_height_px = self.temp_grid_ob.pxpm_y * depth
		try:
			# Work out which dimension is the limiting one from the point of view of fitting the imageinto the available panel space:
			grid_aspect = width/depth
			panel_aspect = float(w)/float(h)
			if grid_aspect >= panel_aspect:  # grid is wider and shallower than the GUI window, so width is the limiting factor
				zoomx = w / grid_width_px
				zoomy = zoomx * float(self.temp_grid_ob.pxpm_x / self.temp_grid_ob.pxpm_y)

				self.draw_grid_size = [int(w), int(w/grid_aspect)]		# This is the size of the grid in PANEL pixels (rather than texture pixels, which may not be the same scale in both axes)
			else:
				zoomy = h / grid_height_px
				zoomx = zoomy * float(self.temp_grid_ob.pxpm_y / self.temp_grid_ob.pxpm_x)

				self.draw_grid_size = [int(h*grid_aspect), int(h)]		# This is the size of the grid in PANEL pixels (rather than texture pixels, which may not be the same scale in both axes)

			self.draw_image_data = (self.temp_grid_ob.image, x, y, zoomx, zoomy, int(self.temp_grid_ob.origin_uv[0] * self.temp_grid_ob.image_size_pix[0]), int(self.temp_grid_ob.origin_uv[1] * self.temp_grid_ob.image_size_pix[1]), int(grid_width_px), int(grid_height_px))

		except ZeroDivisionError:
			PupMenu("ERROR: zero divison%t")

	def drawEdgeTypeIcons(self, data, x, y, w, h, th, spacing, colourRoad, colourNoRoad, colourBackground, halfBlock=0):

		MUTING_RATIO = 0.25
		rects = []
		inner_rects = []
		edges = []
		rects_muted = []
		edges_muted = []

		if not halfBlock:
			all_icons = [[1, 1, 1, 1]]
		else:
			all_icons = [[1, 0, 0, 0], [1, 1, 0, 0], [1, 0, 1, 0], [1, 1, 1, 0], [1, 0, 0, 1], [1, 1, 0, 1], [1, 0, 1, 1], [1, 1, 1, 1]]

		for d in all_icons:

			flag = 1 if d in data else 0
			x1, y1 = x+w, y+h

			(rects if flag else rects_muted).append([x, y, x1, y1])
			inner_rects.append([x+th, y+th, x1-th, y1-th])

			if d[0]:
				(edges if flag else edges_muted).append([x, 	y, 		x1,		y+th])
			if d[1]:
				(edges if flag else edges_muted).append([x1-th,	y,		x1,		y1])
			if d[2]:
				(edges if flag else edges_muted).append([x, 	y1-th, 	x1,		y1])
			if d[3]:
				(edges if flag else edges_muted).append([x,		y,		x+th,	y1])

			x += spacing

		BGL.glColor3f(*colourNoRoad)
		for r in rects:
			BGL.glRecti(*r)

		BGL.glColor3f(*mix(colourBackground, colourNoRoad, MUTING_RATIO))
		for rm in rects_muted:
			BGL.glRecti(*rm)

		BGL.glColor3f(*colourBackground)
		for r in inner_rects:
			BGL.glRecti(*r)

		BGL.glColor3f(*colourRoad)
		for e in edges:
			BGL.glRecti(*e)

		BGL.glColor3f(*mix(colourBackground, colourRoad, MUTING_RATIO))
		for em in edges_muted:
			BGL.glRecti(*em)

	def buildGridImageMenu(self):
		""" Builds a menu string from the available images in blender
		"""
		self.all_image_names		= [img.name for img in Blender.Image.Get()]
		self.gridimagemenuname		= "Image Name %t|REFRESH MENU %x998"
		for i, v in enumerate(self.all_image_names):
			self.gridimagemenuname += ("|%s %%x%d" % (v, i))

	def buildGridObsMenuString(self):
		""" Builds a menu string from the available AGBGrid obs
		"""
		self.grid_obs_menu_string 	= "Active sub grid %t"

		if self.AGB_ob != None:
			for i, v in enumerate(self.AGB_ob.grid_obs):
				self.grid_obs_menu_string += ("|%s %%x%d" % (v.name, i))
			if len(self.AGB_ob.grid_obs) > 0:
				if self.active_grid_index != None:
					self.mn_active_grid_index.val = self.active_grid_index
				else:
					self.mn_active_grid_index.val = len(self.AGB_ob.grid_obs)-1

	def buildMatchGroupMenuString(self, axis):

		if axis == 'h':
			self.match_groups_h_mnstr = "Horz%t"
			if self.temp_grid_ob != None:
				for g in self.temp_grid_ob.match_groups_h:
					self.match_groups_h_mnstr += ("|%s%%x%d" % (g.name, g.index))

		elif axis == 'v':
			self.match_groups_v_mnstr = "Vert%t"
			if self.temp_grid_ob != None:
				for g in self.temp_grid_ob.match_groups_v:
					self.match_groups_v_mnstr += ("|%s%%x%d" % (g.name, g.index))

	def buildSpellingSetMenuString(self):

		self.mn_sp_set_act_menustring = "Spelling Set %t"
		if self.AGB_ob != None and self.AGB_ob.spelling_sets:
			for i, ss in enumerate(self.AGB_ob.spelling_sets):
				nm = ss.comment if ss.comment else str(i+1)
				self.mn_sp_set_act_menustring += "|%s%%x%d" % (nm, i+1)

	def getActiveImage(self):
		""" Gets the acurrently active image from the UV/image editor and selects the corresponding item in the menu_tex_image
		"""
		if not self.gridimagemenuname:
			self.buildGridImageMenu()
		image_act = Blender.Image.GetCurrent()
		if not image_act:
			PupMenu("No active image in UV/image editor window!%t")
			return
		if image_act.name in self.all_image_names:
			self.mn_tex_image.val = self.all_image_names.index(image_act.name)

	def analyseNewGrid(self, ob=None, calledFromTexPanel=False):
		"""
			The init mesh we are given may possibly be a complete, textured grid mesh,
			or else a bare schematic set of vertices denoting the positions of cuts.
			We need to be able to cope with all comers.
		"""
		scn = bpy.data.scenes.active

		if not calledFromTexPanel:
			if ob == None:
				ob_act = scn.objects.active

				#
				# Check for OB sanity
				#
				if not ob_act:
					PupMenu("No active object to analyse!%t")
					return False
				if not ob_act.type == 'Mesh':
					PupMenu("ERROR:The active object is not a mesh!%t")
					return False
				self.temp_grid_ob = AGBGrid(ob_act)
			else:
				self.temp_grid_ob = AGBGrid(ob)

		pukka = self.temp_grid_ob.checkIsPukka()

		print 'pukka', pukka
		#
		# If we don't already have vertex data, get it:
		#
		if not self.temp_grid_ob.got_geom:
			self.temp_grid_ob.readBasicGeomFromMesh()

		#
		# If we have no image or uv data yet:
		#
		if not self.temp_grid_ob.has_image:

			#
			# If the mesh has no faceUV data, prompt user to set texture data in the GUI.
			#
			if not self.temp_grid_ob.has_uvs:
				self.buttonEvent(self.evt_open_tex_panel)
				PupMenu("Please enter texture image and scale info.%t")
				return False
			#
			#  Else get tex scale and image info from the mesh itself:
			#
			else:
				self.temp_grid_ob.getTextureParamsFromMesh()

				#
				# Update the GUI tex settings if we've managed to get them from the analysed grid:
				#
				if self.temp_grid_ob.has_image:
					if self.temp_grid_ob.image.name in self.all_image_names:
						self.mn_tex_image.val = self.all_image_names.index(self.temp_grid_ob.image.name)
					else:
						self.buildGridImageMenu()
						self.mn_tex_image.val = self.all_image_names.index(self.temp_grid_ob.image.name)
				if self.temp_grid_ob.has_uvs:
					if self.temp_grid_ob.tex_scaleX and self.temp_grid_ob.tex_scaleX > 0.0:
						self.nb_texscale_x.val	=	1.0/self.temp_grid_ob.tex_scaleX
					if self.temp_grid_ob.tex_scaleY and self.temp_grid_ob.tex_scaleY > 0.0:
						self.nb_texscale_y.val	=	1.0/self.temp_grid_ob.tex_scaleY
					self.nb_texorig_x.val = self.temp_grid_ob.origin_uv[0] * self.nb_texscale_x.val
					self.nb_texorig_y.val = self.temp_grid_ob.origin_uv[1] * self.nb_texscale_y.val
		#
		# Now calculate UV coords and update our virtual tile grid:
		#
		self.temp_grid_ob.createUVCoords()

		#
		# Build coord, edge and face arrays, and if necessary create a new blender mesh:
		#
		self.temp_grid_ob.createDefaultMesh(DEF_TILE_MESH_LAYER)

		#
		# Check for an existing wrapper - if none, add one
		#
		self.temp_grid_ob.wrapGrid(DEF_TILE_WRAPPER_LAYER)

		#
		# Scrape the mesh for facemode info:
		#
		self.temp_grid_ob.getMeshFaceData()

		self.temp_grid_ob.getOverlapValuesFromWrapper()

		self.temp_grid_ob.getMatchGroupsFromMesh()

		self.temp_grid_ob.getEdgeGroupsFromMesh()

		got_rects = self.temp_grid_ob.getMetaRectData()

		if not got_rects:

			retval = PupMenu("Do you want me to add meta-rects now?%t| NO | YES ")
			if retval == 2:
				block = []
				mbest = Create(3.0)
				mmin  = Create(4.0)
				medge = Create(6.0)
				block.append(("Edge best: ",  mbest, 0.0, 20.0))
				block.append(("Edge min: ",   mmin, mbest.val, 20.0))
				block.append(("Block edge: ", medge, mmin.val, 20.0))
				retval = PupBlock("Enter offsets...", block)
				if retval:
					self.temp_grid_ob.addMetaRects(mbest.val, mmin.val, medge.val, DEF_METARECT_LAYER)

		#
		# Calculate the zoom and stencil sizes of the GUI grid image, using vals set from  drawAGBEditor()
		#

		self.setDrawImageData(*self.GUIimageWindow)

		#self.temp_grid_ob.add3DLabels()

		scn.objects.selected = []
		scn.objects.active = self.temp_grid_ob.wrapper

		return True

	def renameGrid(self):

		self.st_new_grid_name.val = incrementName(self.st_new_grid_name.val, (ob.name for ob in Object.Get()), blenderStyle=True)
		newname = self.st_new_grid_name.val

		if self.temp_grid_ob != None:
			if self.temp_grid_ob.wrapper != None:
				prnt = self.temp_grid_ob.wrapper.parent
				if prnt != None and prnt.name.startswith('AGB'):
					if self.AGB_ob and self.AGB_ob.wrapper == prnt:
						if self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
							if self.AGB_ob.spelling_sets[self.active_sp_set_index].grid_name == self.temp_grid_ob.name:
								self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
							for ss in self.AGB_ob.spelling_sets:
								if ss.grid_name == self.temp_grid_ob.name:
									ss.grid_name = newname
									setIDProp(prnt, 'XPToolsIDProps/AGB/SpellingSets/' + ss.comment + '/grid_name', newname)
									setIDProp(prnt, 'XPToolsIDProps/AGB/SpellingSets/' + ss.comment + '/constraints/grid_name', newname)
				self.temp_grid_ob.wrapper.name = newname
			self.temp_grid_ob.name = newname
			self.buildGridObsMenuString()

	def addTempGridToAGB(self):

		scn = bpy.data.scenes.active

		if self.temp_grid_ob == None:
			PupMenu("No grid mesh in the editor to add.%t")
			return

# 		if self.AGB_ob == None:
# 			PupMenu("No AGB in the editor to which to attach this grid.%t")
# 			return

		if self.temp_grid_ob.wrapper != None:
			p = self.temp_grid_ob.wrapper.parent
			if p != None and p.type == 'Empty' and p.name.startswith('AGB'):
				if self.AGB_ob == None:
					retval = PupMenu("This grid is already part of an AGB. Would you like to load it now?%t| NO | YES")
					if retval == 2:
						self.loadAGB(p, dialogs=False)
					return
				else:
					if self.AGB_ob.wrapper!= None:
						if self.AGB_ob.wrapper == p:
							PupMenu("This grid appears to be already part of the currently loaded AGB! Doing nothing.%t")
							return
						else:
							PupMenu("This grid appears to be part of a different AGB. Doing nothing.%t")
							return
			elif self.AGB_ob != None and self.AGB_ob.wrapper != None:
				self.AGB_ob.wrapper.makeParent([self.temp_grid_ob.wrapper])
				self.AGB_ob.grid_obs.append(self.temp_grid_ob)
				if self.AGB_ob.grid_obs:
					self.active_grid_index = len(self.AGB_ob.grid_obs) - 1
					self.buildGridObsMenuString()
					self.switchGridObs(self.active_grid_index)
					for g in self.AGB_ob.grid_obs:
						print g.name, g.def_mesh_ob.name
				Draw()

	def createDefaultMesh(self):
		"""
			Tell the active grid object to build a mesh from it's data
		"""
		if self.temp_grid_ob:
			self.temp_grid_ob.createDefaultMesh(DEF_TILE_MESH_LAYER, DEF_TILE_WRAPPER_LAYER)

	def createNewAGB(self):

		self.querySaveAGB()

		scn = bpy.data.scenes.active
		obact = scn.objects.active
		editmode = Window.EditMode()
		if editmode:
			Window.EditMode(0)
		if not obact:
			PupMenu("No active object to develop! Please select a tile grid wrapper.%t")
			return

		metalayer = DEF_TILE_WRAPPER_LAYER

		Window.SetActiveLayer(1<<metalayer)
		Window.RedrawAll()
		cursor = Window.GetCursorPos()
		cursor[2] = 0.0
		wrapper = scn.objects.new("Empty")
		if self.st_AGB_name.val == "":
			self.st_AGB_name.val = "AGBNewName"
		elif not self.st_AGB_name.val.startswith("AGB"):
			self.st_AGB_name.val = "AGB" + self.st_AGB_name.val
		self.st_AGB_name.val = self.st_AGB_name.val[:17]
		wrapper.setName(self.st_AGB_name.val)
		if wrapper.name != self.st_AGB_name.val:	# In case of Blender auto name increment because of a name conflict
			self.st_AGB_name.val = wrapper.name

		wrapper.setLocation(cursor)
		wrapper.drawSize = 4.0

		wrapper.addProperty("VNAME", 		"", 			'STRING')
		setIDProp(wrapper, 'XPToolsIDProps/AGB/name', self.st_AGB_name.val)
		# Now look for any other selected orphan empty objects in the scene, and parent them to the one we just added.

		selected_empties = [ob for ob in scn.objects.selected if ob.type == 'Empty' and not ob.parent]
		wrapped_tile_meshes	= [tm for tm in scn.objects if tm.type == 'Mesh' and tm.name.startswith('TILE') and tm.parent]
		candidates = [t.parent for t in wrapped_tile_meshes if t.sel or t.parent in selected_empties]
		wrapper.makeParent(candidates)
		scn.update()
		scn.objects.active = wrapper
		self.loadAGB(wrapper)

	def loadAGB(self, obj=None, dialogs=True):
		"""
			Load a blender AGB hierarchy into the editor, given any active blender object which is either an AGB wrapper or a child thereof.
		"""
		scn = bpy.data.scenes.active
		if obj == None:
			obact = scn.objects.active
			self.querySaveAGB()
		else:
			obact = obj

		editmode = Window.EditMode()
		if editmode:
			Window.EditMode(0)

		if not obact:
			PupMenu("No active object to load!%t")
			return

		parent = obact.getParent()
		while parent != None:
			obact = parent
			parent = obact.parent

		if obact.type != 'Empty':
			PupMenu("Please select an AGB 'empty' wrapper!%t")
			return
		if not obact.name.startswith("AGB"):
			PupMenu("The active object doesn't appear to be part of an AGB.%t")
			return

		# Delete existing goo various

		if self.AGB_ob != None:
			if self.AGB_ob.spelling_sets != None:
				del self.AGB_ob.spelling_sets
			if self.AGB_ob.grid_obs:
				del self.AGB_ob.grid_obs
			del self.AGB_ob
			self.active_grid_index	= None
			self.active_sp_set_index = None
			if self.temp_sp_set:
				self.temp_sp_set = AGBSpellingSet()
			self.mn_sp_set_act_menustring	= "Spelling Set %t"

		# Create new AGB_ob
		self.AGB_ob	= AGBObject(obact)
		self.st_AGB_name.val = obact.name
		obs_level2 = getGrandChildren(obact, scn.objects)
		for ob in [ob for ob in obs_level2 if ob.type == 'Mesh']:
			if ob.name.startswith("TILE"):
				prnt = ob.parent
				if prnt.type == 'Empty':
					mesh_ok = self.analyseNewGrid(ob)
					if mesh_ok:
						self.AGB_ob.grid_obs.append(self.temp_grid_ob)
						self.active_grid_index = len(self.AGB_ob.grid_obs) - 1

		self.buildGridObsMenuString()
		if self.AGB_ob.grid_obs:
			self.active_grid_index = 0
			self.switchGridObs(0)
			for g in self.AGB_ob.grid_obs:
				print g.name, g.def_mesh_ob.name

		if self.AGB_ob.spelling_sets:
			for ss in self.AGB_ob.spelling_sets:
				print "bp1", ss.comment, ss.grid_name
				if ss.grid_name not in [g.name for g in self.AGB_ob.grid_obs]:
					if dialogs:
						retval = PupMenu("ERROR: grid '%s' appears to be missing - what to do?%%t| Wait! - I'll find the grid and reparent it to the AGB before proceeding further! | Go ahead and delete all spelling sets which reference '%s'!" % (ss.grid_name,ss.grid_name))
						if retval != 2:
							return
					else:
						for s in list(self.AGB_ob.spelling_sets):
							if s.grid_name not in (g.name for g in self.AGB_ob.grid_obs):
								self.deleteSpellingSet(s)

			self.buildSpellingSetMenuString()
			self.active_sp_set_index = len(self.AGB_ob.spelling_sets) - 1
			self.mn_sp_set_act.val = self.active_sp_set_index + 1
			self.AGB_ob.updateSpellingSetMetrics()
			for ss in self.AGB_ob.spelling_sets:
				ss.sortSpellings()

			if getIDProp(self.AGB_ob.wrapper, 'XPToolsIDProps', None) != None:
				for ss in self.AGB_ob.spelling_sets:
					cc = getIDProp(self.AGB_ob.wrapper, 'XPToolsIDProps/AGB/SpellingSets/' + ss.comment, None)
					if cc != None:
						message = ss.setConstraintsConfig(self.AGB_ob, cc)
						if message != 'CONSTRAINTS_OK':
							retval = PupMenu("ERROR: AGB grid '%s' not found. Possibly it has been removed from the AGB? Delete spelling sets which reference the missing grid?%%t| NO| YES" % message )
							if retval == 2:
								for ss in self.AGB_ob.spelling_sets:
									if ss.grid_name == message:
										self.deleteSpellingSet(ss)
										try:
											removeIDProp(self.AGB_ob.wrapper,  'XPToolsIDProps/AGB/SpellingSets/'+message)
										except KeyError, detail:
											print "ERROR when trying to delete spelling set config dict %s:" % 'XPToolsIDProps/AGB/SpellingSets/'+ss.comment, detail
							else:
								PupMenu("Notice: Spelling sets which reference grid %s have been de-activated. They may be re-activated if the missing grid is re-parented to the AGB, and the AGB reloaded.%%t| UNDERSTOOD" % message)
				try:
					all_ss_names = [ss.comment for ss in self.AGB_ob.spelling_sets]
					for k in self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'].keys():
						if k not in all_ss_names:
							del self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'][k]
				except KeyError, detail:
					print "ERROR: ", detail

		self.mn_AGB_depth_range.val = self.AGB_ob.depth_menu_index
		self.mn_AGB_height_bucket.val = self.AGB_ob.height_menu_index
		self.tg_AGB_full_half.val  = self.AGB_ob.half_block
		scn.objects.selected = [obact]
		scn.objects.active = obact
		self.refreshGUI(self.evt_load_AGB)
		Window.RedrawAll()

	def duplicateAGB(self, deep=1):
		"""
			Duplicate an entire AGB hierarchy.
		"""
		# Duplicating an AGB in blender is a bit of a pain because of ob name-changes breaking various references.
		# We basically duplicate stuff in batches, noting the old and new names as we go.
		# Afterwards, nip round and search/replace the critical names in the new AGB text object and IDProps, then load the new AGB and hope for the best.

		if self.AGB_ob == None or self.AGB_ob.wrapper == None:
			PupMenu("Nothing to duplicate.%t")
			return

		oldwrappers = []
		newwrappers	= []
		scn = bpy.data.scenes.active

		scn.objects.selected = [self.AGB_ob.wrapper]
		Object.Duplicate()
		AGBwrappernew = scn.objects.active
		# We have to make sure that the new wrapper name doesn't conflict with existing ob and TEXT names across all scenes:
		AGBwrappernew.name = incrementName(AGBwrappernew.name, [ob.name for ob in bpy.data.objects] + [t.name for t in bpy.data.texts], blenderStyle=True)



		for g in self.AGB_ob.grid_obs:
			oldwrappers.append(g.wrapper)
		for ow in oldwrappers:
			scn.objects.selected = getGrandChildren(ow, scn.objects)
			ow.sel = 1
			scn.objects.active = ow
			Object.Duplicate(mesh=deep)
			nw = scn.objects.active
			newwrappers.append(nw)
			nw.clrParent(2)

		AGBwrappernew.makeParent(newwrappers)
		AGBold_real_name = self.AGB_ob.wrapper.name[3:]+".agb"
		AGBnew_real_name = AGBwrappernew.name[3:]+".agb"
		ow_names = [ow.name for ow in oldwrappers]
		nw_names = [nw.name for nw in newwrappers]

		# Text object
		newtext = bpy.data.texts.new(AGBwrappernew.name)
		for line in self.AGB_ob.text.asLines():
			if line[:2] == "# " and line[2:].split()[0] == AGBold_real_name:
				line = "# " + AGBnew_real_name
			elif line.startswith('USE_GRID'):
				toks = line.split()
				if len(toks) > 1 and toks[1] in ow_names:
					line = line.replace(toks[1], nw_names[ow_names.index(toks[1])])
			newtext.write(line + "\n")

		# IDProps
		if self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
			for ss in self.AGB_ob.spelling_sets:
				try:
					oldgridname = AGBwrappernew.properties['XPToolsIDProps']['AGB']['SpellingSets'][ss.comment]['grid_name']
					if oldgridname in ow_names:
						newgridname = nw_names[ow_names.index(oldgridname)]
						AGBwrappernew.properties['XPToolsIDProps']['AGB']['SpellingSets'][ss.comment]['grid_name'] = newgridname
						AGBwrappernew.properties['XPToolsIDProps']['AGB']['SpellingSets'][ss.comment]['constraints']['grid_name'] = newgridname
				except KeyError, detail:
					print "ERROR:", detail
		setIDProp(AGBwrappernew, 'XPToolsIDProps/AGB/name', AGBwrappernew.name)

		scn.objects.selected = [AGBwrappernew]
		scn.objects.active = AGBwrappernew
		AGBwrappernew.LocX += 50.0
		Window.RedrawAll()

	def duplicateGridOb(self):

		scn = bpy.data.scenes.active
		Window.EditMode(0)

		SPACING = 10.0

		def get_rect(bb):
			return [bb[0], bb[2], bb[4], bb[6]]

		def BB_overlap_2D(bb1, bb2):
			br1 = get_rect(bb1)
			br2 = get_rect(bb2)
			for p in br2:
				if min(br1, key=lambda x:x[0])[0] <= p[0] <= max(br1, key=lambda x:x[0])[0] and min(br1, key=lambda x:x[1])[1] <= p[1] <= max(br1, key=lambda x:x[1])[1]:
					return True
			return False

		def next_BB(bb1, bb2):
			bb1_ymax = max(bb1, key = lambda x: x[1])[1]
			bb2_ymin = min(bb2, key = lambda x: x[1])[1]
			if BB_overlap_2D(bb1, bb2):
				dy 		 = bb1_ymax - bb2_ymin + SPACING
				for v in bb2:
					v[1] += dy
				return bb2
			else:
				return bb2

		if self.AGB_ob and self.AGB_ob.grid_obs and self.active_grid_index != None:
			new_bb = self.AGB_ob.grid_obs[self.active_grid_index].def_mesh_ob.boundingBox
			dy = max(new_bb, key = lambda x: x[1])[1] - min(new_bb, key = lambda x: x[1])[1] + SPACING
			for v in new_bb:
				v[1] += dy

			all_bbs = []
			for g in self.AGB_ob.grid_obs:
				all_bbs.append(g.def_mesh_ob.boundingBox)

			collide = True
			madness = 0
			while collide and madness < 20:
				madness += 1
				for bb in all_bbs:
					if BB_overlap_2D(bb, new_bb):
						new_bb = next_BB(bb, new_bb)
						continue
				else:
					collide = False

			delta_y = min(new_bb, key = lambda x: x[1])[1] - min(self.AGB_ob.grid_obs[self.active_grid_index].def_mesh_ob.boundingBox, key = lambda x: x[1])[1]

			if self.temp_grid_ob != None:

				# Check for spelling sets which use this grid:
				ss_to_copy = []
				if self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
					for i, ss in enumerate(self.AGB_ob.spelling_sets):
						if ss.grid_name == self.AGB_ob.grid_obs[self.active_grid_index].wrapper.name:
							ss_to_copy.append(i)

				prnt = self.temp_grid_ob.def_mesh_ob.parent
				if prnt != None:
					children = getGrandChildren(prnt, scn.objects)
					objs = []
					trees = []
					facades = []
					pins = []
					to_copy = [self.temp_grid_ob.def_mesh_ob, prnt]

					for o in children:
						if o.type == 'Mesh':
							if o.name.startswith('META'):
								to_copy.append(o)
							elif o.name.startswith('TREE'):
								trees.append(o)
							elif o.name.startswith('FAC'):
								facades.append(o)
						elif o.type == 'Empty' and o.name.startswith('OBJ'):
							objs.extend(getGrandChildren(o, children))
						elif o.type == 'Lamp' and o.name.startswith('PIN'):
							pins.append(o)

					if objs or trees or facades or pins or ss_to_copy:
						do_objs = Create(1)
						do_trees = Create(1)
						do_facades = Create(1)
						do_pins = Create(1)
						do_ss = Create(1)
						block = []
						if objs:
							block.append(("objects", do_objs, "Also duplicate OBJs."))
						if facades:
							block.append(("facades", do_facades, "Also duplicate facades."))
						if trees:
							block.append(("trees", do_trees, "Also duplicate trees."))
						if pins:
							block.append(("meta-pins", do_pins, "Also duplicate meta-pins."))
						if ss_to_copy:
							block.append(("spelling sets", do_ss, "Also duplicate spelling sets."))
						retval = PupBlock("Also duplicate...", block)
						if retval == 0:
							return
						if do_objs.val:
							to_copy.extend(objs)
						if do_trees.val:
							to_copy.extend(trees)
						if do_facades.val:
							to_copy.extend(facades)
						if do_pins.val:
							to_copy.extend(pins)
						if not do_ss.val:
							ss_to_copy = []

					# All goo must be on visible layers:
					oblayers = 	set(scn.layers)
					for ob in to_copy:
						oblayers.update(ob.layers)
					oblayers.update(prnt.layers)
					scn.layers = list(oblayers)

					scn.objects.selected = to_copy
					scn.objects.active = prnt
					Object.Duplicate(mesh=1)
					prnt_dup = scn.objects.active
					prnt_dup.clrParent(2)
					prnt_dup.LocX = round(prnt_dup.LocX)
					prnt_dup.LocY = round(prnt_dup.LocY + delta_y)
					prnt_dup.LocZ = 0.0
					for ch in getChildren(prnt_dup, scn.objects):
						if ch.type == 'Mesh' and ch.name.startswith('TILE'):
							self.analyseNewGrid(ch)
							self.addTempGridToAGB()
							if ss_to_copy:
								for i in ss_to_copy:
									new_ss = self.AGB_ob.spelling_sets[i].copy(incrementName(self.AGB_ob.spelling_sets[i].comment, [ss.comment for ss in self.AGB_ob.spelling_sets]))
									self.AGB_ob.spelling_sets.append(new_ss)
									new_ss.grid_name = prnt_dup.name
									self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'][new_ss.comment] = {}
									self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'][new_ss.comment].update(self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'][self.AGB_ob.spelling_sets[i].comment].convert_to_pyobject())
									self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'][new_ss.comment]['grid_name'] =  prnt_dup.name
									self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'][new_ss.comment]['constraints']['grid_name'] =  prnt_dup.name
									self.AGB_ob.wrapper.properties['XPToolsIDProps']['AGB']['SpellingSets'][new_ss.comment]['spelling_set_name'] =  new_ss.comment
								self.buildSpellingSetMenuString()
								self.active_sp_set_index = len(self.AGB_ob.spelling_sets) - 1
								self.mn_sp_set_act.val = self.active_sp_set_index + 1
							break

	def deleteGridOb(self):
		"""
			Delete the active grid object, associated spelling sets and, optionally, any attached annotations.
		"""
		retval = PupMenu("Delete the active grid and associated spelling sets?%t|NO|YES")
		if retval < 2:
			return
		retval = PupMenu("Also delete attached annotations?%t|NO|YES")

		scn = bpy.data.scenes.active
		Window.EditMode(0)
		if self.AGB_ob and self.AGB_ob.grid_obs and self.active_grid_index != None:
			children = getGrandChildren(self.temp_grid_ob.wrapper, scn.objects)
			to_delete = []
			to_unparent = []

			if retval == 1:		# Keep annotations
				for o in children:
					if o.type == 'Mesh':
						if o.name.startswith('META') or o.name.startswith('TILE'):
							to_delete.append(o)
						elif o.name.startswith('TREE'):
							to_unparent.append(o)
						elif o.name.startswith('FAC'):
							to_unparent.append(o)
					elif o.type == 'Empty' and o.name.startswith('OBJ'):
						to_unparent.append(o)
					elif o.type == 'Lamp' and o.name.startswith('PIN'):
						to_unparent.append(o)
				to_delete.append(self.temp_grid_ob.wrapper)
				for ob in to_unparent:
					ob.clrParent(2, 1)
				self.temp_grid_ob.wrapper.makeParent([])
			elif retval == 2:
				to_delete = children
			else:
				return
			if self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
				for i, ss in enumerate(self.AGB_ob.spelling_sets):
					if ss.grid_name == self.temp_grid_ob.name:
						self.deleteSpellingSet(ss)
			for ob in to_delete:
				scn.objects.unlink(ob)
			self.AGB_ob.saveAGB(dialogs=False)
			self.unsaved_changes = False
			scn.objects.active = self.AGB_ob.wrapper
			self.loadAGB(self.AGB_ob.wrapper, dialogs=False)

	def getCommonConstraintsSettings(self):
		"""
			Make a dict from the various spelling-set-wide panel vars
			"""
		d = {'horizontal': {}, 'vertical': {}}
		for axis in self.common_constraints.iterkeys():
			for k, v in self.common_constraints[axis].iteritems():
				d[axis][k] = v
		return d

	def setCommonConstraintsSettings(self, data):
		try:
			for axis in self.common_constraints.iterkeys():
				for k in self.common_constraints[axis].iterkeys():
					self.common_constraints[axis][k] = data[axis][k]
		except (KeyError, ValueError), detail:
			print "ERROR:", detail
		else:
			self.updateCommonConstraintButtons()

	def updateCommonConstraintButtons(self):
		a = 'vertical' if self.spellings_axis else 'horizontal'
		self.nb_max_length.val			= self.common_constraints[a]['max_length']
		self.nb_min_length.val			= self.common_constraints[a]['min_length']
		self.nb_max_dupes.val			= self.common_constraints[a]['max_dupes']
		self.nb_max_cuts.val			= self.common_constraints[a]['max_cuts']
		self.nb_max_iterations.val		= self.common_constraints[a]['max_iterations']
		self.nb_center_precision.val	= self.common_constraints[a]['center_precision']
		self.sl_sp_variety.val			= self.common_constraints[a]['sp_variety']

	def addSpellingSet(self):
		"""
		Adds a new spelling set to the current AGB, using the GUI values for height and freq
		"""
		if not self.AGB_ob:
			PupMenu("An AGB must be loaded or created before a new spelling set can be added to it.%t")
			return
		if not self.temp_grid_ob:
			PupMenu("Can't make a spelling set without a tile grid mesh object.")
			return
		if self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
			self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
			namestr =  incrementName(self.AGB_ob.spelling_sets[self.active_sp_set_index].comment, [n.comment for n in self.AGB_ob.spelling_sets])
		else:
			namestr = "Spelling Set 1"
		newname = Create(namestr)
		block = []
		block.append(("Name: ", newname, 0, 32, "Enter a name for the new spelling set."))
		retval = PupBlock('Enter a name', block)
		if retval != 1:
			return
		newset = AGBSpellingSet()
		allnames = [ss.comment for ss in self.AGB_ob.spelling_sets]
		if newname.val in allnames:
			newname.val = incrementName(newname.val, allnames)
		newset.comment = newname.val
		newset.height_min = self.nb_sp_set_height_min.val
		newset.height_max = self.nb_sp_set_height_max.val
		newset.frequency  = self.nb_sp_set_frequency.val
		newset.grid_name  = self.temp_grid_ob.name
		self.AGB_ob.spelling_sets.append(newset)
		self.buildSpellingSetMenuString()
		self.active_sp_set_index = len(self.AGB_ob.spelling_sets) - 1
		self.mn_sp_set_act.val = self.active_sp_set_index + 1
		self.refreshGUI(self.evt_sp_set_act)
		self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)

	def deleteSpellingSet(self, spellingSet=None):
		"""
			Delete the active spelling set from the AGB
		"""

		if self.AGB_ob and self.AGB_ob.spelling_sets:
			# Direct call from another function
			if spellingSet != None:
				for i, ss in enumerate(self.AGB_ob.spelling_sets):
					if ss.comment == spellingSet.comment:
						try:
							removeIDProp(self.AGB_ob.wrapper,  'XPToolsIDProps/AGB/SpellingSets/'+ss.comment)
						except KeyError, detail:
							print "ERROR when trying to delete spelling set config dict %s:" % 'XPToolsIDProps/AGB/SpellingSets/'+ss.comment, detail
						else:
							self.AGB_ob.spelling_sets[i:i+1] = []
							if len(self.AGB_ob.spelling_sets) > 0:
								if self.active_sp_set_index != None:
									self.active_sp_set_index = clamp(0, len(self.AGB_ob.spelling_sets)-1, self.active_sp_set_index - 1)
								else:
									self.active_sp_set_index = 0
							else:
								self.active_sp_set_index = None
							self.buildSpellingSetMenuString()
							self.mn_sp_set_act.val = 0 if self.active_sp_set_index == None else self.active_sp_set_index + 1

					 		self.refreshGUI(self.evt_sp_set_act)
 							if self.active_sp_set_index != None:
 								self.AGB_ob.loadActiveConstraintsIDProps(self.active_sp_set_index)
						break
				else:
					print "ERROR: can't find spelling set."
					return
			# Called by user:
			else:
				retval = PupMenu("Delete the active spelling set?%t| NO| YES")
				if retval != 2:
					return
			act = self.active_sp_set_index
			if act == None:
				PupMenu("Nothing to Delete!%t")
				return

			try:
				removeIDProp(self.AGB_ob.wrapper,  'XPToolsIDProps/AGB/SpellingSets/'+self.AGB_ob.spelling_sets[act].comment)
				del self.AGB_ob.spelling_sets[act]
			except IndexError, detail:
				print "ERROR Index out of range:", detail
			else:
				if len(self.AGB_ob.spelling_sets) == 0:
					self.active_sp_set_index = None

				elif act == len(self.AGB_ob.spelling_sets):
					self.active_sp_set_index -= 1

				self.buildSpellingSetMenuString()
				self.mn_sp_set_act.val = 0 if self.active_sp_set_index == None else self.active_sp_set_index + 1
				self.refreshGUI(self.evt_sp_set_act)
				if self.active_sp_set_index != None:
					self.AGB_ob.loadActiveConstraintsIDProps(self.active_sp_set_index)

	def duplicateSpellingSet(self):
		"""
			Make a copy of the active spelling set and add it to the AGB
		"""
		if self.AGB_ob and self.AGB_ob.spelling_sets:

			self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
			self.AGB_ob.spelling_sets.append(self.AGB_ob.spelling_sets[self.active_sp_set_index].copy(incrementName(self.AGB_ob.spelling_sets[self.active_sp_set_index].comment, [ss.comment for ss in self.AGB_ob.spelling_sets])))
			self.active_sp_set_index = len(self.AGB_ob.spelling_sets) - 1
			self.buildSpellingSetMenuString()
			self.mn_sp_set_act.val = self.active_sp_set_index + 1
			self.refreshGUI(self.evt_sp_set_act)
			self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)

	def makeSpellings(self):

		if not self.temp_grid_ob:
			return
		self.temp_sp_set.grid_name = self.temp_grid_ob.name

		self.MAX_SPELLINGS_PER_COUNT	= self.sl_sp_variety.val
		self.MIN_TILES 					= 2
		self.MAX_DUPES					= self.nb_max_dupes.val
		self.MAX_CUTS 					= self.nb_max_cuts.val
		self.MAX_ITERATIONS				= self.nb_max_iterations.val
		self.CENTER_PRECISION			= self.nb_center_precision.val
		self.MAX_SPELLING_LENGTH		= self.nb_max_length.val
		self.MIN_SPELLING_LENGTH		= self.nb_min_length.val
		self.start_tiles 				= []
		self.end_tiles 					= []
		self.RF_tiles					= []
		self.NRF_tiles					= []
		self.required_tiles 			= []
		self.xor_tiles					= []
		self.slop_tiles					= []
		self.centered_tiles				= []
		self.xor_required				= False		# True if the set of xor tiles are also required - i.e. exactly ONE of the set of xor tiles MUST be included in all spellings
		self.pool 						= []
		self.edge_names					=	['VN', 'VR', 'HN', 'HR']
		self.tilelines 					= self.temp_grid_ob.columns if self.spellings_axis == 0 else self.temp_grid_ob.rows

		for t in [t for t in self.tilelines if not t.ignore and not t.exclude]:
			if t.require:
				self.required_tiles.append(t.index)
			if t.xor:
				self.xor_tiles.append(t.index)
			if t.xor and t.require:
				self.xor_required = True
			if t.is_slop:
				self.slop_tiles.append(t.index)
			if t.center:
				self.centered_tiles.append(t.index)
			if t.match_trail in self.edge_names:
				self.start_tiles.append(t.index)
				if t.match_trail in ['VR', 'HR']:
					self.RF_tiles.append(t.index)
				else:
					self.NRF_tiles.append(t.index)
				continue
			if t.match_lead in self.edge_names:
				self.end_tiles.append(t.index)
				if t.match_lead in ['VR', 'HR']:
					self.RF_tiles.append(t.index)
				else:
					self.NRF_tiles.append(t.index)
			self.pool.append(t.index)

		self.genSpell()

	def numCuts(self, seq):
		if seq != None and len(seq) > 1:
			t = 0
			for i in range(1, len(seq)):
				if seq[i] - seq[i-1] != 1 or seq[i-1] in self.slop_tiles:
					t +=1
			return t
		return 0

	def hasRequired(self, seq):
		if not self.xor_required:
			return all([r in seq for r in self.required_tiles])
		required_only = [r for r in self.required_tiles if r not in self.xor_tiles]
		return all([r in seq for r in required_only]) and (len([rx for rx in self.xor_tiles if rx in seq]) == 1)

	def isCentered(self, spellingOb):
		ct = [t for t in spellingOb.spelling if t in self.centered_tiles]
		ct = []
		for j, t in enumerate(spellingOb.spelling):
			if t in self.centered_tiles:
				ct.append([t, j])

		if not ct:
			return  True
		if len(ct) == 1:
			return abs((spellingOb.min + spellingOb.max)*0.5 - sum(self.temp_grid_ob.getLengthAt(self.spellings_axis, spellingOb.spelling, ct[0][1] )))*0.5 < self.CENTER_PRECISION
		else:
			# If we have multiple centered tiles, make sure that their order is symmetrical, e.g. tile numbers 2 5 5 2 or 1 3 1
			# Pop a tile off each end, and if they're different, we fail the test.
			while len(ct) > 1:
				if ct.pop()[0] != ct.pop(0)[0]:
					return False
			locs = []
			# Make a list of mean distances to the middle of the tile in question
			for i in range(len(spellingOb.spelling)):
				if spellingOb.spelling[i] in self.centered_tiles:
					locs.append(sum(self.temp_grid_ob.getLengthAt(self.spellings_axis, spellingOb.spelling, i)) * 0.5)

			while locs:
				p = [locs.pop(0)]
				if locs:
					p.append(locs.pop())
				if not abs((spellingOb.min + spellingOb.max)*0.25  - sum(p)/len(p)) <  self.CENTER_PRECISION:
					return False
			return True

	def hasAtLeastOneOf(self, seq):
		if self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['xor_is_or'] and self.xor_tiles:
			return any((t for t in seq if t in self.xor_tiles))
		else:
			return True

	def matchesLinks(self, previous, current):

		if self.tilelines[previous].only_before and current not in self.tilelines[previous].only_before:
			return False
		if self.tilelines[current].only_after and previous not in self.tilelines[current].only_after:
			return False
		return True

	def genSpell(self):

		self.spellings_axis = self.tg_axis_vertical.val
		if self.spellings_axis == 0:
			self.tilelines 	= self.temp_grid_ob.columns
		elif self.spellings_axis == 1:
			self.tilelines 	= self.temp_grid_ob.rows

		time = Blender.sys.time()
		spell = []
		for st in self.start_tiles:
			spell.append([st])
		done = 0
		iterations = 0
		do_timing = True
		completed_spellings = []
		set_xor_tiles = set(self.xor_tiles)
		newspell = []
		mean_length = 0.0
		AVERAGING_FACTOR = 0.05
		self.updateProgress("STARTING SPELLING GENERATION", mode='generating')

		while not self.maxed_out_iterations and not done:

			for s in spell:
				iterations += 1
				if not iterations % 10000:
					self.updateProgress("LENGTH: %.1fm   %d ITERATIONS" % (round(mean_length,1), iterations), mode='generating')
					if Window.TestBreak():
						done = True
						break
					if iterations >= self.MAX_ITERATIONS:
						done = True
						self.maxed_out_iterations = True
						break

				if s[-1] in self.end_tiles:
					for t in s:
						if self.tilelines[t].dupes_min != None and self.tilelines[t].dupes_min > s.count(t):
							continue
					else:
						completed_spellings.append(tuple(s))
						actual_length = self.temp_grid_ob.computeSpellingMetrics(self.spellings_axis, s)[0]
						mean_length = AVERAGING_FACTOR * actual_length  + (1.0 - AVERAGING_FACTOR) * mean_length
 						if mean_length > self.MAX_SPELLING_LENGTH:
							done = 1
							break
					continue
				else:
					for t in (t for t in self.pool if self.tilelines[t].match_trail == self.tilelines[s[-1]].match_lead):
						if not self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['xor_is_or'] and t in set_xor_tiles:
							if any((i for i in s if i in set_xor_tiles and not i == t)):
								continue
						if self.tilelines[t].only_after:
							for lst in (lst for lst in self.tilelines[t].only_after if len(lst) <= len(s)):
								for i in range(1, 1 + len(lst)):
									if lst[-i] != s[-i]:
										break
								else:
									break
							else:
								continue
						tn = s + [t]
						if self.numCuts(tn) > self.MAX_CUTS:
							continue
						if  s.count(t) > (self.MAX_DUPES if self.tilelines[t].dupes_max == None else self.tilelines[t].dupes_max):
							continue

						if self.tilelines[t].only_before:
							for lst in self.tilelines[t].only_before:
								snew = tn[:]
								snew.extend(lst)
								newspell.append(snew)
						else:
							newspell.append(tn)

			spell = newspell[:]
			if not spell:
				done = True
			newspell[:] = []
# 		print 'time bp 2:', Blender.sys.time() - time
		self.last_iterations = str(iterations)
		print 'total iterations:', iterations
		print 'number of results:', len(completed_spellings)

		self.updateProgress("FILTERING  AND  SORTING...", mode='generating')
		SORT_MODE = 2
		s_len = self.MIN_TILES
		counter = 0
		done = 0
		filtered = set()

		def sortByLength(sList, numResults, length):
			def numDupesTotal(sp):
				return len(sp.spelling) - len(set(sp.spelling))

			bucket = [x for x in sList if x.max == length]
			all_mins = list(set([x.min for x in bucket]))
			all_mins.sort()
			print "length, all_mins:", length, all_mins
			result = []
			for lmin in all_mins:
				result.extend(heapq.nsmallest(numResults, [b for b in bucket if b.min == lmin] , key = numDupesTotal))
			return result

		def sortByBucket(sList, numResults, low, high):
			bucket = [x for x in sList if low < x.min < high and x.max > high]
			for b in bucket:
				sList.remove(b)
			return heapq.nsmallest(numResults, bucket, key = lambda x: high - x.min)

		self.temp_sp_set.clearSpellings(self.spellings_axis)

		if SORT_MODE == 2:

			templist = []
			for s in ( s for s in completed_spellings if self.hasRequired(s) and self.hasAtLeastOneOf(s)):
				templist.append(AGBSpelling( s,  self.numCuts(s), self.temp_grid_ob.computeSpellingMetrics(self.spellings_axis, s), s[0] in self.RF_tiles, s[-1] in self.RF_tiles ))
			if not templist:
				return

			# sort by minimum length
			templist.sort(key = lambda x : x.min)
			if self.centered_tiles:
				templist = [t for t in templist if self.isCentered(t)]

			runout = False
			all_lengths = list(set([s.max for s in templist if s.max >= self.MIN_SPELLING_LENGTH and s.min <= self.MAX_SPELLING_LENGTH ]))
			if not all_lengths: return
			all_lengths.sort()

			for l in all_lengths:
				batch = set(sortByLength(templist, self.MAX_SPELLINGS_PER_COUNT, l	))
				if not batch:
					print "No batch for length %f!" % l
				else:
					batch = list(batch)
					self.temp_sp_set.extendSpellings(self.spellings_axis, batch)
			self.temp_sp_set.encodeEdges()
			del templist
			del spell
			del completed_spellings

		self.printDivider()
		self.temp_sp_set.printSet()
		print 'time bp 3',  Blender.sys.time() - time
		print "total spelling count =", len(self.temp_sp_set.spellings_t) if self.spellings_axis else len(self.temp_sp_set.spellings_s)
		self.updateProgress("COMPLETE", mode='generating')

	def printDivider(self, destination=None):
		if destination == None:
			if not AGBEditor.temp_text:
				self.getTempText()
			txt = AGBEditor.temp_text
		else:
			txt = destination
		txt.write("\n\n#" + 24 * "   #" + "\n\n")
		txt.write("# " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")	)

	def getSpellingByLength(self, axis, spellingSet, length):
		if not axis:
			cands = [s for s in spellingSet.spellings_s if s.min < length <= s.max]
		else:
			cands = [s for s in spellingSet.spellings_t if s.min < length <= s.max]
		if cands:
			return max(cands, key = lambda x : x.max)
		else:
			return None

	def updateProgress(self, message, mode='make_previews'):
		if mode == 'make_previews':
			self.progress_message = message
		elif mode == 'generating':
			self.progress_message2 = message
		Draw()

	def resetProgress(self):
		self.progress_message = ""
		self.progress_message2 = ""
		self.last_iterations = ""

	def clearAGBPreviews(self):
		"""
			Unlink all the preview grids and dupli empty objects and append them to lists.
			Keeping lists of strays makes for quicker ob recycling during the next run of makeAGBPreviews().
		"""
		elapsed("top of clearAGBPreviews")

		scn = bpy.data.scenes.active
		for ob in list(scn.objects):
			if ob.name.startswith("XPTAGBPreview"):
				scn.objects.unlink(ob)
				self.preview_obs.append(ob)
			elif ob.name.startswith("XPToolsAGBDupl"):
				scn.objects.unlink(ob)
				self.dupli_obs.append(ob)
		for g in [g for g in Blender.Group.Get() if g.name.startswith("XPToolsDupGrp")]:
			Blender.Group.Unlink(g)
		self.preview_source = 'undefined'
		elapsed("bottom of clearAGBPreviews")

	def makeAGBPreviews(self, source='temp', mode='full'):
		"""
			source states which spellings to use for the basis of the previews:
			'temp' -> latest output from the generator
			'active'  -> active spelling set
		"""
		self.preview_source = 'undefined'
		if not self.temp_grid_ob:
			return

		self.clearAGBPreviews()

		SPACING = 10.0

		scn = bpy.data.scenes.active

		xyz = Window.GetCursorPos()
		xy = []
		gloc = Vector(xyz[0], xyz[1], 0.0)
		editmode = Window.EditMode()
		if editmode:
			Window.EditMode(0)

		if self.AGB_ob:
			for s in self.temp_grid_ob.columns:
				s.updateMetrics()
			for t in self.temp_grid_ob.rows:
				t.updateMetrics()
			if self.tg_pv_annotations.val:
				self.temp_grid_ob.doAnnotations()
			pv_spellings_s = []
			pv_spellings_t = []
			lengths_s = []
			lengths_t = []

			if source == 'temp':
				spelling_set = self.temp_sp_set
			elif source == 'active' and self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
				spelling_set = self.AGB_ob.spelling_sets[self.active_sp_set_index]
			else:
				return

			# If we want to build a series of preview grids based on physical length and depth (e.g. build previews at intervals of n metres within specified ranges of width and depth):
			if self.tg_by_size_pv.val:

				xs = float(self.nb_min_length_pv.val)
				while xs <= float(self.nb_max_length_pv.val):
					lengths_s.append(xs)
					xs += self.pv_increment_metres

				xt = float(self.nb_min_depth_pv.val)
				inc = self.pv_increment_metres if mode == 'full' else (float(self.nb_max_depth_pv.val) - float(self.nb_min_depth_pv.val))/max(len(lengths_s), 1)
				while xt <= float(self.nb_max_depth_pv.val):
					lengths_t.append(xt)
					xt += inc

				for len_s in lengths_s:
					res_s = self.getSpellingByLength(0, spelling_set, len_s)
					if res_s:
						pv_spellings_s.append(res_s)
					else:
						pv_spellings_s.append(None)
				for len_t in lengths_t:
					res_t = self.getSpellingByLength(1, spelling_set, len_t)
					if res_t:
						pv_spellings_t.append(res_t)
					else:
						pv_spellings_t.append(None)

			#If we want to build a series of preview grids based on spelling list index (e.g. preview every nth spelling in the lists)
			elif self.tg_by_spelling_pv.val:

				min_l = float(self.nb_min_length_pv.val)
				s_0	  =	None
				while s_0 == None and min_l < float(self.nb_max_length_pv.val):
					s_0 = self.getSpellingByLength(0, spelling_set, min_l)
					min_l += self.pv_increment_metres
				if s_0 == None:
					PupMenu("Requested minimum width out of range.%t")
					return

				min_d = float(self.nb_min_depth_pv.val)
				t_0   = None
				while t_0 == None and min_d < float(self.nb_max_depth_pv.val):
					t_0 = self.getSpellingByLength(1, spelling_set, min_d)
					min_d +=  self.pv_increment_metres
				if t_0 == None:
					PupMenu( "Requested minimum depth out of range.%t")
					return

				s_i0 = spelling_set.spellings_s.index(s_0)
				t_i0 = spelling_set.spellings_t.index(t_0)

				for i in range(s_i0, len(spelling_set.spellings_s), self.pv_increment_spellings):
					sps = spelling_set.spellings_s[i]
					pv_spellings_s.append(sps)
					lengths_s.append((sps.min + sps.max) * 0.5)
					if sps.max > float(self.nb_max_length_pv.val):
						break

				if mode == 'full':
					for i in range(t_i0, len(spelling_set.spellings_t), self.pv_increment_spellings):
						spt = spelling_set.spellings_t[i]
						pv_spellings_t.append(spt)
						lengths_t.append((spt.min + spt.max) * 0.5)
						if spt.max > float(self.nb_max_depth_pv.val):
							break
				else:
					# In the 'compact case, we want the same number of t spellings as s spellings, so that the previews expand simultaneously in both dimensions:
					# Get ALL the t spellings in the requested range:
					all_sp_t = [sp for sp in spelling_set.spellings_t if sp.max >= float(self.nb_min_depth_pv.val) and sp.min <= float(self.nb_max_depth_pv.val)]
					# Get the ratio of numbers of spellings s/t:
					num_sp_t = len(all_sp_t)
					ratio = float(num_sp_t)/len([p for p in pv_spellings_s if p != None])
					# Build a list of t spellings, incrementing at the appropriate index spacing:
					for i in range(len(pv_spellings_s)):
						spt = all_sp_t[int(ratio * i)]
						pv_spellings_t.append(spt)
						lengths_t.append(clamp( float(self.nb_min_depth_pv.val) + (float(self.nb_max_depth_pv.val)-float(self.nb_min_depth_pv.val))*ratio*i, spt.min, spt.max))
			num_pv_spellings_s = len([p for p in pv_spellings_s if p != None])
			total_num_spellings = num_pv_spellings_s * len([p for p in pv_spellings_t if p != None]) if mode == 'full' else num_pv_spellings_s
			retval = PupMenu("I am about to create %d AGB previews, which will occupy an area approximately %dm wide and %dm deep. This may take a few seconds. Proceed?%%t| NO | YES" % \
			(total_num_spellings, int(sum(lengths_s)) + (len(pv_spellings_s)+1) * int(SPACING), int(sum(lengths_t)) + (len(pv_spellings_t)+1) * int(SPACING) ))

			if retval != 2:
				return
			else:
				if total_num_spellings > 300:
					retval2 = PupMenu("Are you sure about this? %d is a LOT of previews!%%t| No, now you've got me worried| YES, just make them already!" % total_num_spellings)
					if retval2 != 2:
						return
			xy = [0.0, 0.0]
			counter = 0
			if mode == 'full':
				for j, sp_t in zip(lengths_t, pv_spellings_t):
					for i, sp_s in zip(lengths_s, pv_spellings_s):
						if sp_s != None and sp_t != None:
							xy = self.temp_grid_ob.createPreviewMeshFromSpellings(gloc, sp_s.spelling, sp_t.spelling, i, j, self.preview_obs, self.tg_pv_annotations.val, self.dupli_obs, source)
							new_x = gloc[0] + xy[0] + SPACING
							gloc = Vector(new_x, gloc[1], gloc[2])
							if not counter % 10:
								self.updateProgress("BUILDING %d OF %d PREVIEWS" % (counter, total_num_spellings))
							counter += 1
					gloc = Vector(xyz[0], gloc[1] + SPACING + xy[1] , 0.0)
			else:
				for i, j, sp_s, sp_t in zip(lengths_s, lengths_t, pv_spellings_s, pv_spellings_t):
					if sp_s != None and sp_t != None:
						xy = self.temp_grid_ob.createPreviewMeshFromSpellings(gloc, sp_s.spelling, sp_t.spelling, i, j, self.preview_obs, self.tg_pv_annotations.val, self.dupli_obs, source)
						new_x = gloc[0] + xy[0] + SPACING
						gloc = Vector(new_x, gloc[1], gloc[2])
						if not counter % 10:
							self.updateProgress("BUILDING %d OF %d PREVIEWS" % (counter, total_num_spellings))
						counter += 1

			scn.objects.selected = []

			Window.RedrawAll()
			self.updateProgress("")
			self.preview_source = source

	def addSpellingsToSet(self):

		if self.AGB_ob and self.AGB_ob.spelling_sets and self.temp_sp_set:
			if self.active_sp_set_index == None:
				PupMenu("There is no spelling set to add the spellings to! Click 'NEW' to create one.%t")
				return

			ax = self.spellings_axis
			ss = self.AGB_ob.spelling_sets[self.active_sp_set_index]
			# Check to see if there are currently any selected preview obs in the scene. If so, check whether they're 'temp' previews or 'active'.
			# If 'temp' (i.e. previews of as yet uncommitted spellings) and any are selected, add a menu option of adding only the spellings of the selected preview obs to the active list.
			scn = bpy.data.scenes.active
			pvobs = [ob for ob in scn.objects.context if ob.name.startswith('XPTAGBPreview')]
			need_selected_option = need_replace_options = False
			if pvobs:
				if getProp(pvobs[-1], 'XPToolsAGBSrc', '') == 'temp':
					need_selected_option = True

			if ss.hasSpellings(ax):
				need_replace_options = True

			axisstring = 'vertical' if ax else 'horizontal'
			submenustring = ""
			if need_replace_options:
				submenustring += ("|Replace all in active %x2|Replace same length in active %x3" )
			submenustring += ("|Append to active %x4|Cancel%x1")

			if not need_selected_option:
				menustring = "Use latest results to:%t"
				selection = PupMenu(menustring + submenustring)
				if selection > 1:
					retval = PupMenu("Sure?%t|NO|YES")
					if retval == 2:
						if selection < 4:
							ss.replaceSpellings(ax, self.temp_sp_set, True if selection == 3 else False)
						elif selection == 4:
							ss.extendSpellings(ax, self.temp_sp_set.spellings_t if ax else self.temp_sp_set.spellings_s)
						ss.encodeEdges()
			elif need_selected_option:
				selection1 = PupMenu("Use:%t|Latest results to...|Latest in selected previews to...|Cancel" )

				if selection1 == 2: 		# selected preview obs
					propname = 'XPToolsAGBSp_S' if ax == 0 else 'XPToolsAGBSp_T'
					tuplist = []
					for ob in pvobs:
						res = getProp(ob, propname, "")
						if res:
							tuplist.append(tuple(map(lambda x: int(x), res.split())))
					ss_sel = AGBSpellingSet()
					for s in (s for s in (self.temp_sp_set.spellings_t if ax else self.temp_sp_set.spellings_s) if s.spelling in tuplist):
						ss_sel.extendSpellings(ax, [s])

					selection = PupMenu(submenustring[1:])
					if selection > 1:
						retval = PupMenu("Sure?%t|NO|YES")
						if retval == 2:
							if selection < 4:
								ss.replaceSpellings(ax, ss_sel, True if selection == 3 else False)
							elif selection == 4:
								ss.extendSpellings(ax, ss_sel.spellings_t if ax else ss_sel.spellings_s)
							ss.encodeEdges()
				elif selection1 == 1:
					selection = PupMenu(submenustring[1:])
					if selection > 1:
						retval = PupMenu("Sure?%t|NO|YES")
						if retval == 2:
							if selection < 4:
								ss.replaceSpellings(ax, self.temp_sp_set, True if selection == 3 else False)
							elif selection == 4:
								ss.extendSpellings(ax, self.temp_sp_set.spellings_t if ax else self.temp_sp_set.spellings_s)
							ss.encodeEdges()

	def filterSpellings(self, mode):
		"""
			We use the user selection of preview AGB grids to denote spellings that they wish to either keep or reject as they develop a spelling set.
		"""
		# Protect those of selected preview grids %%x2| Reject those of selected preview grids %%x3| Reset protected %%x4| Reset rejected %%x5|

		ss = self.AGB_ob.spelling_sets[self.active_sp_set_index]
		ax = self.spellings_axis

		if mode == 4:
			ss.unprotectProtected(ax)

		elif mode == 5:
			ss.resetRejected(ax)

		elif mode in [2, 3]:
			propname = 'XPToolsAGBSp_S' if ax == 0 else 'XPToolsAGBSp_T'
			scn = bpy.data.scenes.active

			strlist = []
			for ob in (ob for ob in scn.objects.context if ob.name.startswith("XPTAGBPreview")):
				strlist.append(getProp(ob, propname, ""))

			if mode == 2:
				ss.defineProtected(ax, strlist)
			elif mode == 3:
				ss.extendRejected(ax, strlist)

	def editMetaFaces(self, evt):

		scn = bpy.data.scenes.active

		if  self.temp_grid_ob == None or self.temp_grid_ob.def_mesh_ob == None:
			PupMenu("Please load an AGB grid mesh into the editor!%t")
			self.panels.closeAny()
			return

		if not self.temp_grid_ob.def_mesh.faceUV:
			PupMenu("Please first assign a texture to the mesh!%t")
			self.panels.closeAny()
			return
		#
		# On opening any of the three face-mode panels, put the mesh into face edit mode:
		#
		if evt in [self.evt_open_pnl_slop, self.evt_open_pnl_antislop, self.evt_open_pnl_ignore]:
			Mesh.Mode(Mesh.SelectModes['FACE'])
			Window.EditMode(1)

		elif evt == self.evt_panel_apply:
			Window.EditMode(0)
			for f in self.temp_grid_ob.def_mesh.faces:
				if f.sel:
					if self.panels.slop:
						f.mode |=  Mesh.FaceModes['TWOSIDE']
					elif self.panels.antislop:
						f.mode &= ~Mesh.FaceModes['DYNAMIC']
					elif self.panels.ignore:
						f.mode |=  Mesh.FaceModes['INVISIBLE']

			# Update our AGBtilelines:
			self.temp_grid_ob.getMeshFaceData()
			Window.EditMode(1)

		elif evt == self.evt_panel_clear:
			Window.EditMode(0)
			self.resetBugs()
			for f in self.temp_grid_ob.def_mesh.faces:
				if self.panels.slop:
					f.mode &=  ~Mesh.FaceModes['TWOSIDE']
				elif self.panels.antislop:
					f.mode |=   Mesh.FaceModes['DYNAMIC']
				elif self.panels.ignore:
					f.mode &=  ~Mesh.FaceModes['INVISIBLE']
			self.temp_grid_ob.getMeshFaceData()
			Window.EditMode(1)

		elif evt == self.evt_panel_show_faces:
			Window.EditMode(0)
			if self.panels.slop:
				for f in self.temp_grid_ob.def_mesh.faces:
					if bool(f.mode&Mesh.FaceModes['TWOSIDE']):
						f.sel = 1
					else:
						f.sel = 0

			elif self.panels.antislop:
				for f in self.temp_grid_ob.def_mesh.faces:
					if bool(f.mode&Mesh.FaceModes['DYNAMIC']):
						f.sel = 0
					else:
						f.sel = 1

			elif self.panels.ignore:
				for f in self.temp_grid_ob.def_mesh.faces:
					if bool(f.mode&Mesh.FaceModes['INVISIBLE']):
						f.sel = 1
					else:
						f.sel = 0
			elif self.panels.match_grp or self.panels.edge_type:
				for f in self.temp_grid_ob.def_mesh.faces:
					f.sel = 0
			Window.EditMode(1)

		elif evt == self.evt_panel_cancel:
			Window.EditMode(0)

	def switchGridObs(self, gridIndex, direct=False):

		clear_spellings = False
		if GUIpage == 13 and direct == True:	# page AGB 2 (spellings)
			if self.AGB_ob and self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
				if self.AGB_ob.spelling_sets[self.active_sp_set_index].hasSpellings():
					if not self.temp_grid_ob.sameCardinality(self.AGB_ob.grid_obs[self.mn_active_grid_index.val]):
						retval = PupMenu("CAUTION: The current and requested grids have differing row or column counts. Switching will clear the current spelling set's spellings. Proceed anyway?%t|NO|YES")
						if retval != 2:
							self.mn_active_grid_index.val = self.active_grid_index
							return
						else:
							clear_spellings = True

		scn = bpy.data.scenes.active
		if Window.EditMode():
			Window.EditMode(0)
		if len(self.AGB_ob.grid_obs) > gridIndex:
			if self.AGB_ob.grid_obs[gridIndex].def_mesh_ob == None or self.AGB_ob.grid_obs[gridIndex].def_mesh_ob not in scn.objects:
				PupMenu("ERROR: The requested subgrid doesn't appear to have a mesh object.")
				self.mn_active_grid_index.val = self.active_grid_index
				self.refreshGUI(self.evt_active_grid_index)
				return
			self.active_grid_index = gridIndex
			self.temp_grid_ob = self.AGB_ob.grid_obs[gridIndex]
			scn.objects.selected = []
			if self.temp_grid_ob.wrapper != None:
				self.temp_grid_ob.wrapper.sel = 1
			if self.temp_grid_ob.def_mesh_ob != None  and self.temp_grid_ob.def_mesh_ob in scn.objects:
				self.temp_grid_ob.def_mesh_ob.sel = 1
				scn.objects.active = self.temp_grid_ob.def_mesh_ob

			if GUIpage == 3: # AGB 1
				if self.panels.edge_type:
					self.temp_grid_ob.selectEdgeGroup(self.tg_road_facing.val)
					self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'edge_type', 0 if self.tg_road_facing.val else 1 )
				else:
					self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5],  self.panels.getMode())

			elif GUIpage == 13 and direct == True: # AGB 2
				if self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
					ss = self.AGB_ob.spelling_sets[self.active_sp_set_index]
					ss.grid_name = self.temp_grid_ob.name
					try:
						setIDProp(self.AGB_ob.wrapper, 'XPToolsIDProps/AGB/SpellingSets/' + ss.comment + '/grid_name', ss.grid_name)
						setIDProp(self.AGB_ob.wrapper, 'XPToolsIDProps/AGB/SpellingSets/' + ss.comment + '/constraints/grid_name', ss.grid_name)
					except StandardError, detail:
						print "ERROR in switchGrids():", detail
					if clear_spellings:
						self.AGB_ob.spelling_sets[self.active_sp_set_index].clearSpellings()
						if self.temp_sp_set:
							self.temp_sp_set.clearSpellings()
		else:
			return
		self.refreshGUI(self.evt_active_grid_index)

	def refreshGUI(self, evt=None):
		"""
			Called whenever the GUI values need updating
		"""
		if self.temp_grid_ob != None and self.temp_grid_ob.def_mesh_ob != None:
			if evt in [self.evt_active_grid_index, self.evt_load_AGB, self.evt_analyse_active, self.evt_delete_grid]:

				self.st_mesh_ob_name.val = self.temp_grid_ob.def_mesh_ob.name
				self.st_new_grid_name.val = self.temp_grid_ob.name
				if self.active_grid_index != None:
					self.mn_active_grid_index.val = self.active_grid_index
				else:
					self.mn_active_grid_index.val = 0

			if evt == self.evt_open_pnl_slop or  self.panels.slop:
				all_vals = self.temp_grid_ob.overlap_h +  self.temp_grid_ob.overlap_v
				for o, b in zip(all_vals, self.st_panel_data_bugs):
 					b.val = niceFround(o,4)
				if  evt == self.evt_open_pnl_slop:
					Window.EditMode(1)

			elif  evt == self.evt_open_pnl_antislop or self.panels.antislop:
				all_vals = self.temp_grid_ob.antislop_h +  self.temp_grid_ob.antislop_v
				for o, b in zip(all_vals, self.st_panel_data_bugs):
					b.val = niceFround(o,4)
				if  evt == self.evt_open_pnl_antislop:
					Window.EditMode(1)

			elif evt == self.evt_open_pnl_match_grp or self.panels.match_grp:
				self.buildMatchGroupMenuString('h')
				self.buildMatchGroupMenuString('v')
				if evt == self.evt_open_pnl_match_grp:
					Window.EditMode(1)

			elif evt == self.evt_open_pnl_edge_type or self.panels.edge_type:
				if evt == self.evt_open_pnl_edge_type:
					Window.EditMode(1)

		if evt in [206, self.evt_sp_set_act]:

			# Copy spelling set var values to the GUI fields:
			if self.AGB_ob and self.AGB_ob.spelling_sets  :
				try:
					if self.active_sp_set_index != None:
						ss = self.AGB_ob.spelling_sets[self.active_sp_set_index]
						self.nb_sp_set_height_min.val	= ss.height_min	if ss.height_min != None else 0
						self.nb_sp_set_height_max.val	= ss.height_max	if ss.height_max != None else 999
						self.nb_sp_set_frequency.val	= ss.frequency	if ss.frequency	!= None else 5
						grid_names = [g.name for g in self.AGB_ob.grid_obs]
						if ss.grid_name	in grid_names:
							self.active_grid_index = grid_names.index(ss.grid_name)
							self.mn_active_grid_index.val = self.active_grid_index
							self.switchGridObs(self.active_grid_index)
						self.tg_override_params.val		= ss.override_params
				except IndexError, detail:
					print "ERROR: %r" % self.active_sp_set_index, detail
		if evt in [201, 206]:
			self.nb_wrapper_layer.val = DEF_TILE_WRAPPER_LAYER
			self.nb_mesh_layer.val 	  = DEF_TILE_MESH_LAYER
			self.nb_meta_layer.val    = DEF_METARECT_LAYER

		Draw()

	def setTileRunData(self, index):
		"""
			Called from buttonEvent() when a tile buttoon is shift-clicked
			Preceding or following runs of tiles are set by entering comma-separated lists of space-separated tiles, e.g. 1 2 3, 4 5 6, 7 8 9
		"""

		tilelines = self.temp_grid_ob.rows if self.spellings_axis else self.temp_grid_ob.columns
		if self.spellings_axis:
			index -= self.temp_grid_ob.num_columns
		if 0 <= index < len(tilelines):
			only_after  = Create(intlists2string(tilelines[index].only_after))
			only_before = Create(intlists2string(tilelines[index].only_before))


			self.st_only_after.val  = intlists2string(tilelines[index].only_after)
			self.st_only_before.val = intlists2string(tilelines[index].only_before)

			self.close_pup = 0
			if tilelines[index].dupes_min == None:
				self.dupes_min_on = False
				self.nb_dupes_min.val = 0
			else:
				self.dupes_min_on = True
				self.nb_dupes_min.val = tilelines[index].dupes_min

			if tilelines[index].dupes_max == None:
				self.dupes_max_on = False
				self.nb_dupes_max.val = 0
			else:
				self.dupes_max_on = True
				self.nb_dupes_max.val = tilelines[index].dupes_max

			self.screen_size = Window.GetScreenSize()
			self.pup_xy = Window.GetMouseCoords()

			panic = 0
			while panic < 10 and not self.close_pup:

				UIBlock(self.drawPup, 0)
				panic += 1

			if self.close_pup == 1:  # Apply settings

				oa = str2intlists(self.st_only_after.val)
				if all((0 <= v < len(tilelines) for lst in oa for v in lst)):
					tilelines[index].only_after = oa
				ob = str2intlists(self.st_only_before.val)
				if all((0 <= v < len(tilelines) for lst in ob for v in lst)):
					tilelines[index].only_before = ob

				if self.dupes_min_on:
					upper_limit = self.nb_dupes_max.val if self.dupes_max_on else 9
					tilelines[index].dupes_min = min(self.nb_dupes_min.val, upper_limit)
				else:
					tilelines[index].dupes_min = None

				if self.dupes_max_on:
					lower_limit = self.nb_dupes_min.val if self.dupes_min_on else 0
					tilelines[index].dupes_max = max(self.nb_dupes_max.val, lower_limit)
				else:
					tilelines[index].dupes_max = None

			self.close_pup == 0

			print tilelines[index].only_after
			print tilelines[index].only_before
			print tilelines[index].dupes_min
			print tilelines[index].dupes_max

	def cb_close_pup(self, event, value):
		self.close_pup = 1 # Close, apply settings
	def cb_dupes_max_on_off(self, event, value):
		self.dupes_max_on = not self.dupes_max_on
	def cb_dupes_min_on_off(self, event, value):
		self.dupes_min_on = not self.dupes_min_on
	def cb_reset_pup_vals(self, event, value):
		self.nb_dupes_max.val 	= 0
		self.nb_dupes_min.val 	= 0
		self.dupes_min_on	  	= False
		self.dupes_max_on	  	= False
		self.st_only_before.val	= ""
		self.st_only_after.val	= ""
	def cb_cancel_pup(self, event, value):
		self.close_pup = -1	# Cancel, don't apply settings
	def drawPup(self):
		st_width = max(GetStringWidth(self.st_only_after.val,'normal'), GetStringWidth(self.st_only_before.val,'normal'))
		bw	= 100
		stw = max(bw+50, st_width + 60)
		h   = 58
		pup_width = stw+bw+17+66
		x0  = min(self.pup_xy[0] - pup_width/2,  self.screen_size[0] - pup_width)
		xbt = x0 + stw/2 - bw/2
		y0  = max(self.pup_xy[1] - h - 10, 5)
		wdmin  = wdmax = bw/2 - 16
		axis_str = 'row' if self.spellings_axis else 'column'
		Label(									 '^',														self.pup_xy[0]-7, y0+60, 20, 10)
		Label(									 "Only place this %s..." % axis_str,	 					x0+bw+15,	y0+48,  150,	16)
		self.st_only_before   		= 	  String("before: ",					9998,						x0+bw+17,	y0+22,	stw,	22, 	self.st_only_before.val,	100,		"Groups of tile numbers, one of which MUST come before the current tile e.g: '2', '2 3' , '5 5, 3 4 5'")
		self.st_only_after   		= 	  String("after   : ",					9999,						x0+bw+17,	y0,		stw,	22, 	self.st_only_after.val,		100,		"Groups of tile numbers, one of which MUST follow the current tile e.g: '4', '1 2 3'")
		Label(									 "Duplicates",												x0+24 ,		y0+48,	100,	16)
		Label(									 "Max",														x0+bw-33,	y0+22,	30,		22)
		Label(									 "Min",														x0+bw-31,	y0,		30,		22)
		if self.dupes_max_on:
			dmax_name = "Disable"
			self.nb_dupes_max	 	= 	  Number("",							9996,						x0+bw,		y0+24,	12,		18,		self.nb_dupes_max.val, 		0,	9,		"Set the maximum number of times that this tile may be duplicated per spelling. Overrides the global setting.")
		else:
			dmax_name = "Enable"

		if self.dupes_min_on:
			dmin_name = "Disable"
			self.nb_dupes_min	 	=	  Number("",							9997,						x0+bw,		y0+2,		12,		18,		self.nb_dupes_min.val, 		0,	9,		"Set the minimum number of times that this tile may be duplicated per spelling. Overrides the global setting.")
		else:
			dmin_name = "Enable"

		self.bt_dupes_max_on_off	= PushButton(dmax_name,						9995,						x0,			y0+22,	bw-34,	22,												"",  self.cb_dupes_max_on_off)
		self.bt_dupes_min_on_off	= PushButton(dmin_name,						9994,						x0,			y0,		bw-34,	22,												"",	 self.cb_dupes_min_on_off)
		self.bt_reset_pup_vals		= PushButton('RESET',						9993,					    x0+pup_width-60,	y0+44,	60, 	22, 									"Reset all values to default settings.",  self.cb_reset_pup_vals)
		self.bt_cancel_pup    		= PushButton('CANCEL',						9991,					    x0+pup_width-60,	y0+22,	60, 	22, 									"Close pop up, do not apply changes.",  self.cb_cancel_pup)
		self.bt_close_pup    		= PushButton('APPLY',						9992,					    x0+pup_width-60,	y0,		60, 	22, 									"Close pop up, apply changes",  		self.cb_close_pup)

	def refreshConstraints(self, evt):

		if not self.temp_grid_ob:
			return
		tilelines =  self.temp_grid_ob.rows if self.spellings_axis else self.temp_grid_ob.columns
		numh   =	self.temp_grid_ob.num_columns
		modes  = ['exclude', 'require', 'center', 'xor']
		modes2 = ['only_after', 'only_before']
		modes3 = ['dupes_min', 'dupes_max']

		def get_data(param):
			for i, t in enumerate(tilelines):
				self.ButtonNumber_val[str(i + numh * self.spellings_axis)] = int(getattr(t, param)) if not t.ignore else -1

		def set_data(param):
			for i, t in enumerate(tilelines):
				val = True if self.ButtonNumber_val[str(i + numh * self.spellings_axis)] > 0 else False
				setattr(t, param, val)

		if evt in [206, self.evt_axis_vertical, self.evt_axis_horizontal]:
			get_data(modes[self.constraints_mode])
			self.updateCommonConstraintButtons()

		elif evt == self.evt_sp_exclude:
			get_data('exclude')

		elif evt == self.evt_sp_require:
			get_data('require')

		elif evt == self.evt_sp_center:
			get_data('center')

		elif evt == self.evt_sp_xor:
			get_data('xor')

		elif evt == self.evt_sp_clr_constraints:
			shift = True if g_qualifs & Window.Qual.SHIFT else False
			for t in tilelines:
				if not shift:
					setattr(t, modes[self.constraints_mode], False)
				else:
					for m in modes:
						setattr(t, m, False)
					for m in modes2:
						setattr(t, m, [])
					for m in modes3:
						setattr(t, m, None)
			if not shift:
				get_data(modes[self.constraints_mode])
			else:
				for m in modes:
					get_data(m)

		tt = "row" if self.spellings_axis else "column"
		evt0 = self.evt_ButtonNumber[str(numh * self.spellings_axis)]	# This gives us the event int of the first ButtonNumber button allocated to the current axis.
		for i in range(len(tilelines)):
			if evt == evt0+i:
				m_now = modes[self.constraints_mode]
				val   = self.ButtonNumber_val[str(i + numh * self.spellings_axis)]
				if val > 0 and m_now in ['require', 'center'] and tilelines[i].exclude == True:
					self.ButtonNumber_val[str(i + numh * self.spellings_axis)] = 0
					PupMenu("The selected constraint cannot be applied to an EXCLUDED %s.%%t" % tt)
				elif val > 0 and m_now == 'exclude' and tilelines[i].require == True:
					self.ButtonNumber_val[str(i + numh * self.spellings_axis)] = 0
					PupMenu("This %s is currently 'required' - you can't 'exclude' a 'required' %s.%%t" % (tt, tt))

				elif m_now == 'require' and tilelines[i].xor:
					for t in (t for t in tilelines if t.xor):
						t.require = bool(val)

				elif m_now == 'xor':
					tilelines[i].xor = bool(val)
					for t in (t for t in tilelines if t.xor):
							t.require = tilelines[i].require
				else:
					setattr(tilelines[i], m_now, val)

	#	elapsed("bp 1 refreshConstraints")
		if self.show_constraints:
			self.temp_grid_ob.selectConstraintFaces(self.spellings_axis, self.constraints_mode)
	#	elapsed("bp 2 refreshConstraints")
		Redraw()

	def resetBugs(self):
		for b in self.st_panel_data_bugs:
			b.val = ""

	def querySaveAGB(self):

		if self.AGB_ob != None:
			if self.active_sp_set_index != None:
				self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
			if self.unsaved_changes == True:
				if self.AGB_ob.wrapper:
					real_name = self.AGB_ob.name[3:] + ".agb"
				retval = PupMenu("There are unsaved changes to %s. Save them now? (Note: the .blend file must also be saved).%%t| NO | YES" % real_name)
				if retval == 2:
					self.AGB_ob.saveAGB()
					self.unsaved_changes = False

	def exitEditor(self):

		self.querySaveAGB()
		self.deleteQuitText()

	def eventHandler(self, evt, val):
		global g_qualifs
# 		if g_qualifs & Window.Qual.SHIFT and g_qualifs & Window.Qual.CTRL and evt in [LEFTMOUSE, MIDDLEMOUSE, RIGHTMOUSE] and val:
# 			mouse_xy = Window.GetMouseCoords()
# 			PupMenu("%d, %d" % (mouse_xy[0], mouse_xy[1]))
# 			Draw()
#		if evt in [LEFTSHIFTKEY, RIGHTSHIFTKEY,	LEFTCTRLKEY, RIGHTCTRLKEY]:
# #		print "got event!"
#			g_qualifs = Window.GetKeyQualifiers()

		if evt == LEFTMOUSE:
			self.resetProgress()

		if not self.show_histograms:
			return

		elif evt == PAGEUPKEY and val:
			self.scroll_offset = min(max(0, self.scroll_offset - 4), self.scroll_limit)
			Draw()
		elif evt == PAGEDOWNKEY and val:
			self.scroll_offset = min(self.scroll_offset + 4 , self.scroll_limit)
			Draw()
		elif evt == HOMEKEY and val:
			self.scroll_offset = 0
			Draw()
		elif evt == WHEELUPMOUSE and val:
			self.scroll_offset = min(max(0, self.scroll_offset - 1), self.scroll_limit)
			Draw()
		elif evt == WHEELDOWNMOUSE and val:
			self.scroll_offset =  min(self.scroll_offset + 1 , self.scroll_limit)
			Draw()

	def buttonEvent(self, evt):

		global g_qualifs
		global GUIpage, GUIchanged, CL_ANN_RED, CL_ANN_GREEN, CL_ANN_AMBER, COL_ARMED, CL4_HISTO_GREEN, CL4_HISTO_ACT, CL4_HISTO_RED, CL_ROAD_BKGND, CL4_HISTO_TEMP
		global DEF_TILE_MESH_LAYER, DEF_TILE_WRAPPER_LAYER, DEF_METARECT_LAYER

		# if 	 evt == 201:	# AGB page 1
# 			pass
		if evt in range(200, 206):
			if GUIpage == 3:
				print "self.active_sp_set_index", self.active_sp_set_index
				if self.AGB_ob != None and self.active_sp_set_index != None:
					self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
					print "saved "
			self.show_histograms = False
		elif evt == 206:	# AGB page 2
			print "self.active_sp_set_index", self.active_sp_set_index
			if self.AGB_ob != None and self.active_sp_set_index != None:
				self.AGB_ob.loadActiveConstraintsIDProps(self.active_sp_set_index)
				print "loaded"
			self.refreshConstraints(evt)
			self.refreshGUI(evt)
		elif evt == self.evt_load_AGB:
			self.resetBugs()
			self.panels.closeAny()
			self.loadAGB()
			self.refreshGUI(evt)
		#	self.buttonEvent(self.evt_open_pnl_slop)
		elif evt == self.evt_save_AGB:
			if self.AGB_ob != None:
				if self.active_sp_set_index != None:
					self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
				self.AGB_ob.saveAGB()
				self.unsaved_changes = False
		elif evt == self.evt_new_AGB:
			self.createNewAGB()
		elif evt == self.evt_duplicate_AGB:
			self.duplicateAGB()
		elif evt == self.evt_AGB_name:
			if self.AGB_ob != None:
				self.st_AGB_name.val = self.AGB_ob.renameAGB(self.st_AGB_name.val)
				Draw()
		elif evt == self.evt_new_grid_name:
			self.renameGrid()
			self.refreshGUI(evt)
		elif evt == self.evt_mesh_ob_name:
			if self.temp_grid_ob != None and self.temp_grid_ob.def_mesh_ob != None:
				self.temp_grid_ob.def_mesh_ob.name = self.st_mesh_ob_name.val
		elif evt == self.evt_add_grid_to_AGB:
			self.addTempGridToAGB()
			self.refreshGUI(evt)
		elif evt == self.evt_duplicate_grid:
			self.duplicateGridOb()
		elif evt == self.evt_delete_grid:
			self.deleteGridOb()
		elif evt == self.evt_active_grid_index:
			if self.mn_active_grid_index.val != None:
				flag = True if GUIpage == 13 else False
				self.switchGridObs(self.mn_active_grid_index.val, direct=flag)
# 			self.refreshGUI(evt)
		elif evt == self.evt_AGB_depth_range and self.AGB_ob:
			self.AGB_ob.depth_menu_index = self.mn_AGB_depth_range.val
			self.AGB_ob.depth_min = AGBObject.depth_buckets[str(self.AGB_ob.depth_menu_index)][0]
			self.AGB_ob.depth_max = AGBObject.depth_buckets[str(self.AGB_ob.depth_menu_index)][1]
		elif evt == self.evt_AGB_height_bucket and self.AGB_ob:
			self.AGB_ob.height_menu_index = self.mn_AGB_height_bucket.val
			if self.AGB_ob.height_menu_index == 1:
				self.AGB_ob.height_max = 999
				self.AGB_ob.height_min = 0
			elif self.AGB_ob.height_menu_index > 1:
				self.AGB_ob.height_max = AGBObject.height_buckets[str(self.AGB_ob.height_menu_index)]
				self.AGB_ob.height_min = AGBObject.height_buckets[str(self.AGB_ob.height_menu_index - 1)]
		elif evt == self.evt_AGB_full_half:
			if self.AGB_ob != None:
				self.AGB_ob.half_block = self.tg_AGB_full_half.val

		# Texture settings panel
		elif evt == self.evt_texscale_x:
			pass
		elif evt == self.evt_texscale_y:
			pass
		elif evt == self.evt_tex_image:
			if self.mn_tex_image.val == 998:
				self.buildGridImageMenu()
				self.mn_tex_image.val = 0
		elif evt == self.evt_get_tex_image:
			self.getActiveImage()
		elif evt == self.evt_open_tex_panel:
			self.tex_panel = True
		elif evt == self.evt_exit_tex_panel:
			self.tex_panel = False
		elif evt == self.evt_set_tex_data:
			if self.temp_grid_ob:
				self.temp_grid_ob.setTextureData(self.all_image_names[self.mn_tex_image.val], self.nb_texscale_x.val, self.nb_texscale_y.val, self.nb_texorig_x.val, self.nb_texorig_y.val)
				self.analyseNewGrid(calledFromTexPanel=True)
			self.tex_panel = False

		# Page 1 sub-panels
		if self.temp_grid_ob != None:

			if evt   == self.evt_open_pnl_slop:
				self.panels.setMode('slop')
				self.resetBugs()
				self.editMetaFaces(self.evt_open_pnl_slop)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'slop')
				self.editMetaFaces(self.evt_panel_show_faces)

			elif evt == self.evt_open_pnl_antislop:
				self.panels.setMode('antislop')
				self.resetBugs()
				self.editMetaFaces(self.evt_open_pnl_antislop)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'antislop')
				self.editMetaFaces(self.evt_panel_show_faces)

			elif evt == self.evt_open_pnl_ignore:
				self.panels.setMode('ignore')
				self.resetBugs()
				self.editMetaFaces(self.evt_open_pnl_ignore)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'ignore')
				self.editMetaFaces(self.evt_panel_show_faces)

			elif evt == self.evt_open_pnl_edge_type:
				self.panels.setMode('edge_type')
				self.temp_grid_ob.setEdgeGroups(self.tg_road_facing.val, 'evt_open_pnl_edge_type')
				self.temp_grid_ob.selectEdgeGroup(self.tg_road_facing.val)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'edge_type', 0 if self.tg_road_facing.val else 1 )

			elif evt == self.evt_open_pnl_match_grp:
				self.panels.setMode('match_grp')
				self.temp_grid_ob.setMatchGroups('evt_open_pnl_match_grp')
				#self.editMetaFaces(self.evt_panel_show_faces)
				self.temp_grid_ob.selectMatchGroup('h', self.mn_match_groups_h.val)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'match_grp')

			elif evt == self.evt_match_groups_h:
				self.temp_grid_ob.selectMatchGroup('h', self.mn_match_groups_h.val)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'match_grp')

			elif evt == self.evt_match_groups_v:
				self.temp_grid_ob.selectMatchGroup('v', self.mn_match_groups_v.val)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'match_grp')

			elif evt == self.evt_add_match_group_h:
				index = self.temp_grid_ob.setMatchGroups('evt_add_match_group_h', len(self.temp_grid_ob.match_groups_h))
				self.buildMatchGroupMenuString('h')
				if index != None:
					self.mn_match_groups_h.val = index
					self.temp_grid_ob.selectMatchGroup('h', index)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'match_grp')

			elif evt == self.evt_add_match_group_v:
				index = self.temp_grid_ob.setMatchGroups('evt_add_match_group_v', len(self.temp_grid_ob.match_groups_v))
				self.buildMatchGroupMenuString('v')
				if index != None:
					self.mn_match_groups_v.val = index
					self.temp_grid_ob.selectMatchGroup('v', index)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'match_grp')

			elif evt == self.evt_del_match_group_h:
				i = self.mn_match_groups_h.val
				ret = self.temp_grid_ob.deleteMatchGroup('evt_del_match_group_h', i)
				self.buildMatchGroupMenuString('h')
				if ret != None:
					self.mn_match_groups_h.val = ret
					self.temp_grid_ob.selectMatchGroup('h', ret)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'match_grp')

			elif evt == self.evt_del_match_group_v:
				i = self.mn_match_groups_v.val
				ret = self.temp_grid_ob.deleteMatchGroup('evt_del_match_group_v', i)
				self.buildMatchGroupMenuString('v')
				if ret != None:
					self.mn_match_groups_v.val = ret
					self.temp_grid_ob.selectMatchGroup('v', ret)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'match_grp')

			elif evt == self.evt_add_edge_group:
				self.temp_grid_ob.setEdgeGroups(self.tg_road_facing.val, 'evt_add_edge_group')
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'edge_type', 0 if self.tg_road_facing.val else 1 )

			elif evt == self.evt_del_edge_group:
				self.temp_grid_ob.clearEdgeGroup(self.tg_road_facing.val)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'edge_type', 0 if self.tg_road_facing.val else 1)

			elif evt == self.evt_road_facing:
				if self.tg_non_road_facing.val == False:  # i.e. we're True
					self.tg_road_facing.val = True
				self.tg_non_road_facing.val = not self.tg_road_facing.val
				self.temp_grid_ob.selectEdgeGroup(self.tg_road_facing.val)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'edge_type', 0 if self.tg_road_facing.val else 1)

			elif evt == self.evt_non_road_facing:
				if self.tg_road_facing.val == False:  # i.e. we're True
					self.tg_non_road_facing.val = True
				self.tg_road_facing.val = not self.tg_non_road_facing.val
				self.temp_grid_ob.selectEdgeGroup(self.tg_road_facing.val)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], 'edge_type', 0 if self.tg_road_facing.val else 1)

			elif evt == self.evt_panel_apply:
				self.editMetaFaces(self.evt_panel_apply)
				# Update the Tilelines:
				if self.panels.slop or self.panels.antislop or self.panels.ignore:
					self.temp_grid_ob.getMeshFaceData()
					self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], self.panels.getMode())

			elif evt == self.evt_panel_show_faces:
				self.editMetaFaces(self.evt_panel_show_faces)

			elif evt == self.evt_panel_clear:
				self.resetBugs()
				self.editMetaFaces(self.evt_panel_clear)
				self.temp_grid_ob.setGUITileData(self.draw_image_data[1], self.draw_image_data[2], self.draw_grid_size[0], self.draw_grid_size[1], self.GUIimageWindow[4], self.GUIimageWindow[5], self.panels.getMode())

			elif evt == self.evt_panel_cancel:
				self.editMetaFaces(self.evt_panel_cancel)

			elif evt in self.evt_panel_data_bugs:
				bug_index = self.evt_panel_data_bugs.index(evt)
				lenbugx = len(self.temp_grid_ob.GUI_bugs_x)
				if bug_index < lenbugx:
					self.temp_grid_ob.GUI_bugs_x[bug_index][2] = self.st_panel_data_bugs[bug_index].val
				elif  lenbugx <= bug_index < len(self.temp_grid_ob.GUI_bugs_y) + lenbugx:
					self.temp_grid_ob.GUI_bugs_y[bug_index - lenbugx][2] = self.st_panel_data_bugs[bug_index].val
				self.temp_grid_ob.setOverlapValuesFromGUI(self.panels.getMode(), self.evt_panel_data_bugs.index(evt),  self.st_panel_data_bugs[self.evt_panel_data_bugs.index(evt)])

			elif evt == self.evt_bug:

				result = PupFloatInput("min overlap", self.bug_value, 0.0, 100.0, 50, 2)
				if result != None:
					self.bug_value = result

			self.refreshGUI(evt)

		# Layer fields
		if   evt == self.evt_wrapper_layer:
			DEF_TILE_WRAPPER_LAYER = self.nb_wrapper_layer.val
		elif evt == self.evt_mesh_layer:
			DEF_TILE_MESH_LAYER = self.nb_mesh_layer.val
		elif evt == self.evt_meta_layer:
			DEF_METARECT_LAYER = self.nb_meta_layer.val
		elif evt == self.evt_analyse_active:
			self.resetBugs()
			self.analyseNewGrid()
			self.refreshGUI(evt)
			Window.RedrawAll()
		elif evt == self.evt_build_grid:
			pass
		elif evt == self.evt_wrap_grid:
			pass
		elif evt == self.evt_sp_set_previous:
			if self.AGB_ob and self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
				self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
				if self.active_sp_set_index > 0:
					self.active_sp_set_index -= 1
				self.mn_sp_set_act.val = self.active_sp_set_index + 1
				self.refreshGUI(self.evt_sp_set_act)
				self.AGB_ob.loadActiveConstraintsIDProps(self.active_sp_set_index)
		elif evt == self.evt_sp_set_next:
			if self.AGB_ob and self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
				self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
				if self.active_sp_set_index + 1 < len(self.AGB_ob.spelling_sets):
					self.active_sp_set_index += 1
				self.mn_sp_set_act.val = self.active_sp_set_index + 1
				self.refreshGUI(self.evt_sp_set_act)
				self.AGB_ob.loadActiveConstraintsIDProps(self.active_sp_set_index)
		elif evt == self.evt_sp_set_act:
			if self.AGB_ob != None and self.active_sp_set_index != None:
				self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
				if 0 < self.mn_sp_set_act.val <= len(self.AGB_ob.spelling_sets):
					self.active_sp_set_index = self.mn_sp_set_act.val - 1
				self.refreshGUI(evt)
				self.AGB_ob.loadActiveConstraintsIDProps(self.active_sp_set_index)
		elif evt == self.evt_override_params:
			if self.AGB_ob and self.AGB_ob.spelling_sets:
				self.AGB_ob.spelling_sets[self.active_sp_set_index].override_params = self.tg_override_params.val
				self.AGB_ob.spelling_sets[self.active_sp_set_index].encodeEdges()
				self.refreshGUI(self.evt_sp_set_act)
				self.unsaved_changes = True
		elif evt == self.evt_rename_sp_set:
			if self.AGB_ob and self.AGB_ob.spelling_sets and self.active_sp_set_index != None:
				block = []
				newname = Create(self.AGB_ob.spelling_sets[self.active_sp_set_index].comment)
				block.append(("", newname, 0, 32))
				retval = PupBlock("Rename Spelling Set", block)
				if retval == 1:
					allnames = [n.comment for n in self.AGB_ob.spelling_sets if self.AGB_ob.spelling_sets.index(n) != self.active_sp_set_index]
					if newname.val in allnames:
						newname.val = incrementName(newname.val, allnames)
					self.AGB_ob.spelling_sets[self.active_sp_set_index].comment = newname.val
					self.buildSpellingSetMenuString()
					#self.active_sp_set_index = len(self.AGB_ob.spelling_sets) - 1
					self.mn_sp_set_act.val = self.active_sp_set_index + 1
					self.refreshGUI(self.evt_sp_set_act)
				else:
					pass
				self.unsaved_changes = True
		elif evt == self.evt_new_sp_set:
			self.addSpellingSet()
			self.unsaved_changes = True
		elif evt == self.evt_delete_sp_set:
			self.deleteSpellingSet()
			self.unsaved_changes = True
		elif evt == self.evt_copy_sp_set:
			self.duplicateSpellingSet()
			self.unsaved_changes = True
		elif evt == self.evt_define_heights:
			if self.AGB_ob and self.AGB_ob.spelling_sets:
				 self.AGB_ob.spelling_sets[self.active_sp_set_index].has_height_range = True
		elif evt == self.evt_sp_set_height_min:
			if self.AGB_ob and self.AGB_ob.spelling_sets:
				self.AGB_ob.spelling_sets[self.active_sp_set_index].height_min = self.nb_sp_set_height_min.val
				if self.nb_sp_set_height_min.val == self.AGB_ob.height_min and self.nb_sp_set_height_max.val == self.AGB_ob.height_max:
					self.AGB_ob.spelling_sets[self.active_sp_set_index].has_height_range = False
				else:
					self.AGB_ob.spelling_sets[self.active_sp_set_index].has_height_range = True
					self.AGB_ob.spelling_sets[self.active_sp_set_index].height_max = self.nb_sp_set_height_max.val
				self.unsaved_changes = True
		elif evt == self.evt_sp_set_height_max:
			if self.AGB_ob and self.AGB_ob.spelling_sets:
				self.AGB_ob.spelling_sets[self.active_sp_set_index].height_max = self.nb_sp_set_height_max.val
				if self.nb_sp_set_height_min.val == self.AGB_ob.height_min and self.nb_sp_set_height_max.val == self.AGB_ob.height_max:
					self.AGB_ob.spelling_sets[self.active_sp_set_index].has_height_range = False
				else:
					self.AGB_ob.spelling_sets[self.active_sp_set_index].has_height_range = True
					self.AGB_ob.spelling_sets[self.active_sp_set_index].height_min = self.nb_sp_set_height_min.val
				self.unsaved_changes = True
		elif evt == self.evt_sp_set_frequency:
			if self.AGB_ob and self.AGB_ob.spelling_sets:
				self.AGB_ob.spelling_sets[self.active_sp_set_index].frequency = self.nb_sp_set_frequency.val
			self.unsaved_changes = True
		elif evt == self.evt_make_prev_grids:
			retval = PupMenu("What would you like to preview?%t|Latest generated spellings|Active spelling set")
			if retval == 2:
				source = 'active'
			elif retval == 1:
				source = 'temp'
			else:
				return
			retval2 = PupMenu("Which preview mode would you like to use?%t| Compact - a single row of previews, progressively increasing in both dimensions. | Full - a grid of previews, covering all combinations of widths and depths.")
			if retval2 == 2:
				mode = 'full'
			elif retval2 == 1:
				mode = 'compact'
			else:
				return
			self.makeAGBPreviews(source, mode)
		elif evt == self.evt_clear_prev_grids:
			self.clearAGBPreviews()
			Window.RedrawAll()
		elif evt == self.evt_axis_horizontal:
			self.tg_axis_horizontal.val = True
			self.tg_axis_vertical.val = False
			self.spellings_axis = 0
			self.refreshConstraints(evt)
		elif evt == self.evt_axis_vertical:
			self.tg_axis_vertical.val = True
			self.tg_axis_horizontal.val = False
			self.spellings_axis = 1
			self.refreshConstraints(evt)
		for i in range(len(self.ButtonNumber)):
			if evt == self.evt_ButtonNumber[str(i)]:
				if g_qualifs & Window.Qual.SHIFT:
					self.setTileRunData(i)
					g_qualifs = Window.GetKeyQualifiers()
					Draw()
				else:
					self.ButtonNumber_val[str(i)] = 1 - self.ButtonNumber_val[str(i)]
					self.refreshConstraints(evt)
		if   evt == self.evt_sp_exclude:
			self.constraints_mode = 0
			self.tg_sp_exclude.val = 1
			self.tg_sp_require.val = 0
			self.tg_sp_center.val  = 0
			self.tg_sp_xor.val =	0
			self.refreshConstraints(evt)
		elif evt == self.evt_sp_require:

			self.constraints_mode = 1
			self.tg_sp_require.val = 1
			self.tg_sp_exclude.val = 0
			self.tg_sp_center.val  = 0
			self.tg_sp_xor.val =	0
			self.refreshConstraints(evt)
		elif evt == self.evt_sp_center:

			self.constraints_mode = 2
			self.tg_sp_xor.val	  =	0
			self.tg_sp_center.val = 1
			self.tg_sp_require.val = 0
			self.tg_sp_exclude.val = 0
			self.refreshConstraints(evt)
		elif evt == self.evt_sp_xor:

			self.constraints_mode = 3
			self.tg_sp_xor.val = 1
			self.tg_sp_center.val = 0
			self.tg_sp_require.val = 0
			self.tg_sp_exclude.val = 0
			self.refreshConstraints(evt)
		elif evt == self.evt_toggle_or_xor:
			self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['xor_is_or'] = 1 - self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['xor_is_or']
			Draw()
		elif evt == self.evt_sp_clr_constraints:
			self.refreshConstraints(evt)
		elif evt == self.evt_sp_sel_constraints:
			self.show_constraints = not self.show_constraints
			if self.temp_grid_ob != None:
				if self.show_constraints:
					if not Window.EditMode():
						Window.EditMode(1)
					self.temp_grid_ob.sel = 1
					self.temp_grid_ob.selectConstraintFaces(self.spellings_axis, self.constraints_mode)
				else:
					Window.EditMode(0)

		elif evt == self.evt_min_length:
			self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['min_length'] 		= self.nb_min_length.val
		elif evt == self.evt_max_length:
			self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['max_length'] 		= self.nb_max_length.val
		elif evt == self.evt_max_dupes:
			 self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['max_dupes']		= self.nb_max_dupes.val
		elif evt == self.evt_max_cuts:
			self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['max_cuts']		= self.nb_max_cuts.val
		elif evt == self.evt_max_iterations:
			self.nb_max_iterations.val = 10000 * (self.nb_max_iterations.val/10000)
			self.maxed_out_iterations = False
			self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['max_iterations']	= self.nb_max_iterations.val
		elif evt == self.evt_center_precision:
			self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['center_precision']= self.nb_center_precision.val
		elif evt == self.evt_sp_variety:
			self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['sp_variety']		= self.sl_sp_variety.val

		elif evt == self.evt_make_spellings:
			if self.AGB_ob != None and self.active_sp_set_index != None:
				self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
			# self.saveQuitText()
			self.makeSpellings()
		elif evt == self.evt_add_to_sp_set:
			self.addSpellingsToSet()
			self.unsaved_changes = True
		elif evt == self.evt_filter_spellings:
			if self.AGB_ob and self.AGB_ob.spelling_sets:
				ax = self.spellings_axis
				axisstring = 'vertical' if ax else 'horizontal'
				if self.active_sp_set_index == None:
					PupMenu("There is no spelling set to work on! Click 'NEW' to create one.%t")
					return
				retval = PupMenu("Filter %s spellings in Active Set:%%t|Protect those in selected Previews %%x2|Reset all Protected flags%%x4|%%l|Reject those in selected Previews %%x3|Reset all Rejected flags%%x5|%%l|Cancel %%x1" % axisstring)
				if retval > 1:
					self.filterSpellings(retval)
		elif evt == self.evt_print_spellings:
			retval = PupMenu("What would you like to print?%t|Latest generated spellings|Active Spelling Set")
			if retval == 1:
				if self.temp_sp_set:
					self.printDivider()
					self.temp_sp_set.printSet()
					Window.RedrawAll()
				else:
					print "No temp_sp_set"
			elif retval == 2:
				try:
					self.printDivider()
					self.AGB_ob.spelling_sets[self.active_sp_set_index].printSet()
					Window.RedrawAll()
				except (TypeError, ValueError, AttributeError), detail:
					print "ERROR:", detail
		elif evt == self.evt_clear_temp_text:
			self.clearTempText()
			Window.RedrawAll()
		elif evt == self.evt_show_histograms:
			if self.show_histograms:
				self.show_histograms = False
				GUIpage = 13
				GUIchanged = True
			elif self.AGB_ob:
				self.show_histograms = True
				ret = self.showHistograms()
				if ret == False:
					self.show_histograms = False
					return
				elif ret == True:
					GUIpage = 14
					GUIchanged = True
			Draw()
		elif evt == self.evt_show_temp:
			if self.temp_sp_set and not self.show_temp:
				if self.temp_sp_set.spellings_s or self.temp_sp_set.spellings_t:
					self.show_temp = True
					self.showHistograms()
					Draw()
				else:
					PupMenu("There are no recent results to display.%t|Understood")
			elif self.temp_sp_set and self.show_temp:
				self.show_temp = False
				self.showHistograms()
				Draw()
		elif evt == self.evt_reset_scale:
			self.histo_prefs['custom_ranges'] = False
			self.nb_min_length_pv.val	= self.AGB_ob.depth_min - 5
			self.nb_max_length_pv.val	= self.AGB_ob.depth_max * 2.0 + 5
			self.nb_min_depth_pv.val	= self.AGB_ob.depth_min - 5
			self.nb_max_depth_pv.val	= self.AGB_ob.depth_max + 5
			self.showHistograms()
			Draw()
		elif evt == self.evt_scroll_up:
			self.scroll_offset =  min(self.scroll_offset + 1 , self.scroll_limit)
			Draw()
		elif evt == self.evt_scroll_down:
			self.scroll_offset = min(max(0, self.scroll_offset - 1), self.scroll_limit)
			Draw()
		elif evt in self.evt_sp_set_select:
			self.AGB_ob.saveActiveConstraintsIDProps(self.active_sp_set_index)
			self.active_sp_set_index = self.evt_sp_set_select.index(evt)
			self.mn_sp_set_act.val	= self.active_sp_set_index + 1
			self.refreshGUI(self.evt_sp_set_act)
			self.AGB_ob.loadActiveConstraintsIDProps(self.active_sp_set_index)
			self.show_histograms = False
			GUIpage = 13
			GUIchanged = 1
			Draw()
		elif evt == self.evt_by_spelling_pv:
			self.tg_by_spelling_pv.val  = 1
			self.tg_by_size_pv.val 		= 0
			self.nb_increment_pv.val	= self.pv_increment_spellings
		elif evt == self.evt_by_size_pv:
			self.tg_by_size_pv.val 		= 1
			self.tg_by_spelling_pv.val  = 0
			self.nb_increment_pv.val	= int(self.pv_increment_metres)
		elif evt == self.evt_increment_pv:
			if self.tg_by_spelling_pv.val:
				self.pv_increment_spellings = self.nb_increment_pv.val
			else:
				self.pv_increment_metres = float(self.nb_increment_pv.val)
		elif evt == self.evt_min_length_pv:
			self.histo_prefs['custom_ranges'] = True
			if self.show_histograms:
				self.histo_prefs['custom_ranges'] = True
				self.showHistograms()
				Draw()
		elif evt == self.evt_max_length_pv:
			self.histo_prefs['custom_ranges'] = True
			if self.show_histograms:
				self.histo_prefs['custom_ranges'] = True
				self.showHistograms()
				Draw()
		elif evt == self.evt_min_depth_pv:
			self.histo_prefs['custom_ranges'] = True
			if self.show_histograms:
				self.histo_prefs['custom_ranges'] = True
				self.showHistograms()
				Draw()
		elif evt == self.evt_max_depth_pv:
			self.histo_prefs['custom_ranges'] = True
			if self.show_histograms:
				self.histo_prefs['custom_ranges'] = True
				self.showHistograms()
				Draw()

		elif evt in [self.evt_test_colour, self.evt_test_alpha]:
			if evt == self.evt_test_colour:
				self.test_colour[:3] = list(self.cp_test_colour.val)
			elif evt == self.evt_test_alpha:
				self.test_colour[3] = self.sl_test_alpha.val

	text_tips = {
				"slop_zones":
				"Define slop zone locations by selecting continuous runs of quads - rows for vertical slop, and columns for horizontal slop - and clicking 'add'. Then enter the minimum permitted tile overlap distance for each zone (in metres) in the respective fields.",

				"antislop":
				"Antislop tiles are defined in the same way as slop zones, by selecting continuous rows or columns of quads. They may be added all at once or line by line, as preferred. Then specify the overshoot distances for each antislop zone.",

				"ignore":
				"Any rows or columns of tiles which are not intended to be used in spellings but which are necessary to prevent texture-bleed between adjacent non-matching tiles should be defined here, by selecting continuous rows and/or columns and clicking 'add'.",

				"edge_types":
				"The automagic spelling generator needs to know which rows and columns are suitable for use around the edges of the block. There are two types: 'road-facing' (happy when facing a residential road of some kind) and 'non-road-facing' (happy when facing railway lines, lakes, elevated highways etc).\
				Define road-facing tiles by clicking 'road', selecting a single outer edge in every row or column of road-facing tiles, and clicking 'add'. For non-road-facing tiles, select 'non' and follow the same procedure as before.",

				"match_groups":
				"The spelling generator works by building sequences of tiles whose 'trailing' edges match the preceding tiles' 'leading' edges (in this context, 'trailing' is defined as 'nearest to the grid's origin'). Specify groups of matching tile cuts by selecting at least one mesh edge per cut and clicking 'add'.\
				'HORIZONTAL' is for columns of tiles (horizontal spellings), and 'VERTICAL' is for rows of tiles (vertical spellings). Each cut must be a member of exactly one group. The objective is to end up with no red lines in the grid. "
				}

	def drawAGBEditor(self, page):

#		self.cp_test_colour 			= ColorPicker(							self.evt_test_colour,		825,	60,   	100,	100,	self.cp_test_colour.val,											"")
#		self.sl_test_alpha				= Slider("alpha ",						self.evt_test_alpha,		825,	38,   	100,	22,		self.sl_test_alpha.val,			0.0,	1.0,	1)

		self.GUIimageWindow		=	(																			234,	20,		328,	131,	29, 	13)															# Available grid size and entry field bug size: x, y, w, h, bug_width, bug_height

		if page == 1:

			drawString(									 "TILE GRID GEOMETRY",									105, 	154, 					CL_TEXT_LT,						'center',	'small')
			self.nb_mesh_layer				= Number(	 "TILE MESH LAYER:", 		self.evt_mesh_layer, 	 	10,		10,		190, 	22, 	self.nb_mesh_layer.val, 		1,		20,							"Blender layer for new grid meshes to be located in.")
			self.bt_analyse_active			= PushButton("LOAD ACTIVE MESH",		self.evt_analyse_active,	10, 	126,	190,	22,																			"Analyse the active mesh object as the basis for an AGB tile grid.")
			if not self.gridimagemenuname:
				self.buildGridImageMenu()
			if self.tex_panel:
				self.mn_tex_image			= Menu(	self.gridimagemenuname,			self.evt_tex_image,			10,		102, 	150, 	22,		self.mn_tex_image.val, 												"")
				self.bt_get_tex_image		= PushButton("Get",						self.evt_get_tex_image,		160, 	102, 	40, 	22,																			"Select all mesh objects which use the same texture image as the active object.")
				self.nb_texscale_x			= Number("size x",						self.evt_texscale_x,		10,		80,		95,		22,		self.nb_texscale_x.val,			1.0,	2048.0,						"Texture width, in metres.")
				self.nb_texscale_y			= Number("size y",						self.evt_texscale_y,		105,	80,		95,		22,		self.nb_texscale_y.val,			1.0,	2048.0,						"Texture height, in metres.")
				self.nb_texorig_x			= Number("orgn x",						self.evt_texorig_x,			10,		58,		95,		22,		self.nb_texorig_x.val,			0.0,	2048.0,						"X-axis position of the mesh object's origin, relative to the left of the texture, in metres.")
				self.nb_texorig_y			= Number("orgn y",						self.evt_texorig_y,			105,	58,		95,		22,		self.nb_texorig_y.val,			0.0,	2048.0,						"Y-axis position of the mesh object's origin, relative to the bottom of the texture, in metres.")
				self.bt_set_tex_data		= PushButton("CONFIRM",					self.evt_set_tex_data,		10, 	36,		120,	22,																			"Apply the editor texture image, size and origin settings to the current grid mesh.")
				self.bt_exit_tex_panel		= PushButton("cancel",					self.evt_exit_tex_panel,	130, 	36,		70,		22,																			"Apply the editor texture image, size and origin settings to the current grid mesh.")
			else:
				self.bt_open_tex_panel		= PushButton("texture settings",		self.evt_open_tex_panel,	10, 	82,		190,	22,																			"Apply the editor texture image, size and origin settings to the active grid mesh.")
				self.st_mesh_ob_name		= String("TILE: ",						self.evt_mesh_ob_name,		10,		104,	190,	22,		self.st_mesh_ob_name.val, 		21,									"Set the name of the active grid mesh object. Must have prefix 'TILE'.")
				self.st_new_grid_name		= String("GRID: ",						self.evt_new_grid_name,		10,		60,		190,	22,		self.st_new_grid_name.val, 		21,								 	"Set the name of the active grid wrapper object.")
				self.bt_add_grid_to_AGB		= PushButton("ADD TO AGB",				self.evt_add_grid_to_AGB,	10, 	38,		190,	22,																			"Add the active grid to the currently loaded AGB.")

			if not self.panels.slop:
				self.bt_open_pnl_slop		= PushButton("slop",					self.evt_open_pnl_slop,		210, 	155,	65,		20,																			"Edit slop zones and minimum overlap values.")
			else:
				drawString(							 	"slop",													243, 	161, 					CL_GUI_ORANGE,					'center',	'normal')
			if not self.panels.antislop:
				self.bt_open_pnl_antislop	= PushButton("antislop",				self.evt_open_pnl_antislop,	275, 	155,	70,		20,																			"Edit anti-slop zones and values.")
			else:
				drawString(								 "antislop",											310, 	161, 					CL_GUI_ORANGE,					'center',	'normal')
			if not self.panels.ignore:
				self.bt_open_pnl_ignore		= PushButton("ignored",					self.evt_open_pnl_ignore,	345, 	155,	70,		20,																			"Edit redundant rows and columns.")
			else:
				drawString(								 "ignored",												380, 	161, 					CL_GUI_ORANGE,					'center',	'normal')
			if not self.panels.edge_type:
				self.bt_open_pnl_edge_type	= PushButton("edge types",				self.evt_open_pnl_edge_type,415, 	155,	85,		20,																			"Edit road-facing and non-road-facing block-edge types")
			else:
				drawString(								 "edge types",											458, 	161, 					CL_GUI_ORANGE,					'center',	'normal')
			if not self.panels.match_grp:
				self.bt_open_pnl_match_grp	= PushButton("match groups",			self.evt_open_pnl_match_grp,500, 	155,	95,		20,																			"Edit tile-matching groups for the automatic spelling generator.")
			else:
				drawString(								 "match groups",										548, 	161, 					CL_GUI_ORANGE,					'center',	'normal')
			self.tg_panel_help				= Toggle(	 "?",						self.evt_panel_help,		595,	155,	20,		20,		self.tg_panel_help.val,												"Huh? Just kidding - a few notes and pointers.")

			drawString(									 "AGBLOCK MASTER",										720, 	154, 	CL_TEXT_LT,		'center',	'small')
			drawString(									 "AGB  DEPTH",											657, 	86, 	CL_TEXT_LT,		'center',	'tiny')
			drawString(									 "AGB  HEIGHT",											737, 	86, 	CL_TEXT_LT,		'center',	'tiny')
			drawString(									 "TYPE",												800, 	86, 	CL_TEXT_LT,		'center',	'tiny')
			self.st_AGB_name				= String(	 "AGB: ",					self.evt_AGB_name,			625,	104,	190,	22,		self.st_AGB_name.val, 			21,								 	"")
			self.bt_load_AGB		 		= PushButton("LOAD",					self.evt_load_AGB,			625, 	126, 	47, 	22,																			"Load the selected AGB into the editor. ")
			self.bt_save_AGB		 		= PushButton("SAVE",					self.evt_save_AGB,			672, 	126, 	48, 	22,																			"Overwrite the current AGB text file with the contents of the editor.")
			self.bt_new_AGB		 			= PushButton("NEW",						self.evt_new_AGB,			720, 	126, 	47, 	22,																			"Add a new AGB wrapper at the cursor location and parent the selected grid wrapper(s) to it.")
			self.bt_duplicate_AGB			= PushButton("COPY",					self.evt_duplicate_AGB,		767, 	126, 	48, 	22,																			"Duplicate the entire currently loaded AGB.")
			self.mn_AGB_depth_range			= Menu(	 	 self.AGB_depth_menustring,	self.evt_AGB_depth_range,	625,	60,		80,		22,		self.mn_AGB_depth_range.val,										"Defines the block depth requirement for this AGB")
			self.mn_AGB_height_bucket		= Menu(	 	 self.AGB_height_menustring,self.evt_AGB_height_bucket,	705,	60,		80,		22,		self.mn_AGB_height_bucket.val,										"Defines the block height bucket for this AGB, if any.")
			self.tg_AGB_full_half			= Toggle(	"half", 					self.evt_AGB_full_half, 	785,	60,		30, 	22, 	self.tg_AGB_full_half.val, 											"'half' or 'full' block status for this AGB. Type 'half' is required to address all edge combinations.")
			self.bt_duplicate_grid			= PushButton("COPY GRID",				self.evt_duplicate_grid,	720, 	10, 	95, 	22,																			"Duplicate the active grid.")
			self.bt_delete_grid				= PushButton("DELETE GRID",				self.evt_delete_grid,		625, 	10, 	95, 	22,																			"Delete the active grid.")

			self.mn_active_grid_index		= Menu(		 self.grid_obs_menu_string,	self.evt_active_grid_index,	625,	38,		190,	22,		self.mn_active_grid_index.val,										"Select an AGB subgrid to edit.")

			if not self.panels.anyPanel():

				if self.temp_grid_ob == None:
					drawString(								"NO TILE GRID - NOTHING TO EDIT",					413, 	111, 					CL_TEXT_LT,				'center',	'large')
# 					for i in range(45, 134, 22):
# 						drawString(	'No Grid',																	390, 	i, 		CL_TEXT_LT,															'center',	'normal')
				else:
					drawString(							"SELECT A SUB-TAB TO DEVELOP THE CURRENT GRID MESH",			413, 	111, 					CL_TEXT_LT,				'center',	'large')
#					drawString(							"GO TO AGB 2 TO MAKE SPELLING SETS AND PREVIEWS",		413, 	89, 					CL_GREY95,				'center',	'large')
#					drawString(		"SLOP AND OPTIONALLY ANTISLOP MUST BE DEFINED FOR EVERY AGB",				413, 	67, 					CL_GREY95,				'center',	'small')
#					drawString(		"EDGE TYPES AND MATCH GROUPS ARE FOR AUTO SPELLING GENERATION.",			413, 	45, 					CL_GREY95,				'center',	'small')

				self.nb_meta_layer			= Number("META-RECT LAYER:", 			self.evt_meta_layer,  		415,	10,	 	200, 	22, 	self.nb_meta_layer.val, 		1,		20,							"Blender layer for new AGB meta-rects to be located in.")
				self.nb_wrapper_layer		= Number("AG WRAPPER LAYER:", 			self.evt_wrapper_layer,  	210,	10,	 	195, 	22, 	self.nb_wrapper_layer.val, 		1,		20,							"Blender layer for new AGB wrappers to be located in.")
			else:
				w_g, h_g = self.draw_grid_size
				if not self.tg_panel_help.val:

					BGL.glEnable( BGL.GL_BLEND )
					BGL.glBlendFunc(BGL.GL_SRC_ALPHA, BGL.GL_ONE_MINUS_SRC_ALPHA)
					Blender.Draw.Image(*self.draw_image_data)

					drawRectangle(																				234,	20,		w_g,	h_g,	[0.62, 0.62, 0.62, 0.45])
					for r in self.temp_grid_ob.GUI_chequerboard:
						drawRectangle(																			r[0],	r[1],	r[2],	r[3],	[0.00, 0.00, 0.00, 0.07])

					if   self.panels.slop:
						col	= CL4_SLOP
					elif self.panels.ignore:
						col = CL4_IGNORE
					elif self.panels.antislop:
						col	= CL4_ANTISLOP

					if self.panels.match_grp or self.panels.edge_type:
						for match_type, match_colour in zip([ 'match_group', 'edge_road', 'edge_no_road', 'untagged', 'match_group_active', 'edge_road_active', 'edge_no_road_active'],[CL4_MATCH_GRP2, CL4_EDGE_ROAD2, CL4_EDGE_NOROAD2, CL4_UNTAGGED, CL4_MATCH_GRP, CL4_EDGE_ROAD, CL4_EDGE_NOROAD]):
							for r in self.temp_grid_ob.GUI_rects[match_type]:
								drawRectangle(																	r[0],	r[1],	r[2],	r[3],	match_colour)
					else:
						for r in self.temp_grid_ob.GUI_zone_rects:
							drawRectangle(																		r[0],	r[1],	r[2],	r[3],	col)

					BGL.glDisable( BGL.GL_BLEND )

					for n in self.temp_grid_ob.GUI_label_locs_x:
						drawString(	str(n[2]),																	n[0],	10-n[1]*7 , 			CL_WHITE,							'center',	'tiny')
					for n in self.temp_grid_ob.GUI_label_locs_y:
						drawString(	str(n[2]),																	224-n[1]*7, n[0], 				CL_WHITE,							'center',	'tiny')
					if self.panels.getMode() in ['slop', 'antislop', 'ignore']:
						self.bt_panel_apply		= PushButton("add",			self.evt_panel_apply,				569, 	104,	46,		22,																			"Make the selected grid quads into '%s' tiles." % self.panels.getMode())
						self.bt_panel_clear		= PushButton("clear",		self.evt_panel_clear,				569, 	82,		46,		22,																			"Clear all '%s' tiles from the grid." % self.panels.getMode())
				else:
					drawRectangle(																				210,	10,		405,	144,	CL_MACRO)

				if self.panels.slop:
					if self.tg_panel_help.val:
						drawParagraph(	AGBEditor.text_tips['slop_zones'],										413,	134,	395,	16,		CL_TEXT_MD,							'center',	'normal')
					else:
						i = 0
						while  i < len(self.temp_grid_ob.GUI_bugs_x) < 8:
							h = self.temp_grid_ob.GUI_bugs_x[i]
							self.st_panel_data_bugs[i].val = h[2]
							self.st_panel_data_bugs[i]	= String("",		self.evt_panel_data_bugs[i],		h[1]-15, 4,		27,		12,		self.st_panel_data_bugs[i].val, 	8,						 		"Enter the minimum overlap value for this slop zone, in metres.")
							i += 1
						j = 0
						while j < len(self.temp_grid_ob.GUI_bugs_y) and i < 8:
							v = self.temp_grid_ob.GUI_bugs_y[j]
							self.st_panel_data_bugs[i].val = v[2]
							self.st_panel_data_bugs[i]	= String("",		self.evt_panel_data_bugs[i],		204,	v[1]-8,	27,		12,		self.st_panel_data_bugs[i].val, 	8,							 	"Enter the minimum overlap value for this slop zone, in metres.")
							i += 1
							j += 1

				elif self.panels.antislop:
					if self.tg_panel_help.val:
						drawParagraph(	AGBEditor.text_tips['antislop'],										413,	134,	395,	16,		CL_TEXT_MD,							'center',	'normal')
					else:
						i = 0
						while  i < len(self.temp_grid_ob.GUI_bugs_x) < 8:
							h = self.temp_grid_ob.GUI_bugs_x[i]
							self.st_panel_data_bugs[i].val = h[2]
							self.st_panel_data_bugs[i]	= String("",		self.evt_panel_data_bugs[i],		h[1]-15,  4,	27,		12,		self.st_panel_data_bugs[i].val,  	8,			 					"Enter the anti-slop overlap for this column of tiles, in metres.")
							i += 1
						j = 0
						while j < len(self.temp_grid_ob.GUI_bugs_y) and i < 8:
							v = self.temp_grid_ob.GUI_bugs_y[j]
							self.st_panel_data_bugs[i].val = v[2]
							self.st_panel_data_bugs[i]	= String("",		self.evt_panel_data_bugs[i],		203,	v[1]-8,	27,		12,		self.st_panel_data_bugs[i].val,  	8,							 	"Enter the anti-slop overlap for this row of tiles, in metres.")
							i += 1
							j += 1

				elif self.panels.ignore:

					if self.tg_panel_help.val:
						drawParagraph(	AGBEditor.text_tips['ignore'],											413,	134,	395,	16,		CL_TEXT_MD,							'center',	'normal')

				elif self.panels.edge_type:

					if self.tg_panel_help.val:
						drawParagraph(	AGBEditor.text_tips['edge_types'],										413,	134,	395,	16,		CL_TEXT_MD,							'center',	'normal')
					else:
						drawString(							'EDGE TYPE',												592, 	124, 					CL_TEXT_LT,							'center',	'tiny')
						self.tg_road_facing		= Toggle(	 "road",				self.evt_road_facing,		569, 	101,	46,		19,		self.tg_road_facing.val,											"Select to edit road-facing edges.")
						self.tg_non_road_facing	= Toggle(	 "non",					self.evt_non_road_facing,	569, 	82,		46,		19,		self.tg_non_road_facing.val,										"Select to edit non-road-facing edges")
						self.bt_add_edge_group	= PushButton("add",					self.evt_add_edge_group,	569, 	63,		46,		19,																			"Make selected edges %s." % ("road-facing" if self.tg_road_facing.val else "non-road-facing") )
						self.bt_del_edge_group	= PushButton("delete",					self.evt_del_edge_group,	569, 	44,		46,		19,																			"Clear all %s edges." % ("road-facing" if self.tg_road_facing.val else "non-road-facing"))

				elif self.panels.match_grp:

					if self.tg_panel_help.val:
						drawParagraph(	AGBEditor.text_tips['match_groups'],									413,	134,	395,	16,		CL_TEXT_MD,							'center',	'normal')
					else:
						mg = self.temp_grid_ob.match_group_active
						if mg == None:
							mg = 'O'
						drawString(							'HORIZONTAL',											592, 	142, 	[0.00, 0.90, 0.00] if mg[0] == 'H' else CL_TEXT_LT,	'center',	'tiny')
						drawString(							'VERTICAL',												592, 	70, 	[0.20, 0.90, 0.00] if mg[0] == 'V' else CL_TEXT_LT,	'center',	'tiny')
						self.mn_match_groups_h	= Menu(	self.match_groups_h_mnstr,	self.evt_match_groups_h,	569, 	120,	46,		18,		self.mn_match_groups_h.val,											"Select a horizontal spelling match group to edit.")
						self.bt_add_match_group_h= PushButton("add",				self.evt_add_match_group_h,	569, 	101,	46,		19,																			"Make a new horizontal match group from the selected mesh edges.")
						self.bt_del_match_group_h= PushButton("delete",				self.evt_del_match_group_h,	569, 	82,		46,		19,																			"Delete the current horizontal match group")

						self.mn_match_groups_v	= Menu(	self.match_groups_v_mnstr,	self.evt_match_groups_v,	569,	48,		46,		18,		self.mn_match_groups_v.val,											"Select a vertical spelling match group to edit")
						self.bt_add_match_group_v= PushButton("add",				self.evt_add_match_group_v,	569, 	29,		46,		19,																			"Make a new vertical match group from the selected mesh edges.")
						self.bt_del_match_group_v= PushButton("delete",				self.evt_del_match_group_v,	569, 	10,		46,		19,																			"Delete the current vertical match group")

		elif page == 2 and not self.show_histograms:

			axis_str = "column" if self.tg_axis_horizontal.val else "row"
			drawString(								 	 "SPELLING SET",										105, 	154,					CL_TEXT_LT,	 			'center', 	'small')
			drawString(									 "MAKE PREVIEWS",										720, 	136,					CL_TEXT_LT,	 			'center', 	'small')
			drawString(								 	 "FREQ",												175, 	65,						CL_TEXT_LT,	 			'center', 	'tiny')

			self.bt_sp_set_previous			= PushButton("",						self.evt_sp_set_previous,	10, 	137, 	18, 	11,																			"Select previous spelling set.")
			self.bt_sp_set_next				= PushButton("",						self.evt_sp_set_next,		10, 	126, 	18, 	11,																			"Select next spelling set.")
			self.mn_sp_set_act				= Menu(	 self.mn_sp_set_act_menustring,	self.evt_sp_set_act,		28,		126,	172,	22,		self.mn_sp_set_act.val,												"The currently loaded AGB's spelling sets. Select one to edit it, or click 'NEW' to add a new one to the list.")
			self.mn_active_grid_index		= Menu(	 self.grid_obs_menu_string,		self.evt_active_grid_index,	10,		82,		190,	22,		self.mn_active_grid_index.val,										"The sub-grid associated with the current spelling set.")
			self.tg_override_params			= Toggle(	"o",						self.evt_override_params,	182,	104,	18,		22,		self.tg_override_params.val,										"Override params. When checked, all spellings are assumed to suit all block edge combinations.")

			if self.AGB_ob and self.AGB_ob.spelling_sets and self.active_sp_set_index != None:

				self.drawEdgeTypeIcons(self.AGB_ob.spelling_sets[self.active_sp_set_index].road_edges, 			11, 	110, 	15, 	11, 	2, 		21, 	CL_ROAD, CL_NOROAD, CL_ROAD_BKGND, halfBlock=self.AGB_ob.half_block)


			AGBht_min = self.AGB_ob.height_min if self.AGB_ob else 0
			AGBht_max = self.AGB_ob.height_max if self.AGB_ob else 999

			if self.AGB_ob and self.AGB_ob.spelling_sets and self.active_sp_set_index != None and self.AGB_ob.spelling_sets[self.active_sp_set_index].has_height_range:
				drawString(								  "MIN     HEIGHT     MAX",								80, 	65,						CL_TEXT_LT,	 			'center', 	'tiny')
				self.nb_sp_set_height_min	= Number(	 "", 						self.evt_sp_set_height_min, 10,		38,		70, 	22, 	self.nb_sp_set_height_min.val, 	AGBht_min,	min(self.nb_sp_set_height_max.val,	AGBht_max),		"Set the minimum height for this spelling set. To undefine a height range, drag min and max fully left and right respectively.")
				self.nb_sp_set_height_max	= Number(	 "", 						self.evt_sp_set_height_max, 80,		38,		70, 	22, 	self.nb_sp_set_height_max.val, 	max(self.nb_sp_set_height_min.val, AGBht_min),	AGBht_max,		"Set the maximum height for this spelling set. To undefine a height range, drag min and max fully left and right respectively.")
			else:
				self.bt_define_heights		= PushButton("define height range",		self.evt_define_heights,	10,		38,		140,	22,																			"Click to define a height range for this spelling set.")

			self.nb_sp_set_frequency		= Number(	 "", 						self.evt_sp_set_frequency, 	150,	38,		50, 	22, 	self.nb_sp_set_frequency.val, 	0,		20,							"Set the deployment frequency of the active spelling set, relative to others in the current AGB definition.")
			self.bt_delete_sp_set			= PushButton("DEL",						self.evt_delete_sp_set,		10, 	10, 	35, 	22,																			"Delete the active spelling set.")
			self.bt_copy_sp_set				= PushButton("COPY",					self.evt_copy_sp_set,		45, 	10, 	48, 	22,																			"Make a duplicate copy of the active spelling set.")
			self.bt_new_sp_set				= PushButton("NEW",						self.evt_new_sp_set,		93, 	10, 	40, 	22,																			"Add a new, empty spelling set.")
			self.bt_rename_sp_set			= PushButton("RENAME",					self.evt_rename_sp_set,		133,	10,		67,		22,																			"Rename the active spelling set.")
			self.tg_axis_horizontal			= Toggle("Horizontal", 					self.evt_axis_horizontal, 	210,	155,	120, 	20, 	self.tg_axis_horizontal.val, 										"Check to edit constraints affecting horizontal spellings (columns of tiles).")
			self.tg_axis_vertical			= Toggle("Vertical", 					self.evt_axis_vertical, 	495,	155,	120, 	20, 	self.tg_axis_vertical.val, 											"Check to edit constraints affecting vertical spellings (rows of tiles).")

			def threeWay(i):
				if self.tg_axis_horizontal.val:
					if self.temp_grid_ob.columns[i].ignore or self.temp_grid_ob.columns[i].exclude:
						return CL_GREY70
					elif self.temp_grid_ob.columns[i].require and self.temp_grid_ob.columns[i].xor:
						return CL_ANN_AMBER
					elif self.temp_grid_ob.columns[i].require:
						return CL_ANN_GREEN
					elif self.temp_grid_ob.columns[i].xor:
						return CL_ANN_RED
					else:
						return CL_TEXT_MD
				else:
					if self.temp_grid_ob.rows[i].ignore or self.temp_grid_ob.rows[i].exclude:
						return CL_GREY70
					elif self.temp_grid_ob.rows[i].require and self.temp_grid_ob.rows[i].xor:
						return CL_ANN_AMBER
					elif self.temp_grid_ob.rows[i].xor:
						return CL_ANN_RED
					elif self.temp_grid_ob.rows[i].require:
						return CL_ANN_GREEN
					else:
						return CL_TEXT_MD

			def buttonSpace(x, totalWidth, numButtons, index):
				px 		= int( index 	* float(totalWidth)/numButtons)
				px_next = int((index+1) * float(totalWidth)/numButtons)
				wdth	= px_next - px
				return [x+px, wdth, x+int((px+px_next)*0.5)]

			num_buttons_h = 0 if self.temp_grid_ob == None else self.temp_grid_ob.num_columns
			num_buttons_v = 0 if self.temp_grid_ob == None else self.temp_grid_ob.num_rows

			if not self.temp_grid_ob:
				drawString(								"AGB NOT LOADED - NOTHING TO EDIT",						413, 	122, 					CL_TEXT_LT,				'center',	'large')

			elif self.tg_axis_horizontal.val and num_buttons_h:
				for i in range(num_buttons_h):
					u = buttonSpace(210, 405, num_buttons_h, i)
					bnstr = "<"  if self.temp_grid_ob.columns[i].only_after else " "
					bnstr += (max(u[1] - 16, 0)/4) * " "
					bnstr += ">" if self.temp_grid_ob.columns[i].only_before else " "
					if self.temp_grid_ob.columns[i].ignore:
						drawButtonFrame(																		u[0],	114,	u[1],  	12,		CL_ACT_DIM)
					else:
						self.ButtonNumber[str(i)] 	=	 		PushButton(bnstr, self.evt_ButtonNumber[str(i)],u[0],	114,	u[1],  	12,																			"Click to toggle the state of the currently selected constraint for the respective column of tiles. Shift click for advanced settings.")

					drawString(							str(i),													u[2], 	129,					threeWay(i),			'center', 'small')
					if self.temp_grid_ob.columns[i].center :
						drawString(						'><',													u[2], 	138,					CL_ANN_GREEN ,			'center', 'small')
					if self.temp_grid_ob.columns[i].is_slop:
						drawRectangle(																			u[0]+1,	110,	u[1]-2,	2,		CL_ANN_GREEN)
					elif self.temp_grid_ob.columns[i].is_antislop:
						drawRectangle(																			u[0]+1,	110,	u[1]-2, 2,		CL_VECT_ORANGE)
					if self.temp_grid_ob.columns[i].dupes_min != None or self.temp_grid_ob.columns[i].dupes_max != None:
						nmstr = ('%s:' % str(self.temp_grid_ob.columns[i].dupes_min)) if self.temp_grid_ob.columns[i].dupes_min != None else '  :'
						nmstr += str(self.temp_grid_ob.columns[i].dupes_max) if self.temp_grid_ob.columns[i].dupes_max != None else '  '
						drawString(						nmstr, 													u[2]-1, 146,					CL_ANN_GREEN ,			'center', 'tiny')

			elif self.tg_axis_vertical.val and num_buttons_v:
				for i in range(num_buttons_v):
					u = buttonSpace(210, 405, num_buttons_v, i)
					bnstr = "<"  if self.temp_grid_ob.rows[i].only_after else " "
					bnstr += (max(u[1] - 16, 0)/4) * " "
					bnstr += ">" if self.temp_grid_ob.rows[i].only_before else " "
					if self.temp_grid_ob.rows[i].ignore:
						drawButtonFrame(																		u[0],	114,	u[1],  	12,		CL_ACT_DIM)
					else:
						self.ButtonNumber[str(num_buttons_h+i)] = PushButton(bnstr, self.evt_ButtonNumber[str(num_buttons_h+i)],u[0], 114,	u[1],  	12,																"Click to toggle the state of the currently selected constraint for the respective row of tiles. Shift click for advanced settings.")
					drawString(							str(i),													u[2], 	129,					threeWay(i),			'center', 'small')

					if self.temp_grid_ob.rows[i].center:
						drawString(						'><',													u[2], 	138,					CL_ANN_GREEN,			'center', 'small')
					if self.temp_grid_ob.rows[i].is_slop:
						drawRectangle(																			u[0]+1,	110,	u[1]-2,  2,		CL_ANN_GREEN)
					elif self.temp_grid_ob.rows[i].is_antislop:
						drawRectangle(																			u[0]+1,	110,	u[1]-2,  2,		CL_VECT_ORANGE)
					if self.temp_grid_ob.rows[i].dupes_min != None or self.temp_grid_ob.rows[i].dupes_max != None:
						nmstr = ('%s:' % str(self.temp_grid_ob.rows[i].dupes_min)) if self.temp_grid_ob.rows[i].dupes_min != None else '  :'
						nmstr += str(self.temp_grid_ob.rows[i].dupes_max) if self.temp_grid_ob.rows[i].dupes_max != None else '  '
						drawString(						nmstr, 													u[2]-1, 146,					CL_ANN_GREEN ,			'center', 'tiny')

			show_hide_name = "hide" if self.show_constraints else "show"
			reset_name = 'reset all' if g_qualifs & Window.Qual.SHIFT else 'reset'
			self.bt_sp_clr_constraints		= PushButton(reset_name,					self.evt_sp_clr_constraints,210, 	82, 	81, 	22,																		"Resets the current constraint to 'off' for all %ss of tiles. Shift-click to reset all constraints and common settings for this spelling set." % axis_str)
			self.tg_sp_exclude				= Toggle("exclude", 					self.evt_sp_exclude, 		291,	82,		61, 	22, 	self.tg_sp_exclude.val, 											"Toggle any of the above buttons to exclude the corresponding %s(s) of tiles from new spellings." % axis_str)
			self.tg_sp_require				= Toggle("require", 					self.evt_sp_require, 		352,	82,		61, 	22, 	self.tg_sp_require.val, 											"Toggle any of the above buttons to require the corresponding %s(s) of tiles to be included in new spellings." % axis_str)
			self.bt_toggle_or_xor			= PushButton("",						self.evt_toggle_or_xor,		455,	82,		18, 	22,																			"Toggles the 'xor' constraint behaviour between 'or' ('at least one of') and 'xor' ('at most one of').")
			xor_bt_name = 'or' if self.common_constraints['vertical' if self.spellings_axis else 'horizontal']['xor_is_or'] else 'xor'
			self.tg_sp_xor					= Toggle(xor_bt_name, 					self.evt_sp_xor, 			413,	82,		42, 	22, 	self.tg_sp_xor.val, 												"Only one XOR ('exclusive or') %s is allowed to appear in a given spelling, effectively making a selection of two or more %ss mutually exclusive." % (axis_str, axis_str))
			self.tg_sp_center				= Toggle("center", 						self.evt_sp_center, 		473,	82,		61, 	22, 	self.tg_sp_center.val, 												"Toggle any of the above buttons to request for the corresponding %s(s) of tiles to be located as near as possible to the physical center of a spelling, when included." % axis_str)
			self.bt_sp_sel_constraints		= PushButton(show_hide_name,			self.evt_sp_sel_constraints,534, 	82, 	81, 	22,																			"View the state of the current constraint by means of mesh face selection. Can sometimes make the interface sluggish.")

			drawString(								"SPELLING CONSTRAINTS",										413, 	162, 						CL_TEXT_LT,								'center',	'small')
			drawString(								"MIN     LENGTH     MAX",									282, 	65, 						CL_TEXT_LT,								'center',	'tiny')


			if self.last_iterations:
				drawString(							self.last_iterations,										574, 	65, CL_ANN_RED if self.maxed_out_iterations else CL_TEXT_LT,		'center',	'tiny')
			else:
				drawString(							"MAX  ITERATIONS",											574, 	65, CL_ANN_RED if self.maxed_out_iterations else CL_TEXT_LT,		'center',	'tiny')
			if self.progress_message2:
				drawString(							self.progress_message2,										441, 	65,							CL_WHITE,	 							'center', 	'tiny')
			else:
				drawString(							"DUPES",													372, 	65, 						CL_TEXT_LT,								'center',	'tiny')
				drawString(							"CUTS",														412, 	65, 						CL_TEXT_LT,								'center',	'tiny')
				drawString(							"VARIANTS",													452, 	65, 						CL_TEXT_LT,								'center',	'tiny')
				drawString(							"TOLERANCE",												503, 	65, 						CL_TEXT_LT,								'center',	'tiny')

			if self.progress_message:
				drawString(							self.progress_message,										720, 	103,						CL_WHITE,	 							'center', 	'tiny')
			else:
				drawString(							"WIDTH  RANGE",												672, 	103,						CL_TEXT_LT,								'center',	'tiny')
				drawString(							"DEPTH  RANGE",												767, 	103,						CL_TEXT_LT,								'center',	'tiny')

			def cb(ev, val):
				return
			plural = "s" if self.nb_increment_pv.val > 1 else ""
			if self.histo_prefs['custom_ranges']:
				self.bt_reset_scale			= PushButton("",						self.evt_reset_scale,		805,	101,	10,		10,																				"Reset to default ranges.")
			self.nb_min_length_pv			= Number("", 							self.evt_min_length_pv, 	625,	82,		47, 	16, 	self.nb_min_length_pv.val, 	0,	min(1000,self.nb_max_length_pv.val-1),	"Sets the minimum width for preview spellings and histograms.", cb, 1000)
			self.nb_max_length_pv			= Number("", 							self.evt_max_length_pv, 	672,	82,		48, 	16, 	self.nb_max_length_pv.val, 	max(self.nb_min_length_pv.val+1,0),	1000,	"Sets the maximum width for preview spellings and histograms.", cb, 1000)
			self.nb_min_depth_pv			= Number("", 							self.evt_min_depth_pv, 		720,	82,		47, 	16, 	self.nb_min_depth_pv.val, 	0,	min(1000,self.nb_max_depth_pv.val-1),	"Sets the minimum depth for preview spellings and histograms.", cb, 1000)
			self.nb_max_depth_pv			= Number("", 							self.evt_max_depth_pv, 		767,	82,		48, 	16, 	self.nb_max_depth_pv.val, 	max(self.nb_min_depth_pv.val+1,0),	1000,	"Sets the maximum depth for preview spellings and histograms.", cb, 1000)

			self.tg_by_spelling_pv			= Toggle("spelling" + plural,			self.evt_by_spelling_pv, 	750,	114,	65, 	16, 	self.tg_by_spelling_pv.val, 											"When selected, previews are made from every nth spelling in the list, where n is the number set in the adjacent field.")
			self.tg_by_size_pv				= Toggle("metre" + plural, 				self.evt_by_size_pv, 		700,	114,	50, 	16, 	self.tg_by_size_pv.val, 												"When selected, successive preview grids will be n metres wider/deeper than their predecessors, where n is the number set in the adjacent field.")
			self.nb_increment_pv			= Number("every ", 						self.evt_increment_pv, 		625,	114,	75, 	16, 	self.nb_increment_pv.val, 	1,	100,									"Increment between successive preview grids. Be aware that very low numbers can result in a very large number of previews.", cb, 100)
			self.tg_pv_annotations			= Toggle("Include Annotations", 		self.evt_pv_annotations, 	625,	60,		190, 	22, 	self.tg_pv_annotations.val, 											"When selected, objs, facades and trees are included in previews, otherwise only the grid mesh is included.")

			self.nb_min_length				= Number("", 							self.evt_min_length, 	 	210,	38,		71, 	22, 	self.nb_min_length.val, 	0.0,	min(self.nb_max_length.val-1.0, 5000.0), "Minimum target length for the next run of the spelling generator.", cb, 1000, 1.0)
			self.nb_max_length				= Number("", 							self.evt_max_length, 	 	281,	38,		71, 	22, 	self.nb_max_length.val, 	max(self.nb_min_length.val+1.0, 10.0),	5000.0,	 "Maximum target length for the next run of the spelling generator.", cb, 1000, 1.0)

			self.nb_max_dupes				= Number("", 							self.evt_max_dupes, 	 	352,	38,		40, 	22, 	self.nb_max_dupes.val, 		0,		8,								"The maximum number of times any given tile may be repeated in a generated spelling.")
			self.nb_max_cuts				= Number("", 							self.evt_max_cuts, 			392,	38,		40, 	22, 	self.nb_max_cuts.val, 		1,		8,								"The maximum number of horizontal or vertical cuts allowed in generated spellings.")
			self.sl_sp_variety				= Number("", 							self.evt_sp_variety, 		432,	38,		41, 	22, 	self.sl_sp_variety.val,		1,		8,								"The number of different spellings that will be generated per spelling length.")
			self.nb_center_precision		= Number("", 							self.evt_center_precision, 	473,	38,		61, 	22, 	self.nb_center_precision.val,0.0,	100.0,							"'Centering' tolerance, in metres.", cb, 100, 1.0)
			self.nb_max_iterations			= Number("", 							self.evt_max_iterations, 	534,	38,		81, 	22, 	self.nb_max_iterations.val, 50000,	5000000,						"The maximum allowed number of iterations of the spelling generator algorithm. Increase if 'MAX ITERATIONS' shows red, indicating the limit has been reached.")

			self.bt_make_spellings			= PushButton("GENERATE SPELLINGS",		self.evt_make_spellings,	352, 	10, 	182, 	22,																			"Starts the spelling generator routine. Depending on your settings, this may take a little while to complete. 'Esc' to cancel")
			self.bt_add_to_sp_set			= PushButton("COMMIT...",				self.evt_add_to_sp_set,		534, 	10, 	81, 	22,																			"Add the latest generated spellings to the active spelling set.")

			self.bt_show_histograms			= PushButton("SHOW HISTOGRAMS",			self.evt_show_histograms,	625, 	153,	190,	22,																			"View a histogram display of the current AGB's metrics performance.")
			self.bt_clear_temp_text			= PushButton("CLEAR",					self.evt_clear_temp_text,	281, 	10,		71,		22,																			"Erase all content from the Blender text object XPTAGBSpellings.")
			self.bt_print_spellings			= PushButton("PRINT",					self.evt_print_spellings,	210, 	10,		71,		22,																			"Print spellings to the Blender text object XPTAGBSpellings.")
			self.bt_make_prev_grids			= PushButton("CREATE...",				self.evt_make_prev_grids,	720, 	38,		95,		22,																			"Build multiple AGB previews covering the specified size ranges. BE ADVISED: This can potentially result in a LOT of objects - if in any doubt, be conservative with your range settings!")
			self.bt_clear_prev_grids		= PushButton("CLEAR",					self.evt_clear_prev_grids,	625, 	38,		95,		22,																			"Use this button to conveniently clear all AGB previews from the scene - AVOID DELETING THEM BY HAND, it prevents XPTools conserving memory by re-cycling previously deleted objects.")
			self.bt_filter_spellings		= PushButton("FILTER SPELLINGS...",		self.evt_filter_spellings,	625,	10,		190,	22,																			"Use selected preview grids to filter the active spelling set...")

		elif page == 2 and self.show_histograms:


#			drawRectangle(																						132,	10,		527,	160,	[0.40,0.40,0.40])
			if self.AGB_ob:
				displayname = self.AGB_ob.name[3:] + '.agb'
			drawString(									"Results for %s" % displayname,							137, 	184,					[0.8, 0.8, 0.8],				'ljust',	'normal')
	#		drawRectangle(
		#	self.bt_toggle_temp_histo		= PushButton("toggle temp",				self.evt_show_histograms,	670, 	10,		145,	22,																			"")
			showstr = "Hide Latest Results" if self.show_temp else "Show Latest Results"
			self.bt_show_temp				= PushButton(showstr,					self.evt_show_temp,		  	670,	35,		145,	18,																			"Show results for the most recent generated spellings as well as those for the current AGB.")
			self.bt_show_histograms			= PushButton("CLOSE",					self.evt_show_histograms,	670, 	10,		145,	22,																			"")
#			self.tg_edge_type_splits		= Toggle("split by edge-type",			self.evt_edge_type_splits,  10,		10,		115,	22,		self.tg_edge_type_splits.val,										"Make the AGB histogram sensitive to differences in coverage for different edge-type configurations.")

			def cb(ev, val):
				return
			plural = "s" if self.nb_increment_pv.val > 1 else ""
			self.nb_min_length_pv			= Number("", 							self.evt_min_length_pv, 	467,	182,	45, 	13, 	self.nb_min_length_pv.val, 	0,	min(1000,self.nb_max_length_pv.val-1),	"Sets the minimum width for preview spellings and histograms.", cb, 1000)
			self.nb_max_length_pv			= Number("", 							self.evt_max_length_pv, 	512,	182,	45, 	13, 	self.nb_max_length_pv.val, 	max(self.nb_min_length_pv.val+1,0),	1000,	"Sets the maximum width for preview spellings and histograms.", cb, 1000)
			self.nb_min_depth_pv			= Number("", 							self.evt_min_depth_pv, 		560,	182,	45, 	13, 	self.nb_min_depth_pv.val, 	0,	min(1000,self.nb_max_depth_pv.val-1),	"Sets the minimum depth for preview spellings and histograms.", cb, 1000)
			self.nb_max_depth_pv			= Number("", 							self.evt_max_depth_pv, 		605,	182,	45, 	13, 	self.nb_max_depth_pv.val, 	max(self.nb_min_depth_pv.val+1,0),	1000,	"Sets the maximum depth for preview spellings and histograms.", cb, 1000)
			if self.histo_prefs['custom_ranges']:
				self.bt_reset_scale			= PushButton("",						self.evt_reset_scale,		655,	183,	10,		11,																				"Reset histogram scale to the default required ranges of the AGB.")

			for h in self.histograms:
				h.drawHistogram()
			for hs in self.histograms_scrolling:
				hs.drawHistogram(-self.scroll_offset*self.SPELLING_SET_HT, self.scroll_limit*self.SPELLING_SET_HT)
			if self.scroll_limit:
				bh = self.scroll_bt_h
				bw = 14
				if self.scroll_offset < self.scroll_limit:
					self.bt_scroll_up		= PushButton("/\\",						self.evt_scroll_up,			653, self.scroll_up_bt_y,			bw,		bh,													"Scroll the histogram display up.")
				if self.scroll_offset > 0:
					self.bt_scroll_down		= PushButton("\\/",						self.evt_scroll_down,		653, self.scroll_down_bt_y,			bw,		bh,													"Scroll the histogram display down.")
# # # _______________________________________________________________________________________________________________________________________________________________________________________

class Bitfield(object):

	def __init__(self,value=0):
		self._d = value

	def __getitem__(self, index):
		return (self._d >> index) & 1

	def __setitem__(self,index,value):
		value	 = (value&1)<<index
		mask	 = (1)<<index
		self._d	 = (self._d & ~mask) | value

	def __getslice__(self, start, end):
		mask = 2**(end - start) -1
		return ((self._d >> start) & mask)

	def __setslice__(self, start, end, value):
		mask = 2**(end - start) -1
		value = (value & mask) << start
		mask = mask << start
		self._d = (self._d & ~mask) | value
		return (self._d >> start) & mask

	def __int__(self):
		return self._d

	def int(self):
		return self._d
# # # _______________________________________________________________________________________________________________________________________________________________________________________

g_snap_angle = 15.0
DEF_TILE_WRAPPER_LAYER 	= 4
DEF_METAPIN_LAYER 		= 6
DEF_OBJ_WRAPPER_LAYER	= 3
DEF_METARECT_LAYER		= 6
DEF_TILE_MESH_LAYER		= 5
DEF_LIGHT_LAYER			= 1


GUIpage = 1
GUIpage_prev = 0
GUIchanged = 1
g_qualifs	= 0
sResults = ["result"]
tasksdone = {	'tcnt':0, 'rnam':0, 'cent':0, 'wrap':0, 'agsw':0, 'pins':0, 'pina':0, 'pinp':0, 'ping':0, 'shrY':0, 'proj':0, 'nonm':0,
				'objprops_added':0, 'objprops_untouched':0, 'objprops_updated':0, 'objprops_stripped':0, 'objs_stripped_from':0,
				'OBJatt':0, 'FACatt':0, 'TREEatt':0,
				'TILES':0, 'DYNAMIC':0, 'TWOSIDE':0, 'SHADOW':0, 'INVISIBLE':0, 'tilesobjs':0, 'dynamicobjs':0, 'twosideobjs':0, 'shadowobjs':0, 'invisibleobjs':0,
				'SOLID':0, 'ALPHA':0, 'CLIP':0, 'transpobjs':0,
				'OBJs_selected':0, 'OBJ_meshes_selected':0, 'assign_OBJ_image':0, 'moveOBJCentre':0, 'shared_image_obs':0}

iRnam = 0
iIdin = 0
sTexn = "Image.png"
fTexh = 256
fTexw = 256
fOrgx = 0
fOrgy = 0
iPins = 0
fMrga = 2.0
fMrgf = 3.0
fMrgr = 2.0
fMrgb = 3.0
fMrgl = 2.0
fSpac = 7
iPina = 0
iPinp = 0
iShrY = 0
iPing = 0
iLayr = DEF_METAPIN_LAYER
iProj = 0
iCent = 0
iWrap = 0
iAgsw = 0
iLayw = DEF_TILE_WRAPPER_LAYER
fDmin = 15.0
#fDmax = 100.0

iSgrp = 0

sdata =	{
			'min_is_rel'		: 0,
			'max_is_rel'		: 0,
			'actual_min'		: 0,
			'actual_max'		: 100,
			'declared_min'		: 0,
			'declared_max'		: 10000,
			'declared_pre_min'	: 0,
			'declared_pre_max'	: 10000,
			'dist_slop'			: 2.0,
			'depth_min'			: 0,
			'depth_max'			: 0,
			'angle_start'		: 0,
			'angle_end'			: 0,
			'tile_variety'		: 0,
			'depth_variety'		: 0,
			'match_group'		: 0,
			'member_group'		: 0,
			'comment'			: "comment",
			'use_group_members' : 0,
			'first_tile'		: 0,
			'last_tile'			: 0,
			'atomic'			: 0
		}

enum_objshape		= {0:'A: Rectangular', 1:'B: L-shape', 2:'C: Reverse L-shape', 3:'D: U-shape', 4:'E: Inverted U-shape', 5:'F: H-shape', 6:'G: Sideways H-shape', 7:'H: Round', 8:'I: Rectangular donut', 9:'J: Irregular'}
enum_objpart		= {0:'A: Default',
						1:'B: Articulated corner part 1', 2:'C: Articulated corner part 2', 3:'D: Articulated corner part 3',
						4:'E: Reflex corner pair part 1', 5:'F: Reflex corner pair part 2',
						6:'G: Atomic concatenation start', 7:'H: Atomic concatenation middle', 8:'I: Atomic concatenation end',
						9:'J: AGB Compression pair front', 10:'K: AGB Compression pair back',
						11:'L: AGB Compression trio front', 12:'M: AGB Compression trio middle', 13:'N: AGB Compression trio back',
						14:'O: FACScraper tower',	15:'P: FACScraper base',	16:'Q: FACScraper pad', 17:'R: OBJScraper tower', 18:'S: OBJScraper base',
						19:'T: SE-facing block corner', 20:'U: SW-facing block corner'}

# for k, v in enum_objshape.iteritems():
# 	print v
#
# for k, v in enum_objpart.iteritems():
# 	print v

OBJRENAME				= 0
OBJSHAPE				= 0
OBJPART					= 0
OBJVARIANT				= 0
AUTOVARIANT				= 1
VARSTOLIB				= 0
EXPORTMODE				= 1						#	EXPORTMODE = 0 -> export OBJs "in place" from parent AG entity using the AG exporter. Wrapper name is used for VIRTUAL name, OB and optionally RNAME property carries real name.
												#	EXPORTMODE = 1 -> export OBJs directly to library, using the multi-OBJ exporter.	  Wrapper name is used for REAL name, VNAME property carries virtual name.
USEHEIGHTBUCKETS		= 1						#	Round OBJ height metrics into established AG buckets.
BUCKETSVIRTUALONLY		= 1						#	Use height buckets only when building the virtual name, and use the real height for the real name.
OBJPROPSOVERWRITE		= 0						#	'0' leaves existing OBJ wrapper properties unchanged, '1' updates them to the user-defined vals
OBJWRITEPROPS			= 0
OBJVIRTUALPATH			= ""					#	User-defined path to the current virtual folder. I.e. this is the virtual path minus the object name
OBJINSTANCED			= 0
HASDRAPED				= 0
facetranspmodes			= {'SOLID':1, 'ALPHA':0, 'CLIP':0}											# states of GUI buttons
facemodes		 		= {'TILES':0, 'TWOSIDE':0, 'DYNAMIC':1, 'SHADOW':0, 'INVISIBLE':0}										# states of GUI buttons
modestosearch			= {'TILES':0, 'TWOSIDE':0, 'DYNAMIC':0, 'SHADOW':0, 'INVISIBLE':0, 'TRANSP':0}		# states of GUI buttons - whether to include the repsective facemodes and facetranspmodes in searches
facematchall			= 1

objprops = {
			"PATH":						"",
			"SHOW_LEVEL":				"",
			"GLOBAL_no_blend":			"",
			"GLOBAL_shadow_blend":		"",
			"GLOBAL_specular":			"",
			"GLOBAL_tint":				"",
			"ATTR_LOD_draped":			"",
			"ATTR_layer_group":			"",
			"ATTR_layer_group_draped":	"",
			"SLOPE_LIMIT":				"",
			"PARAMS":					"",
			"VNAME":					"",
			"Virtual Path":				OBJVIRTUALPATH,
			"STEP":						""
			}

initval = 0
for i, k in enumerate(objprops):
	if k == "Virtual Path":
		initval = i

objprops2 = {"":						"",
			"lod_1":					"",
			"lod_2":					"",
			"TILTED":					"",
			"REQUIRE_WET":				"",
			"REQUIRE_DRY":				"",
			"ATTR_draped":				0,
			"INSTANCED":				1,
			"GRADED":					""
			}
objpropkeys = {
			"tilted":	"TILTED",
			"req wet":	"REQUIRE_WET",
			"req dry":	"REQUIRE_DRY",
			"graded" :	"GRADED"
			}

bt_pageAGS  = Create(0)
bt_pageOBJ  = Create(0)
bt_pageOBJ2 = Create(0)
bt_pageAGB  = Create(0)
bt_pageAGB2 = Create(0)
bt_pageANN  = Create(0)
bt_pageMSC  = Create(0)

#AGS stuff:

rnam = Create(iRnam)
idin = Create(iIdin)
cent = Create(iCent)
wrap = Create(iWrap)
agsw = Create(iAgsw)
pins = Create(iPins)
mrgf = Create(fMrgf)
mrgr = Create(fMrgr)
mrgb = Create(fMrgb)
mrgl = Create(fMrgl)
layr = Create(iLayr)
spac = Create(fSpac)
pina = Create(iPina)
mrga = Create(fMrga)
pinp = Create(iPinp)
ping = Create(iPing)
shrY = Create(iShrY)
layw = Create(iLayw)
texw = Create(fTexw)
texh = Create(fTexh)
orgx = Create(fOrgx)
orgy = Create(fOrgy)
texn = Create(sTexn)
proj = Create(iProj)
dmin = Create(fDmin)
mtch = Create(0)
dslp = Create(0)
sgrp = Create(0)
curr = Create(0)
appl = Create(0)
canc = Create(0)
rtrn = Create(0)
optm = Create(0)

# OBJ stuff:

EVENT_mn_objpart = 110
EVENT_mn_objshape = 111
EVENT_mn_objvariant = 112
EVENT_tg_objbucketsvirtualonly = 113
EVENT_tg_objautovariant = 114
EVENT_tg_objbuckets = 115
EVENT_tg_objrename = 116
EVENT_mn_EXPORTMODE = 117
EVENT_nb_objwrapperlayer = 118
EVENT_mn_propname0 = 119
EVENT_st_propval0 = 120
EVENT_st_propval1 = 121
EVENT_st_propval2 = 122
EVENT_mn_propname1 = 123
EVENT_mn_propname2 = 124
EVENT_tg_objpropsoverwrite = 125
EVENT_bt_objselallfaces = 126
EVENT_bt_objsetfaces = 127
EVENT_bt_objdeselallfaces = 128
EVENT_bt_objselwrappers = 129
EVENT_bt_objshowfaces = 130
EVENT_bt_objshowfacesvis = 131
EVENT_bt_objresetbuttons = 132
EVENT_tg_objvarstolib = 133
EVENT_bt_objgetsameimage = 257
EVENT_bt_objgetfromactive = 258
EVENT_bt_restrictselect = 259
EVENT_bt_enableselect = 260
EVENT_bt_dupehierarchy = 261
EVENT_tg_AWS_enable		= 262
evt_sort_imports		= 263

EVENT_tg_objfmdynamic  	= 300
EVENT_tg_objfmtiles  	= 301
EVENT_tg_objfmtwoside 	= 302
EVENT_tg_objfmshadow  	= 303
EVENT_tg_objfminvisible = 304
EVENT_tg_objftmsolid  	= 305
EVENT_tg_objftmclip  	= 306
EVENT_tg_objftmalpha  	= 307

EVENT_bt_objenableall= 308
EVENT_bt_objdefaultmodes = 309
EVENT_tg_objmatchany	= 310
EVENT_tg_objmatchall	= 311
EVENT_bt_objfindmixedalpha = 312
EVENT_tg_objfacemodesel = []
for i in range(6):
	EVENT_tg_objfacemodesel.append(315 + i)
EVENT_bt_objmatchlogic = 321

Object_mn_objpart = Create(0)
Object_mn_objshape = Create(0)
Object_mn_objvariant = Create(0)
Object_tg_objautovariant = Create(1)
Object_tg_objvarstolib = Create(0)
Object_tg_objbuckets = Create(1)
Object_tg_objrename = Create(0)
Object_mn_EXPORTMODE = Create(0)
Object_tg_objbucketsvirtualonly = Create(1)
Object_nb_objwrapperlayer = Create(DEF_OBJ_WRAPPER_LAYER)
Object_mn_propname0 = Create(initval)
Object_st_propval0 = Create("lib/g10/dev/")
Object_st_propval1 = Create("")
Object_st_propval2 = Create("")
Object_mn_propname1 = Create(999)
Object_mn_propname2 = Create(999)
Object_tg_objpropsoverwrite = Create(0)
Object_tg_objwriteprops = Create(0)
Object_bt_objselwrappers = Create(0)
Object_bt_objselgeom = Create(0)
Object_tg_AWS_enable = Create(0)
bt_sort_imports		= Create(0)

Object_bt_objselallfaces = Create(0)
Object_bt_objdeselallfaces = Create(0)
Object_bt_objshowfaces = Create(0)
Object_bt_objsetfaces = Create(0)
Object_bt_objshowfacesvis = Create(0)
Object_bt_objfindmixedalpha = Create(0)
Object_tg_objfmdynamic  = Create(1)
Object_tg_objfmtiles  = Create(0)
Object_tg_objfmtwoside = Create(0)
Object_tg_objfmshadow  	= Create(0)
Object_tg_objfminvisible = Create(0)

Object_tg_objftmsolid  = Create(1)
Object_tg_objftmclip  = Create(0)
Object_tg_objftmalpha  = Create(0)
Object_tg_objmatchany	 = Create(0)
Object_tg_objmatchall	 = Create(1)
Object_tg_objfacemodesel = []
for i in range(6):
	Object_tg_objfacemodesel.append(Create(0))
Object_bt_objmatchlogic = Create(0)
Object_bt_objenableall = Create(0)
Object_bt_objdefaultmodes  = Create(0)

Object_bt_objresetbuttons = Create(0)
Object_tg_objinstanced = Create(1)
Object_mn_objpropname3 = Create(999)
Object_mn_objpropname4 = Create(999)
Object_tg_objhasdraped = Create(0)
Object_nb_objlod1 = Create(0)
Object_nb_objlod2 = Create(0)
Object_mn_objimage = Create(999)
Object_bt_objassignimage = Create(0)
Object_bt_objgetsameimage = Create(0)
Object_bt_objgetfromactive = Create(0)
Object_bt_restrictselect = Create(0)
Object_bt_enableselect = Create(0)
Object_bt_dupehierarchy	= Create(0)
Object_objmoveOBJcentre = Create(0)
Object_bt_objstripprops = Create(0)
OBJIMAGELIST = []
OBJIMAGE_LIT = None
OBJIMAGE_ALB = None
objimagemenuname = ""

# annotation stuff:

ob_OBJatt  = 0
ob_FACatt  = 0
ob_TREEatt = 0
tg_OBJatt  = Create(ob_OBJatt)
tg_FACatt  = Create(ob_FACatt)
tg_TREEatt = Create(ob_TREEatt)
bt_appann  = Create(0)
mgrp	   = Create(0)
cmmt	   = Create("")
rlmn =		Create(0)
rlmx =		Create(0)
acmn =		Create(0)
acmx =		Create(0)
dcmn =		Create(0)
dcmx =		Create(0)
angs =		Create(0)
ange =		Create(0)
dpmn =		Create(0)
dpmx =		Create(0)
vari =		Create(0)
vard =		Create(0)
tfst =		Create(0)
tlst =		Create(0)
atmc =		Create(0)

# ----------

AWS_scriptlink = None
AWS_scriptlink_text = """# SPACEHANDLER.VIEW3D.EVENT.ALL
import bpy
import Blender
from Blender import Draw, Object
from Blender import Registry, Window
from math import pi

def exportConfig(value):
	conf = Blender.Registry.GetKey("XPTSelectWrapper")
	if conf:
		conf['LAST_EVENT']=value
	else:
		conf = {'LAST_EVENT':value, 'ENABLE':1}
	Blender.Registry.SetKey("XPTSelectWrapper", conf)

def importConfig():
	LAST_EVENT = 0
	ENABLE = 0
	conf = Blender.Registry.GetKey("XPTSelectWrapper")
	if conf:
		try:
			LAST_EVENT = conf['LAST_EVENT']
			ENABLE	   = conf['ENABLE']
		except KeyError, detail:
			print 'ERROR:', detail
	return 	LAST_EVENT and ENABLE

def getChildren(parent, all):
	children = []
	for ob in all:
		if ob.parent == parent:
			children.append(ob)
	return children

if Blender.event == Draw.RIGHTMOUSE and Blender.eventValue == 1:
	exportConfig(1)
	if importConfig() and Window.GetKeyQualifiers() & Window.Qual.ALT:
		Window.SetKeyQualifiers(0)

elif Blender.eventValue == 0:
	q = Window.GetKeyQualifiers()

	if importConfig():
		scn = bpy.data.scenes.active
		obact = scn.objects.active
		if obact != None:
			if obact.type in ['Mesh', 'Lamp']:
				p = obact.parent
				if p != None and p.type == 'Empty' and p.name[:3] =='OBJ':
					obact.sel = 0
					if q & Window.Qual.SHIFT:
						p.sel = 1 - p.sel
					elif q & Window.Qual.ALT:
						p.sel = 1
						retval =Draw.PupMenu("OBJ %t|rot CW%x1|rot 180%x2|rot CCW%x3|%l|copy%x5|delete%x6")
						if retval == 1:
							x,y,z = p.getEuler()
							p.setEuler([x,y,z-pi/2.0])
							p.makeDisplayList()
						elif retval == 2:
							x,y,z = p.getEuler()
							p.setEuler([x,y,z+pi])
							p.makeDisplayList()
						elif retval == 3:
							x,y,z = p.getEuler()
							p.setEuler([x,y,z+pi/2])
							p.makeDisplayList()
						elif retval == 5:
							children = getChildren(p, scn.objects)
							children.append(p)
							scn.objects.selected = children
							scn.objects.active = p
							Object.Duplicate()
							p_dup = scn.objects.active
							scn.objects.selected = [p_dup]
							p_dup.clrParent(2)
							p_dup.LocY -= 10.0
							p_dup.makeDisplayList
						elif retval == 6:
							children = getChildren(p, scn.objects)
							for c in children:
								scn.objects.unlink(c)
							scn.objects.unlink(p)
			#			Blender.event = None
					else:
						p.sel = 1

		exportConfig(0)
"""
def autoWrapperSelect(mode):

	global AWS_scriptlink
	"""
		This fiendish routine, when enabled, results in the following 3-d view behaviour:
		If any child of an OBJ wrapper is right-clicked with the mouse, the WRAPPER will be selected instead of the child itself.
		Works by setting a space-handler script in the 3-d window.
	"""
	def exportAWSConfig(value):
		conf = {'LAST_EVENT':0, 'ENABLE':value}
		Blender.Registry.SetKey("XPTSelectWrapper", conf)

	if mode == 1:
		if AWS_scriptlink == None:
			try:
				AWS_scriptlink = Blender.Text.Get("XPToolsAWS")
				Blender.Text.unlink(AWS_scriptlink)
				AWS_scriptlink = Blender.Text.New("XPToolsAWS")
				AWS_scriptlink.write(AWS_scriptlink_text)
			except:
				AWS_scriptlink = Blender.Text.New("XPToolsAWS")
				AWS_scriptlink.write(AWS_scriptlink_text)
		exportAWSConfig(1)

	elif mode == 0:
		exportAWSConfig(0)

# ----------
sce = 0
obsel = []
obsorted = []

def ExportConfig():

	conf = {}
	conf['OBJRENAME']=			OBJRENAME
	conf['OBJSHAPE']=			OBJSHAPE
	conf['OBJPART']=			OBJPART
	conf['OBJVARIANT']=			OBJVARIANT
	conf['AUTOVARIANT']=		AUTOVARIANT
	conf['VARSTOLIB']=			VARSTOLIB
	conf['EXPORTMODE']=			EXPORTMODE
	conf['USEHEIGHTBUCKETS']=	USEHEIGHTBUCKETS
	conf['BUCKETSVIRTUALONLY']=	BUCKETSVIRTUALONLY
	conf['OBJPROPSOVERWRITE']=	OBJPROPSOVERWRITE
	conf['OBJWRITEPROPS']=		OBJWRITEPROPS
	conf['OBJVIRTUALPATH']=		objprops["Virtual Path"]
	conf['OBJINSTANCED']=		OBJINSTANCED
	conf['HASDRAPED']=			HASDRAPED
	conf['LIGHTS_TXT_DIR'] =	LIGHTS_TXT_DIR
	conf['LIGHTS_TEXTURE_DIR'] = LIGHTS_TEXTURE_DIR
	conf['LIGHT_EDITOR_CONFIG'] = LIGHT_EDITOR_CONFIG
	conf['FACETRANSPMODES'] 	= facetranspmodes
	conf['FACEMODES'] 			= facemodes
	conf['MODESTOSEARCH']       = modestosearch
	conf['GUIPAGE']				= GUIpage
	conf['FACEMATCHALL']		= facematchall
	conf['TILE_WRAPPER_LAYER'] 	= DEF_TILE_WRAPPER_LAYER
	conf['OBJ_WRAPPER_LAYER'] 	= DEF_OBJ_WRAPPER_LAYER
	conf['METAPIN_LAYER'] 		= DEF_METAPIN_LAYER
	conf['METARECT_LAYER'] 		= DEF_METARECT_LAYER
	conf['TILE_MESH_LAYER']		= DEF_TILE_MESH_LAYER
	conf['LIGHT_LAYER'] 		= DEF_LIGHT_LAYER
	Blender.Registry.SetKey("XPlaneAGTools", conf, True)

def ImportConfig():

	global OBJRENAME, OBJSHAPE,	OBJPART,	OBJVARIANT,	AUTOVARIANT, VARSTOLIB,	EXPORTMODE,	USEHEIGHTBUCKETS,	BUCKETSVIRTUALONLY,	OBJPROPSOVERWRITE,	OBJWRITEPROPS,	OBJVIRTUALPATH,	OBJINSTANCED,	HASDRAPED
	global LIGHTS_TXT_DIR, LIGHTS_TEXTURE_DIR, LIGHT_EDITOR_CONFIG
	global facetranspmodes, facemodes, modestosearch, facematchall, GUIpage
	global DEF_TILE_WRAPPER_LAYER, DEF_OBJ_WRAPPER_LAYER,	DEF_METAPIN_LAYER,	DEF_METARECT_LAYER,	DEF_TILE_MESH_LAYER,	DEF_LIGHT_LAYER, layr, layw

	conf = Blender.Registry.GetKey("XPlaneAGTools", True)

	if not conf:
		print "Config file not found \n"
		ExportConfig()
		return
	try:
		OBJRENAME 					=	conf['OBJRENAME']
		OBJSHAPE 					=	conf['OBJSHAPE']
		OBJPART 					=	conf['OBJPART']
		OBJVARIANT 					=	conf['OBJVARIANT']
		AUTOVARIANT 				=	conf['AUTOVARIANT']
		VARSTOLIB					=	conf['VARSTOLIB']
		EXPORTMODE 					=	conf['EXPORTMODE']
		USEHEIGHTBUCKETS =				conf['USEHEIGHTBUCKETS']
		BUCKETSVIRTUALONLY =			conf['BUCKETSVIRTUALONLY']
		OBJPROPSOVERWRITE =				conf['OBJPROPSOVERWRITE']
		OBJWRITEPROPS =					conf['OBJWRITEPROPS']
		OBJVIRTUALPATH =				conf['OBJVIRTUALPATH']
		OBJINSTANCED	=				conf['OBJINSTANCED']
		HASDRAPED		=				conf['HASDRAPED']
		LIGHTS_TXT_DIR = 				conf['LIGHTS_TXT_DIR']
		LIGHTS_TEXTURE_DIR			=   conf['LIGHTS_TEXTURE_DIR']
		LIGHT_EDITOR_CONFIG			=	conf['LIGHT_EDITOR_CONFIG']
		facetranspmodes				=  	conf['FACETRANSPMODES']
		facemodes					=	conf['FACEMODES']
		modestosearch				=	conf['MODESTOSEARCH']
		GUIpage						=	conf['GUIPAGE']
		facematchall				=	conf['FACEMATCHALL']
		DEF_TILE_WRAPPER_LAYER 		=	conf['TILE_WRAPPER_LAYER']
		DEF_OBJ_WRAPPER_LAYER 		=	conf['OBJ_WRAPPER_LAYER']
		DEF_METAPIN_LAYER			=	conf['METAPIN_LAYER']
		DEF_METARECT_LAYER			=	conf['METARECT_LAYER']
		DEF_TILE_MESH_LAYER			=	conf['TILE_MESH_LAYER']
		DEF_LIGHT_LAYER				=	conf['LIGHT_LAYER']


		# If the last active page was AGB 2, we switch to AGB 1, since AGB 2 makes no sense without an AGB loaded!
		if GUIpage == 13:
			GUIpage = 3

	except KeyError:
		print "ERROR: ImportConfig() Key error"
	layr.val					=	DEF_METAPIN_LAYER
	layw.val					=   DEF_TILE_WRAPPER_LAYER
	Object_st_propval0.val 		= 	OBJVIRTUALPATH
	Object_tg_objinstanced.val 	= 	int(OBJINSTANCED)
	Object_tg_objhasdraped.val 	= 	int(HASDRAPED)

	if GUIpage == 9 and not g_lighteditor:
		lightEditorStart()

ImportConfig()

def niceFround(x, denom=4):
	""" This rounds x to the nearest multiple of 1/denom and returns it as a string """
	a = round(x * denom, 0)/float(denom)
	if not math.modf(a)[0]:
		return str(int(a))
	return str(a)

def get_sign(x):
	if not x:
		return 1
	return int(x / abs(x))

def xform(v,x):
	return x.rotationPart() * v.co + x.translationPart()

def pluralize(quantity):
	return " was" if quantity == 1 else "s were"

def pluralize2(quantity):
	return "" if  quantity == 1 else "s"

def isGrandParent(child, grandparent):
	while child != None:
		if child == grandparent:
			return True
		child = child.getParent()
	return False

def isParent(child, parent):
	return child.getParent() == parent

def isDepth(child, depth):
	while depth > 0:
		if child == None:
			return False
		child = child.getParent()
		depth = depth - 1
	if child == None:
		return False
	return child.getParent() == None

def getChildren(parent, all):
	children = []
	for ob in all:
		if isParent(ob, parent):
			children.append(ob)
	return children

def getGrandChildren(parent, all):
	children = []
	for ob in all:
		if isGrandParent(ob, parent):
			children.append(ob)
	return children

def getProp(obj, name, default):

	for prp in obj.getAllProperties():
		if prp.name.upper() == name.upper():
			return str(prp.data)
	if obj.getParent() != None:
		return getProp(obj.getParent(), name, default)
	else:
		return default

def setProp(obj, name, value):
	for p in obj.getAllProperties():
		if p.name.upper() == name.upper():
			p.data = value
			break
	else:
		obj.addProperty(name, value)

def setIDProp(ob, name, value):
	"""
		Set a Blender IDProp on an object.
		"""
		# Nested dictionary keys are passed as a single slash-separated string,
		# or, alternatively, a list of strings.
		# e.g. to set ob.properties['first']['second']['third'] = val, do this:
		# setIDProp(ob, "first/second/third", val)
		#  - or -
		# setIDProp(ob, ["first", "second", "third"], val)

	if ob != None:
		props = ob.properties
		toks = name.split('/') if isinstance(name, str) else name
		k = toks.pop()
		while toks:
			t = toks.pop(0)
			if t not in props.keys():
				props[t] = {}
			props = props[t]
		try:
			props[k] = value
		except (RuntimeError, KeyError), detail:
			print "ERROR: setIDProp:", detail, "k:", k

def getIDProp(ob, name, default=0):
	"""
		Get a Blender IDProp from an object.
		"""
	if ob != None:
		toks = name.split('/') if isinstance(name, str) else name
		try:
			return reduce(lambda dct, k: dct[k], toks, ob.properties)
		except (KeyError, TypeError), detail:
			print "ERROR:", detail, ":", name
		except StandardError, detail:
			print "ERROR:", detail, ":", name
		return default

def removeIDProp(ob, name):
	"""
		Remove a blender IDProp from an object.
		"""
	if ob != None:
		try:
			toks = name.split('/') if isinstance(name, str) else name
			del reduce(lambda dct, k: dct[k], toks[:-1], ob.properties)[toks[-1]]
		except (KeyError, TypeError), detail:
			print "ERROR:", detail, ":", name
			return 0
		except StandardError, detail:
			print "ERROR:", detail, ":", name
			return 0
		else:
			return 1

def removeAllIDProps(ob):
	"""
		Remove all Blender IDProps from an object.
		"""
	if ob != None:
		for k in ob.properties.keys():
			del ob.properties[k]

def annotationInsideTile(ann, tile):
	""" Determines if the centre of annotation object 'ann' is contained by quad tile object 'tile' """

	mq = tile.getMatrix('worldspace')
	me = tile.getData(mesh=1)
	if len(me.verts) != 4:
		PupMenu("ERROR: tile %s is not a quad!" % tile.name)
		return 0
	q = []
	for pt in me.verts:
		q.append(xform(pt, mq))
	return Blender.Geometry.PointInQuad2D(Vector(ann.loc), q[0], q[1], q[2], q[3])

def collectAnnotations(annType):
	""" Search for annotations of the specified type (0 -> OBJ, 1 -> FAC, 2 -> TREE)
		located within the boundaries of child tiles of the selected AG top-level wrapper object (AGS, AGB or AGP) and
		attach them to the respective tile's wrapper.
		"""

	scn = Scene.GetCurrent()
	selected = scn.objects.context
	if len(selected) != 1:
		PupMenu("ERROR: More than one item is selected!")
		return 0
	AGwrapper = selected[0]
	AGname = AGwrapper.name

	keys = ['AGS', 'AGB', 'AGP']
	count = 0

	AGtype = AGname[:3]
	if AGtype not in keys:
		PupMenu("ERROR: No such AG type.\n")
		return 0
	for t in scn.objects:
		if t.name.startswith('TILE') and t.getType() == 'Mesh':
			if t.getParent():
				twrapper = t.parent
				if twrapper.parent:
					if twrapper.parent.getName() == AGname:		# This tile is a child of the AG entity.

						# Get the 2D bounding rect if we're an AGB:
						if AGtype == 'AGB':
							xmin = 9999.0
							xmax =-9999.0
							ymin = 9999.0
							ymax = -9999.0
							bb = t.getBoundBox(1)
							for v in bb:
								if v[0] < xmin:
									xmin = v[0]
								if v[0] > xmax:
									xmax = v[0]
								if v[1] < ymin:
									ymin = v[1]
								if v[1] > ymax:
									ymax = v[1]
						# Now look for enclosed, orphan annotations:
						for o in scn.objects:
							if not o.getParent():
								if (annType == 0 and o.getType() == 'Empty' and o.name.startswith('OBJ')) or \
								(annType == 1 and o.getType() == 'Mesh' and o.name.startswith('FAC')) or \
								(annType == 2 and o.getType() == 'Mesh' and o.name.startswith('TREE')):

									#if we're an AGB, check if the potential annotation is inside the BB:
									if AGtype == 'AGB':
										if xmin <= o.LocX <= xmax and ymin <= o.LocY <= ymax:
											twrapper.makeParent([o])
											count += 1
									# else if we're an AGP or an AGS, the quad may be a weird shape, so use this function:
									elif AGtype == 'AGS' or  AGtype == 'AGP':
										if annotationInsideTile(o, t):
											twrapper.makeParent([o])
											count += 1
	return count

def selectFilteredChildren(parent, filter='all'):

	scn = Blender.Scene.GetCurrent()
	children = getChildren(parent, scn.objects)
	for ob in children:
		if filter == 'all':
			ob.sel = 1
		elif filter == 'pins':
			ob.sel = 1 if ob.type == 'Lamp' and ob.name.strip_suffix().endswith('_PT') else 0
		elif filter == 'trees':
			ob.sel = 1 if ob.type == 'Mesh' and ob.name.startswith('TREE') else 0

def selectOBJs():

	sce = Scene.GetCurrent()
	if sce.objects.context:
		pool = sce.objects.context
	else:
		pool = sce.objects
	obs = [ob for ob in pool if ob.type == 'Empty' and ob.name.startswith("OBJ") and not ob.lib]
	sce.objects.selected = obs
	Window.RedrawAll()
	return len(obs)

def selectOBJMeshes(prefixList):
	""" Selects all child meshes of the selected wrappers and deselects the wrappers.
		If a mesh is on a non-visible layer, that layer is made visible.
	"""
	sce = Scene.GetCurrent()
	s_layers = sce.layers

	children = []
	if sce.objects.context:
		pool = sce.objects.context
	else:
		pool = sce.objects
	for ob in [ob for ob in pool if ob.type == 'Empty' and (ob.name[:3] in prefixList) and not ob.lib]:
		children.extend([c for c in getGrandChildren(ob, sce.objects) if c.type != 'Empty'])
	for ch in children:
		for l in ch.layers:
			if l in s_layers:
				break
		else:
			s_layers.append(l)
	sce.layers = s_layers
	sce.objects.selected = children
	Window.RedrawAll()
	return len(children)

def restrictSelectChildren(mode):
	"""
		This prevents/enables selection of OBJ meshes and lamps in the 3d view. Empty wrappers remain selectable
	"""
	sce = bpy.data.scenes.active

	init_selection = []
	if sce.objects.context:
		pool = sce.objects.context
		init_selection = list(sce.objects.context)
	else:
		pool = sce.objects
	children = []

	if mode == 'restrict':
		for ob in [ob for ob in pool if ob.type != 'Empty']:
			ob.restrictSelect = True
		sce.objects.selected = filter(lambda x : x.type == 'Empty', init_selection)
	elif mode == 'enable':
		for ob in [ob for ob in pool if ob.type == 'Empty' ]:
			children.extend([c for c in getGrandChildren(ob, sce.objects)])
		for ch in children:
			ch.restrictSelect = False
		sce.objects.selected = init_selection

	Window.RedrawAll()
	return len(children)

def duplicateOBJHierarchy():
	"""
		Make a shallow copy of all parts of an OBJ hierarchy (wrapper, mesh(es) and lamps), of which at least one part is selected, and offset it by 10m.
	"""
	scn = bpy.data.scenes.active

	children = []
	parents = []

	for ob in scn.objects.context:
		if ob.type != 'Empty':
			while ob.parent != None:
				ob = ob.parent
				if ob.name[:3] in ['OBJ', 'FAC']:
					children.extend(getGrandChildren(ob,scn.objects))
					parents.append(ob)
					break
		elif ob.type == 'Empty' and ob.name[:3] in ['OBJ', 'FAC']:
			parents.append(ob)
			children.extend(getGrandChildren(ob, scn.objects))
	scn.objects.selected = children + parents
	Object.Duplicate()
	scn.objects.selected = parents
	for ob in parents:
		ob.clrParent(2)
		ob.LocY -= 10.0
		ob.makeDisplayList()
	Window.RedrawAll()
	return len(children)

def isStandardName(inStr):
	if not isinstance(inStr, str):
		return False
	if len(inStr) > 12 and inStr[0:4].isdigit() and inStr[4].isalpha() and inStr[5:9].isdigit() and inStr[9].isalpha() and inStr[10:13].isdigit():
		return True
	else:
		return False

def getSortKey(inValue, mode):
	v = inValue

	leaf = v.rpartition("/")[2]
	leaf = leaf.partition(".")[0]
	if isStandardName(leaf):
		splits = [leaf[0:4], leaf[5:9], leaf[10:13]]
		if mode - 1 < len(splits):
			return splits[mode - 1]
		else:
			return splits[0]
	else:
		splits = leaf.split('_')
		if mode - 1 < len(splits):
			term = splits[mode - 1]
			if not term.strip('0123456789.-+'):
				return float(term)
			else:
				return term
		else:
			return splits[0]

def orderImportedBySize(atCursor=True):
	#			 ^
	#			etc					H2
	#---------------------------...>
	#	w0d2	w1d2	d2d2
	#	w0d1	w1d1	w2d1		H1
	#	w0d0	w1d0   	w2d0
	#----------------------------...>
	#	w0d2	w1d2	d2d2
	# 	w0d1	w1d1	w2d1	etc	H0
	# 	w0d0	w1d0   	w2d0
	SPACE_X = 10.0
	SPACE_Y = 10.0

	dbuckets = [15.0, 22.5, 30.0, 37.5, 45.0, 52.5, 60.0]

	def get_depth_bucket(ob):
		if ob[1] <= 15.0:
			return 15.0
		for i in range(1, len(dbuckets)):
			if dbuckets[i-1] < ob[1] <= dbuckets[i]:
				return dbuckets[i]

	scn = bpy.data.scenes.active
	all_obs = []
	all_names = set()
	for ob in scn.objects.selected:
		name_strip = strip_suffix(ob.name)
		if ob.type == 'Empty' and ob.name[:3] == 'OBJ' and not ob.parent:	# We don't want to mess with objs parented to anything else
			kids = getChildren(ob, scn.objects.selected)
			for k in [k for k in kids if k.type == 'Mesh']:
				kname_strip = strip_suffix(k.name)
				if kname_strip not in all_names:
					all_names.add(kname_strip)
					wdh = [0.0, 0.0, 0.0]
					for c in kids:
						if c.type == 'Mesh':
							m = getBuildingMetrics(c, True, False)
							if len(m) == 3:
								wdh = map(max, wdh, m)
					all_obs.append([fractionize(wdh[0], 2, 1), fractionize(wdh[1], 2, 1), round(wdh[2]), ob])
					break
	widths = set()
	for ob in all_obs:
		widths.add(ob[0])
	w_buckets = sorted(list(widths))
	h_buckets = {10:{}, 16:{}, 24:{}, 32:{}, 40:{}, 80:{}, 120:{}, 999:{}}

	for kh in h_buckets.iterkeys():
		for kw in w_buckets:
			h_buckets[kh][kw] = {}
			for kd in dbuckets:
				h_buckets[kh][kw][kd] = []
				print "h_buckets[%s][%s][%s]:" % (str(kh), str(kw), str(kd)), h_buckets[kh][kw][kd]

	for ob in all_obs:
		hb = getHeightBucket(ob[2])
		db = get_depth_bucket(ob)
		print ob[3].name, hb, db
		try:
			h_buckets[hb][ob[0]][db].append(ob)
		except KeyError, detail:
			print 'ERROR:', detail
	offsets_w = [0.0]

	for w in w_buckets:
		offsets_w.append(ceil(offsets_w[-1] + w + SPACE_X))

	offsets_h = []

	for kh in sorted(h_buckets.keys()):
		offsets_d = [SPACE_Y]
		for kd in dbuckets:
			max_depth = 0.0
			for kw in w_buckets:
				max_depth = max(len(h_buckets[kh][kw][kd]) * (kd + SPACE_Y), max_depth)
			offsets_d.append(max_depth)
		offsets_h.append(offsets_d)

	running_h = 0
	for h in offsets_h:
		h[0] = running_h
		for i in range(1, len(h)):
			h[i] = h[i] + h[i-1]
		running_h = h[-1] + SPACE_Y

	if atCursor:
		x0, y0, z0 = Window.GetCursorPos()
	else:
		x0 = 30000.0
		y0 = 30000.0
		for ob in scn.objects.selected:
			x0 = min(x0, ob.LocX)
			x0 = min(y0, ob.LocY)
	x0 = int(x0/10) * 10.0
	y0 = int(y0/10) * 10.0
	z0 = 0.0

	scn.objects.selected = []
	for h, kh in enumerate(sorted(h_buckets.keys())):
		for w, kw in enumerate(sorted(h_buckets[kh].keys())):
			for d, kd in enumerate(sorted(h_buckets[kh][kw].keys())):
				for i, ob in enumerate(h_buckets[kh][kw][kd]):
					ob[3].LocX = x0 + offsets_w[w]
					ob[3].LocY = y0 + offsets_h[h][d] + i * (kd + SPACE_Y)
					ob[3].makeDisplayList()
	Window.RedrawAll()

def moveOBJCentre():

	sce = Scene.GetCurrent()
	obs = [ob for ob in sce.objects.context if ob.type == 'Mesh']
	changed = 0

	for ob in obs:

		# get the mesh data:
		me = ob.getData(mesh=1)

	#	We want to move the obj centre to the middle of the front wall, or more precisely, the frontmost ground-level part of the object, on the ground-level front-back bisector.
	#	We'll use the below-ground verts as our sample set, and assume the walls are vertical.

		if isScaled(ob):
			PupMenu("ERROR: OB %s has scaling.\n" % ob.name)
			Window.RedrawAll()
			return changed

		subzero = []
		xmin = 9999.0
		xmax =-9999.0
		ymin = 9999.0
		ymax = -9999.0
		zmax = -9999.0

		for v in me.verts:
			if v.co[2] <= 0.0:
				subzero.append(v)
		if not subzero:
			PupMenu("WARNING: OB %s doesn't reach the ground! Is it perhaps in the wrong Blender layer?\n" % ob.name)
			return []

		for v in subzero:
			if v.co[0] < xmin:
				xmin = v.co[0]
			if v.co[0] > xmax:
				xmax = v.co[0]
			if v.co[1] < ymin:
				ymin = v.co[1]
			if v.co[1] > ymax:
				ymax = v.co[1]
		dx = xmin + 0.5 * (xmax - xmin)
		dy = ymin

		# move ob:

		ob.LocX += dx
		ob.LocY += dy

		# move verts back:

		for v in me.verts:
			co_new = Vector(v.co[0] - dx, 	v.co[1] - dy, v.co[2])
			v.co = co_new


		if dx or dy:
			changed += 1
	Window.RedrawAll()
	return changed

def alphaToClip():
	pass

def clipToAlpha():
	pass

def getHeightBucket(height):
	buckets = [10, 16, 24, 32, 40, 80, 120, 999]
	for i in range(1, len(buckets)):
		if buckets[i - 1] < height <= buckets[i]:
			return buckets[i]
	if height <= 10:
		return 10
	else:
		return 999

def strip_suffix(name):
	e = name.rfind('.')
	if e == -1:
		return name
	return name[:e]

def strip_variant(name):

	if name[-1].isalpha() and name[-4:-1].isdigit():  	# New naming convention case
		return name[:-1]
	elif 'v' in name[-3:-1] and name[-1].isdigit():		# Old naming convention case
		s = name[:name.rfind('v')]
		if s[-1] == '_':
			return s[:-1]
		else:
			return s
	else:
		return name

def nearUnity(x, tol):
	return x > 1.0 - tol and x < 1.0 + tol

def enumToChar(inVal):
	if not inVal:
		return '_'
	if 0 < inVal < 27:
		return chr(inVal + 64)
	return ''

def charToEnum(inChar):
	if inChar == '_':
		return 0
	x = ord(inChar)
	if 64 < x < 91:
		return x - 64
	return -1

def isValidEnum(x):
	""" OBJ name code letters may include 'A...Z' and '_'"""
	if not type(x) == IntType:
		PupMenu("\nERROR: enum must be an integer!.\n")
		return 0
	if x < 0 or x > 26:
		# enum 0 -> '_', enum 1 -> 'A', enum 2 -> 'B' etc
		PupMenu("\nERROR: enum must be an integer in the range 0 to 26!.\n")
		return 0
	return 1

def isScaled(obj):
	mat=obj.getMatrix('localspace')
	sp = mat.scalePart()
	t = 0.005
	for i in range(3):
		if not nearUnity(sp[i], t):
			return 1
	return 0

def getBuildingMetrics(object, potentialHeight=False, dialogs=True):

	""" For close-fitting object concatenation, we need to know the size of the _ground-level_ bounding rectangle.
		This may be smaller than the overall bounding box, due to overhanging features higher up the building, which should be ignored for fitting purposes.
		A safe bet is to use the bounding rect of the basement geometry, i.e. all the negative-z verts.
		Width, depth and height are returned as a list."""

	if isScaled(object):
		PupMenu("ERROR: OB %s has scaling.\n")
		return []

	xmin = 9999.0
	xmax =-9999.0
	ymin = 9999.0
	ymax = -9999.0
	zmax = -9999.0
	zmin = 9999.0
	ground_verts = 0

	me = object.getData(mesh=1)

	for v in me.verts:
		if v.co[2] > zmax:
			zmax = v.co[2]
		if v.co[2] < zmin:
			zmin = v.co[2]

		if v.co[2] <= 0.0:
			ground_verts += 1
			if v.co[0] < xmin:
				xmin = v.co[0]
			if v.co[0] > xmax:
				xmax = v.co[0]
			if v.co[1] < ymin:
				ymin = v.co[1]
			if v.co[1] > ymax:
				ymax = v.co[1]
	if not ground_verts:		# no mesh below ground level - suggests that this is a stray second LOD mesh.
		if dialogs:
			PupMenu("WARNING: OB %s doesn't reach the ground! Is it perhaps in the wrong Blender layer?\n" % object.getName)
		return []
	height = zmax - zmin - 2 if potentialHeight else zmax
	return [xmax - xmin, ymax - ymin, height]

def makeOBJName(shape, part, variant, metrics):

	""" This constructs an OBJ name from three code enumerations and a list of three floats.
		The format is:	wwwwSddddPhhhV	where wwww, dddd, and hhh are leading-zero padded integers and S, P and V are single alpha chars."""

	if len(metrics) != 3:
		PupMenu("\nERROR: Wrong number of metrics args.\n")
		return ""
	w, d, h = metrics

	name = ""

	#	Convert width to integer decimetres and pad with leading zeroes to fill four digits:
	name += str(int(round(min(w, 999.0) * 10.0))).zfill(4)

	#	Convert "shape" enum to alpha char:
	if not isValidEnum(shape):
		return ""
	name += chr(shape + 65)

	#	Convert depth to integer decimetres and pad with leading zeroes to fill four digits:
	name += str(int(round(min(d, 999.0) * 10.0))).zfill(4)

	#	Convert "part" enum to alpha char:
	if not isValidEnum(part):
		return ""
	name += chr(part + 65)

	#	Round "height" to nearest integer and pad with leading zeroes to fill three digits:
	name += str(int(round(min(h, 999.0)))).zfill(3)

	#	Convert "variant" enum to alpha char:
	if not isValidEnum(variant):
		return ""
	name += chr(variant + 65)

	return name

def incrementVariant(name):
	""" This just increases the variant letter suffix by one """
	s = ord(name[-1])
	return name[:-1] + chr(s + 1)

def refreshProps(obj, propName, propValue, overwrite):
	""" Check for existing property: if yes - update it, if no - add new.
		Return '-1' if we update an existing prop, '0' if we do nothing, and '1' if we add a new one."""
	exists = 0
	for p in obj.getAllProperties():
		if p.getName().upper() == propName.upper():
			exists = 1
			if overwrite:
				if propValue == "0" or not propValue:
					obj.removeProperty(p)
				else:
					p.setData(propValue)
				return -1
			else:
				return 0
	if not exists:
		obj.addProperty(propName, propValue, 'STRING')
		return 1

def stripProperties():
	""" Remove all properties from selected OBJ wrappers, apart from 'rname' and 'vname' properties
	"""
	sce = Scene.GetCurrent()
	obs = [ob for ob in sce.objects.context if not ob.lib]
	changed = 0
	propsgone = 0
	for ob in obs:
		props = ob.getAllProperties()
		flag = 0
		for p in props:
			if p.name.lower() not in ['rname', 'vname']:
				ob.removeProperty(p)
				propsgone += 1
				flag = 1
		if flag:
			changed += 1
	return [propsgone, changed]

def applyOBJProps():

	""" This updates all selected OBJ wrappers/parents with the properties selected in the GUI."""

	global OBJPROPSOVERWRITE
	global OBJWRITEPROPS
	global objprops
	global objprops2

	global objpropkeys

	sce = Scene.GetCurrent()
	obs = [ob for ob in sce.objects.selected if ob.type == 'Empty']
	for ob in obs:
		for k, v in objprops2.iteritems():
			if v:
				retval = refreshProps(ob, k, v, OBJPROPSOVERWRITE)
				if retval == -1:
					tasksdone['objprops_added'] += 1
				elif not retval:
					tasksdone['objprops_untouched'] += 1
				elif retval == 1:
					tasksdone['objprops_updated'] += 1
		for k, v in objprops.iteritems():
			if v and not k == 'Virtual Path':
				retval = refreshProps(ob, k, v, OBJPROPSOVERWRITE)
				if retval == -1:
					tasksdone['objprops_added'] += 1
				elif not retval:
					tasksdone['objprops_untouched'] += 1
				elif retval == 1:
					tasksdone['objprops_updated'] += 1

		for i, k in enumerate(objpropkeys):
			if i == Object_mn_objpropname3.val:
				retval = refreshProps(ob, objpropkeys[k], "", OBJPROPSOVERWRITE)
				if retval == -1:
					tasksdone['objprops_added'] += 1
				elif not retval:
					tasksdone['objprops_untouched'] += 1
				elif retval == 1:
					tasksdone['objprops_updated'] += 1
		for i, k in enumerate(objpropkeys):
			if i == Object_mn_objpropname4.val:
				retval = refreshProps(ob, objpropkeys[k], "", OBJPROPSOVERWRITE)
				if retval == -1:
					tasksdone['objprops_added'] += 1
				elif not retval:
					tasksdone['objprops_untouched'] += 1
				elif retval == 1:
					tasksdone['objprops_updated'] += 1

def applyOBJSettings():

	global DEF_OBJ_WRAPPER_LAYER
	global OBJVIRTUALPATH
	global OBJRENAME
	global OBJSHAPE
	global OBJPART
	global OBJVARIANT
	global AUTOVARIANT
	global VARSTOLIB
	global OBJPROPSOVERWRITE
	global OBJWRITEPROPS
	global EXPORTMODE
	global USEHEIGHTBUCKETS
	global BUCKETSVIRTUALONLY
	global tasksdone
	global objprops
	global objprops2
	#PupMenu("rename: %d, autovar: %d, usebuckets: %d" % (OBJRENAME, AUTOVARIANT, USEHEIGHTBUCKETS))
	try:
		PRPCSTM1NM = ""
		PRPCSTM2NM = ""
		PRPCSTM1 = ""
		PRPCSTM2 = ""

		obs1 = []			# selected objects on layer 1
		obs2 = []			# selected objects on layer 2
		scn = Blender.Scene.GetCurrent()
		Window.SetActiveLayer(1<<(DEF_OBJ_WRAPPER_LAYER-1))
		Window.RedrawAll()

		for ob in scn.objects.context:
			if ob.getType() == "Mesh":
				if 1 in ob.layers:
					obs1.append(ob)
				elif 2 in ob.layers:
					obs2.append(ob)
		for ob2 in obs2:
			w = d = h = 0.0
			loc2 = ob2.getLocation("worldspace")
			mets = []
			mets = getBuildingMetrics(ob2)
			if mets:

				w, d, h = mets
				heightbucket = getHeightBucket(h)
				metstring = niceFround(w, 4) + " " + niceFround(d, 2) + " " + niceFround(h, 1)
				found = 0
				for p in ob2.getAllProperties():
					if p.name == 'METRICS':
						p.setData(metstring)
						found = 1
						break
				if not found:
					ob2.addProperty("METRICS", metstring, "STRING")

				hb_name = ""
				real_name = ""
				virtual_name = ""
				my_variant = OBJVARIANT
				if OBJRENAME:
					# Generate an automagic name from the mesh metrics and user prefs

					if AUTOVARIANT:
						# We search the scene for other objs with the same metrics as ours.
						# If found and it turns out that we share the same mesh data, we just copy the other objs name and blender appends its .xxx suffix.
						# If found and we have unique mesh data, we find the lowest available variant letter and give ourselves a unique name by adding the variant as our own suffix.
						# Unless VARS > LIB has been checked, we use the non-suffixed name as the virtual name.
					#	PupMenu("break2!")
						nameinit = makeOBJName(OBJSHAPE, OBJPART, 0, mets)
					#	PupMenu("break3!")
						# get ALL scene objects, not just those selected:
						allobs = [ob for ob in scn.objects if not ob.lib]
						allmobs = []

						for ob in allobs:
							if ob.getType() == 'Mesh' and (2 in ob.layers) and (ob.getLocation("worldspace") != loc2):		# All mesh objects in layer 2, excluding ourselves.
								allmobs.append(ob)

						# Build a list of all objects whose names contain our width and depth (the first nine characters) This means that they share our metrics and could be variants:
						alike = []
						for o in allmobs:

							if nameinit[:9] in o.getName():
								# now check if our heights match:
								if USEHEIGHTBUCKETS:
									if heightbucket == getHeightBucket(float(o.getName()[10:13])):
										alike.append(o)
								else:
									if nameinit[10:13] == o.getName()[10:13]:
										alike.append(o)

						# Now check to see if we actually share mesh data:
						dupe = 0
						for al in alike:
							if al.getData(name_only=1) == ob2.getData(name_only=1):		# shared mesh means we're a dupe, so copy the other object's name and go home happy:
								ob2_name = strip_suffix(al.getName())
								dupe = 1
								break

						# If none of the matches are dupes, we need to give ourselves a unique variant suffix.
						# We search for the lowest available letter by building a list of those already used:
						if not dupe:
							vrs = []
							for a in alike:
								nm = a.getName()
								vrs.append(ord(nm[-1]))
							v = ord('A')
							while v in vrs:
								v += 1
							my_variant = v - 64
							# Strip our provisional name of its suffix and add the new one which we now know is unique:
							ob2_name = nameinit[:-1] + chr(v)

					else:
						# If AUTOVARIANT has been disabled by the user, we use a user-defined variant letter:
						ob2_name = makeOBJName(OBJSHAPE, OBJPART, my_variant, mets)
					ob2.setName(ob2_name)

					# Thats the name for the mesh object, now we need the name for the wrapper:
					newmets = [w, d, float(heightbucket)]
					hb_name = makeOBJName(OBJSHAPE, OBJPART, my_variant, newmets )

					if USEHEIGHTBUCKETS:
						virtual_name =	hb_name
						if not BUCKETSVIRTUALONLY:
							real_name = hb_name
						else:
							real_name = ob2_name
					else:
						virtual_name = ob2_name
						real_name = ob2_name
					if not VARSTOLIB:
						virtual_name = strip_variant(virtual_name)

				else:	# Use the name already on the object
					real_name = strip_suffix(ob2.getName())
					virtual_name = strip_variant(real_name)
				if EXPORTMODE == 0:
					wrapper_name = "OBJ" + virtual_name
				elif EXPORTMODE == 1:
					wrapper_name = "OBJ" + real_name
				if len(wrapper_name) > 17:
					PupMenu("\nERROR: Mesh %s has an overlong name.\n" % wrapper_name)
					break

				# Check for existing wrapper, add a new one if necessary:
				wrapper = ob2.parent
				if not wrapper:
					wrapper = scn.objects.new("Empty")

				wrapper.setName(wrapper_name)
				wrapper.setLocation(loc2)
				wrapper.drawSize = 2.0
				wrapper.makeParent([ob2])
				print ("wrapper name: %s" % wrapper.name)
				# wrapper properties

				retval = refreshProps(wrapper, "RNAME",  real_name, OBJPROPSOVERWRITE)
				if retval == -1:
					tasksdone['objprops_added'] += 1
				elif not retval:
					tasksdone['objprops_untouched'] += 1
				elif retval == 1:
					tasksdone['objprops_updated'] += 1

				if objprops['Virtual Path']:
					OBJVIRTUALPATH = objprops['Virtual Path']
					if not OBJVIRTUALPATH.endswith('/'):
						OBJVIRTUALPATH += '/'

				retval = refreshProps(wrapper, "VNAME", OBJVIRTUALPATH + virtual_name, OBJPROPSOVERWRITE)
				if retval == -1:
					tasksdone['objprops_added'] += 1
				elif not retval:
					tasksdone['objprops_untouched'] += 1
				elif retval == 1:
					tasksdone['objprops_updated'] += 1


				# Look for any possible LOD_1 detail mesh objects on layer 1
				for i, ob1 in enumerate(obs1):
					if ob1.getLocation("worldspace") == loc2:
						ob1.setName(real_name + "LD1")
						wrapper.makeParent([ob1])
						del obs1[i]
						break
			else:
				PupMenu("\nERROR: Couldn't get OBJ metrics\n")
	except (RuntimeError, IOError):
		PupMenu("\nError in applyOBJSettings()\n")
		return -1

	scn.update()
	Draw()
	return 0

def buildObjImageMenu():
	"""
		Do this only when the menu button is used, rather than per GUI redraw.
	"""
	global OBJIMAGELIST, objimagemenuname
	OBJIMAGELIST = Blender.Image.Get()
	objimagemenuname = "Image name %t|REFRESH MENU %x998|Texture Menu  %x999"
	for i, v in enumerate(OBJIMAGELIST):
		objimagemenuname += ("|%s %%x%d" % (v.name, i))

def assignImageToSelected(image):

	in_emode = Window.EditMode()
	if in_emode:
		Window.EditMode(0)

	sce = Scene.GetCurrent()
	changed = 0
	obs = [ob for ob in sce.objects.context if not ob.lib]
	meshes = set()
	for obj in obs:
		if obj.getType() == 'Mesh':
			msh = obj.getData(mesh=1)

			# This avoid needlessly repeating operations on shared meshes:
			if msh in meshes:
				changed += 1
				continue
			meshes.add(msh)

			if msh.faceUV == 0:
				msh.faceUV = 1
			orig_uvlayer = msh.activeUVLayer
			for uvlayer in msh.getUVLayerNames():
				msh.activeUVLayer = uvlayer
				for f in msh.faces:
					f.image = image
			msh.activeUVLayer = orig_uvlayer
			obj.select(1)
			Window.EditMode(1)
			Window.EditMode(0)
			changed += 1
			sce.update()
	if in_emode:
		Window.EditMode(1)
	Window.RedrawAll()
	return changed

def getObImageName(ob):
	"""
		Returns the name of the first image it finds on the passed-in object's mesh.
	"""
	if ob.type == 'Mesh':
		me = ob.getData(mesh=1)
		if me.faceUV:
			for f in [f for f in me.faces if f.image]:
				return f.image.name
	return None

def isImageUser(ob, imageToMatch):
	"""
		Returns True if the passed-in mesh is a user of the passed-in image.
	"""
	if ob.type != 'Mesh':
		return False
	me = ob.getData(mesh=1)
	if me.faceUV:
		for f in [f for f in me.faces if f.image]:
			if f.image == imageToMatch:
				return True
	return False

def showWithLIT(litImageName=None):
	"""
		This attempts to find a LIT counterpart to the selected object's daytime texture and,
		if successfull, substitute it in all mesh users of the original image.
		It can also be called with an argument specifying the image name to use.
	"""
	global OBJIMAGELIST
	scn = Blender.Scene.GetCurrent()

	obact = scn.objects.active
	my_image_name = getObImageName(obact)

	# If we're called with no argument, try to guess the LIT name:
	if not litImageName:
		e = my_image_name.find('.')
		if e != -1:
			my_image_name_clean = my_image_name[:e]

		for s in ['_ALB', 'ALB']:
			if my_image_name_clean.upper().endswith(s):
				my_image_name_clean = my_image_name_clean[:-len(s)]

		# Look for a lit variant of our current image:
		all_names = [tex.name for tex in OBJIMAGELIST]
		for tex in all_names:
			if my_image_name_clean.upper() in tex.upper() and 'LIT' in tex.upper():
				my_lit_name = tex
				break
		else:
			my_lit_name = None
			PupMenu("Can't guess the LIT texture%t")
			return False
		print "my lit name:", my_lit_name
	else:
		my_lit_name = litImageName

	# Find all the obs which share our albedo:
	image_obs = []
	for ob in [ob for ob in scn.objects if ob.type == 'Mesh']:
		if getObImageName(ob) == my_image_name:
			image_obs.append(ob)

	# Cache the daylight texture
	OBJIMAGE_ALB = Image.Get(my_image_name)

def getImageFromOb():
	"""
		Gets the name of the texface image of the active mesh.
	"""
	global objimagemenuname, OBJIMAGELIST

	in_emode = Window.EditMode()
	if in_emode:
		Window.EditMode(0)

	if not objimagemenuname:
		buildObjImageMenu()

	scn = Blender.Scene.GetCurrent()
	obact = scn.objects.active
	try:
		if obact and obact.type == 'Mesh':
			my_image_name = getObImageName(obact)

			if my_image_name:
				# Display this name in the menu: find its index by searching for its name in the menuname string.
				i = objimagemenuname.find(my_image_name)
				if i != -1:
					j = objimagemenuname.find('%x', i)
					j += 2
					f = ""
					while (len(objimagemenuname) - 1 > j) and objimagemenuname[j].isdigit():
						f += objimagemenuname[j]
						j += 1
					if f:
						Object_mn_objimage.val = int(f)
				return my_image_name
			else:
				PupMenu("Could not get image.%t")
				return None
		else:
			PupMenu("No active mesh object.%t")
			return None

	except (ValueError, RuntimeError):
		PupMenu("Error: encountered a problem finding an image.%t")
		print "Error: encountered a problem finding an image."
		return None

def selectSameImageObs(imageName):
	"""
		Finds and selects all meshes in the current scene which use the image whose name is passed in.
	"""
	scn = Blender.Scene.GetCurrent()
	in_emode = Window.EditMode()
	if in_emode:
		Window.EditMode(0)

	image_obs = []
	for ob in [ob for ob in scn.objects if ob.type == 'Mesh']:
		if getObImageName(ob) == imageName:
			image_obs.append(ob)

	if len(scn.objects.selected) > 1:
		rep = PupMenu("Add to current selection? :%t|    NO|  YES")
		if rep == 2 :
			obs_sel = set(scn.objects.selected)
			scn.objects.selected = list(obs_sel | set(image_obs))
		else:
			scn.objects.selected = image_obs
	else:
		scn.objects.selected = image_obs

	if in_emode:
		Window.EditMode(1)
	Window.RedrawAll()
	print len(image_obs)
	return len(image_obs)

def searchFaceModes(searchIn="selected"):
	"""
		Search for specified mesh face modes in multiple obs.
	"""
	#
	# if searchIn == 'selected', we search in selected mesh objects.
	# if searchIn == 'visible', we search in all visible mesh objects.
	#
	global facemodes, facetranspmodes, modestosearch, facematchall
	scn = Blender.Scene.GetCurrent()

	#Set face select mode in the 3d window to keep face selections tidy.
	Mesh.Mode(Mesh.SelectModes['FACE'])
	in_emode = Window.EditMode()
	if in_emode:
		Window.EditMode(0)

	if searchIn == 'selected':
		obs = [ob for ob in scn.objects.context if ob.type == 'Mesh']
		if not obs:
			PupMenu("No meshes are selected to operate on!%t")
			return
	elif searchIn == 'visible':
		obs = [ob for ob in scn.objects if ob.type == 'Mesh' and any([item in scn.layers for item in ob.layers])]
		if not obs:
			PupMenu("No visible meshes to operate on!%t")
			return
	else:
		return

	# Build a list of key/value pairs of the search terms we're interested in
	# Preconvert the key strings into blender enumerations for speed
	match_modes = []
	for k, v in [(k, v) for (k, v) in facemodes.iteritems() if modestosearch[k]]:
		match_modes.append([Mesh.FaceModes[k], v])

	# check for facetranspmode:
	if modestosearch['TRANSP']:
		for k, v in facetranspmodes.iteritems():
			# These modes are mutually exclusive, so we just look for the first one with value = True.
			if v:
				# preconvert to enum, as before:
				match_transp = Mesh.FaceTranspModes[k]
				break
	else:
		match_transp = None

	if not match_modes and match_transp == None:
		PupMenu('Please select one or more face modes to search for, using the "target" buttons.%t')
		return

	for ob in obs:
		ob.select(1)
		deselect_ob = 1
		me = ob.getData(mesh=1)
		if me.faceUV:
			for f in me.faces:
				if facematchall:
					# We want only faces whose modes match ALL those enabled for search:
					# check for matching facemodes
					for kv in match_modes:
						if bool(f.mode&kv[0]) != bool(kv[1]):
							f.sel = 0
							break
					else:
						# if all our enabled facemodes have matched, move onto the facetranspmode, if enabled:
						if modestosearch['TRANSP'] and  f.transp != match_transp:
							f.sel = 0
						else:
							f.sel = 1
							deselect_ob = 0
				else:
					# In this case, we only care if ANY of the enabled modes match:
					for kv in match_modes:
						if bool(f.mode&kv[0]) == bool(kv[1]):
							f.sel = 1
							deselect_ob = 0
							break
					else:
						if modestosearch['TRANSP'] and f.transp == match_transp:
							f.sel = 1
							deselect_ob = 0
						else:
							f.sel = 0
		if deselect_ob:
			ob.select(0)

	if in_emode:
		Window.EditMode(1)
	Window.RedrawAll()

def findMixedAlpha():
	"""
		Search for meshes having both alpha and clip-alpha faces.
	"""
	global modestosearch, facetranspmodes
	# Do this by searching first for all objs with 'alpha' faces, and caching them
	# Search again this time for objects with 'clip alpha' faces, and find the union of the two searches.

	# Set up for searches - dlear all targets except transparency:
	for k in modestosearch.iterkeys():
		modestosearch[k] = 0
	modestosearch['TRANSP'] = 1

	scn = Blender.Scene.GetCurrent()

	# Set up for searching alpha faces:
	facetranspmodes['SOLID'] = 0
	facetranspmodes['ALPHA'] = 1
	facetranspmodes['CLIP']  = 0

	# Deselect all objects to start:
	scn.objects.selected = []
	# Do the alpha search, this will leave any alpha obs selected. Make a list copy of them:
	searchFaceModes('visible')
	alpha_obs = list(scn.objects.selected)

	# Now repeat for alpha clip faces:
	facetranspmodes['SOLID'] = 0
	facetranspmodes['ALPHA'] = 0
	facetranspmodes['CLIP']  = 1
	refreshMeshModeButtons()

	scn.objects.selected = []
	searchFaceModes('visible')

	# Now deselect any objects in the clip search result which aren't also in the alpha result:
	for ob in scn.objects.selected:
		if ob not in alpha_obs:
			ob.select(0)

def setDefaultFaceModes():
	"""
		Sets all faces of selected objects to defaullt face modes.
		Does not effect transparency modes.
	"""
	global facemodes, facetranspmodes, modestosearch, tasksdone
	faces_changed = 0
	in_emode = Window.EditMode()
	if in_emode:
		Window.EditMode(0)
	Mesh.Mode(Mesh.SelectModes['FACE'])
	scene = Scene.GetCurrent()

def refreshMeshModeButtons():
	global Object_tg_objfmtiles, Object_tg_objfmtwoside, Object_tg_objfmdynamic, Object_tg_objfmshadow,  Object_tg_objfminvisible
	global Object_tg_objftmsolid, Object_tg_objftmalpha, Object_tg_objftmclip
	global Object_tg_objfacemodesel
	"""
		This just updates the GUI buttons after we've messed with the vars in the background
	"""
	Object_tg_objfmtiles.val		= facemodes['TILES']
	Object_tg_objfmtwoside.val		= facemodes['TWOSIDE']
	Object_tg_objfmdynamic.val		= facemodes['DYNAMIC']
	Object_tg_objfmshadow.val		= facemodes['SHADOW']
	Object_tg_objfminvisible.val	= facemodes['INVISIBLE']

	Object_tg_objftmsolid.val		= facetranspmodes['SOLID']
	Object_tg_objftmalpha.val		= facetranspmodes['ALPHA']
	Object_tg_objftmclip.val		= facetranspmodes['CLIP']

	Object_tg_objfacemodesel[0].val	= modestosearch['TILES']
	Object_tg_objfacemodesel[1].val	= modestosearch['TWOSIDE']
	Object_tg_objfacemodesel[2].val	= modestosearch['DYNAMIC']
	Object_tg_objfacemodesel[3].val	= modestosearch['SHADOW']
	Object_tg_objfacemodesel[4].val	= modestosearch['INVISIBLE']
	Object_tg_objfacemodesel[5].val	= modestosearch['TRANSP']

def doFaceModes(action):
	"""
		Here we perform all face and transparency mode modification actions
	"""
	global facemodes, facetranspmodes, modestosearch, tasksdone
	faces_changed = 0
	in_emode = Window.EditMode()
	if in_emode:
		Window.EditMode(0)
	Mesh.Mode(Mesh.SelectModes['FACE'])
	scene = bpy.data.scenes.active

	# Build a list of key/value pairs of the face modes we're interested in
	# Preconvert the key strings into blender enumerations for speed
	apply_modes = []
	for k, v in [(k, v) for (k, v) in facemodes.iteritems() if modestosearch[k]]:
		apply_modes.append([Mesh.FaceModes[k], v, k])
	print 'apply_modes', apply_modes
	apply_transp = None
	if modestosearch['TRANSP']:
		for k, v in facetranspmodes.iteritems():
			if v:
				apply_transp = [Mesh.FaceTranspModes[k], k] 	# preconvert to enum, as before. Also include key string.
				break

	# This just translates a face transparency mode enum from blender back into its key string:
	facetranspmodes_inv = {Mesh.FaceTranspModes['SOLID']:'SOLID', Mesh.FaceTranspModes['ALPHA']:'ALPHA', Mesh.FaceTranspModes['CLIP']:'CLIP'}

	# For SETTING fade modes, we're always working on visible selected obs
	obs = [ob for ob in scene.objects.context if ob.type == 'Mesh']
	for ob in obs:
		me = ob.getData(mesh=1)
		if me.faceUV:
			for kvk in apply_modes:
				faces_changed = 0
				for f in [f for f in me.faces if f.sel]:

					if (kvk[1] and not f.mode&kvk[0]):
						f.mode |= kvk[0]
						tasksdone[kvk[2]] += 1
						faces_changed = 1

					elif (f.mode&kvk[0] and not kvk[1]):
						f.mode &= ~kvk[0]
						tasksdone[kvk[2]] -= 1
						faces_changed = 1

				if faces_changed:
					if kvk[2]	== 'TILES':
						tasksdone['tilesobjs']	   += 1
					elif kvk[2] == 'DYNAMIC':
						tasksdone['dynamicobjs']   += 1
					elif kvk[2] == 'TWOSIDE':
						tasksdone['twosideobjs']   += 1
					elif kvk[2] == 'SHADOW':
						tasksdone['shadowobjs']    += 1
					elif kvk[2] == 'INVISIBLE':
						tasksdone['invisibleobjs'] += 1

			if apply_transp != None:
				faces_changed = 0
				for f in [f for f in me.faces if f.sel]:
					if 	f.transp != apply_transp[0]:
						try:
							tasksdone[facetranspmodes_inv[f.transp]] -= 1	# decrement the face counter of the current alpha mode
							f.transp  = apply_transp[0]
							tasksdone[apply_transp[1]] += 1				# increment the face counter of the NEW alpha mode
							faces_changed = 1
						except (KeyError, ValueError, RuntimeError):
							print "Unexpected value for f.transp:", f.transp
				if faces_changed:
					tasksdone['transpobjs'] += 1
	if in_emode:
		Window.EditMode(1)
	Window.RedrawAll()

def multiObjectFaceSelection(action):
	"""
		Selects all faces in visible selected obs.
	"""
	scn = Blender.Scene.GetCurrent()
	Mesh.Mode(Mesh.SelectModes['FACE'])
	in_emode = Window.EditMode()
	if in_emode:
		Window.EditMode(0)

	obs = scn.objects.context
	for ob in [ob for ob in obs if ob.type == 'Mesh']:
		me = ob.getData(mesh=1)
		for f in me.faces:
			if action   == 'select':
				f.sel = 1
			elif action == 'deselect':
				f.sel = 0

	if in_emode:
		Window.EditMode(1)
	Window.RedrawAll()

def isQuadOrFgon(ob):
	"""
		Returns True if the passed-in mesh-type object is a single quad or a Blender fgon
	"""
	me = ob.getData(mesh=1)
	num_faces = len(me.faces)
	if num_faces == 1 and len(me.verts) == 4:
		return True
	elif num_faces == len([ed for ed in me.edges if ed.flag & Mesh.EdgeFlags.FGON]) + 1:
		return True
	else:
		return False

def getSelection():
	"""
		Get selected objects and make a L -> R location-sorted list.
	"""
	global  obsorted, sce
	sce = Scene.GetCurrent()
	obsel = [[ob.LocX, ob.name, ob] for ob in sce.objects.context if not ob.lib]

	# Check that the selected objects are either all empties, or all legal tile meshes (quads or fgons).
	num_empties = 0
	num_meshes  = 0

	for loc, nm, ob in obsel:
		if ob.type == "Mesh":
			if num_empties:
				sce.objects.selected = [ob]
				return "multiple_types"
			elif not isQuadOrFgon(ob):
				sce.objects.selected = [ob]
				return "bad_mesh"
			else:
				num_meshes += 1
		elif ob.type == "Empty":
			if num_meshes:
				sce.objects.selected = [ob]
				return "multiple_types"
			else:
				num_empties += 1
		else:
			return "bad_type"

	if num_meshes and not num_empties:
		obsorted[:] = sorted(obsel)
		return "mesh_type"
	elif num_empties and not num_meshes:
		obsorted[:] = sorted(obsel)
		return "empty_type"

def renameTiles(initID):

	""" Rename tiles from left to right, incrementing the ID number by 1 each tile.	"""

	global obsorted, sce
	tile_number = 0
	i = initID

	for obsort in obsorted:
		tile_id = str(i + tile_number)
		newname = "TILE" + tile_id
		obsort[2].name = newname
		meshdat = obsort[2].getData(mesh=1)
		meshdat.name = "MESH" + tile_id
		obsort[1] = newname
		tile_number += 1

		if len(meshdat.verts) != 4:
			PupBlock("ERROR MESSAGE",["%s is not a quad!" % newname])
			sce.objects.selected = []
			obsort[2].select(1)
			return tile_number

		# Check to see if we already have a parent "empty", if so, rename that too.
		prt = obsort[2].parent
		if not prt:
			continue
		prt.setName(tile_id)
	sce.update()

	return tile_number

def moveTileCentre():

	global obsorted
	changed = 0

	for ob in obsorted:

		# get the mesh data:
		me = ob[2].getData(mesh=1)
		bbox = ob[2].getBoundBox(1)

	#	we want to move the tile centre to the bottom left hand corner of the tile.
	# 	So move the tile to the bottom LH corner of its bounding box, then move the verts back again.
	# 	these are the xy compents of the vector to restore the position of the verts:
	# 	we retain z=0
		dx = ob[2].loc[0] - bbox[0][0]
		dy = ob[2].loc[1] - bbox[0][1]

		# move ob:
		ob[2].setLocation(bbox[0][0], bbox[0][1], 0.0)
		ob[0] = bbox[0][0]

		# move verts back:
		for v in me.verts:
			v.co.x += dx
			v.co.y += dy
			v.co.z = 0.0
		if dx or dy:
			changed += 1
	return changed

def wrapTiles(layer):

	# Set the active layer to the user-defined destination for wrappers
	global sce, obsorted

	metalayer = layer - 1
	Window.SetActiveLayer(1<<metalayer)
	Window.RedrawAll()
	wrapped = 0
#	tiles = [[ob.LocX, ob.name, ob] for ob in sce.objects.context if not ob.lib]
#	srtd = sorted(tiles)
#	for obj in srtd:
	for obj in obsorted:
		prt = obj[2].parent
		if prt:
			continue
		wrapper = sce.objects.new("Empty")
		tile_id = obj[1][4:]
		wrapper_name = tile_id
		if len(tile_id) > 21:
			print "ERROR: Tile %s has an overlong name.\n" %obj[1]
			PupMenu("\nERROR: Tile %s has an overlong name.%t" %obj[1])
		else:
			wrapper.setName(tile_id)
			wrapper.setLocation(obj[2].getLocation("worldspace"))
			wrapper.drawSize = 2.0
			wrapper.makeParent([obj[2]])
			wrapped += 1
	sce.update()
	return wrapped

def addAGStringWrapper(layer):

	""" add an empty object at the 3d cursor x,y position with z=0. """

	sce = Blender.Scene.GetCurrent()
	metalayer = layer - 1

	Window.SetActiveLayer(1<<metalayer)
	Window.RedrawAll()
	cursor = Window.GetCursorPos()
	cursor[2] = 0.0
	wrapper = sce.objects.new("Empty")
	wrapper.setName("AGSnew_name")
	wrapper.setLocation(cursor)
	wrapper.drawSize = 4.0

	wrapper.addProperty("VNAME", 		"", 			'STRING')
	wrapper.addProperty("TILE_IDS", 	1, 				'BOOL'  )
	wrapper.addProperty("ROTATION", 	"0", 			'STRING')
	wrapper.addProperty("CREASE_ANGLE", "0", 			'STRING')
	wrapper.addProperty("LAYER_GROUP", 	"roads -1", 	'STRING')
	wrapper.addProperty("path", 		"objects", 		'STRING')
	wrapper.addProperty("VEGETATION",	"",				'STRING')
	wrapper.addProperty("FILL_LAYER",	"",				'STRING')

	# Now look for any other selected orphan empty objects in the scene, and parent them to the one we just added.

	selected_empties = [ob for ob in sce.objects.selected if ob.type == 'Empty' and not ob.parent]
	wrapped_tile_meshes	= [tm for tm in sce.objects if tm.type == 'Mesh' and tm.name.startswith('TILE') and tm.parent]
	candidates = [t.parent for t in wrapped_tile_meshes if t.sel or t.parent in selected_empties]
	wrapper.makeParent(candidates)
	sce.update()
	return 1

def placeOnePin(location, name, datablock, layer):

	"""  Create and place a new 'pin' lamp. For better performance, we link to a shared lamp datablock. """

	pin = Blender.Object.New('Lamp')
	pin.link(datablock)
	sce.link(pin)
	pin.setName(name)
	pin.setLocation(location)
	pin.layers = [layer]
	return pin

def doStopPins(A, B, spacing, parent, datablock, layer):

	""" Two points are passed in - we divide the distance between them by the requested spacing distance,
	then round up to determine how many pins to use. Dividing the original distance by this number gives us equi-spaced pins
	at roughly the desired spacing. We return the number of pins placed."""

	space = spacing
	length = math.sqrt((A[0] - B[0])*(A[0] - B[0]) + (A[1] - B[1])*(A[1] - B[1]))
	num_points = math.ceil(length/space)
	if not num_points:
		num_points = 1
	ratio = 1.0/num_points
	prt = parent
	i = 0
	while (i < num_points):
		stoppin = placeOnePin([A[0] + (B[0] - A[0]) * ratio * i, A[1] + (B[1] - A[1]) * ratio * i, 0.0], "PINSTOP_PT", datablock, layer)
		prt.makeParent([stoppin])
		i+=1
	return i

def doMetaPins(wantStopPins, wantAnchorPins, wantPivotPins, wantGroundPins, wantShareY, layer):

	global sce
	global mrgr, mrgf, mrgl, mrgb, mrga, spac

	counter = [0, 0, 0, 0, 0, 0]
	# Set the active layer to the user-defined destination for pins

	metalayer = layer - 1
	Window.SetActiveLayer(1<<metalayer)
	Window.RedrawAll()

	# Next, check to see if there are already any pin objects in the scene...
	# if yes, we can make shallow copies to avoid bogging blender down, otherwise we make one.

	pindata = None
	try:
		pindata = Blender.Lamp.Get('XPToolsPinData')
		print 'XPToolsPinData found.'
	except:
		print 'XPToolsPinData not found. Creating new.'
		pindata = Blender.Lamp.New('Lamp', 'XPToolsPinData')
		# strip it down to bare minimum features for better blender performance:
		pindata.mode |= Lamp.Modes["Layer"]
		pindata.mode |= Lamp.Modes["NoDiffuse"]
		pindata.mode |= Lamp.Modes["NoSpecular"]
		pindata.setDist(0.01)
		pindata.energy = 0

	# ojs = [o.name for o in sce.objects if not o.lib]
# 	anypin = "_PT"
# 	exists = 0
# 	for oj in ojs:
# 		if anypin in oj:
# 			exists = 1
# 			pinobj = Blender.Object.Get(oj)
# 			pindata = pinobj.getData()
# 			break
# 	if not exists:
# 		pindata = Blender.Lamp.New('Lamp', 'Pindata')
# 		# strip it down to bare minimum features for better blender performance:
# 		pindata.mode |= Lamp.Modes["Layer"]
# 		pindata.mode |= Lamp.Modes["NoDiffuse"]
# 		pindata.mode |= Lamp.Modes["NoSpecular"]
# 		pindata.setDist(0.01)
# 		pindata.energy = 0

	for object in sce.objects.context:
		t = Blender.Object.Get(object.name)
		if (t.getType() == "Mesh"):

			if not t.getParent(): # bail if the mesh has no parent.
				return [0, 0, 0, 0, 0, -1]
			else:
				m = t.getData(mesh=1)
				crds = []
				for v in m.verts:
					crds.append(v.co)
				l = len(crds)
				if l == 3 or l > 4:
					PupBlock("ERROR MESSAGE",["%s is not a quad!" % t.name])
					return [0,0,0,0,0,0]

				tloc = Vector(t.getLocation("worldspace"))
				p = t.getParent()

				if wantStopPins:
					# stop-pins.
					# turn the inset distances into vectors:
					mr = Vector(-mrgr.val, 0.0, 0.0)
					mf = Vector(0.0, mrgf.val, 0.0)
					ml = Vector(mrgl.val, 0.0, 0.0)
					mb = Vector(0.0, -mrgb.val, 0.0)

					vrb = Vector(crds[0])
					vrf = Vector(crds[1])
					vlf = Vector(crds[2])
					vlb = Vector(crds[3])

					ret_val = []
					ret_val = Mathutils.LineIntersect(vlb + mb, vrb + mb, vrf + mr, vrb + mr)
					rb = ret_val[0] + tloc
					ret_val = Mathutils.LineIntersect(vlf + mf, vrf + mf, vrf + mr, vrb + mr)
					rf = ret_val[0] + tloc
					ret_val = Mathutils.LineIntersect(vlf + mf, vrf + mf, vlf + ml, vlb + ml)
					lf = ret_val[0] + tloc
					ret_val = Mathutils.LineIntersect(vlb + mb, vrb + mb, vlf + ml, vlb + ml)
					lb = ret_val[0]  + tloc

					s = float(spac.val)

					counter[0] += doStopPins(rb, rf, s, p, pindata, layer)
					counter[0] += doStopPins(rf, lf, s, p, pindata, layer)
					counter[0] += doStopPins(lf, lb, s, p, pindata, layer)
					counter[0] += doStopPins(lb, rb, s, p, pindata, layer)

				if wantAnchorPins:
					# anchor-pin - we place this 65% along the front edge at the specified distance back from it.
					# 65% is chosen so as not to coincide with stop pins.
					loca = Vector([(crds[1][0] + crds[2][0]) * 0.65, (crds[1][1] + crds[2][1]) * 0.65 + mrga.val, 0.0]) + tloc
					anc = placeOnePin(loca, 'PINANCHOR_PT', pindata, layer)
					p.makeParent([anc])
					counter[1] += 1

				if wantPivotPins:
					# pivot-pins: we place these near the front of the left and right tile edges, for manual relocation.
					locl = Vector([crds[1][0], crds[1][1] + mrga.val, 0.0]) + tloc
					pivlead = placeOnePin(locl, 'PINPIVOT_LEAD_PT', pindata, layer)
					p.makeParent([pivlead])

					loct = Vector([crds[2][0], crds[2][1] + mrga.val, 0.0]) + tloc
					pivtrail = placeOnePin(loct, 'PINPIVOT_TRAIL_PT', pindata, layer)
					p.makeParent([pivtrail])
					counter[2] += 1

	 			if wantGroundPins:
					# ground-pin - we place this halfway along the front edge at the specified distance back from it.
					locg = Vector([(crds[1][0] + crds[2][0]) * 0.5, (crds[1][1] + crds[2][1]) * 0.5 + mrga.val, 0.0]) + tloc
					gnd = placeOnePin(locg, 'PINGROUND_PT', pindata, layer)
					p.makeParent([gnd])
					#p.addProperty("SHARE_Y", 1, 'BOOL')
					counter[3] += 1

				if wantShareY:
					refreshProps(p, "SHARE_Y", "1", 1)
					counter[4] += 1

				counter[5] += 1
	return	counter

def optimisePins():
	"""
		Find all pin lamp objects in the scene and link them all to a single pindata datablock for efficiency
	"""

	# Find an exisitng pindata lamp datablock - if none, create one:
	pindata = None
	try:
		for lamp in Blender.Lamp.Get():
			if lamp.name.startswith('XPToolsPinData'):
				pindata = lamp
				pindata.name = 'XPToolsPinData'
				print 'XPToolsPinData found.'
				break
		else:
			print 'XPToolsPinData not found. Creating new.'
			pindata = Blender.Lamp.New('Lamp', 'XPToolsPinData')
			# strip it down to bare minimum features for better blender performance:
			pindata.mode |= Lamp.Modes["Layer"]
			pindata.mode |= Lamp.Modes["NoDiffuse"]
			pindata.mode |= Lamp.Modes["NoSpecular"]
			pindata.setDist(0.01)
			pindata.energy = 0
	except:
		print 'ERROR: Could not find or create pindata'
		return[None, None]

	# Search for all pin objects and relink them to the pindata datablock:
	scn = bpy.data.scenes.active
	data_names = set()
	pin_count = 0
	changed_count = 0
	for ob in [ob for ob in scn.objects if ob.type == 'Lamp']:
		if strip_suffix(ob.name) in ['PINANCHOR_PT', 'PINSTOP_PT', 'PINGROUND_PT', 'PINPIVOT_LEAD_PT', 'PINPIVOT_TRAIL_PT', 'PIN']:
			data_names.add(ob.data.name)
			pin_count += 1
			if ob.data.name != 'XPToolsPinData':
				ob.link(pindata)
				changed_count += 1
	print 'Relinked %d of %d "pin" lamps to a common "XPToolsPinData" datablock' % (changed_count, pin_count)
	print 'Unlinked %d redundant lamp datablocks which will be deleted on reload of this .blend file' % len(data_names)
	for lamp in Blender.Lamp.Get():
		if lamp.name.startswith('XPToolsPinData'):
			print lamp.name, lamp.users
	PupMenu('Relinked %d of %d "pin" lamps to a common "XPToolsPinData" datablock.\t|Unlinked %d redundant lamp datablocks which will be deleted on save and reload of this .blend file.\t'\
	% (changed_count, pin_count, len(data_names)))

	return [pin_count, changed_count]

def projectUVMap():

	""" Often it is more convenient to establish tile shapes and sizes ahead of texture creation.
	The downside is that it's then necessary to individually UV map a ton of tiles at a later date.
	This function projects a texture vertically downwards onto multiple selected objects in the 3-d view,
	saving hours of truly mindless tedium... """
	global sce
	sce = Blender.Scene.GetCurrent()
	fOrgx, fOrgy, fTexw, fTexh, sTexn
	ox = float(fOrgx)
	oy = float(fOrgy)
	width = float(fTexw)
	height = float(fTexh)
	changed = 0

	if width == 0.0 or height == 0.0:
		PupBlock("ERROR MESSAGE", ["Texture cannot have","zero width or height!"])
		return -2
	try:
		myimage = Blender.Image.Get(sTexn)

	except:
		return -1

	obs = [ob for ob in sce.objects.context if not ob.lib]
	for obj in obs:
		if obj.getType() == 'Mesh':
			locw = obj.getLocation('worldspace')
			msh = obj.getData(mesh=1)
			if msh.faceUV == 0:
				msh.faceUV = 1
			orig_uvlayer = msh.activeUVLayer
			for uvlayer in msh.getUVLayerNames():
				msh.activeUVLayer = uvlayer
				for f in msh.faces:
					f.image = myimage
					for (uv, vt) in zip(f.uv, f.verts):
						uv[0] = (vt.co.x + locw[0] - ox)/width
						uv[1] = (vt.co.y + locw[1] - oy)/height
					f.transp = Mesh.FaceTranspModes.ALPHA
			msh.activeUVLayer = orig_uvlayer
			obj.select(1)
			Window.EditMode(1)
			Window.EditMode(0)
			changed += 1
			sce.update()
	Window.RedrawAll()

	return	changed

def getAtomicTileLength(tileWrapper, scene):

	""" Here we're looking for the distance between the leftmost edge of the tile quad, and the rightmost stop pin.
		This is the effective length of the tile when a component of an atomic spelling """

	kids = getChildren(tileWrapper, scene.objects)

	# get the tile:
	bb = []
	for k in kids:
		if k.name.startswith('TILE'):
			bb = k.getBoundBox()
			break
	else:
		PupMenu("Wrapper %s contains no tile!" % tileWrapper.name)
		return 0.0

	txmin = 9999.0

	for vt in bb:
		if vt[0] < txmin:
			txmin = vt[0]

	# get the stop pins:
	pxmax = -9999.0

	for p in kids:
		if p.name.startswith('PINSTOP_PT'):
			if p.LocX > pxmax:
				pxmax = p.LocX
	if pxmax == -9999.0:
		PupMenu("Wrapper %s contains no stop pins!" % tileWrapper.name)
		return 0.0
	l = pxmax - txmin

	return l

def getAtomicTileMetrics(tileWrapper, scene):

	""" Here we're looking for the distance between the leftmost edge of the tile quad, and the rightmost stop pin.
		This is the effective length of the tile when a component of an atomic spelling.
		For the depth value, we measure the distance from the anchor pin to the rearmost stop pin, since this is the
		effective depth of the obstruction created when the tile is laid down, regardless of the overall size of the quad"""

	kids = getChildren(tileWrapper, scene.objects)

	# get the tile:
	bb = []
	for k in kids:
		if k.name.startswith('TILE'):
			bb = k.getBoundBox()
			break
	else:
		PupMenu("Wrapper %s contains no tile!" % tileWrapper.name)
		return [0.0, 0.0]

	txmin = 9999.0
	tymax = -9999.0

	for vt in bb:
		if vt[0] < txmin:
			txmin = vt[0]
		if vt[1] > tymax:
			tymax = vt[1]

	# get the stop pins and anchor pin:
	pxmax = -9999.0
	pymax = -9999.0
	apy = 9999.0
	for p in kids:
		if p.name.startswith('PINSTOP_PT'):
			if p.LocX > pxmax:
				pxmax = p.LocX
			if p.LocY > pymax:
				pymax = p.LocY
		elif p.name.startswith('PINANCHOR_PT'):
			apy = p.LocY
	if pxmax == -9999.0 or pymax == -9999.0:
		PupMenu("Wrapper %s contains no stop pins!" % tileWrapper.name)
		return [0.0, 0.0]
	if apy == 9999.0:
		PupMenu("Wrapper %s contains no anchor pin!" % tileWrapper.name)
		return [0.0, 0.0]

	l = round(pxmax - txmin, 3)
	d = round(pymax - apy, 3)

	return [l, d]

def getDepth(distance, angle, depthMin):

	# We assume that the distance between the incoming and outgoing edges at the point where the corner tiles end (and, consequently, where we start) will be greater or equal to d_min * 2.
	# It may turn out necessary to add a fudge factor for _really_ sharp corners.

	if angle >= -90.0:
		return 10000.0
	return depthMin + distance * math.tan(math.radians(angle))

def	generateSpellings2():

	""" An attempt to improve ground coverage by regulating tile depth with corner angle info.
		We can bind atomic spelling sets to specific match groups. If the match group consists only of atomic corners with narrow angle ranges,
		we are effectively forcing the atomic spelling set to be used only on edges emerging from those ranges of angles.
		If we are not the first edge to be filled, there will (hopefully) be tiles on the previous one, and we can rely on the AGS algo to fill the new edge as deeply as possible to the back of them.
		However, if we ARE the first edge, there's a danger that we will fill too deep, making it impossible to lay tiles on the back edge when we eventually reach it.
		Therefore, we must keep the depth of the new edge as low as possible until we are far enough from the corner to theoretically not clog the previous one.
		The AGS will naturally fill as deeply as it can, so long as we give it deep spellings before shallow ones. What we should be aiming to do is to END spellings in a tapered manner,
		to try to avoid choking the corner and making it impossible to lay ANY tiles on the next edge, if the corner angle < -90 degrees.
		We have six parameters to play with - min and max depths, start and end angles, and the option to specify particular tiles for the start and end of the spelling.
		Zero values indicate that the parameters are ignored.
		depth_min - the depth of the skinniest tiles available to us.
		depth_max - the target maximum depth we'd like to fill to if possible.
		angle_start - the most negative (sharpest) angle the previous corner can cope with.
		angle_end - ditto for the next corner.
		distance - how far we have progressed through the current spelling. This doesn't include the length of the corner itself, of course.
	"""
	global sdata
	d_min = sdata['depth_min']
	d_max = sdata['depth_max']
	a_start = sdata['angle_start']
	a_end = sdata['angle_end']
	first_tile = sdata['first_tile']
	last_tile = sdata['last_tile']

	do_depth = d_min or d_max or a_start or a_end
	consider_equal_length = float(sdata['tile_variety']) * 0.2 + 1.0	# gives us a range of  1:1 ... 3:1 for length selection ratio
	consider_equal_depth  = float(sdata['depth_variety']) * 0.2 + 1.0   # ditto for tile depth

	lmax  = float(sdata['actual_max'] )
	lslop = sdata['dist_slop']
	tiles = []
	spelling = []

	for t in Blender.Text.Get():
		if t.name == "Spellings_out":
			txt = t
			break
	else:
		txt = Blender.Text.New("Spellings_out")
	txt.write("\n\n")
	sc = Scene.GetCurrent()
	selected = sc.objects.selected

	# These are optional, manually added, first and last tiles:
	len_first_tile = len_last_tile = 0.0
	if first_tile:
		len_first_tile = getAtomicTileLength(Object.Get(str(first_tile)), sc)
	if last_tile:
		len_last_tile = getAtomicTileLength(Object.Get(str(last_tile)), sc)

	for ob in selected:
		if ob.type == 'Empty':
			length, depth = getAtomicTileMetrics(ob, sc)
			if length and depth:
				if length < 0.0:
					PupMenu("ERROR: Tile %s has a negative length - how did this happen?\n" % ob.name)
					return 0
				elif depth < 0.0:
					PupMenu("ERROR: Tile %s has a negative depth - how did this happen?\n"% ob.name)
					return 0
				else:
					tiles.append([length, depth, ob.name])
			else:
				PupMenu("ERROR: Tile %s has a zero value length or depth - how did this happen?\n" % ob.name)
				return 0

	if len(tiles):
		tiles.sort()
		tiles.reverse()

		txt.write("# tile ID\t\tlength\tdepth\n\n")
		for t in tiles:
			txt.write("# %s\t\t%7.3f\t%7.3f\n" % (t[2], t[0], t[1]))
		txt.write("\n#\t")

		for t in reversed(tiles):
			txt.write("%s\t" % t[2])
		txt.write("\n\n")
		rh_col = 0

		while (tiles[-1][0] + len_first_tile + len_last_tile) <= lmax:

			lspell = 0.0
			spelling = []
			spelling_best = []
			gap_best = 1000.0
			s = 0
			attempt = 0
			MAX_ATTEMPTS = 10
			backtrack = 3
			i = 0
			j = 0

			while (lspell + lslop) < lmax:

				if (lspell + tiles[i][0]) <= lmax:
					alts = [tiles[i]]
					alts.extend( [a for a in tiles[i:] if attempt ==0 and ((tiles[i][0] / a[0]) <= consider_equal_length)])

					# Now look at available tile _depths_. Find the minimum available depth ... could be either the start or end angle which affects us most at this point.

					if do_depth:
						depth_alts = [[da[1], da[0], da[2]] for da in alts if da[1] < d_max and da[1] < min(getDepth(lspell, a_start, d_min),  getDepth(lmax - lspell - da[0], a_end, d_min))]
						depth_alts.sort()
						all_alts = [[aa[1], aa[0], aa[2]] for aa in depth_alts if (depth_alts[0][0]/aa[0]) <= consider_equal_depth]
						irand = tiles.index(random.choice(all_alts))
					else:
						irand = tiles.index(random.choice(alts))

					spelling.append(irand)
					lspell += tiles[irand][0]
				else:
					i += 1
					if i == len(tiles):								# we've already used our smallest tile
						if attempt < MAX_ATTEMPTS:		# we're still short of the target by more than the slop amoumt
							gap = lmax - lspell
							if gap < gap_best:
								gap_best = gap
								spelling_best = spelling[:]
							else:
								spelling = spelling_best[:]
								lspell = 0.0
								for t in spelling:
									lspell += tiles[t][0]

							attempt += 1

							while i + 1 >= len(tiles):
								while len(spelling) > 0 and (len(tiles) - i < backtrack):
									i = spelling.pop()				# get the id of the last used tile and remove it from the spelling
									lspell -= tiles[i][0]			# subtract the length of the tile we just removed
								backtrack +=1
							i += 1						# move on to the next smallest tile after the last used one.
							if i >= len(tiles):
								PupMenu("ERROR: Insufficient slop available to cover range without gaps.\n")
								return 0
						else:
							if gap > lslop:
								PupMenu("ERROR: Failed to find spelling for distance %7.3f after %d attempts." % (lmax, attempt))
								return 0

			outlength = len_first_tile +  len_last_tile
			for tile_idx in spelling:
				outlength += tiles[tile_idx][0]

			declared_min = sdata['declared_min']
			if sdata['min_is_rel']:
				declared_min += outlength
			declared_max = sdata['declared_max']
			if sdata['max_is_rel']:
				declared_max += outlength

			if sdata['atomic'] == 0:
				spelling_str = ("%s\t %d \t %d \t %d \t" % ("ATOMIC_SPELLING", int(declared_min), int(declared_max), sdata['match_group']))

			elif sdata['atomic'] == 1:
				spelling_str = ("%s\t %d \t %d \t %d \t %d \t %d \t %d \t %d \t" % ("ATOMIC_SPELLING_CORNER", sdata['declared_pre_min'], sdata['declared_pre_max'], int(a_start), int(a_end), int(declared_min), int(declared_max), sdata['match_group']))

			else:
				spelling_str = "SPELLING\t\t"

			comment_col = len(spelling_str.expandtabs(4)) - 3

			if first_tile:
				spelling_str += (str(first_tile) + "\t")
			for tile_idx in spelling:
				spelling_str += (tiles[tile_idx][2] + "    ")
			if last_tile:
				spelling_str += (str(last_tile) + "\t")
			spelling_str += ("\n")
			txt.write(spelling_str)

			if len(spelling_str) + 8 > rh_col:
				rh_col = len(spelling_str) + 8
			comment_str = "#" + " " * comment_col
			if first_tile:
				comment_str += ("%7.2f " % len_first_tile)
			for tile_idx in spelling:
				comment_str += ("%7.2f " % tiles[tile_idx][0])
			if last_tile:
				comment_str += ("%7.2f " % len_last_tile)
			pad = rh_col - len(comment_str)
			if pad > 0:
				comment_str += (" " * pad)
			comment_str += ("Length range: %7.2f...%7.2f\n" % (outlength, outlength + lslop))
			txt.write(comment_str)

#			chart = "# "
#			for s in spelling:
#				chart += 'X' * int(math.ceil(tiles[s][0]))
#				chart += ' '
#			chart += '\n'
#			txt.write(chart)

			lmax = lspell - 0.1
	else:
		PupMenu("ERROR: There seem to be no tiles!\n")
		return 0

	Window.Redraw()

def	generateSpellings1(lengthMax, lengthSlop, spellingGroup):

	global sdata

	consider_equal_length = float(sdata['tile_variety']) * 0.2 + 1.0	# gives us a range of  1:1 ... 3:1 for length selection ratio

	lmax  = float(sdata['actual_max'] )
	lslop = sdata['dist_slop']
	tiles = []
	spelling = []

	for t in Blender.Text.Get():
		if t.name == "Spellings_out":
			txt = t
			break
	else:
		txt = Blender.Text.New("Spellings_out")
	txt.write("\n\n")
	sc = Scene.GetCurrent()
	selected = sc.objects.selected

	for ob in selected:
		if ob.type == 'Empty':
			length = getAtomicTileLength(ob, sc)
			if length:
				if length < 0.0:
					PupMenu("ERROR: Tile %s has a negative length how did this happen?\n")
					return 0
				else:
					tiles.append([length, ob.name])
			else:
				return 0

	if len(tiles):
		tiles.sort()
		tiles.reverse()

		txt.write("# tile ID\t\tlength\n\n")
		for t in tiles:
			txt.write("# %s\t\t%7.3f\n" % (t[1], t[0]))
		txt.write("\n#\t")

		for t in reversed(tiles):
			txt.write("%s\t" % t[1])
		txt.write("\n\n")

		rh_col = 0

		while lmax >= tiles[-1][0]:
			lspell = 0.0
			spelling = []
			spelling_best = []
			gap_best = 1000.0
			s = 0
			attempt = 0
			MAX_ATTEMPTS = 10
			backtrack = 3
			i = 0
			j = 0
			while (lspell + lslop) < lmax:

				if (lspell + tiles[i][0]) <= lmax:
					alts = [i]
					j = 1
					while (i + j < len(tiles)) and ((tiles[i][0] / tiles[i+j][0])< consider_equal_length) and attempt == 0:
						alts.append(i+j)
						j += 1
					irand = random.choice(alts)

					spelling.append(irand)
					lspell += tiles[irand][0]
				else:
					i += 1
					if i == len(tiles):								# we've already used our smallest tile
						if attempt < MAX_ATTEMPTS:		# we're still short of the target by more than the slop amoumt
							gap = lmax - lspell
				#			txt.write("gap: %f, gap_best: %f\n" % (gap, gap_best))
							if gap < gap_best:
								gap_best = gap
								spelling_best = spelling[:]
							else:
								spelling = spelling_best[:]
								lspell = 0.0
								for t in spelling:
									lspell += tiles[t][0]

							attempt += 1

							while i + 1 >= len(tiles):
								while len(spelling) > 0 and (len(tiles) - i < backtrack):
									i = spelling.pop()				# get the id of the last used tile and remove it from the spelling
									lspell -= tiles[i][0]			# subtract the length of the tile we just removed
						#			txt.write("breakpoint 2: spelling: %s, i: %d, lspell: %7.1f, attempt: %d, backtrack: %d\n" % (str(spelling), i, lspell, attempt, backtrack))
								backtrack +=1
							i += 1						# move on to the next smallest tile after the last used one.
							if i >= len(tiles):
								PupMenu("ERROR: Insufficient slop available to cover range without gaps.\n")
								return 0
						else:
							if gap > lslop:
								PupMenu("ERROR: Failed to find spelling for distance %7.3f after %d attempts." % (lmax, attempt))
								return 0

			outlength = 0.0
			for tile_idx in spelling:
				outlength += tiles[tile_idx][0]

			declared_min = sdata['declared_min']
			if sdata['min_is_rel']:
				declared_min += outlength
			declared_max = sdata['declared_max']
			if sdata['max_is_rel']:
				declared_max += outlength

			spelling_str = ("ATOMIC_SPELLING\t %d \t %d \t %d \t" % (int(declared_min), int(declared_max), sdata['match_group']))

			for tile_idx in spelling:
				spelling_str += (tiles[tile_idx][1] + "    ")
			spelling_str += ("\n")
			txt.write(spelling_str)

			if len(spelling_str) + 8 > rh_col:
				rh_col = len(spelling_str) + 8
			comment_str = "#                            "
			for tile_idx in spelling:
				comment_str += ("%7.2f " % tiles[tile_idx][0])
			pad = rh_col - len(comment_str)
			if pad > 0:
				comment_str += (" " * pad)
			comment_str += ("Length range: %7.2f...%7.2f\n" % (outlength, outlength + lslop))
			txt.write(comment_str)
			lmax = lspell - 0.1
	else:
		PupMenu("ERROR: There seem to be no tiles!\n")
		return 0

	Window.Redraw()

def makeMemberGroup(groupName):

	""" This just prints out the names of the selected tiles and an optional comment after the keyword GROUP and the passed-in group ID number """

	for t in Blender.Text.Get():
		if t.name == "Spellings_out":
			txt = t
			break
	else:
		txt = Blender.Text.New("Spellings_out")

	scn = Scene.GetCurrent()
	selected = scn.objects.context
	tilenames = []
	for ob in selected:
		if ob.type == 'Empty' and (ob.name.isdigit()):
			tilenames.append(ob.name)
		else:
			PupMenu("ERROR: Illegal tile name or nothing selected.")
			return 0
	tilenames.sort()

	txt.write("\nGROUP\t%d\t" % groupName)
	for n in tilenames:
		txt.write("%s\t" % n)
	txt.write("\n")
	if sdata['comment'] and sdata['comment'] != "comment":
		txt.write("#\t%s\n" % sdata['comment'])
	return len(tilenames)

def clearResults():
	global tasksdone
	for k in tasksdone.iterkeys():
		tasksdone[k] = 0

def showResults():

	global tasksdone
	global sTexn
	global GUIpage, GUIpage_prev, GUIchanged
	global sResults
	sResults = []

	if tasksdone['rnam']:
		if tasksdone['rnam'] == 1:
			sResults.append("1 tile was renamed.\n")
		else:
			sResults.append("%d tiles were renamed.\n" % tasksdone['rnam'])
	if tasksdone['cent']:
		if tasksdone['cent'] == 1:
			sResults.append("1 tile centre was moved to the lower left corner of its bounding rectangle.\n")
		else:
			sResults.append("%d tile centres were moved to the lower left corner of their bounding rectangles.\n" % tasksdone['cent'])
	if tasksdone['wrap']:
		if tasksdone['wrap'] == 1:
			sResults.append("1 tile was parented to a new wrapper.\n")
		else:
			sResults.append("%d tiles were parented to new wrappers.\n" % tasksdone['wrap'])
	if tasksdone['agsw']:
		sResults.append("A new AGS master wrapper was created at the 3d cursor location.\n")
	if tasksdone['tcnt']:
		if tasksdone['tcnt'] == -1:
			sResults.append("WARNING: No pins were added because tile wrappers were not found. The operation was skipped.\n")
		else:
			if tasksdone['pins']:
				if tasksdone['tcnt'] == 1:
					sResults.append("1 tile had a total of %d stop pins added to it.\n" % tasksdone['pins'])
				else:
					sResults.append("%d tiles had a total of %d stop pins added to them.\n" % (tasksdone['tcnt'], tasksdone['pins']))
	if tasksdone['pina']:
		if tasksdone['pina'] == 1:
			sResults.append("1 tile had an anchor pin added to it.\n")
		else:
			sResults.append("%d tiles had anchor pins added to them.\n"  % tasksdone['pina'])
	if tasksdone['pinp']:
		if tasksdone['pinp'] == 1:
			sResults.append("1 tile had a pair of pivot pins added to it.\n")
		else:
			sResults.append("%d tiles had pairs of pivot pins added to them.\n" % tasksdone['pinp'])
	if tasksdone['ping']:
		if tasksdone['ping'] == 1:
			sResults.append("1 tile had a ground elevation reference pin added to it.\n")
		else:
			sResults.append("%d tiles had ground elevation reference pins added to them.\n" % tasksdone['ping'])
	if tasksdone['shrY']:
		if tasksdone['shrY'] == 1:
			sResults.append("1 tile had a 'SHARE_Y' property added to its wrapper.\n")
		else:
			sResults.append("%d tiles had a 'SHARE_Y' property added to their wrappers.\n" % tasksdone['shrY'])
	if tasksdone['proj']:
		if tasksdone['proj'] == -1:
			sResults.append("WARNING: Could not find image for projection. Operation was skipped.\n")
		elif tasksdone['proj'] == 1:
			sResults.append("1 tile had image '%s' projected on it and was set to alpha blend mode.\n" % sTexn)
		else:
			sResults.append("%d tiles had image '%s' projected on them and were set to alpha blend mode.\n" % (tasksdone['proj'], sTexn))
	if tasksdone['OBJatt']:
		if tasksdone['OBJatt'] == 1:
			sResults.append("1 OBJ annotation was added to the selected AG entity.\n")
		else:
			sResults.append("%d OBJ annotations were added to the selected AG entity.\n" % tasksdone['OBJatt'])
	if tasksdone['FACatt']:
		if tasksdone['FACatt'] == 1:
			sResults.append("1 facade annotation was added to the selected AG entity.\n")
		else:
			sResults.append("%d facade annotations were added to the selected AG entity.\n" % tasksdone['FACatt'])
	if tasksdone['TREEatt']:
		if tasksdone['TREEatt'] == 1:
			sResults.append("1 tree annotation was added to the selected AG entity.\n")
		else:
			sResults.append("%d tree annotations were added to the selected AG entity.\n" % tasksdone['TREEatt'])

	subs = {1:'added to', -1:'removed from'}
	if tasksdone['TILES']:
		sResults.append("%d tiled face%s %s %d object%s\n" % (abs(tasksdone['TILES']), pluralize(tasksdone['TILES']), subs[get_sign(tasksdone['TILES'])], tasksdone['tilesobjs'], pluralize2(tasksdone['tilesobjs'])))

	if tasksdone['TWOSIDE']:
		sResults.append("%d two-sided face%s %s %d object%s\n" % (abs(tasksdone['TWOSIDE']), pluralize(tasksdone['TWOSIDE']), subs[get_sign(tasksdone['TWOSIDE'])], tasksdone['twosideobjs'], pluralize2(tasksdone['twosideobjs'])))

	if tasksdone['DYNAMIC']:
		sResults.append("%d collision face%s %s %d object%s\n" % (abs(tasksdone['DYNAMIC']), pluralize(tasksdone['DYNAMIC']), subs[get_sign(tasksdone['DYNAMIC'])], tasksdone['dynamicobjs'], pluralize2(tasksdone['dynamicobjs'])))

	if tasksdone['SHADOW']:
		sResults.append("%d shadow face%s %s %d object%s\n" % (abs(tasksdone['SHADOW']), pluralize(tasksdone['SHADOW']), subs[get_sign(tasksdone['SHADOW'])], tasksdone['shadowobjs'], pluralize2(tasksdone['shadowobjs'])))

	if tasksdone['INVISIBLE']:
		sResults.append("%d invisible face%s %s %d object%s\n" % (abs(tasksdone['INVISIBLE']), pluralize(tasksdone['INVISIBLE']), subs[get_sign(tasksdone['INVISIBLE'])], tasksdone['invisibleobjs'], pluralize2(tasksdone['invisibleobjs'])))

	if tasksdone['SOLID']:
		sResults.append("%d solid face%s %s %d object%s\n" % (abs(tasksdone['SOLID']), pluralize(tasksdone['SOLID']), subs[get_sign(tasksdone['SOLID'])], tasksdone['transpobjs'], pluralize2(tasksdone['transpobjs'])))

	if tasksdone['ALPHA']:
		sResults.append("%d alpha face%s %s %d object%s\n" % (abs(tasksdone['ALPHA']), pluralize(tasksdone['ALPHA']), subs[get_sign(tasksdone['ALPHA'])], tasksdone['transpobjs'], pluralize2(tasksdone['transpobjs'])))

	if tasksdone['CLIP']:
		sResults.append("%d clip alpha face%s %s %d object%s\n" % (abs(tasksdone['CLIP']), pluralize(tasksdone['CLIP']), subs[get_sign(tasksdone['CLIP'])], tasksdone['transpobjs'], pluralize2(tasksdone['transpobjs'])))

	if tasksdone['OBJs_selected']:
		sResults.append("%d OBJ%s selected.\n" % (tasksdone['OBJs_selected'], pluralize(tasksdone['OBJs_selected'])))

	if tasksdone['OBJ_meshes_selected']:
		sResults.append("%d OBJ mesh object%s selected.\n" % (tasksdone['OBJ_meshes_selected'], pluralize(tasksdone['OBJ_meshes_selected'])))

	if tasksdone['assign_OBJ_image']:
		sResults.append("The active image was assigned to %d mesh objects." % tasksdone['assign_OBJ_image'])

	if tasksdone['moveOBJCentre']:
		if tasksdone['moveOBJCentre'] == 1:
			sResults.append("One mesh object had its centre moved to the ground level centre front." )
		elif tasksdone['moveOBJCentre'] > 1:
			sResults.append("%d mesh objects had their centres moved to the ground level centre front." % tasksdone['moveOBJCentre'])

	if tasksdone['shared_image_obs']:
		sResults.append("%d mesh object%s found to be using the selected image." % (tasksdone['shared_image_obs'], pluralize(tasksdone['shared_image_obs'])))


	if tasksdone['objs_stripped_from'] == 1:
		sResults.append("%d properties were stripped from one OBJ" % tasksdone['objprops_stripped'])
	elif tasksdone['objs_stripped_from'] > 1:
		sResults.append("%d properties were stripped from %d OBJs" % (tasksdone['objprops_stripped'], tasksdone['objs_stripped_from']))

	if not sResults:
		sResults.append("No actions were performed")
	GUIpage_prev = GUIpage
	GUIpage = 2
	GUIchanged = 1
	Redraw(1)
	return

def exitXPTools():
	if g_lighteditor:
		g_lighteditor.exitLightEditor()
	if g_AGB_editor:
		g_AGB_editor.exitEditor()
	try:
		for t in Blender.Text.Get():
			if t.name.startswith('XPToolsAWS'):
				Blender.Text.unlink(t)
		AWS_scriptlink = None
	except:
		pass
	ExportConfig()
	Exit()

def event(evt, val):

	global g_lighteditor, g_AGB_editor, g_qualifs

# 	if evt == ESCKEY:
# 		rep = PupMenu(" REALLY EXIT?:%t|    NO|  YES")
# 		if rep == 2:
# 			if g_lighteditor:
# 				g_lighteditor.exitLightEditor()
# 			ExportConfig()
# 			Exit()
# 			return

	if evt in [LEFTSHIFTKEY, RIGHTSHIFTKEY,	LEFTCTRLKEY, RIGHTCTRLKEY]:
#		print "got event!"
		g_qualifs = Window.GetKeyQualifiers()

		Draw()

	if g_lighteditor:
		g_lighteditor.eventHandler(evt, val)

	if g_AGB_editor:
		g_AGB_editor.eventHandler(evt, val)

def button_event(evt):

	global iIdin, iRnam, sTexn, fTexw, fTexh, fOrgx, fOrgy, fMrga, fMrgf, fMrgr, fMrgb, fMrgl, fSpac
	global iPins, iPina, iPinp, iPing, iCent, cent, iProj, iWrap, iAgsw, iLayr, iShrY, iLayw, fDmin, dmin, fDmax, mtch,  dslp, iSgrp, sgrp, mgrp, rlmn, rlmx, acmn, acmx, dcmn, dcmx, GUIpage, GUIpage_prev, GUIchanged
	global angs, ange, dpmn, dpmx, vari, atmc, tfst, tlst, vard
	global ob_OBJatt, ob_FACatt, ob_TREEatt
	global texn, rnam
	global tasksdone
	global facemodes
	global facetranspmodes
	global modestosearch
	global facematchall
	global EXPORTMODE
	global DEF_OBJ_WRAPPER_LAYER, DEF_METAPIN_LAYER, DEF_TILE_WRAPPER_LAYER
	global OBJVIRTUALPATH
	global OBJRENAME
	global OBJSHAPE
	global OBJPART
	global OBJVARIANT
	global AUTOVARIANT
	global VARSTOLIB
	global OBJPROPSOVERWRITE
	global OBJWRITEPROPS
	global USEHEIGHTBUCKETS
	global BUCKETSVIRTUALONLY
	global OBJINSTANCED
	global HASDRAPED
	global objprops
	global objprops2
	global objpropkeys
	global EVENT_mn_objpart
	global EVENT_mn_objshape
	global EVENT_mn_objvariant
	global EVENT_tg_objbucketsvirtualonly
	global EVENT_tg_objautovariant
	global EVENT_tg_objvarstolib
	global EVENT_tg_objbuckets
	global EVENT_tg_objrename
	global EVENT_mn_EXPORTMODE
	global EVENT_nb_objwrapperlayer
	global EVENT_mn_propname0
	global EVENT_st_propval0
	global EVENT_st_propval1
	global EVENT_st_propval2
	global EVENT_mn_propname1
	global EVENT_mn_propname2
	global EVENT_tg_objpropsoverwrite
	global EVENT_bt_objgetsameimage
	global EVENT_bt_objgetfromactive
	global EVENT_bt_objselallfaces
	global EVENT_bt_objsetfaces
	global EVENT_bt_objdeselallfaces
	global EVENT_bt_objselwrappers
	global EVENT_bt_objshowfaces
	global EVENT_bt_objfindmixedalpha
	global EVENT_tg_objmatchany
	global EVENT_tg_objmatchall
	global EVENT_bt_objenableall
	global EVENT_bt_objdefaultmodes
	global EVENT_bt_objmatchlogic
	global EVENT_bt_objshowfacesvis
	global Object_mn_objpart
	global Object_mn_objshape
	global Object_mn_objvariant
	global Object_tg_objbucketsvirtualonly
	global Object_tg_objautovariant
	global Object_tg_objvarstolib
	global Object_tg_objbuckets
	global Object_tg_objrename
	global Object_mn_EXPORTMODE
	global Object_nb_objwrapperlayer
	global Object_mn_propname0
	global Object_st_propval0
	global Object_st_propval1
	global Object_st_propval2
	global Object_mn_propname1
	global Object_mn_propname2
	global Object_tg_objpropsoverwrite
	global Object_tg_objwriteprops
	global Object_bt_objselallfaces
	global Object_bt_objsetfaces
	global Object_bt_objdeselallfaces
	global Object_bt_objselwrappers
	global Object_bt_objselgeom
	global Object_bt_objgetsameimage
	global Object_bt_objgetfromactive
	global Object_bt_objassignimage
	global Object_bt_objshowfaces
	global Object_bt_objfindmixedalpha
	global Object_tg_objmatchany
	global Object_tg_objmatchall
	global Object_bt_objmatchlogic
	global Object_bt_objenableall
	global Object_bt_objdefaultmodes
	global Object_bt_objshowfacesvis
	global EVENT_tg_objfacemodesel
	global Object_tg_objfacemodesel
	global Object_tg_objftmsolid
	global Object_tg_objftmalpha
	global Object_tg_objftmclip
	global Object_bt_objresetbuttons
	global EVENT_bt_objresetbuttons
	global Object_tg_objfmdynamic
	global Object_tg_objfmtiles
	global Object_tg_objfmtwoside
	global Object_tg_objfmshadow
	global Object_tg_objfminvisible
	global Object_tg_objinstanced
	global Object_mn_objpropname3
	global Object_mn_objpropname4
	global Object_tg_objhasdraped
	global Object_nb_objlod1
	global Object_nb_objlod2
	global Object_mn_objimage
	global Object_objmoveOBJcentre
	global Object_bt_objstripprops
	global OBJIMAGELIST
	global objimagemenuname
	global EVENT_bt_restrictselect
	global EVENT_bt_enableselect, EVENT_tg_AWS_enable
	global Object_tg_AWS_enable
	global EVENT_bt_dupehierarchy
	global evt_sort_imports
	global cp_colour, st_sRGBint,	st_sRGBflt,	st_linRGB, bt_nrmlize, sl_hue, sl_saturation, sl_brightness, tg_RGB, tg_HSB, sl_red, sl_green, sl_blue
	global evt_cp_colour, evt_st_sRGBint,  evt_st_sRGBflt,	evt_st_linRGB, evt_bt_nrmlize, evt_sl_hue, evt_sl_saturation, evt_sl_brightness
	global evt_tg_RGB,	evt_tg_HSB,	evt_sl_red,	evt_sl_green, evt_sl_blue, nm_normal, evt_nm_normal, st_normal, evt_st_normal, g_normal
	global mdata
	global LIGHTS_TXT_DIR, LIGHTS_TXT, g_lighteditor, g_AGB_editor

	if evt	 == 1:
		iRnam = 1 - iRnam
	elif evt == 2:
		iIdin  = 	idin.val
	elif evt == 3:
		iCent = 	1 - iCent
	elif evt == 5:
		iWrap = 	1 - iWrap
	elif evt == 6:
		iAgsw = 	1 - iAgsw
	elif evt == 7:
		iLayw = DEF_TILE_WRAPPER_LAYER = layw.val
	elif evt == 8:
		iLayr = DEF_METAPIN_LAYER =	layr.val
	elif evt == 9:
		iPins = 	1 - iPins
	elif evt == 10:
		fMrgr = 	mrgr.val
	elif evt == 11:
		fMrgf = 	mrgf.val
	elif evt == 12:
		fMrgl = 	mrgl.val
	elif evt == 13:
		fMrgb = 	mrgb.val
	elif evt == 14:
		fSpac = 	spac.val
	elif evt == 15:
		fMrga = 	mrga.val
	elif evt == 16:
		iPina = 	1 - iPina
	elif evt == 17:
		iPinp = 	1 - iPinp
	elif evt == 18:
		iPing = 	1 - iPing
	elif evt == 19:
		iShrY =     1 - iShrY
	elif evt == 20:
		ret = optimisePins()

	elif evt == 40:
		sTexn = 	texn.val
	elif evt == 41:
		 iProj =	 1 - iProj
	elif evt == 42:
		fTexw = 	float(texw.val)
	elif evt == 43:
		fTexh = 	float(texh.val)
	elif evt == 44:
		fOrgx = 	float(orgx.val)
	elif evt == 45:
		fOrgy = 	float(orgy.val)
	elif evt == 46:
		current_image = Blender.Image.GetCurrent()
		if current_image:
			sTexn = current_image.name
			texn.val = sTexn
			Redraw(1)
		else:
			PupMenu("\n ERROR: Current image not found! \n", 1)
	elif evt == 47:
		generateSpellings2()
	elif evt == 48:
		sdata['member_group'] =		int(sgrp.val)
	elif evt == 49:
		sdata['match_group'] =		mtch.val
	elif evt == 50:
		sdata['dist_slop'] =		dslp.val
	elif evt == 96:				# reset selections
		iIdin = iRnam =	iPins = iPina = iPinp = iPing = iCent = iProj = iWrap = iAgsw = iShrY = 0
	elif evt == 97:				# close results page
		GUIpage = GUIpage_prev
		GUIchanged = 1
		Redraw(1)
	elif evt == 98:


		rep = PupMenu(" REALLY EXIT?:%t|    NO|  YES")
		if rep == 2:
			exitXPTools()
			return
		else:
			Redraw(1)
			return
	elif evt == 99:	# apply all enabled AGS routines

		clearResults()

		if iPins or iPina or iPinp or iPing or iShrY or iRnam or iCent or iWrap:
			r = getSelection()
			if r == "multiple_types":
				PupMenu("Error: selection contains objects of different types.%t")
				return
			elif r == "bad_mesh":
				PupMenu("Error: tiles must be either quads or fgons.%t")
				return
			elif r in [ "bad_type", "empty_type"]:
				PupMenu("Error: selected objects must be meshes.%t")
				return

			if iRnam and r == "mesh_type":
				tasksdone['rnam'] = renameTiles(iIdin)

			if iCent and r == "mesh_type":
				tasksdone['cent'] = moveTileCentre()

			if iWrap and r == "mesh_type":
				tasksdone['wrap'] = wrapTiles(iLayw)

			if iPins or iPina or iPinp or iPing or iShrY and r == "mesh_type":
				retval = []
				retval = doMetaPins(iPins, iPina, iPinp, iPing, iShrY, iLayr)
				tasksdone['pins'] = retval[0]
				tasksdone['pina'] = retval[1]
				tasksdone['pinp'] = retval[2]
				tasksdone['ping'] = retval[3]
				tasksdone['shrY'] = retval[4]
				tasksdone['tcnt'] = retval[5]

		if iProj:
			tasksdone['proj'] = projectUVMap()

		if iAgsw:
			tasksdone['agsw'] = addAGStringWrapper(iLayw)

		showResults()

	elif evt == 110:
		OBJPART = Object_mn_objpart.val
	elif evt == 111:
		OBJSHAPE = Object_mn_objshape.val
	elif evt == 112:
		OBJVARIANT = Object_mn_objvariant.val
	elif evt == 113:
		BUCKETSVIRTUALONLY = Object_tg_objbucketsvirtualonly.val
	elif evt == 117:
		EXPORTMODE = Object_mn_EXPORTMODE.val
	elif evt == 114:
		AUTOVARIANT = Object_tg_objautovariant.val
	elif evt == 133:
		VARSTOLIB = Object_tg_objvarstolib.val
	elif evt == 115:
		USEHEIGHTBUCKETS = Object_tg_objbuckets.val
	elif evt == 116:
		OBJRENAME = Object_tg_objrename.val
	elif evt == EVENT_nb_objwrapperlayer:
		DEF_OBJ_WRAPPER_LAYER = Object_nb_objwrapperlayer.val
	elif evt == EVENT_mn_propname0:
		pass
	elif evt == EVENT_mn_propname1:
		pass
	elif evt == EVENT_mn_propname2:
		pass
	elif evt == EVENT_st_propval0:
		i = 0
		for k, v in objprops.iteritems():
			if i == Object_mn_propname0.val:
				objprops[k] = Object_st_propval0.val
			i += 1
	elif evt == EVENT_st_propval1:
		i = 0
		for k, v in objprops.iteritems():
			if i == Object_mn_propname1.val:
				objprops[k] = Object_st_propval1.val
			i += 1
	elif evt == EVENT_st_propval2:
		i = 0
		for k, v in objprops.iteritems():
			if i == Object_mn_propname2.val:
				objprops[k] = Object_st_propval2.val
			i += 1
	elif evt == EVENT_tg_objpropsoverwrite:
		OBJPROPSOVERWRITE = Object_tg_objpropsoverwrite.val
	elif evt == 250:
		OBJWRITEPROPS = Object_tg_objwriteprops.val
	elif evt == 256:
		rep = PupMenu("Really strip all but RNAME and VNAME props from selected?:%t|    NO|  YES")
		if rep == 2:
			p, o = stripProperties()
			clearResults()
			tasksdone['objprops_stripped'] = p
			tasksdone['objs_stripped_from'] = o
			showResults()

			return
		else:
			Redraw(1)
			return
	elif evt == EVENT_bt_objshowfaces:
		clearResults()
		searchFaceModes('selected')

	elif evt == EVENT_bt_objshowfacesvis:
		print modestosearch
		clearResults()
		searchFaceModes('visible')

	elif evt == EVENT_bt_objfindmixedalpha:
		findMixedAlpha()

	elif evt == EVENT_bt_objselallfaces:
		multiObjectFaceSelection('select')
	elif evt == EVENT_bt_objdeselallfaces:
		multiObjectFaceSelection('deselect')
	elif evt == EVENT_tg_objmatchany:
		if not facematchall:
			Object_tg_objmatchany.val = 1
		facematchall = 1 - Object_tg_objmatchany.val
		Object_tg_objmatchall.val = facematchall
	elif evt == EVENT_tg_objmatchall:
		if facematchall:
			Object_tg_objmatchall.val = 1
		facematchall = Object_tg_objmatchall.val
		Object_tg_objmatchany.val = 1 - facematchall

	elif evt == EVENT_bt_objmatchlogic:
		facematchall = not facematchall

	elif evt == EVENT_bt_objsetfaces:
		rep = PupMenu("Really apply all included mode settings to selected faces in selected meshes? :%t|    NO|  YES")
		if rep == 2:
			clearResults()
			doFaceModes('SET')
			showResults()

	elif evt == EVENT_bt_objselwrappers:
		clearResults()
		tasksdone['OBJs_selected'] = selectOBJs()
		showResults()

	elif evt == 251:
		clearResults()
		tasksdone['OBJ_meshes_selected'] = selectOBJMeshes(['OBJ'])
		showResults()

	elif evt == 252:
		pass
	elif evt == 253:
		if Object_mn_objimage.val == 998:
			buildObjImageMenu()
			Object_mn_objimage.val = 999
			Redraw()
	elif evt == 254:

		if Object_mn_objimage.val not in [998, 999]:	# 'blank' and 'refresh' menu items respectively
			clearResults()
			tasksdone['assign_OBJ_image'] = assignImageToSelected(OBJIMAGELIST[Object_mn_objimage.val])
			showResults()
	elif evt == EVENT_bt_objgetsameimage:
		if Object_mn_objimage.val not in [998, 999]:
			clearResults()
			tasksdone['shared_image_obs'] = selectSameImageObs(OBJIMAGELIST[Object_mn_objimage.val].name)
			showResults()
	elif evt == EVENT_bt_objgetfromactive:
		getImageFromOb()

	elif evt == 255:
		clearResults()
		tasksdone['moveOBJCentre'] = moveOBJCentre()
		showResults()

	elif evt == EVENT_bt_restrictselect:
		num_obs = restrictSelectChildren('restrict')

	elif evt == EVENT_bt_enableselect:
		num_obs = restrictSelectChildren('enable')

	elif evt == EVENT_tg_AWS_enable:
		autoWrapperSelect(Object_tg_AWS_enable.val)

	elif evt == EVENT_bt_dupehierarchy:
		num_obs = duplicateOBJHierarchy()
	elif evt == evt_sort_imports:
		orderImportedBySize()
	elif evt == EVENT_tg_objfmdynamic:
		facemodes['DYNAMIC'] = Object_tg_objfmdynamic.val
	elif evt == EVENT_tg_objfmtiles:
		facemodes['TILES'] = Object_tg_objfmtiles.val
	elif evt == EVENT_tg_objfmtwoside:
		facemodes['TWOSIDE'] = Object_tg_objfmtwoside.val
	elif evt == EVENT_tg_objfmshadow:
		facemodes['SHADOW'] = Object_tg_objfmshadow.val
	elif evt == EVENT_tg_objfminvisible:
		facemodes['INVISIBLE'] = Object_tg_objfminvisible.val
	elif evt == EVENT_tg_objftmsolid: #The face transparency mode buttons are interlocking,
		#	so clicking on a button which is already 'on' ought to have no effect - it should stay 'on'.
		if facetranspmodes['SOLID']:
			Object_tg_objftmsolid.val = 1
		else:
			facetranspmodes['SOLID'] = Object_tg_objftmsolid.val
		facetranspmodes['ALPHA'] = facetranspmodes['CLIP'] = 0
		Object_tg_objftmalpha.val = Object_tg_objftmclip.val = 0
	elif evt == EVENT_tg_objftmalpha:
		if facetranspmodes['ALPHA']:
			Object_tg_objftmalpha.val = 1
		else:
			facetranspmodes['ALPHA'] = Object_tg_objftmalpha.val
		facetranspmodes['SOLID'] = facetranspmodes['CLIP'] = 0
		Object_tg_objftmsolid.val = Object_tg_objftmclip.val = 0
	elif evt == EVENT_tg_objftmclip:
		if facetranspmodes['CLIP']:
			Object_tg_objftmclip.val = 1
		else:
			facetranspmodes['CLIP'] = Object_tg_objftmclip.val
		facetranspmodes['ALPHA'] = facetranspmodes['SOLID'] = 0
		Object_tg_objftmalpha.val = Object_tg_objftmsolid.val = 0
	elif evt == EVENT_bt_objdefaultmodes:

		facemodes['DYNAMIC'] = Object_tg_objfmdynamic.val = 1
		facemodes['TILES']	 = Object_tg_objfmtiles.val	  = 0
		facemodes['TWOSIDE'] = Object_tg_objfmtwoside.val = 0
		facemodes['SHADOW']	 = Object_tg_objfmshadow.val  = 0
		facemodes['INVISIBLE'] = Object_tg_objfminvisible.val = 0
		facetranspmodes['SOLID'] = Object_tg_objftmsolid.val = 1
		facetranspmodes['ALPHA'] = Object_tg_objftmalpha.val = 0
		facetranspmodes['CLIP']	 = Object_tg_objftmclip.val	 = 0
	elif evt == EVENT_bt_objenableall:
		if all([v.val for v in Object_tg_objfacemodesel]):
			val_to_set = 0
		else:
			val_to_set = 1
		for tg in Object_tg_objfacemodesel:
			tg.val = val_to_set
		for k in modestosearch.iterkeys():
			modestosearch[k] = val_to_set
		Draw()
	elif evt in EVENT_tg_objfacemodesel:
		if evt == EVENT_tg_objfacemodesel[0]:
			modestosearch['TILES'] 	  = Object_tg_objfacemodesel[0].val
		if evt == EVENT_tg_objfacemodesel[1]:
			modestosearch['TWOSIDE']  = Object_tg_objfacemodesel[1].val
		if evt == EVENT_tg_objfacemodesel[2]:
			modestosearch['DYNAMIC']  = Object_tg_objfacemodesel[2].val
		if evt == EVENT_tg_objfacemodesel[3]:
			modestosearch['SHADOW']   = Object_tg_objfacemodesel[3].val
		if evt == EVENT_tg_objfacemodesel[4]:
			modestosearch['INVISIBLE']= Object_tg_objfacemodesel[4].val
		if evt == EVENT_tg_objfacemodesel[5]:
			modestosearch['TRANSP']   = Object_tg_objfacemodesel[5].val
	elif evt == 138:
		objprops2['INSTANCED'] = str(Object_tg_objinstanced.val)
		OBJINSTANCED = Object_tg_objinstanced.val
	elif evt == 139:
		pass

	elif evt == 140:
		objprops2['ATTR_draped'] = str(Object_tg_objhasdraped.val)
		HASDRAPED = Object_tg_objhasdraped.val
	elif evt == 141:
		objprops2["lod_1"] = Object_nb_objlod1.val
	elif evt == 142:
		objprops2["lod_2"] = Object_nb_objlod2.val
	elif evt == 150:
		ob_OBJatt = 1 - ob_OBJatt
	elif evt == 151:
		ob_FACatt = 1 - ob_FACatt
	elif evt == 152:
		ob_TREEatt = 1- ob_TREEatt
	elif evt == 160:
		makeMemberGroup(sdata['member_group'])
	elif evt == 161:
		sdata['comment'] = cmmt.val
	elif evt == 162:
		sdata['min_is_rel'] = rlmn.val
	elif evt == 163:
		sdata['max_is_rel'] = rlmx.val
	elif evt == 164:
		sdata['actual_min'] = acmn.val
	elif evt == 165:
		sdata['actual_max'] = acmx.val
	elif evt == 166:
		sdata['declared_min'] = dcmn.val
	elif evt == 167:
		sdata['declared_max'] = dcmx.val
	elif evt == 170:
		for k in tasksdone.iterkeys():
			tasksdone[k] = 0
		if ob_OBJatt:
			tasksdone['OBJatt'] = collectAnnotations(0)
		if ob_FACatt:
			tasksdone['FACatt'] = collectAnnotations(1)
		if ob_TREEatt:
			tasksdone['TREEatt'] = collectAnnotations(2)
		showResults()

	elif evt == 171:
		sdata['angle_start'] = angs.val
	elif evt == 172:
		sdata['angle_end'] = ange.val
	elif evt == 173:
		sdata['depth_min'] = dpmn.val
	elif evt == 174:
		sdata['depth_max'] = dpmx.val
	elif evt == 175:
		sdata['tile_variety'] = vari.val
	elif evt == 176:
		sdata['atomic'] = atmc.val
	elif evt == 177:
		sdata['first_tile'] = tfst.val
	elif evt == 178:
		sdata['last_tile'] = tlst.val
	elif evt == 179:
		sdata['depth_variety'] = vard.val
	elif evt == 199:
		applyOBJSettings()
		if OBJWRITEPROPS:
			applyOBJProps()
	# AGS 1
	elif evt == 200:
		GUIpage = 1
		GUIchanged = 1
		layw.val = DEF_TILE_WRAPPER_LAYER
		layr.val = DEF_METAPIN_LAYER
	# AGB 1
	elif evt == 201:
		GUIpage = 3
		GUIchanged = 1
	# OBJ
	elif evt == 202:
		GUIpage = 5
		GUIchanged = 1
		Object_nb_objwrapperlayer.val = DEF_OBJ_WRAPPER_LAYER
	# AGS 2
	elif evt == 203:
		GUIpage = 7
		GUIchanged = 1
	# Lights
	elif evt == 204:
		if not g_lighteditor:
			lightEditorStart()
		GUIpage = 9
		GUIchanged = 1
	# Mesh
	elif evt == 205:
		GUIpage = 11
		GUIchanged = 1
	# AGB 2
	elif evt == 206:
		GUIpage = 13
		GUIchanged = 1
	# AGB Histograms
	elif evt == 207:
		GUIpage = 14
		GUIchanged = 1
	if g_lighteditor:
		g_lighteditor.buttonEvent(evt)
	if g_AGB_editor:
		g_AGB_editor.buttonEvent(evt)

	Redraw(1)

def gui():

	global g_lighteditor
	global g_AGB_editor
	global iRnam, iIdin, iCent, iWrap, iAgsw, iPins, iLayr, iPina, iPinp, iPing, iShrY, iLayw, iProj, fDmin, iSgrp, GUIpage, GUIchanged
	global fMrga, fMrgf, fMrgr, fMrgb, fMrgl, fSpac, fTexw, fTexh, fOrgx, fOrgy
	global sTexn
	global rnam, idin, texn, mrgr, mrgf, mrgl, mrgb, mrga, spac, pina, pinp, ping, shrY, layr, proj, cent, orgx, orgy, texw, texh, optm
	global wrap, agsw, layw, rtrn, dmin, mtch, dslp, sgrp, mgrp, cmmt, rlmn, rlmx, acmn, acmx, dcmn, dcmx
	global angs, ange, dpmn, dpmx, vari, atmc, tfst, tlst, vard
	global ob_OBJatt, ob_FACatt, ob_TREEatt
	global tg_OBJatt, tg_FACatt, tg_TREEatt
	global sResults
	global facemodes
	global facematchall
	global EVENT_tg_objfacemodesel
	global bt_applob, bt_pageAGB, bt_pageAGB2, bt_pageAGS, bt_pageOBJ, bt_pageOBJ2, bt_pageANN, bt_appann, bt_pageMSC
	global sdata
	global EVENT_mn_objpart
	global EVENT_mn_objshape
	global EVENT_mn_objvariant
	global EVENT_tg_objbucketsvirtualonly
	global EVENT_tg_objautovariant
	global EVENT_tg_objvarstolib
	global EVENT_tg_objbuckets
	global EVENT_tg_objrename
	global EVENT_mn_EXPORTMODE
	global EVENT_nb_objwrapperlayer
	global EVENT_mn_propname0
	global EVENT_st_propval0
	global EVENT_st_propval1
	global EVENT_st_propval2
	global EVENT_mn_propname1
	global EVENT_mn_propname2
	global EVENT_tg_objpropsoverwrite
	global EVENT_bt_objgetsameimage
	global EVENT_bt_objgetfromactive
	global EVENT_bt_objselallfaces
	global EVENT_bt_objsetfaces
	global EVENT_bt_objdeselallfaces
	global EVENT_bt_objselwrappers
	global EVENT_bt_objshowfaces
	global EVENT_bt_objfindmixedalpha
	global EVENT_tg_objmatchany
	global EVENT_tg_objmatchall
	global EVENT_bt_objenableall
	global EVENT_bt_objdefaultmodes
	global EVENT_bt_objmatchlogic
	global EVENT_bt_objshowfacesvis
	global EVENT_bt_objresetbuttons
	global EVENT_bt_restrictselect, EVENT_bt_enableselect, EVENT_tg_AWS_enable
	global EVENT_bt_dupehierarchy
	global evt_sort_imports
	global EXPORTMODE
	global DEF_OBJ_WRAPPER_LAYER
	global OBJVIRTUALPATH
	global OBJRENAME
	global OBJSHAPE
	global OBJPART
	global OBJVARIANT
	global OBJINSTANCED
	global AUTOVARIANT
	global VARSTOLIB
	global OBJPROPSOVERWRITE
	global USEHEIGHTBUCKETS
	global BUCKETSVIRTUALONLY
	global objprops
	global objprops2
	global objpropkeys
	global Object_mn_objpart
	global Object_mn_objshape
	global Object_mn_objvariant
	global Object_tg_objbucketsvirtualonly
	global Object_tg_objautovariant
	global Object_tg_objvarstolib
	global Object_tg_objbuckets
	global Object_tg_objrename
	global Object_mn_EXPORTMODE
	global Object_nb_objwrapperlayer
	global Object_mn_propname0
	global Object_st_propval0
	global Object_st_propval1
	global Object_st_propval2
	global Object_mn_propname1
	global Object_mn_propname2
	global Object_tg_objpropsoverwrite
	global Object_tg_objwriteprops
	global Object_bt_objselallfaces
	global Object_bt_objsetfaces
	global Object_bt_objdeselallfaces
	global Object_bt_objselwrappers
	global Object_bt_objselgeom
	global Object_bt_objgetsameimage
	global Object_bt_objgetfromactive
	global Object_bt_objassignimage
	global bt_sort_imports

	global Object_bt_restrictselect, Object_bt_enableselect, Object_tg_AWS_enable
	global Object_bt_dupehierarchy
	global Object_bt_objshowfaces
	global Object_bt_objfindmixedalpha
	global Object_tg_objmatchany
	global Object_tg_objmatchall
	global Object_bt_objenableall
	global Object_bt_objdefaultmodes
	global Object_bt_objmatchlogic
	global Object_bt_objshowfacesvis
	global Object_bt_objresetbuttons
	global Object_tg_objfmdynamic
	global Object_tg_objfmtiles
	global Object_tg_objfmtwoside
	global Object_tg_objfmshadow
	global Object_tg_objfminvisible
	global modestosearch
	global Object_tg_objftmsolid
	global Object_tg_objftmalpha
	global Object_tg_objftmclip
	global Object_tg_objfacemodesel
	global Object_tg_objinstanced
	global Object_mn_objpropname3
	global Object_mn_objpropname4
	global Object_tg_objhasdraped
	global Object_nb_objlod1
	global Object_nb_objlod2
	global Object_mn_objimage
	global Object_objmoveOBJcentre
	global Object_bt_objstripprops
	global OBJIMAGELIST
	global objimagemenuname
	global cp_colour, st_sRGBint,	st_sRGBflt,	st_linRGB, bt_nrmlize, sl_hue, sl_saturation, sl_brightness, tg_RGB, tg_HSB, sl_red, sl_green, sl_blue
	global evt_cp_colour, evt_st_sRGBint,  evt_st_sRGBflt,	evt_st_linRGB, evt_bt_nrmlize, evt_sl_hue, evt_sl_saturation, evt_sl_brightness
	global evt_tg_RGB,	evt_tg_HSB,	evt_sl_red,	evt_sl_green, evt_sl_blue
	global tg_RGB, tg_HSB, nm_normal, evt_nm_normal, st_normal, evt_st_normal, g_normal
	global mdata

	global LIGHTS_TXT_DIR

	if GUIpage == 0:
		pass

	elif GUIpage == 1:

		# AGS 1 goo
		BGL.glClearColor(0.60, 0.60, 0.60, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(10, 177, 815, 204)
		BGL.glColor3f(*CL_GUI_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')
		BGL.glColor3f(*CL_GREY60)
		BGL.glRecti(120, 175, 200, 202)
		# BGL.glColor3f(*CL_GUI_ORANGE)
# 		BGL.glRasterPos2i(148, 184)
# 		Text('AGS 1', 'normal')
		drawString(				'EXISTING PINS',				767, 	38, 	CL_TEXT_LT,		'center',		'small')
		drawString(				'AGS 1',						161, 	184, 	CL_GUI_ORANGE,	'center',		'normal')
		bt_pageANN = PushButton("AGS 2", 				203, 	200, 	177, 	80, 	22, 							"Go to the AGS spellings and annotation tools page.")
		bt_pageAGB = PushButton("AGB 1",				201, 	280, 	177, 	80, 	22, 							"Go to the AGB tools page.")
		bt_pageAGB2= PushButton("AGB 2",				206, 	360, 	177, 	80, 	22, 							"Go to the AGB spellings and annotation tools page.")
		bt_pageOBJ = PushButton("OBJ", 					202, 	440, 	177, 	80, 	22, 							"Go to the OBJ tools page.")
		bt_pageOBJ2= PushButton("Mesh", 				205, 	520, 	177, 	80, 	22, 							"Goto the MESH tools page.")
		bt_pageMSC = PushButton("Lights", 				204, 	600, 	177, 	80, 	22, 							"Go to the lights tools page.")

# 		bt_pageAGB = PushButton("AGB",				 	201, 	200, 	177, 	80, 	22, 							"Go to the AGB tools page.")
# 		bt_pageOBJ = PushButton("OBJ", 					202, 	280, 	177, 	80, 	22, 							"Go to the OBJ tools page.")
# 		bt_pageOBJ2= PushButton("Mesh", 				205, 	360, 	177, 	80, 	22, 							"Goto the MESH tools page.")
# 		bt_pageANN = PushButton("Annotation and Spelling", 203, 440, 	177, 	160, 	22, 							"Go to the annotation and spelling tools page.")
# 		bt_pageMSC = PushButton("Lights", 				204, 	600, 	177, 	80, 	22, 							"Go to the lights tools page.")

		rnam = 		Toggle(		"rename tile meshes", 	1, 		10, 	148, 	190, 	22, 	iRnam, 					"Rename selected tiles sequentially from left to right, starting from 'start ID'.")
		idin = 		Number(		"start ID:", 			2, 		10,	 	126, 	190, 	22, 	iIdin, 	0, 		25600, 	"ID number for first tile in renaming sequence.")
		cent = 		Toggle(		"move tile centres", 	3, 		10, 	104, 	190, 	22, 	iCent, 					"Move selected tile centres to lower left-hand corner.")
		wrap = 		Toggle(		"wrap tiles", 			5, 		10, 	60, 	190, 	22, 	iWrap, 					"Add 'empty' object wrappers to selected tiles, name and make parent.")
		agsw = 		Toggle(		"wrap AGS", 			6, 		10, 	38, 	190, 	22, 	iAgsw, 					"Add 'empty' object AGS wrapper at the cursor location, and make parent.")
		layw = 		Number(		"AG WRAPPER LAYER:", 	7,  	10,	 	10,	 	190, 	22, 	layw.val, 	1,		20,		"Blender layer for wrappers to be located in.")
		layr = 		Number(		"META-PIN LAYER:", 		8,  	210, 	10, 	300, 	22, 	layr.val, 	1,		20,		"Blender layer for meta-pins to be located in.")
		pins = 		Toggle(		"stop pins", 			9,  	310, 	82, 	100, 	22, 	iPins, 					"Add stop pins to selected tiles, using specified inset distances.")
		mrgr = 		Number(		"right:", 				10, 	410, 	82, 	100, 	22, 	fMrgr, 	0, 		16.0, 	"Stop pin inset distance, right edge.")
		mrgf = 		Number(		"front:", 				11, 	310, 	60, 	100, 	22, 	fMrgf, 	0, 		16.0, 	"Stop pin inset distance, front edge.")
		mrgl = 		Number(		"left:", 				12, 	210, 	82, 	100, 	22, 	fMrgl, 	0, 		16.0, 	"Stop pin inset distance, left edge.")
		mrgb = 		Number(		"back:", 				13, 	310, 	104,  	100, 	22, 	fMrgb, 	0, 		16.0, 	"Stop pin inset distance, back edge.")
		spac = 		Number(		"spacing:", 			14, 	310, 	38, 	100, 	22, 	fSpac, 	5,	 	50, 	"Nominal stop pin spacing.")
		mrga = 		Number(		"front:", 				15, 	410, 	126,  	100, 	22, 	fMrga, 	0, 		16.0, 	"Anchor pin inset distance, front edge.")
		pina = 		Toggle(		"anchor pin", 			16,  	410, 	148, 	100, 	22, 	iPina, 					"Add an anchor pin to each selected tile.")
		pinp = 		Toggle(		"pivot pins", 			17,  	310, 	148, 	100, 	22, 	iPinp, 					"Add a pair of pivot pins to each selected tile.")
		ping = 		Toggle(		"ground point", 		18,  	210, 	148, 	100, 	22, 	iPing, 					"Add a ground elevation reference point to each selected tile.")
		shrY = 		Toggle(		"share y",				19,  	210, 	126, 	100, 	22, 	iShrY, 					"Add a SHARE_Y property to each selected tile.")
		texn = 		String(		"", 					40, 	520, 	126, 	150, 	22, 	sTexn, 	21, 			"Enter blender image filename")
		proj = 		Toggle(		"project texture", 		41,  	520, 	148, 	190, 	22, 	iProj, 					"Project texture vertically downwards onto all selected objects, at the specified scale.")
		curr = 		PushButton(	"GET", 					46,  	670, 	126, 	40, 	22, 							"Use the image currently displayed in the UV/Image Editor window.")
		texw = 		Number(		"size x:",				42, 	520, 	104,  	95, 	22, 	fTexw, 	8, 		2048, 	"Texture width in metres.")
		texh = 		Number(		"size y:",				43, 	615, 	104,  	95, 	22, 	fTexh, 	8, 		2048, 	"Texture height in metres.")
		orgx = 		Number(		"loc x:", 				44, 	520, 	82,  	95, 	22, 	fOrgx, 	-4096, 	4096, 	"X-axis coordinate of lower left hand corner of texture.")
		orgy = 		Number(		"loc y:", 				45, 	615, 	82,  	95, 	22, 	fOrgy, 	-4096, 	4096, 	"Y-axis coordinate of lower left hand corner of texture.")
		canc = 		PushButton(	"QUIT",					98, 	772, 	181,    38, 	18, 							"Cancel tasks and exit script. Settings will be lost.")
		clrs = 		PushButton(	"RESET", 				96, 	520, 	10, 	84, 	22, 							"Clear all task selector buttons. Value fields will not be cleared.")
		appl = 		PushButton(	"APPLY",				99, 	606, 	10, 	104, 	22, 							"Apply settings to selected tiles.")
		optm = 		PushButton(	"OPTIMISE",				20, 	720, 	10, 	95, 	22, 							"Declutter .blend file by linking all 'PINxxx'-type lamp obs in the current scene to a single lamp datablock and removing redundant datablocks.")

	elif GUIpage == 2:

		# Display results goo
		if len(sResults) > 12:
			PupMenu("Dev error: more than 12 results!",1)
		else:
			rtrn = PushButton(	"Done",					97,  	711, 	10, 	104, 	22,  							"Close results and return to settings.")

			BGL.glClearColor(0.40, 0.40, 0.40, 1)
			BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
			BGL.glColor3f(*CL_GUI_LT_BLUE)
			BGL.glRasterPos2i(20, 184)
			Text('XPLANETOOLS', 'large')
			BGL.glColor3f(0.8, 0.8, 0.8)
			BGL.glRasterPos2i(120, 184)
			Text('Notification', 'normal')

			BGL.glColor3f(1, 1, 1)
			for i in range(min(len(sResults), 12)):
				if "WARNING" in sResults[i]:
					BGL.glColor3f(1.0, 0.61, 0.23)
					BGL.glRasterPos2i(20, 160 - 14 * i)
					Text(sResults[i], 'normal')
					BGL.glColor3f(1, 1, 1)
				else:
					BGL.glRasterPos2i(20, 160 - 14 * i)
					Text(sResults[i], 'normal')

	elif GUIpage == 3:

		# AGB 1 goo

		BGL.glClearColor(0.60, 0.60, 0.60, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(10, 177, 815, 204)
		BGL.glColor3f(*CL_GUI_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')
		BGL.glColor3f(*CL_GREY60)
		BGL.glRecti(280, 175, 360, 202)


		bt_pageAGS = PushButton("AGS 1", 				200, 	120, 	177, 	80, 	22, 							"Goto the AGS tools page.")
		bt_pageANN = PushButton("AGS 2", 				203, 	200, 	177, 	80, 	22, 							"Go to the AGS spellings and annotation tools page.")
		drawString(				'AGB 1',						321, 	184, 	CL_GUI_ORANGE,	'center',		'normal')
		bt_pageAGB2= PushButton("AGB 2",				206, 	360, 	177, 	80, 	22, 							"Go to the AGB spellings and annotation tools page.")
		bt_pageOBJ = PushButton("OBJ", 					202, 	440, 	177, 	80, 	22, 							"Go to the OBJ tools page.")
		bt_pageOBJ2= PushButton("Mesh", 				205, 	520, 	177, 	80, 	22, 							"Goto the MESH tools page.")
		bt_pageMSC = PushButton("Lights", 				204, 	600, 	177, 	80, 	22, 							"Go to the lights tools page.")
		canc 	   = PushButton("QUIT",					98, 	772, 	181,    38, 	18, 							"Cancel tasks and exit script. Settings will be lost.")

		if g_AGB_editor:
			g_AGB_editor.drawAGBEditor(1)
		else:
			g_AGB_editor = AGBEditor(1000)

	elif GUIpage == 13:

		# AGB 2 goo:

		BGL.glClearColor(0.60, 0.60, 0.60, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(10, 177, 815, 204)
		BGL.glColor3f(*CL_GUI_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')
		BGL.glColor3f(*CL_GREY60)
		BGL.glRecti(360, 175, 440, 202)

		bt_pageAGS = PushButton("AGS 1", 				200, 	120, 	177, 	80, 	22, 							"Goto the AGS tools page.")
		bt_pageANN = PushButton("AGS 2", 				203, 	200, 	177, 	80, 	22, 							"Go to the AGS spellings and annotation tools page.")
		bt_pageAGB = PushButton("AGB 1",				201, 	280, 	177, 	80, 	22, 							"Go to the AGB tools page.")
		drawString(				'AGB 2',						401, 	184, 	CL_GUI_ORANGE,	'center',		'normal')
		bt_pageOBJ = PushButton("OBJ", 					202, 	440, 	177, 	80, 	22, 							"Go to the OBJ tools page.")
		bt_pageOBJ2= PushButton("Mesh", 				205, 	520, 	177, 	80, 	22, 							"Goto the MESH tools page.")
		bt_pageMSC = PushButton("Lights", 				204, 	600, 	177, 	80, 	22, 							"Go to the lights tools page.")
		canc 	   = PushButton("QUIT",					98, 	772, 	181,    38, 	18, 							"Cancel tasks and exit script. Settings will be lost.")
		if g_AGB_editor:
			g_AGB_editor.drawAGBEditor(2)
		else:
			g_AGB_editor = AGBEditor(1000)

	elif GUIpage == 14:

		# AGB Histograms:
		BGL.glClearColor(0.40, 0.40, 0.40, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GUI_LT_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')

		if g_AGB_editor:
			g_AGB_editor.drawAGBEditor(2)

	elif GUIpage == 5:

		# OBJ goo
		BGL.glClearColor(0.60, 0.60, 0.60, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(10, 177, 815, 204)
		BGL.glColor3f(*CL_GUI_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')
		BGL.glColor3f(*CL_GREY60)
		BGL.glRecti(440, 175, 520, 202)

		drawString(								 	'SELECT OBJ',											767, 155, 				CL_TEXT_LT,	'center',	'small')
		drawString(								 	'PROPERTIES',											298, 66, 				CL_TEXT_LT,	'center',	'small')
		drawString(								 	'MAKE SELECTABLE',											767, 88, 				CL_TEXT_LT,	'center',	'small')
		bt_pageAGS = PushButton("AGS 1", 				200, 	120, 	177, 	80, 	22, 							"Goto the AGS tools page.")
		bt_pageANN = PushButton("AGS 2", 				203, 	200, 	177, 	80, 	22, 							"Go to the AGS spellings and annotation tools page.")
		bt_pageAGB = PushButton("AGB 1",				201, 	280, 	177, 	80, 	22, 							"Go to the AGB tools page.")
		bt_pageAGB2= PushButton("AGB 2",				206, 	360, 	177, 	80, 	22, 							"Go to the AGB spellings and annotation tools page.")
		drawString(				'OBJ',							481, 	184, 	CL_GUI_ORANGE,	'center',		'normal')
		bt_pageOBJ2= PushButton("Mesh", 				205, 	520, 	177, 	80, 	22, 							"Goto the MESH tools page.")
		bt_pageMSC = PushButton("Lights", 				204, 	600, 	177, 	80, 	22, 							"Go to the lights tools page.")

		partmenuname = "OBJ name part code %t"
		for k in range(len(enum_objpart)):
			partmenuname += ("|%s %%x%d" % (enum_objpart[k], k))
		Object_mn_objpart = Menu(partmenuname,								110,							10, 82, 190, 22,		OBJPART, "")

		shapemenuname = "OBJ name shape code %t"
		for k in range(len(enum_objshape)):
			shapemenuname += ("|%s %%x%d" % (enum_objshape[k], k))
		Object_mn_objshape = Menu(shapemenuname,							111,							10, 104, 190, 22,		OBJSHAPE, "")
		Object_mn_objvariant = Number("var ",								112,							10, 60,  60, 22,		OBJVARIANT, 1, 8, "")
		Object_tg_objautovariant = Toggle("autovars",						114,							70, 60,  60, 22,		AUTOVARIANT, "")
		Object_tg_objvarstolib = Toggle("vars to lib",						133,							130, 60,  70, 22,		VARSTOLIB, "")
		Object_tg_objbuckets = Toggle("use height buckets",					115,							10, 126, 120, 22,		USEHEIGHTBUCKETS, "")
		Object_tg_objbucketsvirtualonly = Toggle("for lib only",			113,							130, 126, 70, 22,		BUCKETSVIRTUALONLY, "")
		Object_tg_objrename = Toggle("auto rename",							116,							10, 148, 190,  22,		OBJRENAME, "")
		exportmodename = "Export mode %t|using AG exporter %x0|Using bulk OBJ exporter %x1"
		Object_mn_EXPORTMODE = Menu(exportmodename,							117,							10,	 38, 190,   22,		EXPORTMODE, "")
		Object_nb_objwrapperlayer = Number("OBJ WRAPPER LAYER",				EVENT_nb_objwrapperlayer,		10,  10,  190,  22,		Object_nb_objwrapperlayer.val, 1, 20, "")
		propkeys = []
		for i, k in enumerate(objprops.iterkeys()):
			propkeys.append([k, i])
		propkeys.sort()
		prop0menuname = "OBJ property name %t|"" %x999"
		for k, i in propkeys:
			prop0menuname += ("|%s %%x%d" % (k, i))

		Object_mn_propname0 = Menu(prop0menuname,							EVENT_mn_propname0,				210, 148, 175, 22,		Object_mn_propname0.val, "")

		prop1menuname = "OBJ property name %t|"" %x999"
		for k, i in propkeys:
			prop1menuname += ("|%s %%x%d" % (k, i))
		Object_mn_propname1 = Menu(prop1menuname,							EVENT_mn_propname1,				210, 126, 175, 22,		Object_mn_propname1.val, "")

		prop2menuname = "OBJ property name %t|"" %x999"
		for k, i in propkeys:
			prop2menuname += ("|%s %%x%d" % (k, i))
		Object_mn_propname2			= Menu(prop2menuname,					EVENT_mn_propname2,				210, 104, 175, 22,		Object_mn_propname2.val, "")
		Object_st_propval0			= String("",							EVENT_st_propval0,				385, 148, 325, 22,		Object_st_propval0.val, 399, "")
		Object_st_propval1			= String("",							EVENT_st_propval1,				385, 126, 325, 22,		Object_st_propval1.val, 399, "")
		Object_st_propval2			= String("",							EVENT_st_propval2,				385, 104, 325, 22,		Object_st_propval2.val, 399, "")

		propkeys2 = []
		for i, k in enumerate(objpropkeys.iterkeys()):
			propkeys2.append([k, i])
		propkeys2.sort()
		prop3menuname = "OBJ property name %t|"" %x999"
		for k, i in propkeys2:
			prop3menuname += ("|%s %%x%d" % (k, i))
		Object_mn_objpropname3		= Menu(		prop3menuname,				139,							210, 82,	88,		22,		Object_mn_objpropname3.val, "")

		prop4menuname = "OBJ property name %t|"" %x999"
		for k, i in propkeys2:
			prop4menuname += ("|%s %%x%d" % (k, i))
		Object_mn_objpropname4		= Menu(		prop4menuname,				252,							298, 82,	87,		22,		Object_mn_objpropname4.val, "")
		Object_tg_objinstanced		= Toggle(	 "instanced",				138,							385, 82,	65,		22,		Object_tg_objinstanced.val, "")
		Object_tg_objhasdraped		= Toggle(	 "has draped",				140,							450, 82,	70,		22,		Object_tg_objhasdraped.val, "")
		Object_tg_objpropsoverwrite = Toggle(	 "overwrite",				EVENT_tg_objpropsoverwrite,		298, 38,   	87, 	22,		OBJPROPSOVERWRITE, 			"Overwrite any existing property values")
		Object_tg_objwriteprops		= Toggle(	 "update",					250,							210, 38,  	88, 	22,		OBJWRITEPROPS,				"If selected, new properties will be written to selected wrappers when settings are applied.")
		Object_bt_objstripprops		= PushButton("STRIP PROPS",				256,							210, 10,  	175,	22,									"Strip all properties except VNAME and RNAME from selected objects.")
		Object_bt_objselwrappers	= PushButton("Wrappers",				EVENT_bt_objselwrappers,		720, 126,  	95,		22,									"Select all wrappers in the current scene which have the prefix 'OBJ' to their names.")
		Object_bt_objselgeom		= PushButton("Meshes",					251,							720, 104,  	95,		22,									"Select all mesh geometry children of selected OBJ wrappers")

		Object_bt_restrictselect    = PushButton("Wrappers Only",			EVENT_bt_restrictselect,		720, 60,  	95,		22,									"Prevent selection of children of 'empty'-type wrappers in the 3-d view. Applies to selected objects if one or more are selected, otherwise all. Handy when working with multiple, closely-packed objects.")
		Object_bt_enableselect    	= PushButton("All Parts",				EVENT_bt_enableselect,			720, 38,  	95,		22,									"Re-enable selection of children of 'empty'-type wrappers. Applies to selected objects if one or more are selected, otherwise all.")
		Object_bt_dupehierarchy		= PushButton("Duplicate OBJ",			EVENT_bt_dupehierarchy,			720, 10,  	95,		22,									"Make a shallow copy of all parts of an OBJ hierarchy (wrapper, mesh(es) and lamps), of which at least one part is selected, unparent it and offset it by 10m.")
		Object_nb_objlod1			= String("LOD 1: ",						141,							520, 82,   	95,  	22,		objprops2['lod_1'],	6, 		"")
		Object_nb_objlod2			= String("LOD 2: ",						142,							615, 82,   	95,		22,		objprops2['lod_2'],	6, 		"")
		Object_objmoveOBJcentre		= PushButton("RECENTER MESH",			255,							395, 10, 	115,  	22,									"Moves the centers of selected meshes to front center, assessed at ground level. Can help to keep building doorways in the right place.")
		canc						= PushButton(	"QUIT",					98, 							772, 181,    38, 	18, 								"Cancel tasks and exit script. Settings will be lost.")
		bt_applob					= PushButton("APPLY SETTINGS",			199,							520, 10, 	190, 	22, 								"Apply all enabled settings to selected objects.")

	elif GUIpage == 11:

		# MESH goo

		BGL.glClearColor(0.60, 0.60, 0.60, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(10, 177, 815, 204)
		BGL.glColor3f(*CL_GUI_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')
		BGL.glColor3f(*CL_GREY60)
		BGL.glRecti(520, 175, 600, 202)

		psn_x = [615, 615, 240, 240, 300, 300, 360, 360, 420, 420, 480, 480, 285, 285, 435, 435, 285, 436, 552, 552, 678, 678, 552, 679, 240, 616]
		psn_y = [ 30, 128,  90, 128,  90, 128,  40, 128,  90, 128,  90, 128,  30,  41,  30,  41,  40,  40, 111, 128, 111, 128, 111, 111,  90, 90]
		BGL.glColor3f(*CL_GREY75)
		BGL.glBegin(BGL.GL_LINES)
		for x, y in zip(psn_x, psn_y):
			BGL.glVertex2i(x, y)
		BGL.glEnd()

		bt_pageAGS = PushButton("AGS 1", 				200, 	120, 	177, 	80, 	22, 							"Goto the AGS tools page.")
		bt_pageANN = PushButton("AGS 2", 				203, 	200, 	177, 	80, 	22, 							"Go to the AGS spellings and annotation tools page.")
		bt_pageAGB = PushButton("AGB 1",				201, 	280, 	177, 	80, 	22, 							"Go to the AGB tools page.")
		bt_pageAGB2= PushButton("AGB 2",				206, 	360, 	177, 	80, 	22, 							"Go to the AGB spellings and annotation tools page.")
		bt_pageOBJ = PushButton("OBJ", 					202, 	440, 	177, 	80, 	22, 							"Go to the OBJ tools page.")
		drawString(				'Mesh',							561, 	184, 	CL_GUI_ORANGE,	'center',		'normal')
		bt_pageMSC = PushButton("Lights", 				204, 	600, 	177, 	80, 	22, 							"Go to the lights tools page.")
		canc	   = PushButton("QUIT",					98, 	772, 	181,    38, 	18, 							"Cancel tasks and exit script. Settings will be lost.")

		if not objimagemenuname:
			buildObjImageMenu()
		Object_mn_objimage			= Menu(			 objimagemenuname,			253,						10,		60, 	190, 	22,		Object_mn_objimage.val, "")
		Object_bt_objgetfromactive 	= PushButton(	"From Active",				EVENT_bt_objgetfromactive,	10, 	38, 	95, 	22,										"Select all mesh objects which use the same texture image as the active object.")
		Object_bt_objgetsameimage 	= PushButton(	"Select Users",			EVENT_bt_objgetsameimage,		105, 	38, 	95, 	22,										"Select all mesh objects which use the same texture image as the active object.")
		Object_bt_objassignimage 	= PushButton(	"TEXTURE TO SELECTED",		254,						10, 	10, 	190, 	22,										"Assign selected image to selected meshes.")
		Object_bt_objsetfaces		= PushButton(	"APPLY TO SELECTED",		EVENT_bt_objsetfaces,		520,  	10,  	190,	22,										"Apply the target mode settings to all selected faces of selected objects.")
		Object_bt_objshowfaces		= PushButton(	"Find In Selected Obs",  	EVENT_bt_objshowfaces,		210,  	10,  	150,   	22,										"Search selected meshes, selecting all faces which match the target mode settings.")
		Object_bt_objshowfacesvis	= PushButton(	"Find In Visible Obs",		EVENT_bt_objshowfacesvis,	360,  	10,  	150,	22,										"Search all meshes in visible layers, selecting all faces which match the target mode settings.")
		Object_bt_objfindmixedalpha = PushButton(	"Find Mixed",				EVENT_bt_objfindmixedalpha,	720, 	126,   	95,	  	22,										"Search for and select meshes containing a mix of Alpha (blend) and Clip Alpha (no_blend) faces.")

		drawString(								 	'MESH FACE MODES',										360,	154, 					CL_TEXT_LT,	'center',	'small')
		drawString(								 	'FACE ALPHA BLENDING',									620,	154, 					CL_TEXT_LT,	'center',	'small')
		drawString(								 	'TRANSPARENCY',											767,	154, 					CL_TEXT_LT,	'center',	'small')
		drawString(								 	'TEXTURE TOOLS',										105,	87, 					CL_TEXT_LT,	'center',	'small')
		drawString(								 	'ALL FACES IN',											767,	80, 					CL_TEXT_LT,	'center',	'small')
		drawString(								 	'SELECTED OBS',											767,	67, 					CL_TEXT_LT,	'center',	'small')

		if sum([v for v in modestosearch.itervalues()]) > 1:
			drawButtonFrame(																				300,	60,		120,  	22,  	0.77)
			drawString(				 "match any targets" if facematchall else "match all targets",			360,	67, 					CL_TEXT_DK,		'center',	'normal')
			Object_bt_objmatchlogic = PushButton(	"",						EVENT_bt_objmatchlogic,			300,	46,		120,  	14,										"Toggle between searching for faces matching any/all target settings.")

		Object_bt_objenableall 		= PushButton("  Toggle Targets All/None  >",   EVENT_bt_objenableall, 	 10,	99,  	190,   	22,										"Shortcut to quickly enable all target buttons. Click again to disable all.")
		Object_bt_objdefaultmodes 	= PushButton("   Reset Buttons To Default  >", EVENT_bt_objdefaultmodes, 10,	126,  	190,   	22,										"Resets the mode buttons to their default values.")
		Object_bt_objselallfaces	= PushButton(   "SELECT",					EVENT_bt_objselallfaces,	720,  	38,   	95,   	22,										"Select all faces in visible selected objects.")
		Object_bt_objdeselallfaces	= PushButton(   "DESELECT",					EVENT_bt_objdeselallfaces,	720,	10,   	95,	  	22,										"Deselect all faces in visible selected objects.")

		Object_tg_objfacemodesel[0] = Toggle(		"target",					EVENT_tg_objfacemodesel[0],	210,	99, 	60,   	22,		modestosearch['TILES'],			"Enable this to include the Tiles face mode in searches and actions.")
		Object_tg_objfacemodesel[1] = Toggle(		"target",					EVENT_tg_objfacemodesel[1],	270,	99, 	60,   	22,		modestosearch['TWOSIDE'],		"Enable this to include the Twoside face mode in searches and actions.")
		Object_tg_objfacemodesel[2] = Toggle(		"target",					EVENT_tg_objfacemodesel[2],	330,	99, 	60,   	22,		modestosearch['DYNAMIC'],		"Enable this to include the Collision face mode in searches and actions.")
		Object_tg_objfacemodesel[3] = Toggle(		"target",					EVENT_tg_objfacemodesel[3],	390,	99, 	60,   	22,		modestosearch['SHADOW'],		"Enable this to include the Shadow face mode in searches and actions.")
		Object_tg_objfacemodesel[4] = Toggle(		"target",					EVENT_tg_objfacemodesel[4],	450,	99, 	60,   	22,		modestosearch['INVISIBLE'],		"Enable this to include the Invisible face mode in searches and actions.")
		Object_tg_objfacemodesel[5] = Toggle(		"target",					EVENT_tg_objfacemodesel[5],	583,	99, 	64,   	22,		modestosearch['TRANSP'],		"Enable this to include face Alpha Blending in searches and actions.")

		Object_tg_objfmtiles		= Toggle(		'Tiles', 					EVENT_tg_objfmtiles,		210,	126, 	60,	 	22, 	facemodes['TILES'],				"The Blender Tiles face mode.")
		Object_tg_objfmtwoside		= Toggle(		"Twoside",					EVENT_tg_objfmtwoside,		270,	126, 	60,  	22,		facemodes['TWOSIDE'],			"The Blender Twoside face mode.")
		Object_tg_objfmdynamic		= Toggle(	 	"Collision",				EVENT_tg_objfmdynamic,		330,	126, 	60,  	22,		facemodes['DYNAMIC'],			"The Blender Collision face mode (turn OFF to create 'hard' surfaces in X-Plane).")
		Object_tg_objfmshadow		= Toggle(	 	"Shadow",					EVENT_tg_objfmshadow,		390,	126, 	60,  	22,		facemodes['SHADOW'],			"The Blender Shadow face mode.")
		Object_tg_objfminvisible	= Toggle(	 	"Invisible",				EVENT_tg_objfminvisible,	450,	126, 	60,  	22,		facemodes['INVISIBLE'],	 		"The Blender Invisible face mode.")

		Object_tg_objftmsolid		= Toggle(	 	"Opaque",					EVENT_tg_objftmsolid,		520,	126, 	63,	 	22,		facetranspmodes['SOLID'], 		"No alpha transparency.")
		Object_tg_objftmalpha		= Toggle(	 	"Alpha",					EVENT_tg_objftmalpha,		583,	126, 	64,	 	22,		facetranspmodes['ALPHA'], 		"Blended alpha transparency (ATTR_blend in X-Plane).")
		Object_tg_objftmclip		= Toggle(	 	"Clip",						EVENT_tg_objftmclip,		647,	126, 	63,	 	22,		facetranspmodes['CLIP'], 		"Clipped alpha transparency (ATTR_no_blend in X-Plane).")

	elif GUIpage == 7:

		# AGS 2 goo
		BGL.glClearColor(0.60, 0.60, 0.60, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(10, 177, 815, 204)
		BGL.glColor3f(*CL_GUI_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')
		BGL.glColor3f(*CL_GREY60)
		BGL.glRecti(200, 175, 280, 202)
		BGL.glColor3f(*CL_TEXT_LT)
		BGL.glRasterPos2i(550, 110)
		Text('ENCLOSED ANNOTATIONS', 'small')
		BGL.glRasterPos2i(258, 132)
		Text('SPELLING LENGTHS', 'small')
		BGL.glRasterPos2i(241, 88)
		Text('DECLARED EDGE LENGTHS', 'small')
		BGL.glRasterPos2i(75, 132)
		Text('TILE DEPTHS', 'small')
		BGL.glRasterPos2i(63, 88)
		Text('CORNER ANGLES', 'small')

		bt_pageAGS = PushButton("AGS 1", 				200, 	120, 	177, 	80, 	22, 							"Goto the AGS tools page.")
		drawString(				'AGS 2',						241, 	184, 	CL_GUI_ORANGE,	'center',		'normal')
		bt_pageAGB = PushButton("AGB 1",				201, 	280, 	177, 	80, 	22, 							"Go to the AGB tools page.")
		bt_pageAGB2= PushButton("AGB 2",				206, 	360, 	177, 	80, 	22, 							"Go to the AGB spellings and annotation tools page.")
		bt_pageOBJ = PushButton("OBJ", 					202, 	440, 	177, 	80, 	22, 							"Go to the OBJ tools page.")
		bt_pageOBJ2= PushButton("Mesh", 				205, 	520, 	177, 	80, 	22, 							"Goto the MESH tools page.")
		bt_pageMSC = PushButton("Lights", 				204, 	600, 	177, 	80, 	22, 							"Go to the lights tools page.")


		gspl = 		PushButton(	"GENERATE",				47,  	410, 	10, 	100, 	22,												"Generate a sequence of atomic spellings for the selected tile wrappers to cover the specified range of distances.")
		sgrp =		Number(		"grp",					48, 	10, 	148, 	100, 	22, 	sdata['member_group'], 	0, 		9999, 	"Atomic spelling group for this batch of spellings.")
		mtch =		Number(		"match",				49, 	410, 	82,		100, 	22, 	sdata['match_group'], 	0, 		9999, 	"")
		dslp =		Number(		"slop", 				50, 	410, 	104, 	100, 	22, 	sdata['dist_slop'], 	0.0,	100.0,	"Maximum available slop distance for atomic spelling.")
		rlmn =		Toggle(		"relative", 			162, 	210, 	38, 	100, 	22, 	sdata['min_is_rel'],					"")
		rlmx =		Toggle(		"relative", 			163, 	310, 	38, 	100, 	22, 	sdata['max_is_rel'],					"")
		acmn =		Number(		"min: ",				164, 	210, 	104, 	100, 	22, 	sdata['actual_min'], 	0,		1000, 	"")
		acmx =		Number(		"max: ",				165, 	310, 	104, 	100, 	22, 	sdata['actual_max'], 	0,		10000, 	"Longest distance to cover with this batch of atomic spellings.")
		dcmn =		Number(		"min: ",				166, 	210, 	60, 	100, 	22, 	sdata['declared_min'], 	0,		1000, 	"")
		dcmx =		Number(		"max: ",				167, 	310, 	60, 	100, 	22, 	sdata['declared_max'], 	0,		10000, 	"")
		atomicmenuname = " spelling type %t|     atomic %x0|atomic corner %x1|      basic %x2"
		atmc =		Menu(		atomicmenuname,			176, 	410, 	60, 	100, 	22,		atmc.val,								"")
		angs =		Number(		"start: ",				171, 	10, 	60, 	95, 	22, 	sdata['angle_start'], 	-179, 	179, 	"")
		ange =		Number(		"end: ",				172, 	105, 	60, 	95, 	22, 	sdata['angle_end'], 	-179, 	179, 	"")
		dpmn =		Number(		"min: ",				173, 	10, 	104, 	95, 	22, 	sdata['depth_min'], 	0,		1000, 	"")
		dpmx =		Number(		"max: ",				174, 	105, 	104, 	95, 	22, 	sdata['depth_max'], 	0,		1000, 	"")
		vari =		Slider(		"length var: ",			175,	210,	10,		200,	22,		sdata['tile_variety'],	0,	10,		1,	"")
		vard =		Slider(		"depth var: ",			179,	10,		10,		190,	22,		sdata['depth_variety'],	0,	10,		1,	"")
		tfst =		Number(		"first: ",				177, 	10, 	38, 	95, 	22, 	sdata['first_tile'], 	0, 	100000, 	"")
		tlst =		Number(		"last: ",				178, 	105, 	38, 	95, 	22, 	sdata['last_tile'], 	0, 	100000, 	"")
		mgrp = 		PushButton(	"MAKE GROUP",			160,  	410, 	148, 	100, 	22,												"Print the names of selected tile wrappers as a match group with the specified ID and comment.")
		cmmt =		String(		"",						161,	110,	148,	300,	22,		sdata['comment'], 399,				 	"")
		tg_OBJatt  = Toggle(	"objects", 				150, 	520, 	82, 	190, 	22, 	ob_OBJatt,								"Attach all OBJs located within boundaries of tiles attached to the selected AG entity.")
		tg_FACatt  = Toggle(	"facades",				151,  	520, 	60, 	190, 	22, 	ob_FACatt, 								"Attach all FACades located within boundaries of tiles attached to the selected AG entity.")
		tg_TREEatt = Toggle(	"trees", 				152, 	520, 	38, 	190, 	22, 	ob_TREEatt,								"Attach all TREE quads located within boundaries of tiles attached to the selected AG entity.")
		canc = 		PushButton(	"QUIT",					98, 	772, 	181,    38, 	18, 											"Cancel tasks and exit script. Settings will be lost.")
		bt_appann =	PushButton(	"COLLECT AND ATTACH",	170, 	520, 	10, 	190, 	22, 											"Apply selected settings.")
		bt_sort_imports = PushButton("SORT SEL",	evt_sort_imports,	720,	10,		95,		22,												"Sort locations of all selected OBJS, ordering by width, depth and height bucket.")

	elif GUIpage == 9:

		# lights goo
		BGL.glClearColor(0.60, 0.60, 0.60, 1)
		BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(10, 177, 815, 204)
		BGL.glColor3f(*CL_GUI_BLUE)
		BGL.glRasterPos2i(20, 184)
		Text('XPLANETOOLS', 'large')
		BGL.glColor3f(*CL_GREY60)
		BGL.glRecti(600, 177, 680, 202)
		BGL.glColor3f(*CL_GUI_ORANGE)

		bt_pageAGS = PushButton("AGS 1", 				200, 	120, 	177, 	80, 	22, 							"Goto the AGS tools page.")
		bt_pageANN = PushButton("AGS 2", 				203, 	200, 	177, 	80, 	22, 							"Go to the AGS spellings and annotation tools page.")
		bt_pageAGB = PushButton("AGB 1",				201, 	280, 	177, 	80, 	22, 							"Go to the AGB tools page.")
		bt_pageAGB2= PushButton("AGB 2",				206, 	360, 	177, 	80, 	22, 							"Go to the AGB spellings and annotation tools page.")
		bt_pageOBJ = PushButton("OBJ", 					202, 	440, 	177, 	80, 	22, 							"Go to the OBJ tools page.")
		bt_pageOBJ2= PushButton("Mesh", 				205, 	520, 	177, 	80, 	22, 							"Goto the MESH tools page.")
		if g_lighteditor:
			if g_lighteditor.dev_mode:
				drawString(		'Lights Dev',					641, 	184, 	CL_GUI_ORANGE,	'center',		'normal')
			else:
				drawString(		'Lights',						641, 	184, 	CL_GUI_ORANGE,	'center',		'normal')

		canc	   = PushButton("QUIT",					98, 	772, 	181,    38, 	18, 							"Cancel tasks and exit script. Settings will be lost.")

		if g_lighteditor:
			g_lighteditor.drawEditor()

	if GUIpage not in [2, 14]:
		Object_tg_AWS_enable = Toggle("AWS",EVENT_tg_AWS_enable, 725,	181,	38,		18,		Object_tg_AWS_enable.val,	"Right-click on any part of an OBJ to select its WRAPPER.  Enable 3-d view menu option: 'View/Space Handler Scripts/XPToolsAWS'.")

	if GUIchanged == 1:
		GUIchanged = 0
		Redraw(1)


Register(gui, event, button_event)


